@@ -30,9 +30,9 @@ ADDR    STD_PROTO(Yap_PreAllocCodeSpace, (void));
 Prop	STD_PROTO(PredPropByFunc,(Functor, Term));
 Prop	STD_PROTO(PredPropByAtom,(Atom, Term));
 #include "Yatom.h"
-#include "Heap.h"
 #include "yapio.h"
 #include <stdio.h>
+#include <wchar.h>
 #if HAVE_STRING_H
 #include <string.h>
 #endif
@@ -63,6 +63,8 @@ InlinedUnlockedMkFunctor(AtomEntry *ae, unsigned int arity)
     return ((Functor) RepProp(p0));
   }
   p = (FunctorEntry *) Yap_AllocAtomSpace(sizeof(*p));
+  if (!p)
+    return NULL;
   p->KindOfPE = FunctorProperty;
   p->NameOfFE = AbsAtom(ae);
   p->ArityOfFE = arity;
@@ -139,6 +141,21 @@ SearchAtom(unsigned char *p, Atom a) {
   return(NIL);
 }
 
+static inline Atom
+SearchWideAtom(wchar_t *p, Atom a) {
+  AtomEntry *ae;
+
+  /* search atom in chain */
+  while (a != NIL) {
+    ae = RepAtom(a);
+    if (wcscmp((wchar_t *)ae->StrOfAE, p) == 0) {
+      return a;
+    }
+    a = ae->NextOfAE;
+  }
+  return(NIL);
+}
+
 static Atom
 LookupAtom(char *atom)
 {				/* lookup atom in atom table            */
@@ -194,12 +211,102 @@ LookupAtom(char *atom)
   return na;
 }
 
+static Atom
+LookupWideAtom(wchar_t *atom)
+{				/* lookup atom in atom table            */
+  CELL hash;
+  wchar_t *p;
+  Atom a, na;
+  AtomEntry *ae;
+  UInt sz;
+  WideAtomEntry *wae;
+
+  /* compute hash */
+  p = atom;
+  hash = WideHashFunction(p) % WideAtomHashTableSize;
+  /* we'll start by holding a read lock in order to avoid contention */
+  READ_LOCK(WideHashChain[hash].AERWLock);
+  a = WideHashChain[hash].Entry;
+  /* search atom in chain */
+  na = SearchWideAtom(atom, a);
+  if (na != NIL) {
+    READ_UNLOCK(WideHashChain[hash].AERWLock);
+    return(na);
+  }
+  READ_UNLOCK(WideHashChain[hash].AERWLock);
+  /* we need a write lock */
+  WRITE_LOCK(WideHashChain[hash].AERWLock);
+  /* concurrent version of Yap, need to take care */
+#if defined(YAPOR) || defined(THREADS)
+  if (a != WideHashChain[hash].Entry) {
+    a = WideHashChain[hash].Entry;
+    na = SearchWideAtom(atom, a);
+    if (na != NIL) {
+      WRITE_UNLOCK(WideHashChain[hash].AERWLock);
+      return na;
+    }
+  }
+#endif  
+  /* add new atom to start of chain */
+  sz = wcslen(atom);
+  ae = (AtomEntry *) Yap_AllocAtomSpace(sizeof(AtomEntry) + sizeof(wchar_t)*(sz + 1)+sizeof(WideAtomEntry));
+  if (ae == NULL) {
+    WRITE_UNLOCK(WideHashChain[hash].AERWLock);
+    return NIL;
+  }
+  wae = (WideAtomEntry *)(ae->StrOfAE+sizeof(wchar_t)*(sz + 1));
+  na = AbsAtom(ae);
+  ae->PropsOfAE = AbsWideAtomProp(wae);
+  wae->NextOfPE = NIL;
+  wae->KindOfPE = WideAtomProperty;
+  wae->SizeOfAtom = sz;
+  if (ae->StrOfAE != (char *)atom)
+    wcscpy((wchar_t *)(ae->StrOfAE), atom);
+  NOfAtoms++;
+  ae->NextOfAE = a;
+  WideHashChain[hash].Entry = na;
+  INIT_RWLOCK(ae->ARWLock);
+  WRITE_UNLOCK(WideHashChain[hash].AERWLock);
+  if (NOfWideAtoms > 2*WideAtomHashTableSize) {
+    Yap_signal(YAP_CDOVF_SIGNAL);
+  }
+  return na;
+}
+
+Atom
+Yap_LookupMaybeWideAtom(wchar_t *atom)
+{				/* lookup atom in atom table            */
+  wchar_t *p = atom, c;
+  size_t len = 0;
+  char *ptr, *ptr0;
+  Atom at;
+
+  while ((c = *p++)) { 
+    if (c > 255) return LookupWideAtom(atom);
+    len++;
+  }
+  /* not really a wide atom */
+  ptr0 = ptr = Yap_AllocCodeSpace(len+1);
+  if (!ptr)
+    return NIL;
+  while ((*ptr++ = *p++));
+  at = LookupAtom(ptr0);
+  Yap_FreeCodeSpace(ptr0);
+  return at;
+}
+
 Atom
 Yap_LookupAtom(char *atom)
 {				/* lookup atom in atom table            */
   return LookupAtom(atom);
 }
 
+Atom
+Yap_LookupWideAtom(wchar_t *atom)
+{				/* lookup atom in atom table            */
+  return LookupWideAtom(atom);
+}
+
 Atom
 Yap_FullLookupAtom(char *atom)
 {				/* lookup atom in atom table            */
@@ -400,31 +507,6 @@ Yap_GetPredPropByAtomInThisModule(Atom at, Term cur_mod)
 }
 
 
-static inline Prop
-GetPredPropByFuncHavingLock(Functor f, Term cur_mod)
-/* get predicate entry for ap/arity; create it if neccessary.              */
-{
-  Prop p0;
-  FunctorEntry *fe = (FunctorEntry *)f;
-
-  p0 = fe->PropsOfFE;
-  while (p0) {
-    PredEntry *p = RepPredProp(p0);
-    if (/* p->KindOfPE != 0 || only props */
-	(p->ModuleOfPred == cur_mod || !(p->ModuleOfPred))) {
-#if THREADS
-      /* Thread Local Predicates */
-      if (p->PredFlags & ThreadLocalPredFlag) {
-	return AbsPredProp(Yap_GetThreadPred(p));
-      }
-#endif
-      return (p0);
-    }
-    p0 = p->NextOfPE;
-  }
-  return(NIL);
-}
-
 Prop
 Yap_GetPredPropByFunc(Functor f, Term cur_mod)
      /* get predicate entry for ap/arity;               */
@@ -432,35 +514,12 @@ Yap_GetPredPropByFunc(Functor f, Term cur_mod)
   Prop p0;
 
   READ_LOCK(f->FRWLock);
+
   p0 = GetPredPropByFuncHavingLock(f, cur_mod);
   READ_UNLOCK(f->FRWLock);
   return (p0);
 }
 
-static inline Prop
-GetPredPropByFuncHavingLockInThisModule(Functor f, Term cur_mod)
-/* get predicate entry for ap/arity; create it if neccessary.              */
-{
-  Prop p0;
-  FunctorEntry *fe = (FunctorEntry *)f;
-
-  p0 = fe->PropsOfFE;
-  while (p0) {
-    PredEntry *p = RepPredProp(p0);
-    if (p->ModuleOfPred == cur_mod) {
-#if THREADS
-      /* Thread Local Predicates */
-      if (p->PredFlags & ThreadLocalPredFlag) {
-	return AbsPredProp(Yap_GetThreadPred(p));
-      }
-#endif
-      return (p0);
-    }
-    p0 = p->NextOfPE;
-  }
-  return(NIL);
-}
-
 Prop
 Yap_GetPredPropByFuncInThisModule(Functor f, Term cur_mod)
      /* get predicate entry for ap/arity;               */
@@ -468,7 +527,7 @@ Yap_GetPredPropByFuncInThisModule(Functor f, Term cur_mod)
   Prop p0;
 
   READ_LOCK(f->FRWLock);
-  p0 = GetPredPropByFuncHavingLockInThisModule(f, cur_mod);
+  p0 = GetPredPropByFuncHavingLock(f, cur_mod);
   READ_UNLOCK(f->FRWLock);
   return (p0);
 }
@@ -517,17 +576,89 @@ Yap_GetExpPropHavingLock(AtomEntry *ae, unsigned int arity)
   p = RepExpProp(p0 = ae->PropsOfAE);
   while (p0 && (p->KindOfPE != ExpProperty || p->ArityOfEE != arity))
     p = RepExpProp(p0 = p->NextOfPE);
+
   return (p0);
 }
 
+static int
+ExpandPredHash(void)
+{
+  UInt new_size = PredHashTableSize+PredHashIncrement;
+  PredEntry **oldp = PredHash;
+  PredEntry **np = (PredEntry **) Yap_AllocAtomSpace(sizeof(PredEntry **)*new_size);
+  UInt i;
+
+  if (!np) {
+    return FALSE;
+  }
+  for (i = 0; i < new_size; i++) {
+    np[i] = NULL;
+  }
+  for (i = 0; i < PredHashTableSize; i++) {
+    PredEntry *p = PredHash[i];
+
+    while (p) {
+      Prop nextp = p->NextOfPE;
+      UInt hsh = PRED_HASH(p->FunctorOfPred, p->ModuleOfPred, new_size);
+      p->NextOfPE = AbsPredProp(np[hsh]);
+      np[hsh] = p;
+      p = RepPredProp(nextp);
+    }
+  }
+  PredHashTableSize = new_size;
+  PredHash = np;
+  Yap_FreeAtomSpace((ADDR)oldp);
+  return TRUE;
+}
+
 /* fe is supposed to be locked */
 Prop
 Yap_NewPredPropByFunctor(FunctorEntry *fe, Term cur_mod)
 {
-  Prop p0;
   PredEntry *p = (PredEntry *) Yap_AllocAtomSpace(sizeof(*p));
 
-  INIT_RWLOCK(p->PRWLock);
+  if (p == NULL) {
+    WRITE_UNLOCK(fe->FRWLock);
+    return NULL;
+  }
+  if (cur_mod == TermProlog)
+    p->ModuleOfPred = 0L;
+  else
+    p->ModuleOfPred = cur_mod;
+  if (fe->PropsOfFE) {
+    UInt hsh = PRED_HASH(fe, cur_mod, PredHashTableSize);
+
+    WRITE_LOCK(PredHashRWLock);
+    if (10*(PredsInHashTable+1) > 6*PredHashTableSize) {
+      if (!ExpandPredHash()) {
+	Yap_FreeCodeSpace((ADDR)p);
+	WRITE_UNLOCK(PredHashRWLock);
+	WRITE_UNLOCK(fe->FRWLock);
+	return NULL;
+      }
+      /* retry hashing */
+      hsh = PRED_HASH(fe, cur_mod, PredHashTableSize);
+    }
+    PredsInHashTable++;
+    if (p->ModuleOfPred == 0L) {
+      PredEntry *pe = RepPredProp(fe->PropsOfFE);
+
+      hsh = PRED_HASH(fe, pe->ModuleOfPred, PredHashTableSize);
+      /* should be the first one */
+      pe->NextOfPE = AbsPredProp(PredHash[hsh]);
+      PredHash[hsh] = pe;
+      fe->PropsOfFE = AbsPredProp(p);
+    } else {
+      p->NextOfPE = AbsPredProp(PredHash[hsh]);
+      PredHash[hsh] = p;
+    }
+    WRITE_UNLOCK(PredHashRWLock);
+    /* make sure that we have something here */
+    RepPredProp(fe->PropsOfFE)->NextOfPE = fe->PropsOfFE;
+  } else {
+    fe->PropsOfFE = AbsPredProp(p);
+    p->NextOfPE = NIL;
+  }
   INIT_LOCK(p->PELock);
   p->KindOfPE = PEProp;
   p->ArityOfPE = fe->ArityOfFE;
@@ -538,6 +669,8 @@ Yap_NewPredPropByFunctor(FunctorEntry *fe, Term cur_mod)
   p->OpcodeOfPred = UNDEF_OPCODE;
   p->CodeOfPred = p->cs.p_code.TrueCodeOfPred = (yamop *)(&(p->OpcodeOfPred)); 
   p->cs.p_code.ExpandCode = EXPAND_OP_CODE; 
+  p->TimeStampOfPred = 0L; 
+  p->LastCallOfPred = LUCALL_ASSERT; 
   if (cur_mod == TermProlog)
     p->ModuleOfPred = 0L;
   else
@@ -554,26 +687,24 @@ Yap_NewPredPropByFunctor(FunctorEntry *fe, Term cur_mod)
   p->beamTable = NULL;
 #endif  /* BEAM */
   /* careful that they don't cross MkFunctor */
-  p->NextOfPE = fe->PropsOfFE;
   if (PRED_GOAL_EXPANSION_FUNC) {
     if (fe->PropsOfFE &&
 	(RepPredProp(fe->PropsOfFE)->PredFlags & GoalExPredFlag)) {
       p->PredFlags |= GoalExPredFlag;
     }
   }
-  fe->PropsOfFE = p0 = AbsPredProp(p);
-  p->FunctorOfPred = (Functor)fe;
+  p->FunctorOfPred = fe;
   WRITE_UNLOCK(fe->FRWLock);
 #ifdef LOW_PROF
-    if (ProfilerOn &&
-	Yap_OffLineProfiler) {
-      Yap_inform_profiler_of_clause((yamop *)&(p->OpcodeOfPred), (yamop *)(&(p->OpcodeOfPred)+1), p, 1);
-      if (!(p->PredFlags & (CPredFlag|AsmPredFlag))) {
-	Yap_inform_profiler_of_clause((yamop *)&(p->cs.p_code.ExpandCode), (yamop *)(&(p->cs.p_code.ExpandCode)+1), p, 1);
-      }
+  if (ProfilerOn &&
+      Yap_OffLineProfiler) {
+    Yap_inform_profiler_of_clause((yamop *)&(p->OpcodeOfPred), (yamop *)(&(p->OpcodeOfPred)+1), p, 1);
+    if (!(p->PredFlags & (CPredFlag|AsmPredFlag))) {
+      Yap_inform_profiler_of_clause((yamop *)&(p->cs.p_code.ExpandCode), (yamop *)(&(p->cs.p_code.ExpandCode)+1), p, 1);
     }
+  }
 #endif /* LOW_PROF */
-  return p0;
+  return AbsPredProp(p);
 }
 
 #if THREADS
@@ -582,7 +713,6 @@ Yap_NewThreadPred(PredEntry *ap)
 {
   PredEntry *p = (PredEntry *) Yap_AllocAtomSpace(sizeof(*p));
 
-  INIT_RWLOCK(p->PRWLock);
   INIT_LOCK(p->PELock);
   p->KindOfPE = PEProp;
   p->ArityOfPE = ap->ArityOfPE;
@@ -595,6 +725,8 @@ Yap_NewThreadPred(PredEntry *ap)
   p->cs.p_code.ExpandCode = EXPAND_OP_CODE; 
   p->ModuleOfPred = ap->ModuleOfPred;
   p->NextPredOfModule = NULL;
+  p->TimeStampOfPred = 0L; 
+  p->LastCallOfPred = LUCALL_ASSERT; 
   INIT_LOCK(p->StatisticsForPred.lock);
   p->StatisticsForPred.NOfEntries = 0;
   p->StatisticsForPred.NOfHeadSuccesses = 0;
@@ -630,7 +762,6 @@ Yap_NewPredPropByAtom(AtomEntry *ae, Term cur_mod)
 
 /* Printf("entering %s:%s/0\n", RepAtom(AtomOfTerm(cur_mod))->StrOfAE, ae->StrOfAE); */
 
-  INIT_RWLOCK(p->PRWLock);
   INIT_LOCK(p->PELock);
   p->KindOfPE = PEProp;
   p->ArityOfPE = 0;
@@ -650,6 +781,8 @@ Yap_NewPredPropByAtom(AtomEntry *ae, Term cur_mod)
   p->StatisticsForPred.NOfEntries = 0;
   p->StatisticsForPred.NOfHeadSuccesses = 0;
   p->StatisticsForPred.NOfRetries = 0;
+  p->TimeStampOfPred = 0L; 
+  p->LastCallOfPred = LUCALL_ASSERT; 
 #ifdef TABLING
   p->TableOfPred = NULL;
 #endif /* TABLING */
@@ -692,21 +825,34 @@ Prop
 Yap_PredPropByFunctorNonThreadLocal(Functor f, Term cur_mod)
 /* get predicate entry for ap/arity; create it if neccessary.              */
 {
-  Prop p0;
-  FunctorEntry *fe = (FunctorEntry *)f;
+  PredEntry *p;
 
-  WRITE_LOCK(fe->FRWLock);
-  p0 = fe->PropsOfFE;
-  while (p0) {
-    PredEntry *p = RepPredProp(p0);
-    if (/* p->KindOfPE != 0 || only props */
-	(p->ModuleOfPred == cur_mod || !(p->ModuleOfPred))) {
-      WRITE_UNLOCK(fe->FRWLock);
-      return (p0);
-    }
-    p0 = p->NextOfPE;
+  WRITE_LOCK(f->FRWLock);
+  if (!(p = RepPredProp(f->PropsOfFE))) 
+    return Yap_NewPredPropByFunctor(f,cur_mod);
+
+  if ((p->ModuleOfPred == cur_mod || !(p->ModuleOfPred))) {
+    WRITE_UNLOCK(f->FRWLock);
+    return AbsPredProp(p);
   }
-  return Yap_NewPredPropByFunctor(fe,cur_mod);
+  if (p->NextOfPE) {
+    UInt hash = PRED_HASH(f,cur_mod,PredHashTableSize);
+    READ_LOCK(PredHashRWLock);
+    p = PredHash[hash];
+    
+    while (p) {
+      if (p->FunctorOfPred == f &&
+	  p->ModuleOfPred == cur_mod)
+	{
+	  READ_UNLOCK(PredHashRWLock);
+	  WRITE_UNLOCK(f->FRWLock);
+	  return AbsPredProp(p);
+	}
+      p = RepPredProp(p->NextOfPE);
+    }
+    READ_UNLOCK(PredHashRWLock);
+  }
+  return Yap_NewPredPropByFunctor(f,cur_mod);
 }
 
 Prop
@@ -862,6 +1008,30 @@ Yap_StringToList(char *s)
   return (t);
 }
 
+Term
+Yap_WStringToList(wchar_t *s)
+{
+  Term t;
+  wchar_t *cp = s + wcslen(s);
+
+  t = MkAtomTerm(AtomNil);
+  while (cp > s) {
+    t = MkPairTerm(MkIntegerTerm(*--cp), t);
+  }
+  return t;
+}
+
+Term
+Yap_StringToDiffList(char *s, Term t)
+{
+  register unsigned char *cp = (unsigned char *)s + strlen(s);
+
+  while (cp > (unsigned char *)s) {
+    t = MkPairTerm(MkIntTerm(*--cp), t);
+  }
+  return t;
+}
+
 Term
 Yap_StringToListOfAtoms(char *s)
 {
@@ -878,6 +1048,22 @@ Yap_StringToListOfAtoms(char *s)
   return (t);
 }
 
+Term
+Yap_WStringToListOfAtoms(wchar_t *s)
+{
+  register Term t;
+  wchar_t so[2];
+  wchar_t *cp = s + wcslen(s);
+
+  so[1] = '\0';
+  t = MkAtomTerm(AtomNil);
+  while (cp > s) {
+    so[0] = *--cp;
+    t = MkPairTerm(MkAtomTerm(LookupWideAtom(so)), t);
+  }
+  return t;
+}
+
 Term
 Yap_ArrayToList(register Term *tp, int nof)
 {
@@ -910,8 +1096,8 @@ Yap_GetName(char *s, UInt max, Term t)
     if (!IsNumTerm(Head))
       return (FALSE);
     i = IntOfTerm(Head);
-    if (i < 0 || i > 255)
-      return (FALSE);
+    if (i < 0 || i > MAX_ISO_LATIN1)
+      return FALSE;
     *s++ = i;
     t = TailOfTerm(t);
     if (--max == 0) {
@@ -979,12 +1165,16 @@ Yap_InitSlot(Term t)
   return((ASP+1)-LCL0);
 }
 
-void
+int
 Yap_RecoverSlots(int n)
 {
   Int old_slots = IntOfTerm(ASP[0]);
+  if (old_slots - n < 0) {
+    return FALSE;
+  }
   ASP += n;
   ASP[0] = MkIntTerm(old_slots-n);
+  return TRUE;
 }
 
 Term
@@ -1005,4 +1195,60 @@ Yap_PutInSlot(long slot, Term t)
   LCL0[slot] = t;
 }
 
+HoldEntry *
+Yap_InitAtomHold(void)
+{
+  HoldEntry *x = (HoldEntry *)Yap_AllocAtomSpace(sizeof(struct hold_entry));
+  x->KindOfPE = HoldProperty;
+  x->NextOfPE = NIL;
+  return x;
+}
 
+int
+Yap_AtomGetHold(Atom at)
+{
+  AtomEntry *ae = RepAtom(at);
+  PropEntry *pp, *opp = NULL;
+
+  WRITE_LOCK(ae->ARWLock);
+  pp = RepProp(ae->PropsOfAE);
+  while (!EndOfPAEntr(pp)) {
+    pp = RepProp(pp->NextOfPE);
+    opp = pp;
+  }
+  if (!pp) {
+    ae->PropsOfAE = AbsHoldProp(GlobalHoldEntry);
+  } else if (opp->KindOfPE != HoldProperty) {
+    opp->NextOfPE = AbsHoldProp(GlobalHoldEntry);
+  } else {
+    WRITE_UNLOCK(ae->ARWLock);
+    return FALSE;
+  }
+  WRITE_UNLOCK(ae->ARWLock);
+  return TRUE;
+}
+
+int
+Yap_AtomReleaseHold(Atom at)
+{
+  AtomEntry *ae = RepAtom(at);
+  PropEntry *pp, *opp = NULL;
+
+  WRITE_LOCK(ae->ARWLock);
+  pp = RepProp(ae->PropsOfAE);
+  while (!EndOfPAEntr(pp)) {
+    if (pp->KindOfPE == HoldProperty) {
+      if (!opp) {
+	ae->PropsOfAE = NIL;
+      } else {
+	opp->NextOfPE = NIL;
+      }
+      WRITE_UNLOCK(ae->ARWLock);
+      return TRUE;
+    }
+    pp = RepProp(pp->NextOfPE);
+    opp = pp;
+  }
+  WRITE_UNLOCK(ae->ARWLock);
+  return FALSE;
+}
