@@ -12,7 +12,7 @@
 * Last rev:								 *
 * mods:									 *
 * comments:	allocating space					 *
-* version:$Id: alloc.c,v 1.82 2006/03/24 17:15:18 vsc Exp $		 *
+* version:$Id: alloc.c,v 1.95 2008/05/10 23:24:11 vsc Exp $		 *
 *************************************************************************/
 #ifdef SCCS
 static char SccsId[] = "%W% %G%";
@@ -51,8 +51,6 @@ static char SccsId[] = "%W% %G%";
 #endif
 #endif
 
-#define K		((Int) 1024)
-
 
 /************************************************************************/
 /* Yap workspace management                                             */
@@ -61,6 +59,11 @@ static char SccsId[] = "%W% %G%";
 #define malloc Yap_dlmalloc
 #define free Yap_dlfree
 #define realloc Yap_dlrealloc
+#else
+void Yap_add_memory_hole(ADDR Start, ADDR End)
+{
+  Yap_HoleSize += Start-End;
+}
 #endif
 
 #if USE_SYSTEM_MALLOC||USE_DL_MALLOC
@@ -104,6 +107,7 @@ void
 Yap_FreeCodeSpace(char *p)
 {
   Yap_PrologMode |= MallocMode;
+
 #if INSTRUMENT_MALLOC
   if (frees % 1024*4 == 0) 
     minfo('F');
@@ -132,6 +136,8 @@ Yap_FreeAtomSpace(char *p)
   Yap_PrologMode &= ~MallocMode;
 }
 
+#endif
+
 /* If you need to dinamically allocate space from the heap, this is
  * the macro you should use */
 ADDR
@@ -165,9 +171,11 @@ Yap_ExpandPreAllocCodeSpace(UInt sz0, void *cip)
   UInt sz = ScratchPad.msz;
   if (sz0 < SCRATCH_INC_SIZE)
     sz0 = SCRATCH_INC_SIZE;
-  ScratchPad.msz =
-    ScratchPad.sz =
-    sz = sz + sz0;
+  if (sz0 < ScratchPad.sz)
+    sz = ScratchPad.sz+sz0;
+  else 
+    sz = sz0;
+  sz = AdjustLargePageSize(sz+sz/4);
 
 #if INSTRUMENT_MALLOC
   if (reallocs % 1024*4 == 0) 
@@ -187,13 +195,12 @@ Yap_ExpandPreAllocCodeSpace(UInt sz0, void *cip)
     Yap_PrologMode |= MallocMode;
   }
   Yap_PrologMode &= ~MallocMode;
+  ScratchPad.sz = ScratchPad.msz = sz;
   ScratchPad.ptr = ptr;
   AuxSp = (CELL *)(AuxTop = ptr+sz);
   return ptr;
 }
 
-#endif
-
 #if USE_SYSTEM_MALLOC
 
 struct various_codes *Yap_heap_regs;
@@ -201,13 +208,15 @@ struct various_codes *Yap_heap_regs;
 static void
 InitHeap(void)
 {
-  Yap_heap_regs = (struct various_codes *)malloc(sizeof(struct various_codes));
+  Yap_heap_regs = (struct various_codes *)calloc(1, sizeof(struct various_codes));
 }
 
 void
 Yap_InitHeap(void *heap_addr)
 {
   InitHeap();
+  Yap_HoleSize = 0;
+  HeapMax = 0;
 }
 
 static void
@@ -224,7 +233,10 @@ InitExStacks(int Trail, int Stack)
 				 * requested         */
   sa = Stack*K;			/* stack area size   */
 
-  Yap_GlobalBase = malloc(pm);
+#ifdef THREADS
+  if (worker_id)
+    Yap_GlobalBase = (ADDR)ThreadHandle[worker_id].stack_address;
+#endif
   Yap_TrailTop = Yap_GlobalBase + pm;
   Yap_LocalBase = Yap_GlobalBase + sa;
   Yap_TrailBase = Yap_LocalBase + sizeof(CELL);
@@ -253,14 +265,14 @@ Yap_InitExStacks(int Trail, int Stack)
   InitExStacks(Trail, Stack);
 }
 
-#if defined(YAPOR) || defined(THREADS)
+#if defined(THREADS)
 void
 Yap_KillStacks(int wid)
 {
-  ADDR gb = Yap_thread_gl[wid].global_base;
+  ADDR gb = ThreadHandle[wid].stack_address;
   if (gb) {
     free(gb);
-    Yap_thread_gl[wid].global_base = NULL;
+    ThreadHandle[wid].stack_address = NULL;
   }
 }
 #else
@@ -288,6 +300,9 @@ Yap_ExtendWorkSpace(Int s)
   nbp = realloc(basebp, s+s0);
   if (nbp == NULL) 
     return FALSE;
+#if defined(THREADS)
+  ThreadHandle[worker_id].stack_address = (char *)nbp;
+#endif
   Yap_GlobalBase = (char *)nbp;
   return TRUE;
 }
@@ -303,6 +318,16 @@ Yap_AllocHole(UInt actual_request, UInt total_size)
 {
 }
 
+#if HAVE_MALLINFO
+UInt
+Yap_givemallinfo(void)
+{
+  struct mallinfo mi = mallinfo();
+  return mi.uordblks;
+}
+#endif
+
+
 #else
 
 #if HAVE_SNPRINTF
@@ -632,15 +657,6 @@ Yap_AllocCodeSpace(unsigned int size)
   return out;
 }
 
-ADDR
-Yap_ExpandPreAllocCodeSpace(UInt sz, void *cip)
-{
-  if (!Yap_growheap((cip!=NULL), sz, cip)) {
-    Yap_Error(OUT_OF_HEAP_ERROR, TermNil, Yap_ErrorMessage);
-    return NULL;
-  }
-  return Addr(HeapTop) + sizeof(CELL);
-}
 #endif
 
 
@@ -675,6 +691,10 @@ ExtendWorkSpace(Int s, int fixed_allocation)
 
 
   Yap_PrologMode = ExtendStackMode;
+
+#if DEBUG_WIN32_ALLOC
+  fprintf(stderr,"trying: %p--%x %d\n",b, s, fixed_allocation);
+#endif
   if (fixed_allocation) {
     b = VirtualAlloc(b, s, MEM_RESERVE, PAGE_NOACCESS);
   } else {
@@ -685,7 +705,16 @@ ExtendWorkSpace(Int s, int fixed_allocation)
   }
   if (!b) {
     Yap_PrologMode = OldPrologMode;
-    /*  fprintf(stderr,"NOT OK1: %p--%p\n",b,brk);*/
+#if DEBUG_WIN32_ALLOC
+    {
+      char msg[256];
+      FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
+		    NULL, GetLastError(), 
+		    MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), msg, 256,
+		    NULL);
+      fprintf(stderr,"NOT OK1: %p--%p %s\n",b, brk, msg);
+    }
+#endif
     return FALSE;
   }
   b = VirtualAlloc(b, s, MEM_COMMIT, PAGE_READWRITE);
@@ -695,11 +724,15 @@ ExtendWorkSpace(Int s, int fixed_allocation)
 	      "VirtualAlloc could not commit %ld bytes",
 	      (long int)s);
     Yap_PrologMode = OldPrologMode;
-    /*  fprintf(stderr,"NOT OK2: %p--%p\n",b,brk);*/
+#if DEBUG_WIN32_ALLOC
+    fprintf(stderr,"NOT OK2: %p--%p\n",b,brk);
+#endif
     return FALSE;
   }
   brk = (LPVOID) ((Int) b + s);
-  /*  fprintf(stderr,"OK: %p--%p\n",b,brk);*/
+#if DEBUG_WIN32_ALLOC
+  fprintf(stderr,"OK: %p--%p\n",b,brk);
+#endif
   Yap_PrologMode = OldPrologMode;
   return TRUE;
 }
@@ -711,7 +744,7 @@ InitWorkSpace(Int s)
 
   GetSystemInfo(&si);
   Yap_page_size = si.dwPageSize;
-  s = ((s+ (ALLOC_SIZE-1))/ALLOC_SIZE)*ALLOC_SIZE;
+  s = ((s+ (YAP_ALLOC_SIZE-1))/YAP_ALLOC_SIZE)*YAP_ALLOC_SIZE;
   brk = (LPVOID)Yap_page_size;
   if (!ExtendWorkSpace(s,0))
     return FALSE;
@@ -1294,6 +1327,7 @@ InitHeap(void *heap_addr)
   /* reserve space for specially allocated functors and atoms so that
      their values can be known statically */
   HeapTop = Yap_HeapBase + AdjustSize(sizeof(all_heap_codes));
+  Yap_HoleSize = 0;
 #if USE_DL_MALLOC
   Yap_initdlmalloc();
 #else
@@ -1322,17 +1356,6 @@ Yap_InitMemory(int Trail, int Heap, int Stack)
 {
   Int pm, sa, ta;
 
-  if (Heap < MinHeapSpace)
-    Heap = MinHeapSpace;
-  /* sanity checking for data areas */
-  if (Trail < MinTrailSpace)
-    Trail = MinTrailSpace;
-  Trail = AdjustPageSize(Trail * K);
-  if (Stack < MinStackSpace)
-    Stack = MinStackSpace;
-  Stack = AdjustPageSize(Stack * K);
-  Heap = AdjustPageSize(Heap * K);
-
   pm = (Trail + Heap + Stack);	/* memory to be
 				 * requested         */
   sa = Stack;			/* stack area size   */
@@ -1411,10 +1434,7 @@ Yap_ExtendWorkSpaceThroughHole(UInt s)
     /* progress 1 MB */
     WorkSpaceTop += 512*1024;
     if (ExtendWorkSpace(s, MAP_FIXED)) {
-#if USE_DL_MALLOC 
-      Yap_hole_start = (ADDR)WorkSpaceTop0;
-      Yap_hole_end = (ADDR)WorkSpaceTop-s;
-#endif
+      Yap_add_memory_hole((ADDR)WorkSpaceTop0, (ADDR)WorkSpaceTop-s);
       Yap_ErrorMessage = NULL;
       return WorkSpaceTop-WorkSpaceTop0;
     }
@@ -1429,6 +1449,7 @@ Yap_ExtendWorkSpaceThroughHole(UInt s)
   WorkSpaceTop = WorkSpaceTop0;
 #endif
   if (ExtendWorkSpace(s, 0)) {
+    Yap_add_memory_hole((ADDR)WorkSpaceTop0, (ADDR)WorkSpaceTop-s);
     Yap_ErrorMessage = NULL;
     return WorkSpaceTop-WorkSpaceTop0;
   }
