@@ -44,7 +44,7 @@ float_to_int(Float v, union arith_ret *o)
   Int i = (Int)v;
   if (i-v == 0.0) {
     o->Int = i;
-    return(long_int_e);
+    return long_int_e;
   } else {
     mpz_init_set_d(o->big, v);
     return big_int_e;
@@ -195,7 +195,21 @@ p_uminus(Term t E_ARGS)
 
   switch (BlobOfFunctor(f)) {
   case long_int_e:
-    RINT(-IntegerOfTerm(t));
+    {
+ #ifdef USE_GMP
+      Int i = IntegerOfTerm(t);
+      
+      if (i == Int_MIN) {
+	MP_INT *new = TMP_BIG();
+
+	mpz_init_set_si(new, i);
+	mpz_neg(new, new);
+	RBIG(new);	
+      }
+      else
+#endif
+	RINT(-IntegerOfTerm(t));
+    }
   case double_e:
     RFLOAT(-FloatOfTerm(t));
 #ifdef USE_GMP
@@ -203,7 +217,7 @@ p_uminus(Term t E_ARGS)
     {
       MP_INT *new = TMP_BIG();
 
-      MPZ_SET(new, Yap_BigIntOfTerm(t));
+      mpz_init_set(new, Yap_BigIntOfTerm(t));
       mpz_neg(new, new);
       RBIG(new);
     }
@@ -219,8 +233,14 @@ p_uminus(Term t E_ARGS)
       RFLOAT(-v.dbl);
 #ifdef USE_GMP
     case big_int_e:
-      mpz_neg(v.big, v.big);
-      RBIG(v.big);
+      {
+	MP_INT *new = TMP_BIG();
+
+	mpz_init_set(new, v.big);
+	mpz_neg(new, new);
+	mpz_clear(v.big);
+	RBIG(new);
+      }
 #endif
     default:
       /* Error */
@@ -363,6 +383,7 @@ p_log(Term t E_ARGS)
       dbl = v.dbl;
       break;
 #ifdef USE_GMP
+    case big_int_e:
       dbl = mpz_get_d(v.big);
       mpz_clear(v.big);
       break;
@@ -1581,7 +1602,6 @@ p_integer(Term t E_ARGS)
 #ifdef USE_GMP
     MP_INT *new = TMP_BIG();
 
-    fprintf(stderr,"dbl is %g\n",dbl);
     mpz_init_set_d(new, dbl);
     RBIG(new);
 #else
