@@ -68,33 +68,36 @@ AddFailToQueue(void)
 }
 
 static int
-CopyAttVar(CELL *orig, CELL ***to_visit_ptr, CELL *res)
+CopyAttVar(CELL *orig, struct cp_frame **to_visit_ptr, CELL *res)
 {
   register attvar_record *attv = (attvar_record *)orig;
   register attvar_record *newv;
-  CELL **to_visit = *to_visit_ptr;
+  struct cp_frame *to_visit = *to_visit_ptr;
   CELL *vt;
 
   /* add a new attributed variable */
   newv = DelayTop();
-  if (H0 - (CELL *)newv < 1024)
+  if ((ADDR)newv - Yap_GlobalBase < 1024*sizeof(CELL))
     return FALSE;
+  newv--;
   RESET_VARIABLE(&(newv->Value));
   RESET_VARIABLE(&(newv->Done));
   vt = &(attv->Atts);
-  to_visit[0] = vt-1;
-  to_visit[1] = vt;
+  to_visit->start_cp = vt-1;
+  to_visit->end_cp = vt;
   if (IsVarTerm(attv->Atts)) {
     newv->Atts = (CELL)H;
-    to_visit[2] = H;
+    to_visit->to = H;
     H++;
   } else {
-    to_visit[2] = &(newv->Atts);
+    to_visit->to = &(newv->Atts);
   }
-  to_visit[3] = (CELL *)vt[-1];
-  *to_visit_ptr = to_visit+4;
+  to_visit->oldv = vt[-1];
+  /* you're coming from a variable */
+  to_visit->ground = FALSE;
+  *to_visit_ptr = to_visit+1;
   *res = (CELL)&(newv->Done);
-  SetDelayTop(newv+1);
+  SetDelayTop(newv);
   return TRUE;
 }
 
@@ -110,13 +113,13 @@ static attvar_record *
 BuildNewAttVar(void)
 {
   attvar_record *attv = DelayTop();
-  if (H0 - (CELL *)(attv+1) < 1024) {
-    return NULL;
-  }
+  if ((ADDR)attv - Yap_GlobalBase < 1024*sizeof(CELL))
+    return FALSE;
+  attv--;
   RESET_VARIABLE(&(attv->Done));
   RESET_VARIABLE(&(attv->Value));
   RESET_VARIABLE(&(attv->Atts));
-  SetDelayTop(attv+1);  
+  SetDelayTop(attv);  
   return attv;
 }
 
@@ -200,13 +203,6 @@ mark_attvar(CELL *orig)
   Yap_mark_external_reference(&(attv->Atts));
 }
 
-#if FROZEN_STACKS
-static Term
-CurrentTime(void) {
-  return(MkIntegerTerm(TR-(tr_fr_ptr)Yap_TrailBase));
-}
-#endif
-
 static Term
 BuildAttTerm(Functor mfun, UInt ar)
 {
@@ -214,6 +210,7 @@ BuildAttTerm(Functor mfun, UInt ar)
   UInt i;
 
   if (H+(1024+ar) > ASP) {
+    Yap_Error_Size=ar*sizeof(CELL);
     return 0L;
   }
   H[0] = (CELL)mfun;
@@ -414,26 +411,29 @@ AllAttVars(attvar_record *attv) {
   CELL *h0 = H;
   attvar_record *max = DelayTop();
 
-  while (attv != max) {
-
+  while (--attv >= max) {
     if (ASP - H < 1024) {
       H = h0;
+      Yap_Error_Size = (ASP-H)*sizeof(CELL);
       return 0L;
     }
     if (IsVarTerm(attv->Done) && IsUnboundVar(&attv->Done)) {
-      if (IsIntegerTerm(attv->Atts)) {
-	/* skip call residue(s) */
-	UInt n = IntegerOfTerm(attv->Atts)-1;
-	attv += n;
-      } else {
-	if (H != h0) {
-	  H[-1] = AbsPair(H);
+      if (IsVarTerm(attv->Atts)) {
+	if (VarOfTerm(attv->Atts) < (CELL *)attv) {
+	  /* skip call residue(s) */
+	  attv = (attvar_record *)(attv->Atts);
+	  continue;
+	} else if (IsUnboundVar(&attv->Atts)) {
+	  /* ignore arena */
+	  continue;
 	}
-	H[0] = (CELL)attv;
-	H += 2;
       }
+      if (H != h0) {
+	H[-1] = AbsPair(H);
+      }
+      H[0] = (CELL)attv;
+      H += 2;
     }
-    attv++;
   }
   if (H != h0) {
     H[-1] = TermNil;
@@ -471,7 +471,7 @@ p_put_att(void) {
     mfun= Yap_MkFunctor(modname,ar);
     if (IsVarTerm(tatts = SearchAttsForModule(attv->Atts,mfun))) {
       while (!(tatts = BuildAttTerm(mfun,ar))) {
-	if (!Yap_gc(5, ENV, P)) {
+	if (!Yap_gcl(Yap_Error_Size, 5, ENV, P)) {
 	  Yap_Error(OUT_OF_STACK_ERROR, TermNil, Yap_ErrorMessage);
 	  return FALSE;
 	}    
@@ -549,7 +549,7 @@ p_rm_att(void) {
     mfun= Yap_MkFunctor(modname,ar);
     if (IsVarTerm(tatts = SearchAttsForModule(attv->Atts,mfun))) {
       while (!(tatts = BuildAttTerm(mfun,ar))) {
-	if (!Yap_gc(4, ENV, P)) {
+	if (!Yap_gcl(Yap_Error_Size, 4, ENV, P)) {
 	  Yap_Error(OUT_OF_STACK_ERROR, TermNil, Yap_ErrorMessage);
 	  return FALSE;
 	}    
@@ -880,9 +880,9 @@ p_all_attvars(void)
     Term out;
     attvar_record *base;
 
-    base = (attvar_record *)Yap_GlobalBase+IntegerOfTerm(Yap_ReadTimedVar(AttsMutableList));
+    base = (attvar_record *)Yap_ReadTimedVar(AttsMutableList);
     if (!(out = AllAttVars(base))) {
-      if (!Yap_gc(1, ENV, P)) {
+      if (!Yap_gcl(Yap_Error_Size, 1, ENV, P)) {
 	Yap_Error(OUT_OF_STACK_ERROR, TermNil, Yap_ErrorMessage);
 	return FALSE;
       }    
