@@ -25,6 +25,7 @@ static char     SccsId[] = "%W% %G%";
 #if HAVE_STRING_H
 #include <string.h>
 #endif
+#include <wchar.h>
 
 STATIC_PROTO(Int compare, (Term, Term));
 STATIC_PROTO(Int p_compare, (void));
@@ -43,6 +44,36 @@ STATIC_PROTO(Int a_gen_ge, (Term,Term));
 
 #define rfloat(X)	( X > 0.0 ? 1 : ( X == 0.0 ? 0 : -1))
 
+static int
+cmp_atoms(Atom a1, Atom a2)
+{
+  if (IsWideAtom(a1)) {
+    if (IsWideAtom(a2)) {
+      return wcscmp((wchar_t *)RepAtom(a1)->StrOfAE,(wchar_t *)RepAtom(a2)->StrOfAE);
+    } else {
+      /* The standard does not seem to have nothing on this */
+      unsigned char *s1 = (unsigned char *)RepAtom(a1)->StrOfAE;
+      wchar_t *s2 = (wchar_t *)RepAtom(a2)->StrOfAE;
+
+      while (*s1 == *s2) {
+	if (!*s1) return 0;
+      }
+      return *s1-*s2;
+    }
+  } else if (IsWideAtom(a2)) {
+    /* The standard does not seem to have nothing on this */
+    wchar_t *s1 = (wchar_t *)RepAtom(a1)->StrOfAE;
+    unsigned char *s2 = (unsigned char *)RepAtom(a2)->StrOfAE;
+
+    while (*s1 == *s2) {
+      if (!*s1) return 0;
+    }
+    return *s1-*s2;
+  } else {
+    return strcmp(RepAtom(a1)->StrOfAE,RepAtom(a2)->StrOfAE);
+  }
+}
+
 static int compare_complex(register CELL *pt0, register CELL *pt0_end, register
 		   CELL *pt1)
 {
@@ -73,10 +104,7 @@ static int compare_complex(register CELL *pt0, register CELL *pt0_end, register
       if (d0 == d1) continue;
       else if (IsAtomTerm(d0)) {
 	if (IsAtomTerm(d1))
-	    out = strcmp(
-			 RepAtom(AtomOfTerm(d0))->StrOfAE,
-			 RepAtom(AtomOfTerm(d1))->StrOfAE
-			 );
+	  out = cmp_atoms(AtomOfTerm(d0), AtomOfTerm(d1));
 	else if (IsPrimitiveTerm(d1))
 	  out = 1;
 	else out = -1;
@@ -92,7 +120,12 @@ static int compare_complex(register CELL *pt0, register CELL *pt0_end, register
 	  out = IntOfTerm(d0) - LongIntOfTerm(d1);
 #ifdef USE_GMP
 	} else if (IsBigIntTerm(d1)) {
-	  out = -mpz_cmp_si(Yap_BigIntOfTerm(d1), IntOfTerm(d0));
+	  MP_INT *b1 = Yap_BigIntOfTerm(d1);
+	  if (!mpz_size(b1)) {
+	    out = -1; 
+	  } else {
+	    out = -mpz_cmp_si(b1, IntOfTerm(d0));
+	  }
 #endif
 	} else if (IsRefTerm(d1))
 		out = 1 ;
@@ -118,7 +151,12 @@ static int compare_complex(register CELL *pt0, register CELL *pt0_end, register
 	  out = LongIntOfTerm(d0) - LongIntOfTerm(d1);
 #ifdef USE_GMP
 	} else if (IsBigIntTerm(d1)) {
-	  out = -mpz_cmp_si(Yap_BigIntOfTerm(d1), LongIntOfTerm(d0));
+	  MP_INT *b1 = Yap_BigIntOfTerm(d1);
+	  if (!mpz_size(b1)) {
+	    out = -1; 
+	  } else {
+	    out = -mpz_cmp_si(b1, LongIntOfTerm(d0));
+	  }
 #endif
 	} else if (IsRefTerm(d1)) {
 	  out = 1 ;
@@ -130,15 +168,30 @@ static int compare_complex(register CELL *pt0, register CELL *pt0_end, register
       }
 #ifdef USE_GMP
       else if (IsBigIntTerm(d0)) {
-	if (IsIntTerm(d1))
-	  out = mpz_cmp_si(Yap_BigIntOfTerm(d0), IntOfTerm(d1));
+	MP_INT *b0 = Yap_BigIntOfTerm(d0);
+
+	if (!mpz_size(b0)) {
+	  if (IsBigIntTerm(d1)) {
+	    MP_INT *b1 = Yap_BigIntOfTerm(d1);
+	    out = b0-b1;
+	  } else {
+	    out = 1;
+	  }
+	} else if (IsIntTerm(d1))
+	  out = mpz_cmp_si(b0, IntOfTerm(d1));
 	else if (IsFloatTerm(d1)) {
 	  out = 1;
 	} else if (IsLongIntTerm(d1))
-	  out = mpz_cmp_si(Yap_BigIntOfTerm(d0), LongIntOfTerm(d1));
-	else if (IsBigIntTerm(d1))
-	  out = mpz_cmp(Yap_BigIntOfTerm(d0), Yap_BigIntOfTerm(d1));
-	else if (IsRefTerm(d1))
+	  out = mpz_cmp_si(b0, LongIntOfTerm(d1));
+	else if (IsBigIntTerm(d1)) {
+	  MP_INT *b1 = Yap_BigIntOfTerm(d1);
+
+	  if (!mpz_size(b1)) {
+	    out = -1; 
+	  } else {
+	    out = mpz_cmp(b0, b1);
+	  }
+	} else if (IsRefTerm(d1))
 	  out = 1 ;
 	else out = -1;
 	if (out != 0)
@@ -207,8 +260,7 @@ static int compare_complex(register CELL *pt0, register CELL *pt0_end, register
 	  /* compare functors */
 	  if (f != (Functor)*ap3) {
 	    if (!(out = ArityOfFunctor(f)-ArityOfFunctor(f2)))
-	       out = strcmp(RepAtom(NameOfFunctor(f))->StrOfAE,
-			    RepAtom(NameOfFunctor(f2))->StrOfAE);
+	       out = cmp_atoms(NameOfFunctor(f), NameOfFunctor(f2));
 	    goto done;
 	  }
 #ifdef RATIONAL_TREES
@@ -285,10 +337,7 @@ compare(Term t1, Term t2) /* compare terms t1 and t2	 */
   if (IsAtomOrIntTerm(t1)) {
     if (IsAtomTerm(t1)) {
       if (IsAtomTerm(t2))
-	return strcmp(
-		      RepAtom(AtomOfTerm(t1))->StrOfAE,
-		      RepAtom(AtomOfTerm(t2))->StrOfAE
-		      );
+	return cmp_atoms(AtomOfTerm(t1),AtomOfTerm(t2));
       if (IsPrimitiveTerm(t2))
 	return 1;
       return -1;
@@ -303,7 +352,12 @@ compare(Term t1, Term t2) /* compare terms t1 and t2	 */
       }
 #ifdef USE_GMP
       if (IsBigIntTerm(t2)) {
-	return -mpz_cmp_si(Yap_BigIntOfTerm(t2),IntOfTerm(t1));
+	MP_INT *b1 = Yap_BigIntOfTerm(t2);
+	if (!mpz_size(b1)) {
+	  return -1; 
+	} else {
+	  return -mpz_cmp_si(b1,IntOfTerm(t1));
+	}
       }
 #endif
       if (IsRefTerm(t2))
@@ -334,53 +388,87 @@ compare(Term t1, Term t2) /* compare terms t1 and t2	 */
 
     if (IsExtensionFunctor(fun1)) {
       /* float, long, big, dbref */
-      if (IsFloatTerm(t1)) {
-	if (IsFloatTerm(t2))
-	  return(rfloat(FloatOfTerm(t1) - FloatOfTerm(t2)));
-	if (IsRefTerm(t2))
-	  return 1;
-	return -1;
-      }
-      if (IsLongIntTerm(t1)) {
-	if (IsIntTerm(t2))
-	  return LongIntOfTerm(t1) - IntOfTerm(t2);
-	if (IsFloatTerm(t2)) {
-	  return 1;
+      switch ((CELL)fun1) {
+      case double_e:
+	{
+	  if (IsFloatTerm(t2))
+	    return(rfloat(FloatOfTerm(t1) - FloatOfTerm(t2)));
+	  if (IsRefTerm(t2))
+	    return 1;
+	  return -1;
 	}
-	if (IsLongIntTerm(t2))
-	  return LongIntOfTerm(t1) - LongIntOfTerm(t2);
+      case long_int_e:
+	{
+	  if (IsIntTerm(t2))
+	    return LongIntOfTerm(t1) - IntOfTerm(t2);
+	  if (IsFloatTerm(t2)) {
+	    return 1;
+	  }
+	  if (IsLongIntTerm(t2))
+	    return LongIntOfTerm(t1) - LongIntOfTerm(t2);
 #ifdef USE_GMP
-	if (IsBigIntTerm(t2))
-	  return -mpz_cmp_si(Yap_BigIntOfTerm(t2), LongIntOfTerm(t1));
+	  if (IsBigIntTerm(t2)) {
+	    MP_INT *b1 = Yap_BigIntOfTerm(t2);
+	    if (!mpz_size(b1)) {
+	      return -1; 
+	    } else {
+	      return -mpz_cmp_si(b1, LongIntOfTerm(t1));
+	    }
+	  }
 #endif
-	if (IsRefTerm(t2))
-	  return 1;
-	return -1;
-      }
+	  if (IsRefTerm(t2))
+	    return 1;
+	  return -1;
+	}
 #ifdef USE_GMP
-      if (IsBigIntTerm(t1)) {
-	if (IsIntTerm(t2))
-	  return mpz_cmp_si(Yap_BigIntOfTerm(t1), IntOfTerm(t2));
-	if (IsFloatTerm(t2)) {
-	  return 1;
+      case big_int_e:
+	{
+	  MP_INT *b0 = Yap_BigIntOfTerm(t1);
+
+	  if (!mpz_size(b0)) {
+	    if (IsBigIntTerm(t2)) {
+	      MP_INT *b1 = Yap_BigIntOfTerm(t2);
+	      return b0-b1;
+	    } else {
+	      return 1;
+	    }
+	  } else if (IsIntTerm(t2))
+	    return mpz_cmp_si(Yap_BigIntOfTerm(t1), IntOfTerm(t2));
+	  if (IsFloatTerm(t2)) {
+	    return 1;
+	  }
+	  if (IsLongIntTerm(t2))
+	    return mpz_cmp_si(Yap_BigIntOfTerm(t1), LongIntOfTerm(t2));
+	  if (IsBigIntTerm(t2)) {
+	    MP_INT *b1 = Yap_BigIntOfTerm(t2);
+
+	    if (!mpz_size(b1)) {
+	      return -1; 
+	    } else {
+	      return mpz_cmp(b0, b1);
+	    }
+	  }
+	  if (IsRefTerm(t2))
+	    return 1;
+	  return -1;
 	}
-	if (IsLongIntTerm(t2))
-	  return mpz_cmp_si(Yap_BigIntOfTerm(t1), LongIntOfTerm(t2));
-	if (IsBigIntTerm(t2))
-	  return mpz_cmp(Yap_BigIntOfTerm(t1), Yap_BigIntOfTerm(t2));
-	if (IsRefTerm(t2))
-	  return 1;
-	return -1;
-      }
 #endif
-      if (IsRefTerm(t1)) {
+      case db_ref_e:
 	if (IsRefTerm(t2))
 	  return Unsigned(RefOfTerm(t2)) -
-		  Unsigned(RefOfTerm(t1));
+	    Unsigned(RefOfTerm(t1));
 	return -1;
       }
     }
     if (!IsApplTerm(t2)) {
+      if (IsPairTerm(t2)) {
+	Int out;
+	Functor f = FunctorOfTerm(t1);
+
+	if (!(out = ArityOfFunctor(f))-2)
+	  out = strcmp(RepAtom(NameOfFunctor(f))->StrOfAE,".");
+	return out;
+      }
       return 1;
     } else {
       Functor fun2 = FunctorOfTerm(t2);
@@ -392,8 +480,7 @@ compare(Term t1, Term t2) /* compare terms t1 and t2	 */
       r = ArityOfFunctor(fun1) - ArityOfFunctor(fun2);
       if (r)
 	return r;
-      r = strcmp(RepAtom(NameOfFunctor(fun1))->StrOfAE,
-		 RepAtom(NameOfFunctor(fun2))->StrOfAE);
+      r = cmp_atoms(NameOfFunctor(fun1), NameOfFunctor(fun2));
       if (r)
 	return r;
       else
@@ -404,7 +491,7 @@ compare(Term t1, Term t2) /* compare terms t1 and t2	 */
   }
 }
 
-int Yap_compare_terms(register CELL d0, register CELL d1)
+int Yap_compare_terms(CELL d0, CELL d1)
 {
   return (compare(Deref(d0),Deref(d1)));
 }
