@@ -11,8 +11,58 @@
 * File:		compiler.c						 *
 * comments:	Clause compiler						 *
 *									 *
-* Last rev:     $Date: 2006/04/13 02:04:24 $,$Author: vsc $						 *
+* Last rev:     $Date: 2008/03/13 14:37:58 $,$Author: vsc $						 *
 * $Log: compiler.c,v $
+* Revision 1.88  2008/03/13 14:37:58  vsc
+* update chr
+*
+* Revision 1.87  2007/12/18 17:46:58  vsc
+* purge_clauses does not need to do anything if there are no clauses
+* fix gprof bugs.
+*
+* Revision 1.86  2007/11/26 23:43:08  vsc
+* fixes to support threads and assert correctly, even if inefficiently.
+*
+* Revision 1.85  2007/11/06 17:02:11  vsc
+* compile ground terms away.
+*
+* Revision 1.84  2007/03/27 13:48:51  vsc
+* fix number of overflows (comments by Bart Demoen).
+*
+* Revision 1.83  2007/03/26 15:18:43  vsc
+* debugging and clause/3 over tabled predicates would kill YAP.
+*
+* Revision 1.82  2006/11/06 18:35:03  vsc
+* 1estranha
+*
+* Revision 1.81  2006/10/11 15:08:03  vsc
+* fix bb entries
+* comment development code for timestamp overflow.
+*
+* Revision 1.80  2006/09/20 20:03:51  vsc
+* improve indexing on floats
+* fix sending large lists to DB
+*
+* Revision 1.79  2006/08/01 13:14:17  vsc
+* fix compilation of |
+*
+* Revision 1.78  2006/07/27 19:04:56  vsc
+* fix nasty overflows in and add some very preliminary support for very large
+* clauses with lots
+* of disjuncts (eg, query packs).
+*
+* Revision 1.77  2006/05/19 14:31:31  vsc
+* get rid of IntArrays and FloatArray code.
+* include holes when calculating memory usage.
+*
+* Revision 1.76  2006/05/19 13:48:11  vsc
+* help to make Yap work with dynamic libs
+*
+* Revision 1.75  2006/05/16 18:37:30  vsc
+* WIN32 fixes
+* compiler bug fixes
+* extend interface
+*
 * Revision 1.74  2006/04/13 02:04:24  vsc
 * fix debugging typo
 *
@@ -179,7 +229,7 @@ STATIC_PROTO(void c_arg, (Int, Term, unsigned int, unsigned int, compiler_struct
 STATIC_PROTO(void c_args, (Term, unsigned int, compiler_struct *));
 STATIC_PROTO(void c_eq, (Term, Term, compiler_struct *));
 STATIC_PROTO(void c_test, (Int, Term, compiler_struct *));
-STATIC_PROTO(void c_bifun, (Int, Term, Term, Term, int, compiler_struct *));
+STATIC_PROTO(void c_bifun, (Int, Term, Term, Term, Term, int, compiler_struct *));
 STATIC_PROTO(void c_goal, (Term, int, compiler_struct *));
 STATIC_PROTO(void c_body, (Term, int, compiler_struct *));
 STATIC_PROTO(void c_head, (Term, compiler_struct *));
@@ -447,8 +497,9 @@ optimize_ce(Term t, unsigned int arity, unsigned int level, compiler_struct *cgl
     return (p->VarOfCE);
   }
   /* first occurrence */
-  if (cglobs->onbranch || level > 1)
+  if (cglobs->onbranch || level > 1) {
     return t;
+  }
   ++(cglobs->n_common_exps);
   p = (CExpEntry *) Yap_AllocCMem(sizeof(CExpEntry), &cglobs->cint);
 
@@ -457,7 +508,7 @@ optimize_ce(Term t, unsigned int arity, unsigned int level, compiler_struct *cgl
   if (H >= (CELL *)cglobs->cint.freep0) {
     /* oops, too many new variables */
     save_machine_regs();
-    longjmp(cglobs->cint.CompilerBotch,4);
+    longjmp(cglobs->cint.CompilerBotch,OUT_OF_TEMPS_BOTCH);
   }
   p->NextCE = cglobs->common_exps;
   cglobs->common_exps = p;
@@ -492,7 +543,7 @@ compile_sf_term(Term t, int argno, int level)
 	Yap_Error_Term = TermNil;
 	Yap_ErrorMessage = "illegal argument of soft functor";
 	save_machine_regs();
-	longjmp(cglobs->cint.CompilerBotch, 2);
+	longjmp(cglobs->cint.CompilerBotch, OUT_OF_HEAP_BOTCH);
       }
       else
 	c_var(t, -argno, arity, level, cglobs);
@@ -519,7 +570,7 @@ c_args(Term app, unsigned int level, compiler_struct *cglobs)
       Yap_Error_Term = TermNil;
       Yap_ErrorMessage = "exceed maximum arity of compiled goal";
       save_machine_regs();
-      longjmp(cglobs->cint.CompilerBotch, 2);
+      longjmp(cglobs->cint.CompilerBotch, OUT_OF_HEAP_BOTCH);
     }
     if (Arity > cglobs->max_args)
       cglobs->max_args = Arity;
@@ -528,6 +579,52 @@ c_args(Term app, unsigned int level, compiler_struct *cglobs)
     c_arg(i, ArgOfTerm(i, app), Arity, level, cglobs);
 }
 
+static int
+try_store_as_dbterm(Term t, Int argno, unsigned int arity, int level, compiler_struct *cglobs)
+{
+  DBTerm *dbt;
+  int g;
+  CELL *h0 = H;
+
+  while ((g=Yap_SizeGroundTerm(t,TRUE)) < 0) {
+    /* oops, too deep a term */
+    save_machine_regs();
+    Yap_Error_Size = 0;
+    longjmp(cglobs->cint.CompilerBotch, OUT_OF_AUX_BOTCH);
+  }
+  if (g < 16)
+    return FALSE;
+  /* store ground term away */
+  H = CellPtr(cglobs->cint.freep);
+  if ((dbt = Yap_StoreTermInDB(t, -1)) == NULL) {
+    H = h0;
+    switch(Yap_Error_TYPE) {
+    case OUT_OF_STACK_ERROR:
+      Yap_Error_TYPE = YAP_NO_ERROR;
+      longjmp(cglobs->cint.CompilerBotch,OUT_OF_STACK_BOTCH);
+    case OUT_OF_TRAIL_ERROR:
+      Yap_Error_TYPE = YAP_NO_ERROR;
+      longjmp(cglobs->cint.CompilerBotch,OUT_OF_TRAIL_BOTCH);
+    case OUT_OF_HEAP_ERROR:
+      Yap_Error_TYPE = YAP_NO_ERROR;
+      longjmp(cglobs->cint.CompilerBotch,OUT_OF_HEAP_BOTCH);
+    case OUT_OF_AUXSPACE_ERROR:
+      Yap_Error_TYPE = YAP_NO_ERROR;
+      longjmp(cglobs->cint.CompilerBotch,OUT_OF_AUX_BOTCH);
+    default:
+      longjmp(cglobs->cint.CompilerBotch,COMPILER_ERR_BOTCH);
+    }
+  }
+  H = h0;
+  if (level == 0)
+    Yap_emit((cglobs->onhead ? get_dbterm_op : put_dbterm_op), dbt->Entry, argno, &cglobs->cint);
+  else
+    Yap_emit((cglobs->onhead ? (argno == (Int)arity ? unify_last_dbterm_op
+				: unify_dbterm_op) :
+	      write_dbterm_op), dbt->Entry, Zero, &cglobs->cint);
+  return TRUE;
+}
+
 static void
 c_arg(Int argno, Term t, unsigned int arity, unsigned int level, compiler_struct *cglobs)
 {
@@ -541,57 +638,43 @@ c_arg(Int argno, Term t, unsigned int arity, unsigned int level, compiler_struct
       Yap_emit((cglobs->onhead ? (argno == (Int)arity ? unify_last_atom_op
 		      : unify_atom_op) :
 	    write_atom_op), (CELL) t, Zero, &cglobs->cint);
-  }
-  else if (IsIntegerTerm(t) || IsFloatTerm(t) || IsBigIntTerm(t)) {
+  } else  if (IsIntegerTerm(t) || IsFloatTerm(t) || IsBigIntTerm(t)) {
     if (!IsIntTerm(t)) {
-      /* we are taking a blob, that is a binary that is supposed to be
-	 guarded in the clause itself. Possible examples include
-	 floats, long ints, bignums, bitmaps.... */
-      CELL l1 = ++cglobs->labelno;
-      CELL *src = RepAppl(t);
-      PInstr *ocpc = cglobs->cint.cpc, *OCodeStart = cglobs->cint.CodeStart;
-
-      /* use a special list to store the blobs */
-      cglobs->cint.cpc = cglobs->cint.icpc;
-      /*      if (IsFloatTerm(t)) {
-	Yap_emit(align_float_op, Zero, Zero, &cglobs->cint);
-	}*/
-      Yap_emit(label_op, l1, Zero, &cglobs->cint);
       if (IsFloatTerm(t)) {
-	/* let us do floats first */
-	CELL *dest = 
-	  Yap_emit_extra_size(blob_op,
-			  (CELL)(SIZEOF_DOUBLE/SIZEOF_LONG_INT+1),
-			  (1+SIZEOF_DOUBLE/SIZEOF_LONG_INT)*CellSize, &cglobs->cint);
-	/* copy the float bit by bit */
-	dest[0] = src[0];
-	dest[1] = src[1];
-#if SIZEOF_DOUBLE == 2*SIZEOF_LONG_INT
-	dest[2] = src[2];
-#endif
-	/* note that we don't need to copy size info, unless we wanted
-	   to garbage collect clauses ;-) */
-	cglobs->cint.icpc = cglobs->cint.cpc;
-	if (cglobs->cint.BlobsStart == NULL)
-	  cglobs->cint.BlobsStart = cglobs->cint.CodeStart;
-	cglobs->cint.cpc = ocpc;
-	cglobs->cint.CodeStart = OCodeStart;
-	/* The argument to pass to the structure is now the label for
-	   where we are storing the blob */
 	if (level == 0)
-	  Yap_emit((cglobs->onhead ? get_float_op : put_float_op), l1, argno, &cglobs->cint);
+	  Yap_emit((cglobs->onhead ? get_float_op : put_float_op), t, argno, &cglobs->cint);
 	else
 	  Yap_emit((cglobs->onhead ? (argno == (Int)arity ? unify_last_float_op
 			  : unify_float_op) :
-		write_float_op), l1, Zero, &cglobs->cint);
-#if USE_GMP
-      } else if (IsBigIntTerm(t)) {
-	/* next, let us do bigints */
+		write_float_op), t, Zero, &cglobs->cint);
+      } else if (IsLongIntTerm(t)) {
+	if (level == 0)
+	  Yap_emit((cglobs->onhead ? get_longint_op : put_longint_op), t, argno, &cglobs->cint);
+	else
+	  Yap_emit((cglobs->onhead ? (argno == (Int)arity ? unify_last_longint_op
+			  : unify_longint_op) :
+		write_longint_op), t, Zero, &cglobs->cint);
+      } else {
+	/* we are taking a blob, that is a binary that is supposed to be
+	 guarded in the clause itself. Possible examples include
+	 floats, long ints, bignums, bitmaps.... */
+	CELL l1 = ++cglobs->labelno;
+	CELL *src = RepAppl(t);
+	PInstr *ocpc = cglobs->cint.cpc, *OCodeStart = cglobs->cint.CodeStart;
 	Int sz = sizeof(CELL)+
 	  sizeof(MP_INT)+
 	   ((((MP_INT *)(RepAppl(t)+1))->_mp_alloc)*sizeof(mp_limb_t));
-	CELL *dest = 
+	CELL *dest;
+
+	/* use a special list to store the blobs */
+	cglobs->cint.cpc = cglobs->cint.icpc;
+	/*      if (IsFloatTerm(t)) {
+		Yap_emit(align_float_op, Zero, Zero, &cglobs->cint);
+		}*/
+	Yap_emit(label_op, l1, Zero, &cglobs->cint);
+	dest = 
 	  Yap_emit_extra_size(blob_op, sz/CellSize, sz, &cglobs->cint);
+
 	/* copy the bignum */
 	memcpy(dest, src, sz);
 	/* note that we don't need to copy size info, unless we wanted
@@ -609,27 +692,6 @@ c_arg(Int argno, Term t, unsigned int arity, unsigned int level, compiler_struct
 	  Yap_emit((cglobs->onhead ? (argno == (Int)arity ? unify_last_bigint_op
 			  : unify_bigint_op) :
 		write_bigint_op), l1, Zero, &cglobs->cint);
-#endif
-      } else {
-	/* for now, it's just a long int */
-	CELL *dest = 
-	  Yap_emit_extra_size(blob_op,
-			  2,
-			  2*CellSize, &cglobs->cint);
-	/* copy the long int in one fell swoop */
-	dest[0] = src[0];
-	dest[1] = src[1];
-	cglobs->cint.icpc = cglobs->cint.cpc;
-	if (cglobs->cint.BlobsStart == NULL)
-	  cglobs->cint.BlobsStart = cglobs->cint.CodeStart;
-	cglobs->cint.cpc = ocpc;
-	cglobs->cint.CodeStart = OCodeStart;
-	if (level == 0)
-	  Yap_emit((cglobs->onhead ? get_longint_op : put_longint_op), l1, argno, &cglobs->cint);
-	else
-	  Yap_emit((cglobs->onhead ? (argno == (Int)arity ? unify_last_longint_op
-			  : unify_longint_op) :
-		write_longint_op), l1, Zero, &cglobs->cint);
       }
       /* That's it folks! */
       return;
@@ -642,6 +704,10 @@ c_arg(Int argno, Term t, unsigned int arity, unsigned int level, compiler_struct
 	    write_num_op), (CELL) t, Zero, &cglobs->cint);
   } else if (IsPairTerm(t)) {
     if (optimizer_on && level < 6) {
+      if (!(cglobs->cint.CurrentPred->PredFlags & (DynamicPredFlag|LogUpdatePredFlag))) {
+	if (try_store_as_dbterm(t, argno, arity, level, cglobs))
+	  return;
+      }      
       t = optimize_ce(t, arity, level, cglobs);
       if (IsVarTerm(t)) {
 	c_var(t, argno, arity, level, cglobs);
@@ -657,7 +723,7 @@ c_arg(Int argno, Term t, unsigned int arity, unsigned int level, compiler_struct
     ++level;
     c_arg(1, HeadOfTerm(t), 2, level, cglobs);
     if (argno == (Int)arity) {
-      /* optimise for tail recursion */
+     /* optimise for tail recursion */
       t = TailOfTerm(t);
       goto restart;
     }
@@ -667,12 +733,12 @@ c_arg(Int argno, Term t, unsigned int arity, unsigned int level, compiler_struct
       pop_code(level, cglobs);
     }
   } else if (IsRefTerm(t)) {
-    READ_LOCK(cglobs->cint.CurrentPred->PRWLock);
+    LOCK(cglobs->cint.CurrentPred->PELock);
     if (!(cglobs->cint.CurrentPred->PredFlags & (DynamicPredFlag|LogUpdatePredFlag))) {
-      READ_UNLOCK(cglobs->cint.CurrentPred->PRWLock);
+      UNLOCK(cglobs->cint.CurrentPred->PELock);
       FAIL("can not compile data base reference",TYPE_ERROR_CALLABLE,t);
     } else {
-      READ_UNLOCK(cglobs->cint.CurrentPred->PRWLock);
+      UNLOCK(cglobs->cint.CurrentPred->PELock);
       cglobs->hasdbrefs = TRUE;
       if (level == 0)
 	Yap_emit((cglobs->onhead ? get_atom_op : put_atom_op), (CELL) t, argno, &cglobs->cint);      
@@ -691,11 +757,14 @@ c_arg(Int argno, Term t, unsigned int arity, unsigned int level, compiler_struct
 #endif
 
     if (optimizer_on) {
+      if (!(cglobs->cint.CurrentPred->PredFlags & (DynamicPredFlag|LogUpdatePredFlag))) {
+	if (try_store_as_dbterm(t, argno, arity, level, cglobs))
+	  return;
+      }      
       t = optimize_ce(t, arity, level, cglobs);
       if (IsVarTerm(t)) {
 	c_var(t, argno, arity, level, cglobs);
 	return;
-
       }
     }
     if (level == 0)
@@ -804,7 +873,7 @@ c_test(Int Op, Term t1, compiler_struct *cglobs) {
     Yap_bip_name(Op, s);
     sprintf(Yap_ErrorMessage, "when compiling %s/1", s);
     save_machine_regs();
-    longjmp(cglobs->cint.CompilerBotch, 1);
+    longjmp(cglobs->cint.CompilerBotch, COMPILER_ERR_BOTCH);
   }
   if (IsNewVar(t)) {
     /* in this case, var trivially succeeds and the others trivially fail */
@@ -844,7 +913,7 @@ bip_cons	   Op,Xk,Ri,C
 
  */
 static void
-c_bifun(Int Op, Term t1, Term t2, Term t3, int mod, compiler_struct *cglobs)
+c_bifun(Int Op, Term t1, Term t2, Term t3, Term Goal, int mod, compiler_struct *cglobs)
 {
   /* compile Z = X Op Y  arithmetic function */
   /* first we fetch the arguments */
@@ -859,7 +928,7 @@ c_bifun(Int Op, Term t1, Term t2, Term t3, int mod, compiler_struct *cglobs)
       Yap_bip_name(Op, s);
       sprintf(Yap_ErrorMessage, "when compiling %s/2",  s);
       save_machine_regs();
-      longjmp(cglobs->cint.CompilerBotch, 1);
+      longjmp(cglobs->cint.CompilerBotch, COMPILER_ERR_BOTCH);
     } else if (IsVarTerm(t2)) {
       if (IsNewVar(t2)) {
 	char s[32];
@@ -870,7 +939,7 @@ c_bifun(Int Op, Term t1, Term t2, Term t3, int mod, compiler_struct *cglobs)
 	Yap_bip_name(Op, s);
 	sprintf(Yap_ErrorMessage, "when compiling %s/2",  s);
 	save_machine_regs();
-	longjmp(cglobs->cint.CompilerBotch, 1);
+	longjmp(cglobs->cint.CompilerBotch, COMPILER_ERR_BOTCH);
       } else {
 	/* first temp */
 	Int v1 = --cglobs->tmpreg;
@@ -885,21 +954,27 @@ c_bifun(Int Op, Term t1, Term t2, Term t3, int mod, compiler_struct *cglobs)
       }
     } else {
       if (Op == _arg) {
-	Term tn = MkVarTerm();
-	Int v1 = --cglobs->tmpreg;
-	Int v2 = --cglobs->tmpreg;
+	/* we know the second argument is bound */
+	if (IsPrimitiveTerm(t2) || IsNumTerm(t2)) {
+	  Yap_emit(fail_op, Zero, Zero, &cglobs->cint);
+	  return;
+	} else {
+	  Term tn = MkVarTerm();
+	  Int v1 = --cglobs->tmpreg;
+	  Int v2 = --cglobs->tmpreg;
 
-	c_arg(t2, v2, 0, 0, cglobs);
-	Yap_emit(fetch_args_vv_op, Zero, Zero, &cglobs->cint);
-	/* these should be the arguments */
-	c_var(t1, v1, 0, 0, cglobs);
-	c_var(tn, v2, 0, 0, cglobs);
+	  c_eq(t2, tn, cglobs);
+	  Yap_emit(fetch_args_vv_op, Zero, Zero, &cglobs->cint);
+	  /* these should be the arguments */
+	  c_var(t1, v1, 0, 0, cglobs);
+	  c_var(tn, v2, 0, 0, cglobs);
+	}
       /* it has to be either an integer or a floating point */
       } else if (IsIntegerTerm(t2)) {
 	/* first temp */
 	Int v1 = --cglobs->tmpreg;
 
-	Yap_emit(fetch_args_vc_op, IntegerOfTerm(t2), Zero, &cglobs->cint);
+	Yap_emit(fetch_args_vi_op, IntegerOfTerm(t2), 0L, &cglobs->cint);
 	/* these should be the arguments */
 	c_var(t1, v1, 0, 0, cglobs);
 	/* now we know where the arguments are */
@@ -982,24 +1057,38 @@ c_bifun(Int Op, Term t1, Term t2, Term t3, int mod, compiler_struct *cglobs)
 	    if (H+2 >= (CELL *)cglobs->cint.freep0) {
 	      /* oops, too many new variables */
 	      save_machine_regs();
-	      longjmp(cglobs->cint.CompilerBotch,4);
+	      longjmp(cglobs->cint.CompilerBotch,OUT_OF_TEMPS_BOTCH);
 	    }
 	    RESET_VARIABLE(H);
 	    RESET_VARIABLE(H+1);
 	    H += 2;
 	    c_eq(AbsPair(H-2),t3, cglobs);
-	  } else if (i2 < 16) {
+	  } else if (i2 < 256) {
 	    *H++ = (CELL)Yap_MkFunctor(AtomOfTerm(t1),i2);
 	    for (i=0; i < i2; i++) {
 	      if (H >= (CELL *)cglobs->cint.freep0) {
 		/* oops, too many new variables */
 		save_machine_regs();
-		longjmp(cglobs->cint.CompilerBotch,4);
+		longjmp(cglobs->cint.CompilerBotch,OUT_OF_TEMPS_BOTCH);
 	      }
 	      RESET_VARIABLE(H);
 	      H++;	    
 	    }
 	    c_eq(AbsAppl(hi),t3, cglobs);
+	  } else {
+	    /* compile as default */
+	    Functor f = FunctorOfTerm(Goal);
+	    Prop p0 = PredPropByFunc(f, mod);
+
+	    if (profiling)
+	      Yap_emit(enter_profiling_op, (CELL)RepPredProp(p0), Zero, &cglobs->cint);
+	    else if (call_counting)
+	      Yap_emit(count_call_op, (CELL)RepPredProp(p0), Zero, &cglobs->cint);
+	    c_args(Goal, 0, cglobs);
+	    Yap_emit(safe_call_op, (CELL)p0 , Zero, &cglobs->cint);
+	    Yap_emit(empty_call_op, Zero, Zero, &cglobs->cint);
+	    Yap_emit(restore_tmps_and_skip_op, Zero, Zero, &cglobs->cint);
+	    return;
 	  }
 	}
       } else if (Op == _arg) {
@@ -1109,7 +1198,7 @@ c_bifun(Int Op, Term t1, Term t2, Term t3, int mod, compiler_struct *cglobs)
 	    if (H+1+arity >= (CELL *)cglobs->cint.freep0) {
 	      /* oops, too many new variables */
 	      save_machine_regs();
-	      longjmp(cglobs->cint.CompilerBotch,4);
+	      longjmp(cglobs->cint.CompilerBotch,OUT_OF_TEMPS_BOTCH);
 	    }
 	    tnew = AbsAppl(H);
 	    *H++ = (CELL)Yap_MkFunctor(AtomOfTerm(t1),arity);
@@ -1135,7 +1224,7 @@ c_bifun(Int Op, Term t1, Term t2, Term t3, int mod, compiler_struct *cglobs)
     } else if (IsIntegerTerm(t1)) {
       /* first temp */
       Int v1 = --cglobs->tmpreg;
-      Yap_emit(fetch_args_cv_op, IntegerOfTerm(t1), Zero, &cglobs->cint);
+      Yap_emit(fetch_args_iv_op, IntegerOfTerm(t1), 0L, &cglobs->cint);
       /* these should be the arguments */
       c_var(t2, v1, 0, 0, cglobs);
       /* now we know where the arguments are */
@@ -1158,7 +1247,7 @@ c_bifun(Int Op, Term t1, Term t2, Term t3, int mod, compiler_struct *cglobs)
       if (H == (CELL *)cglobs->cint.freep0) {
 	/* oops, too many new variables */
 	save_machine_regs();
-	longjmp(cglobs->cint.CompilerBotch,4);
+	longjmp(cglobs->cint.CompilerBotch,OUT_OF_TEMPS_BOTCH);
       }
       c_var(tmpvar,f_flag,(unsigned int)Op, 0, cglobs);
       c_eq(tmpvar,t3, cglobs);
@@ -1185,7 +1274,7 @@ c_bifun(Int Op, Term t1, Term t2, Term t3, int mod, compiler_struct *cglobs)
     if (H == (CELL *)cglobs->cint.freep0) {
       /* oops, too many new variables */
       save_machine_regs();
-      longjmp(cglobs->cint.CompilerBotch,4);
+      longjmp(cglobs->cint.CompilerBotch,OUT_OF_TEMPS_BOTCH);
     }
     c_var(tmpvar,f_flag,(unsigned int)Op, 0, cglobs);
     /* I have to dit here, before I do the unification */
@@ -1205,7 +1294,7 @@ c_functor(Term Goal, int mod, compiler_struct *cglobs)
   Term t3 = ArgOfTerm(3, Goal);
 
   if (IsVarTerm(t1) && IsNewVar(t1)) {
-    c_bifun(_functor, t2, t3, t1, mod, cglobs);
+    c_bifun(_functor, t2, t3, t1, Goal, mod, cglobs);
   } else if (IsNonVarTerm(t1)) {
     /* just split the structure */
     if (IsAtomicTerm(t1)) {
@@ -1222,7 +1311,7 @@ c_functor(Term Goal, int mod, compiler_struct *cglobs)
   } else if (IsVarTerm(t2) && IsNewVar(t2) &&
 	     IsVarTerm(t3) && IsNewVar(t3)) {
     Int v1 = --cglobs->tmpreg;
-    Yap_emit(fetch_args_vc_op, Zero, Zero, &cglobs->cint);
+    Yap_emit(fetch_args_vi_op, Zero, Zero, &cglobs->cint);
     c_var(t1, v1, 0, 0, cglobs);
     c_var(t2,f_flag,(unsigned int)_functor, 0, cglobs);
     c_var(t3,f_flag,(unsigned int)_functor, 0, cglobs);
@@ -1249,7 +1338,7 @@ IsTrueGoal(Term t) {
     if (f == FunctorModule) {
       return(IsTrueGoal(ArgOfTerm(2,t)));
     }
-    if (f == FunctorComma || f == FunctorOr || f == FunctorArrow) {
+    if (f == FunctorComma || f == FunctorOr || f == FunctorVBar || f == FunctorArrow) {
       return(IsTrueGoal(ArgOfTerm(1,t)) && IsTrueGoal(ArgOfTerm(2,t)));
     }
     return(FALSE);
@@ -1279,7 +1368,7 @@ c_goal(Term Goal, int mod, compiler_struct *cglobs)
       Yap_Error_Term = M;
       Yap_ErrorMessage = "in module name";
       save_machine_regs();
-      longjmp(cglobs->cint.CompilerBotch, 1);
+      longjmp(cglobs->cint.CompilerBotch, COMPILER_ERR_BOTCH);
     }
     Goal = ArgOfTerm(2, Goal);
     mod = M;
@@ -1307,14 +1396,14 @@ c_goal(Term Goal, int mod, compiler_struct *cglobs)
       if (cglobs->onlast) {
 	Yap_emit(deallocate_op, Zero, Zero, &cglobs->cint);
 #ifdef TABLING
-	READ_LOCK(cglobs->cint.CurrentPred->PRWLock);
+	LOCK(cglobs->cint.CurrentPred->PELock);
 	if (is_tabled(cglobs->cint.CurrentPred))
 	  Yap_emit(table_new_answer_op, Zero, cglobs->cint.CurrentPred->ArityOfPE, &cglobs->cint);
 	else
 #endif /* TABLING */
 	  Yap_emit(procceed_op, Zero, Zero, &cglobs->cint);
 #ifdef TABLING
-	READ_UNLOCK(cglobs->cint.CurrentPred->PRWLock);
+	UNLOCK(cglobs->cint.CurrentPred->PELock);
 #endif
       }
       return;
@@ -1328,7 +1417,7 @@ c_goal(Term Goal, int mod, compiler_struct *cglobs)
 	/* never a problem here with a -> b, !, c ; d */
 	Yap_emit(deallocate_op, Zero, Zero, &cglobs->cint);
 #ifdef TABLING
-	READ_LOCK(cglobs->cint.CurrentPred->PRWLock);
+	LOCK(cglobs->cint.CurrentPred->PELock);
 	if (is_tabled(cglobs->cint.CurrentPred)) {
 	  Yap_emit_3ops(cut_op, Zero, Zero, Zero, &cglobs->cint);
 	  Yap_emit(table_new_answer_op, Zero, cglobs->cint.CurrentPred->ArityOfPE, &cglobs->cint);
@@ -1339,7 +1428,7 @@ c_goal(Term Goal, int mod, compiler_struct *cglobs)
 	    Yap_emit_3ops(cutexit_op, Zero, Zero, Zero, &cglobs->cint);
 	  }
 #ifdef TABLING
-	READ_UNLOCK(cglobs->cint.CurrentPred->PRWLock);
+	UNLOCK(cglobs->cint.CurrentPred->PELock);
 #endif
       }
       else {
@@ -1376,14 +1465,14 @@ c_goal(Term Goal, int mod, compiler_struct *cglobs)
       Yap_emit(label_op, l2, Zero, &cglobs->cint);
       if (cglobs->onlast) {
 #ifdef TABLING
-	READ_LOCK(cglobs->cint.CurrentPred->PRWLock);
+	LOCK(cglobs->cint.CurrentPred->PELock);
 	if (is_tabled(cglobs->cint.CurrentPred))
 	  Yap_emit(table_new_answer_op, Zero, cglobs->cint.CurrentPred->ArityOfPE, &cglobs->cint);
 	else
 #endif /* TABLING */
 	  Yap_emit(procceed_op, Zero, Zero, &cglobs->cint);
 #ifdef TABLING
-	READ_UNLOCK(cglobs->cint.CurrentPred->PRWLock);
+	UNLOCK(cglobs->cint.CurrentPred->PELock);
 #endif
       }
       else
@@ -1404,7 +1493,7 @@ c_goal(Term Goal, int mod, compiler_struct *cglobs)
   else {
     f = FunctorOfTerm(Goal);
     p = RepPredProp(p0 = Yap_PredPropByFunctorNonThreadLocal(f, mod));
-    if (f == FunctorOr) {
+    if (f == FunctorOr || f == FunctorVBar) {
       Term arg;
       CELL l = ++cglobs->labelno;
       CELL m = ++cglobs->labelno;
@@ -1476,7 +1565,7 @@ c_goal(Term Goal, int mod, compiler_struct *cglobs)
 	    if (H == (CELL *)cglobs->cint.freep0) {
 	      /* oops, too many new variables */
 	      save_machine_regs();
-	      longjmp(cglobs->cint.CompilerBotch,4);
+	      longjmp(cglobs->cint.CompilerBotch,OUT_OF_TEMPS_BOTCH);
 	    }
 	    savecpc = cglobs->cint.cpc;
 	    savencpc = FirstP->nextInst;
@@ -1513,7 +1602,8 @@ c_goal(Term Goal, int mod, compiler_struct *cglobs)
 	++cglobs->curbranch;
 	cglobs->onbranch = cglobs->curbranch;
       } while (IsNonVarTerm(Goal) && IsApplTerm(Goal)
-	       && FunctorOfTerm(Goal) == FunctorOr);
+	       && (FunctorOfTerm(Goal) == FunctorOr
+		   || FunctorOfTerm(Goal) == FunctorVBar));
       Yap_emit(pushpop_or_op, Zero, Zero, &cglobs->cint);
       Yap_emit(label_op, l, Zero, &cglobs->cint);
       if (!optimizing_commit) {
@@ -1554,7 +1644,7 @@ c_goal(Term Goal, int mod, compiler_struct *cglobs)
       if (H == (CELL *)cglobs->cint.freep0) {
 	/* oops, too many new variables */
 	save_machine_regs();
-	longjmp(cglobs->cint.CompilerBotch,4);
+	longjmp(cglobs->cint.CompilerBotch,OUT_OF_TEMPS_BOTCH);
       }
       push_branch(cglobs->onbranch, commitvar, cglobs);
       ++cglobs->curbranch;
@@ -1589,7 +1679,7 @@ c_goal(Term Goal, int mod, compiler_struct *cglobs)
       if (H == (CELL *)cglobs->cint.freep0) {
 	/* oops, too many new variables */
 	save_machine_regs();
-	longjmp(cglobs->cint.CompilerBotch,4);
+	longjmp(cglobs->cint.CompilerBotch,OUT_OF_TEMPS_BOTCH);
       }
       cglobs->onlast = FALSE;
       c_var(commitvar, save_b_flag, 1, 0, cglobs);
@@ -1608,14 +1698,14 @@ c_goal(Term Goal, int mod, compiler_struct *cglobs)
       if (cglobs->onlast) {
 	Yap_emit(deallocate_op, Zero, Zero, &cglobs->cint);
 #ifdef TABLING
-	READ_LOCK(cglobs->cint.CurrentPred->PRWLock);
+	LOCK(cglobs->cint.CurrentPred->PELock);
 	if (is_tabled(cglobs->cint.CurrentPred))
 	  Yap_emit(table_new_answer_op, Zero, cglobs->cint.CurrentPred->ArityOfPE, &cglobs->cint);
 	else
 #endif /* TABLING */
 	  Yap_emit(procceed_op, Zero, Zero, &cglobs->cint);
 #ifdef TABLING
-	READ_UNLOCK(cglobs->cint.CurrentPred->PRWLock);
+	UNLOCK(cglobs->cint.CurrentPred->PELock);
 #endif
       }
       return;
@@ -1632,14 +1722,14 @@ c_goal(Term Goal, int mod, compiler_struct *cglobs)
 	if (cglobs->onlast) {
 	  Yap_emit(deallocate_op, Zero, Zero, &cglobs->cint);
 #ifdef TABLING
-	  READ_LOCK(cglobs->cint.CurrentPred->PRWLock);
+	  LOCK(cglobs->cint.CurrentPred->PELock);
 	  if (is_tabled(cglobs->cint.CurrentPred))
 	    Yap_emit(table_new_answer_op, Zero, cglobs->cint.CurrentPred->ArityOfPE, &cglobs->cint);
 	  else
 #endif /* TABLING */
 	    Yap_emit(procceed_op, Zero, Zero, &cglobs->cint);
 #ifdef TABLING
-	  READ_UNLOCK(cglobs->cint.CurrentPred->PRWLock);
+	  UNLOCK(cglobs->cint.CurrentPred->PELock);
 #endif
 	}
 	return;
@@ -1653,20 +1743,21 @@ c_goal(Term Goal, int mod, compiler_struct *cglobs)
 		  ArgOfTerm(1, Goal),
 		  ArgOfTerm(2, Goal),
 		  ArgOfTerm(3, Goal),
+		  Goal,
 		  mod,
 		  cglobs);
 	}
 	if (cglobs->onlast) {
 	  Yap_emit(deallocate_op, Zero, Zero, &cglobs->cint);
 #ifdef TABLING
-	  READ_LOCK(cglobs->cint.CurrentPred->PRWLock);
+	  LOCK(cglobs->cint.CurrentPred->PELock);
 	  if (is_tabled(cglobs->cint.CurrentPred))
 	    Yap_emit(table_new_answer_op, Zero, cglobs->cint.CurrentPred->ArityOfPE, &cglobs->cint);
 	  else
 #endif /* TABLING */
 	    Yap_emit(procceed_op, Zero, Zero, &cglobs->cint);
 #ifdef TABLING
-	  READ_UNLOCK(cglobs->cint.CurrentPred->PRWLock);
+	  UNLOCK(cglobs->cint.CurrentPred->PELock);
 #endif
 	}
 	return;
@@ -1702,7 +1793,7 @@ c_goal(Term Goal, int mod, compiler_struct *cglobs)
 	  if (H == (CELL *)cglobs->cint.freep0) {
 	    /* oops, too many new variables */
 	    save_machine_regs();
-	    longjmp(cglobs->cint.CompilerBotch,4);
+	    longjmp(cglobs->cint.CompilerBotch,OUT_OF_TEMPS_BOTCH);
 	  }
 	  c_eq(t2, a2, cglobs);
 	  c_var(a1, bt1_flag, 2, 0, cglobs);
@@ -1715,7 +1806,7 @@ c_goal(Term Goal, int mod, compiler_struct *cglobs)
 	if (H == (CELL *)cglobs->cint.freep0) {
 	  /* oops, too many new variables */
 	  save_machine_regs();
-	  longjmp(cglobs->cint.CompilerBotch,4);
+	  longjmp(cglobs->cint.CompilerBotch,OUT_OF_TEMPS_BOTCH);
 	}
 	c_eq(t1, a1, cglobs);
 
@@ -1729,7 +1820,7 @@ c_goal(Term Goal, int mod, compiler_struct *cglobs)
 	  if (H == (CELL *)cglobs->cint.freep0) {
 	    /* oops, too many new variables */
 	    save_machine_regs();
-	    longjmp(cglobs->cint.CompilerBotch,4);
+	    longjmp(cglobs->cint.CompilerBotch,OUT_OF_TEMPS_BOTCH);
 	  }
 	  c_eq(t2, a2, cglobs);
 	  c_var(t1, bt1_flag, 2, 0, cglobs);
@@ -1740,14 +1831,14 @@ c_goal(Term Goal, int mod, compiler_struct *cglobs)
       if (cglobs->onlast) {
 	Yap_emit(deallocate_op, Zero, Zero, &cglobs->cint);
 #ifdef TABLING
-	READ_LOCK(cglobs->cint.CurrentPred->PRWLock);
+	LOCK(cglobs->cint.CurrentPred->PELock);
 	if (is_tabled(cglobs->cint.CurrentPred))
 	  Yap_emit(table_new_answer_op, Zero, cglobs->cint.CurrentPred->ArityOfPE, &cglobs->cint);
 	else
 #endif /* TABLING */
 	  Yap_emit(procceed_op, Zero, Zero, &cglobs->cint);
 #ifdef TABLING
-	READ_UNLOCK(cglobs->cint.CurrentPred->PRWLock);
+	UNLOCK(cglobs->cint.CurrentPred->PELock);
 #endif
       }
       return;
@@ -1775,19 +1866,19 @@ c_goal(Term Goal, int mod, compiler_struct *cglobs)
     if (cglobs->onlast) {
       Yap_emit(deallocate_op, Zero, Zero, &cglobs->cint);
 #ifdef TABLING
-      READ_LOCK(cglobs->cint.CurrentPred->PRWLock);
+      LOCK(cglobs->cint.CurrentPred->PELock);
       if (is_tabled(cglobs->cint.CurrentPred))
 	Yap_emit(table_new_answer_op, Zero, cglobs->cint.CurrentPred->ArityOfPE, &cglobs->cint);
       else
 #endif /* TABLING */
 	Yap_emit(procceed_op, Zero, Zero, &cglobs->cint);
 #ifdef TABLING
-      READ_UNLOCK(cglobs->cint.CurrentPred->PRWLock);
+      UNLOCK(cglobs->cint.CurrentPred->PELock);
 #endif
     }
   }
   else {
-    if (p->PredFlags & (CPredFlag | AsmPredFlag)) {
+    if (p->PredFlags & (CPredFlag | AsmPredFlag | ModuleTransparentPredFlag)) {
 #ifdef YAPOR
       if (p->PredFlags & SyncPredFlag)
 	Yap_emit(sync_op, (CELL)p, (CELL)(p->ArityOfPE), &cglobs->cint);
@@ -1804,14 +1895,14 @@ c_goal(Term Goal, int mod, compiler_struct *cglobs)
 	Yap_emit(deallocate_op, Zero, Zero, &cglobs->cint);
 	cglobs->or_found = TRUE;
 #ifdef TABLING
-	READ_LOCK(cglobs->cint.CurrentPred->PRWLock);
+	LOCK(cglobs->cint.CurrentPred->PELock);
 	if (is_tabled(cglobs->cint.CurrentPred))
 	  Yap_emit(table_new_answer_op, Zero, cglobs->cint.CurrentPred->ArityOfPE, &cglobs->cint);
 	else
 #endif /* TABLING */
 	  Yap_emit(procceed_op, Zero, Zero, &cglobs->cint);
 #ifdef TABLING
-	READ_UNLOCK(cglobs->cint.CurrentPred->PRWLock);
+	UNLOCK(cglobs->cint.CurrentPred->PELock);
 #endif
       }
     }
@@ -1819,7 +1910,7 @@ c_goal(Term Goal, int mod, compiler_struct *cglobs)
       if (cglobs->onlast) {
 	Yap_emit(deallocate_op, Zero, Zero, &cglobs->cint);
 #ifdef TABLING
-	READ_LOCK(cglobs->cint.CurrentPred->PRWLock);
+	LOCK(cglobs->cint.CurrentPred->PELock);
 	if (is_tabled(cglobs->cint.CurrentPred)) {
 	  cglobs->needs_env = TRUE;
 	  Yap_emit_3ops(call_op, (CELL) p0, Zero, Zero, &cglobs->cint);
@@ -1829,7 +1920,7 @@ c_goal(Term Goal, int mod, compiler_struct *cglobs)
 #endif /* TABLING */
 	  Yap_emit(execute_op, (CELL) p0, Zero, &cglobs->cint);
 #ifdef TABLING
-	READ_UNLOCK(cglobs->cint.CurrentPred->PRWLock);
+	UNLOCK(cglobs->cint.CurrentPred->PELock);
 #endif
       }
       else {
@@ -2089,7 +2180,7 @@ clear_bvarray(int var, CELL *bvarray
     Yap_Error_Term = TermNil;
     Yap_ErrorMessage = "compiler internal error: variable initialised twice";
     save_machine_regs();
-    longjmp(cglobs->cint.CompilerBotch, 2);
+    longjmp(cglobs->cint.CompilerBotch, OUT_OF_HEAP_BOTCH);
   }
   cglobs->pbvars++;
 #endif
@@ -2118,7 +2209,7 @@ typedef struct {
   PInstr *pc;
 }  bventry;
 
-#define MAX_DISJUNCTIONS 32
+#define MAX_DISJUNCTIONS 128
 static bventry *bvstack;
 static int bvindex = 0;
 
@@ -2130,7 +2221,7 @@ push_bvmap(int label, PInstr *pcpc, compiler_struct *cglobs)
     Yap_Error_Term = TermNil;
     Yap_ErrorMessage = "Too many embedded disjunctions";
     save_machine_regs();
-    longjmp(cglobs->cint.CompilerBotch, 2);
+    longjmp(cglobs->cint.CompilerBotch, OUT_OF_HEAP_BOTCH);
   }
   /* the label instruction */
   bvstack[bvindex].lab = label;
@@ -2153,7 +2244,7 @@ reset_bvmap(CELL *bvarray, int nperm, compiler_struct *cglobs)
     Yap_Error_Term = TermNil;
     Yap_ErrorMessage = "No embedding in disjunctions";
     save_machine_regs();
-    longjmp(cglobs->cint.CompilerBotch, 2);
+    longjmp(cglobs->cint.CompilerBotch, OUT_OF_HEAP_BOTCH);
   }
   env_size = (bvstack[bvindex-1].pc)->rnd1;
   size = env_size/(8*sizeof(CELL));
@@ -2173,7 +2264,7 @@ pop_bvmap(CELL *bvarray, int nperm, compiler_struct *cglobs)
     Yap_Error_Term = TermNil;
     Yap_ErrorMessage = "Too few embedded disjunctions";
     /*  save_machine_regs();
-	longjmp(cglobs->cint.CompilerBotch, 2); */
+	longjmp(cglobs->cint.CompilerBotch, OUT_OF_HEAP_BOTCH); */
   }
   reset_bvmap(bvarray, nperm, cglobs);
   bvindex--;
@@ -2441,7 +2532,7 @@ checktemp(Int arg, Int rn, compiler_vm_op ic, compiler_struct *cglobs)
     Yap_Error_Term = TermNil;
     Yap_ErrorMessage = "too many temporaries";
     save_machine_regs();
-    longjmp(cglobs->cint.CompilerBotch, 1);
+    longjmp(cglobs->cint.CompilerBotch, COMPILER_ERR_BOTCH);
   }
   v->NoOfVE = cglobs->vadr = vadr = TempVar | target1;
   v->KindOfVE = TempVar;
@@ -2570,7 +2661,7 @@ c_layout(compiler_struct *cglobs)
 	Yap_Error_Term = TermNil;
 	Yap_ErrorMessage = "wrong number of variables found in bitmap";
 	save_machine_regs();
-	longjmp(cglobs->cint.CompilerBotch, 2);
+	longjmp(cglobs->cint.CompilerBotch, OUT_OF_HEAP_BOTCH);
       } 
 #endif
     }
@@ -2615,7 +2706,7 @@ c_layout(compiler_struct *cglobs)
 	cglobs->cint.cpc->op = nop_op;
       } else {
 #ifdef TABLING
-	READ_LOCK(cglobs->cint.CurrentPred->PRWLock);
+	LOCK(cglobs->cint.CurrentPred->PELock);
 	if (is_tabled(cglobs->cint.CurrentPred))
 	  cglobs->cint.cpc->op = nop_op;
 	else
@@ -2623,7 +2714,7 @@ c_layout(compiler_struct *cglobs)
 	  if (cglobs->goalno == 1 && !cglobs->or_found && nperm == 0)
 	    cglobs->cint.cpc->op = nop_op;
 #ifdef TABLING
-	READ_UNLOCK(cglobs->cint.CurrentPred->PRWLock);
+	UNLOCK(cglobs->cint.CurrentPred->PELock);
 #endif
       }
       break;
@@ -2702,6 +2793,8 @@ c_layout(compiler_struct *cglobs)
       break;
     case fetch_args_cv_op:
     case fetch_args_vc_op:
+    case fetch_args_iv_op:
+    case fetch_args_vi_op:
       rn_to_kill[1]=rn_to_kill[0]=0;
       if (cglobs->cint.cpc->nextInst &&
 	  cglobs->cint.cpc->nextInst->op == put_val_op &&
@@ -2781,7 +2874,7 @@ c_layout(compiler_struct *cglobs)
       {
 	up = cglobs->Uses;
 	cop = cglobs->Contents;
-	for (rn = 0; rn <= cglobs->MaxCTemps; ++rn) {
+	for (rn = 0; rn < cglobs->MaxCTemps; ++rn) {
 	  if (*cop != (TempVar | rn)) {
 	    *up++ = *cop++ = NIL;
 	  } else {
@@ -3011,56 +3104,91 @@ Yap_cclause(volatile Term inp_clause, int NOfArgs, int mod, volatile Term src)
   int botch_why;
   /* may botch while doing a different module */
   /* first, initialise cglobs->cint.CompilerBotch to handle all cases of interruptions */
- compiler_struct cglobs;
+  compiler_struct cglobs;
 
- /* make sure we know there was no error yet */
- Yap_ErrorMessage = NULL;
-  if ((botch_why = setjmp(cglobs.cint.CompilerBotch)) == 3) {
-    /* out of local stack, just duplicate the stack */
+  /* make sure we know there was no error yet */
+  Yap_ErrorMessage = NULL;
+  if ((botch_why = setjmp(cglobs.cint.CompilerBotch))) {
     restore_machine_regs();
     reset_vars(cglobs.vtable);
-    {
-      Int osize = 2*sizeof(CELL)*(ASP-H);
-      ARG1 = inp_clause;
-      ARG3 = src;
+    switch(botch_why) {
+    case OUT_OF_STACK_BOTCH:
+      /* out of local stack, just duplicate the stack */
+      {
+	Int osize = 2*sizeof(CELL)*(ASP-H);
+	ARG1 = inp_clause;
+	ARG3 = src;
 
-      YAPLeaveCriticalSection();
-      if (!Yap_gcl(Yap_Error_Size, NOfArgs, ENV, P)) {
-	Yap_Error_TYPE = OUT_OF_STACK_ERROR;
-	Yap_Error_Term = inp_clause;
-      }
-      if (osize > ASP-H) {
-	if (!Yap_growstack(2*sizeof(CELL)*(ASP-H))) {
+	YAPLeaveCriticalSection();
+	if (!Yap_gcl(Yap_Error_Size, NOfArgs, ENV, P)) {
 	  Yap_Error_TYPE = OUT_OF_STACK_ERROR;
 	  Yap_Error_Term = inp_clause;
 	}
+	if (osize > ASP-H) {
+	  if (!Yap_growstack(2*sizeof(CELL)*(ASP-H))) {
+	    Yap_Error_TYPE = OUT_OF_STACK_ERROR;
+	    Yap_Error_Term = inp_clause;
+	  }
+	}
+	YAPEnterCriticalSection();
+	src = ARG3;
+	inp_clause = ARG1;
+      }
+      break;
+    case OUT_OF_AUX_BOTCH:
+      /* out of local stack, just duplicate the stack */
+      YAPLeaveCriticalSection();
+      ARG1 = inp_clause;
+      ARG3 = src;
+      if (!Yap_ExpandPreAllocCodeSpace(Yap_Error_Size, NULL)) {
+	Yap_Error_TYPE = OUT_OF_AUXSPACE_ERROR;
+	Yap_Error_Term = inp_clause;
       }
       YAPEnterCriticalSection();
       src = ARG3;
       inp_clause = ARG1;
+      break;
+    case OUT_OF_TEMPS_BOTCH:
+      /* out of temporary cells */
+      if (maxvnum < 16*1024) {
+	maxvnum *= 2;
+      } else {
+	maxvnum += 4096;
+      }
+      break;
+    case OUT_OF_HEAP_BOTCH:
+      /* not enough heap */
+      ARG1 = inp_clause;
+      ARG3 = src;
+      YAPLeaveCriticalSection();
+      if (!Yap_growheap(FALSE, Yap_Error_Size, NULL)) {
+	Yap_Error_TYPE = OUT_OF_HEAP_ERROR;
+	Yap_Error_Term = inp_clause;
+	return NULL;
+      }
+      YAPEnterCriticalSection();
+      src = ARG3;
+      inp_clause = ARG1;
+      break;
+    case OUT_OF_TRAIL_BOTCH:
+      /* not enough trail */
+      ARG1 = inp_clause;
+      ARG3 = src;
+      YAPLeaveCriticalSection();
+      if (!Yap_growtrail(0L, FALSE)) {
+	Yap_Error_TYPE = OUT_OF_TRAIL_ERROR;
+	Yap_Error_Term = inp_clause;
+	return NULL;
+      }
+      YAPEnterCriticalSection();
+      src = ARG3;
+      inp_clause = ARG1;
+      break;
+    default:
+      return NULL;
     }
-  } else if (botch_why == 4) {
-    /* out of temporary cells */
-    restore_machine_regs();
-    reset_vars(cglobs.vtable);
-    if (maxvnum < 16*1024) {
-      maxvnum *= 2;
-    } else {
-      maxvnum += 4096;
-    }
-  } else if (botch_why == 2) {
-    /* not enough heap */
-    restore_machine_regs();
-    reset_vars(cglobs.vtable);
-    Yap_Error_TYPE = OUT_OF_HEAP_ERROR;
-    Yap_Error_Term = TermNil;
-    return 0;
   }
   my_clause = inp_clause;
-  if (Yap_ErrorMessage) {
-    reset_vars(cglobs.vtable);
-    return (0);
-  }
   HB = H;
   Yap_ErrorMessage = NULL;
   Yap_Error_Size = 0;
@@ -3069,6 +3197,7 @@ Yap_cclause(volatile Term inp_clause, int NOfArgs, int mod, volatile Term src)
   
   cglobs.cint.CodeStart = cglobs.cint.cpc = NULL;
   cglobs.cint.BlobsStart = cglobs.cint.icpc = NULL;
+  cglobs.cint.dbterml = NULL;
   cglobs.cint.freep =
     cglobs.cint.freep0 =
     (char *) (H + maxvnum+(sizeof(Int)/sizeof(CELL))*MaxTemps+MaxTemps);
@@ -3125,7 +3254,7 @@ Yap_cclause(volatile Term inp_clause, int NOfArgs, int mod, volatile Term src)
       cglobs.cint.CurrentPred = RepPredProp(PredPropByFunc(FunctorOfTerm(head),mod));
     }
     /* insert extra instructions to count calls */
-    READ_LOCK(cglobs.cint.CurrentPred->PRWLock);
+    LOCK(cglobs.cint.CurrentPred->PELock);
     if ((cglobs.cint.CurrentPred->PredFlags & ProfiledPredFlag) ||
 	(PROFILING && (cglobs.cint.CurrentPred->cs.p_code.FirstClause == NIL))) {
       profiling = TRUE;
@@ -3138,7 +3267,7 @@ Yap_cclause(volatile Term inp_clause, int NOfArgs, int mod, volatile Term src)
       profiling = FALSE;
       call_counting = FALSE;
     }
-    READ_UNLOCK(cglobs.cint.CurrentPred->PRWLock);
+    UNLOCK(cglobs.cint.CurrentPred->PELock);
   }
   cglobs.is_a_fact = (body == MkAtomTerm(AtomTrue));
   /* phase 1 : produce skeleton code and variable information              */
@@ -3147,14 +3276,14 @@ Yap_cclause(volatile Term inp_clause, int NOfArgs, int mod, volatile Term src)
 
   if (cglobs.is_a_fact && !cglobs.vtable) {
 #ifdef TABLING
-    READ_LOCK(cglobs.cint.CurrentPred->PRWLock);
+    LOCK(cglobs.cint.CurrentPred->PELock);
     if (is_tabled(cglobs.cint.CurrentPred))
       Yap_emit(table_new_answer_op, Zero, cglobs.cint.CurrentPred->ArityOfPE, &cglobs.cint);
     else
 #endif /* TABLING */
       Yap_emit(procceed_op, Zero, Zero, &cglobs.cint);
 #ifdef TABLING
-    READ_UNLOCK(cglobs.cint.CurrentPred->PRWLock);
+    UNLOCK(cglobs.cint.CurrentPred->PELock);
 #endif
     /* ground term, do not need much more work */
     if (cglobs.cint.BlobsStart != NULL) {
@@ -3223,7 +3352,7 @@ Yap_cclause(volatile Term inp_clause, int NOfArgs, int mod, volatile Term src)
 #endif
 
   /* phase 3: assemble code                                                */
-  acode = Yap_assemble(ASSEMBLING_CLAUSE, src, cglobs.cint.CurrentPred, (cglobs.is_a_fact && !cglobs.hasdbrefs), &cglobs.cint);
+  acode = Yap_assemble(ASSEMBLING_CLAUSE, src, cglobs.cint.CurrentPred, (cglobs.is_a_fact && !cglobs.hasdbrefs && !(cglobs.cint.CurrentPred->PredFlags & TabledPredFlag)), &cglobs.cint);
   /* check first if there was space for us */
   if (acode == NULL) {
     return NULL;
