@@ -176,23 +176,39 @@ ChunkPtrAdjust (struct malloc_chunk *ptr)
 
 /* vsc: emulation of sbrk with YAP contiguous memory management */
 
+void
+Yap_add_memory_hole(ADDR start, ADDR end)
+{
+  if (Yap_NOfMemoryHoles == MAX_DLMALLOC_HOLES) {
+    Yap_Error(OPERATING_SYSTEM_ERROR, 0L, "Unexpected Too Much Memory Fragmentation: please contact YAP maintainers");
+    return;
+  }
+  Yap_MemoryHoles[Yap_NOfMemoryHoles].start = start;
+  Yap_MemoryHoles[Yap_NOfMemoryHoles].end = end;
+  Yap_HoleSize += start-end;
+  Yap_NOfMemoryHoles++;
+}
+
 static void *
 yapsbrk(long size)
 {
   ADDR newHeapTop = HeapTop, oldHeapTop = HeapTop;
-  LOCK(HeapUsedLock);
   newHeapTop = HeapTop+size;
-  if (Yap_hole_start && newHeapTop > Yap_hole_start) {
-    HeapTop = oldHeapTop = Yap_hole_end;
+  while (Yap_NOfMemoryHoles && newHeapTop > Yap_MemoryHoles[0].start) {
+    UInt i;
+
+    HeapTop = oldHeapTop = Yap_MemoryHoles[0].end;
     newHeapTop = oldHeapTop+size;
-    Yap_hole_start = Yap_hole_end = NULL;
+    Yap_NOfMemoryHoles--;
+    for (i=0; i < Yap_NOfMemoryHoles; i++) {
+      Yap_MemoryHoles[i].start = Yap_MemoryHoles[i+1].start;
+      Yap_MemoryHoles[i].end = Yap_MemoryHoles[i+1].end;
+    }
   }
   if (newHeapTop > HeapLim - MinHeapGap) {
     if (HeapTop + size < HeapLim) {
       /* small allocations, we can wait */
       HeapTop += size;
-      HeapUsed += size;
-      UNLOCK(HeapUsedLock);
       UNLOCK(HeapTopLock);
       Yap_signal(YAP_CDOVF_SIGNAL);
     } else {
@@ -205,7 +221,6 @@ yapsbrk(long size)
     }
   }
   HeapTop = newHeapTop;
-  HeapUsed += size;
   UNLOCK(HeapTopLock);
   return oldHeapTop;
 }
@@ -1150,6 +1165,7 @@ static Void_t* sYSMALLOc(nb, av) INTERNAL_SIZE_T nb; mstate av;
     
     /* check that one of the above allocation paths succeeded */
     if ((CHUNK_SIZE_T)(size) >= (CHUNK_SIZE_T)(nb + MINSIZE)) {
+
       remainder_size = size - nb;
       remainder = chunk_at_offset(p, nb);
       av->top = remainder;
@@ -1168,7 +1184,6 @@ static Void_t* sYSMALLOc(nb, av) INTERNAL_SIZE_T nb; mstate av;
 
 
 
-
 /*
   sYSTRIm is an inverse of sorts to sYSMALLOc.  It gives memory back
   to the system (via negative arguments to sbrk) if there is unused
@@ -2554,6 +2569,14 @@ struct mallinfo mALLINFo()
   ------------------------------ malloc_stats ------------------------------
 */
 
+UInt
+Yap_givemallinfo(void)
+{
+  struct mallinfo mi = mALLINFo();
+  return mi.uordblks;
+}
+
+
 void mSTATs(void)
 {
   struct mallinfo mi = mALLINFo();
@@ -2901,11 +2924,12 @@ void
 Yap_initdlmalloc(void)
 {
   HeapTop = (ADDR)ALIGN_SIZE(HeapTop,16);
+  Yap_NOfMemoryHoles = 0;
   Yap_av = (struct malloc_state *)HeapTop;
   memset((void *)Yap_av, 0, sizeof(struct malloc_state));
   HeapTop += sizeof(struct malloc_state);
   HeapTop = (ADDR)ALIGN_SIZE(HeapTop,2*SIZEOF_LONG_LONG_INT);
-  HeapMax = HeapUsed = HeapTop-Yap_HeapBase;
+  HeapMax = HeapTop-Yap_HeapBase;
 }
 
 void Yap_RestoreDLMalloc(void)
