@@ -67,12 +67,12 @@ legal_env (CELL *ep)
     return (FALSE);
   ps = *((CELL *) (Addr (cp) - CellSize));
   pe = (PredEntry *) (ps - sizeof (OPREG) - sizeof (Prop));
-  READ_LOCK(pe->PRWLock);
+  LOCK(pe->PELock);
   if (!ONHEAP (pe) || Unsigned (pe) & 3 || pe->KindOfPE & 0xff00) {
-    READ_UNLOCK(pe->PRWLock);
+    UNLOCK(pe->PELock);
     return (FALSE);
   }
-  READ_UNLOCK(pe->PRWLock);
+  UNLOCK(pe->PELock);
   return (TRUE);
 }
 
@@ -100,9 +100,9 @@ DumpActiveGoals (void)
       pe = EnvPreg(cp);
       if (!ONHEAP (pe) || Unsigned (pe) & (sizeof(CELL)-1))
 	break;
-      READ_LOCK(pe->PRWLock);
+      LOCK(pe->PELock);
       if (pe->KindOfPE & 0xff00) {
-	READ_UNLOCK(pe->PRWLock);
+	UNLOCK(pe->PELock);
 	break;
       }
       if (pe->PredFlags & (CompiledPredFlag | DynamicPredFlag))
@@ -110,7 +110,7 @@ DumpActiveGoals (void)
 	  Functor f;
 	  Term mod = TermProlog;
 
-	  READ_UNLOCK(pe->PRWLock);
+	  UNLOCK(pe->PELock);
 	  f = pe->FunctorOfPred;
 	  if (pe->KindOfPE && hidden (NameOfFunctor (f)))
 	    goto next;
@@ -128,7 +128,7 @@ DumpActiveGoals (void)
 	  }
 	  Yap_DebugPutc (Yap_c_error_stream,'\n');
 	} else {
-	  READ_UNLOCK(pe->PRWLock);
+	  UNLOCK(pe->PELock);
 	}
     next:
       ep = (CELL *) ep[E_E];
@@ -142,7 +142,7 @@ DumpActiveGoals (void)
       if (!ONLOCAL (b_ptr) || b_ptr->cp_b == NULL)
 	break;
       pe = Yap_PredForChoicePt(b_ptr);
-      READ_LOCK(pe->PRWLock);
+      LOCK(pe->PELock);
       {
 	Functor f;
 	Term mod = PROLOG_MODULE;
@@ -168,7 +168,7 @@ DumpActiveGoals (void)
 	}
 	Yap_DebugPutc (Yap_c_error_stream,'\n');
       }
-      READ_UNLOCK(pe->PRWLock);
+      UNLOCK(pe->PELock);
       b_ptr = b_ptr->cp_b;
     }
 }
@@ -269,7 +269,7 @@ dump_stack(void)
   
 #if DEBUG
   fprintf(stderr,"%% YAP regs: P=%p, CP=%p, ASP=%p, H=%p, TR=%p, HeapTop=%p\n",P,CP,ASP,H,TR,HeapTop);
-  fprintf(stderr,"%% YAP mode: %x\n",Yap_PrologMode);
+  fprintf(stderr,"%% YAP mode: %ux\n",(unsigned int)Yap_PrologMode);
   if (Yap_ErrorMessage)
     fprintf(stderr,"%% YAP_ErrorMessage: %s\n",Yap_ErrorMessage);
 #endif
@@ -281,8 +281,13 @@ dump_stack(void)
 #if !USE_SYSTEM_MALLOC
     fprintf (stderr,"%ldKB of Code Space (%p--%p)\n",(long int)((CELL)HeapTop-(CELL)Yap_HeapBase)/1024,Yap_HeapBase,HeapTop); 
 #if USE_DL_MALLOC 
-    if (Yap_hole_start) {
-      fprintf (stderr,"  Last hole: %p--%p\n", Yap_hole_start, Yap_hole_end);
+    if (Yap_NOfMemoryHoles) {
+      UInt i;
+
+      for (i=0; i < Yap_NOfMemoryHoles; i++)
+	fprintf(stderr,"  Current hole: %p--%p\n", 
+		Yap_MemoryHoles[i].start,
+		Yap_MemoryHoles[i].end);
     }
 #endif
 #endif
@@ -372,6 +377,8 @@ Yap_Error(yap_error_number type, Term where, char *format,...)
   int psize = YAP_BUF_SIZE;
 
   Yap_Error_TYPE = YAP_NO_ERROR;
+  if (where == 0L)
+    where = TermNil;
 #if DEBUG_STRICT
   if (Yap_heap_regs && !(Yap_PrologMode & BootMode)) 
     fprintf(stderr,"***** Processing Error %d (%lx,%x) %s***\n", type, (unsigned long int)ActiveSignals,Yap_PrologMode,format);
@@ -851,6 +858,21 @@ Yap_Error(yap_error_number type, Term where, char *format,...)
       serious = TRUE;
     }
     break;
+  case EXISTENCE_ERROR_KEY:
+    {
+      int i;
+      Term ti[2];
+
+      i = strlen(tmpbuf);
+      ti[0] = MkAtomTerm(Yap_LookupAtom("key"));
+      ti[1] = where;
+      nt[0] = Yap_MkApplTerm(Yap_MkFunctor(Yap_LookupAtom("existence_error"),2), 2, ti);
+      tp = tmpbuf+i;
+      psize -= i;
+      fun = Yap_MkFunctor(Yap_LookupAtom("error"),2);
+      serious = TRUE;
+    }
+    break;
   case EXISTENCE_ERROR_STREAM:
     {
       int i;
@@ -1321,6 +1343,20 @@ Yap_Error(yap_error_number type, Term where, char *format,...)
       serious = TRUE;
     }
     break;
+  case RESOURCE_ERROR_MAX_THREADS:
+    {
+      int i;
+      Term ti[1];
+
+      i = strlen(tmpbuf);
+      ti[0] = MkAtomTerm(Yap_LookupAtom("threads"));
+      nt[0] = Yap_MkApplTerm(Yap_MkFunctor(Yap_LookupAtom("resource_error"),1), 1, ti);
+      tp = tmpbuf+i;
+      psize -= i;
+      fun = Yap_MkFunctor(Yap_LookupAtom("error"),2);
+      serious = TRUE;
+    }
+    break;
   case SYNTAX_ERROR:
     {
       int i;
