@@ -23,8 +23,8 @@ static char     SccsId[] = "@(#)cdmgr.c	1.1 05/02/98";
 #ifdef CUT_C
 #include "cut_c.h"
 #endif
-#if defined CUT_C && (defined MYDDAS_ODBC || defined MYDDAS_MYSQL)
-#include "myddas_util.h"
+#if defined MYDDAS_ODBC || defined MYDDAS_MYSQL
+#include "myddas.h"
 #endif
 
 STATIC_PROTO(Int  CallPredicate, (PredEntry *, choiceptr, yamop *));
@@ -42,7 +42,7 @@ cp_as_integer(choiceptr cp)
 static choiceptr
 cp_from_integer(Term cpt)
 {
-  return (choiceptr)(LCL0-(CELL *)IntegerOfTerm(cpt));
+  return (choiceptr)(LCL0-IntegerOfTerm(cpt));
 }
 
 Term
@@ -57,12 +57,11 @@ CallPredicate(PredEntry *pen, choiceptr cut_pt, yamop *code) {
   if (Yap_do_low_level_trace)
     low_level_trace(enter_pred,pen,XREGS+1);
 #endif	/* LOW_LEVEL_TRACE */
-  READ_LOCK(pen->PRWLock);
 #ifdef DEPTH_LIMIT
   if (DEPTH <= MkIntTerm(1)) {/* I assume Module==0 is prolog */
     if (pen->ModuleOfPred) {
       if (DEPTH == MkIntTerm(0)) {
-	READ_UNLOCK(pen->PRWLock);
+	UNLOCK(pen->PELock);
 	return FALSE;
       }
       else DEPTH = RESET_DEPTH();
@@ -73,7 +72,6 @@ CallPredicate(PredEntry *pen, choiceptr cut_pt, yamop *code) {
   CP = P;
   P = code;
   /* vsc: increment reduction counter at meta-call entry */
-  READ_UNLOCK(pen->PRWLock);
   if (pen->PredFlags & ProfiledPredFlag) {
     LOCK(pen->StatisticsForPred.lock);
     pen->StatisticsForPred.NOfEntries++;
@@ -82,7 +80,7 @@ CallPredicate(PredEntry *pen, choiceptr cut_pt, yamop *code) {
   ENV = YENV;
   YENV = ASP;
   YENV[E_CB] = (CELL) cut_pt;
-  return (TRUE);
+  return TRUE;
 }
 
 inline static Int
@@ -137,6 +135,35 @@ p_save_cp(void)
   return(TRUE);
 }
 
+static Int
+p_save_env_b(void)
+{
+  Term t = Deref(ARG1);
+  Term td;
+#if SHADOW_HB
+  register CELL *HBREG = HB;
+#endif
+  if (!IsVarTerm(t)) return(FALSE);
+  td = cp_as_integer((choiceptr)YENV[E_CB]);
+  BIND((CELL *)t,td,bind_save_cp);
+#ifdef COROUTINING
+  DO_TRAIL(CellPtr(t), td);
+  if (CellPtr(t) < H0) Yap_WakeUp((CELL *)t);
+ bind_save_cp:
+#endif
+  return(TRUE);
+}
+
+static Int
+p_trail_suspension_marker(void)
+{
+  Term t = Deref(ARG1);
+  
+  TrailTerm(TR) = AbsPair((CELL*)t);
+  TR++;
+  return TRUE;
+}
+
 inline static Int
 do_execute(Term t, Term mod)
 {
@@ -181,7 +208,7 @@ do_execute(Term t, Term mod)
 	  goto restart_exec;
 	}
       } else {
-	return(CallMetaCall(mod));
+	return CallMetaCall(mod);
       }
     }
     /* now let us do what we wanted to do from the beginning !! */
@@ -197,6 +224,8 @@ do_execute(Term t, Term mod)
       else
 	XREGS[i] = d0;
 #else
+
+
       XREGS[i] = *pt++;
 #endif
     }
@@ -233,7 +262,7 @@ copy_execn_to_heap(Functor f, CELL *pt, unsigned int n, unsigned int arity, Term
       *H++ = pt[i];
     }
     for (i=0; i< n; i++) {
-      *H++ = h0[i-n];
+      *H++ = h0[(int)(i-n)];
     }
     tf = AbsPair(h0);
   } else {
@@ -242,7 +271,7 @@ copy_execn_to_heap(Functor f, CELL *pt, unsigned int n, unsigned int arity, Term
       *H++ = pt[i];
     }
     for (i=0; i< n; i++) {
-      *H++ = h0[i-n];
+      *H++ = h0[(int)(i-n)];
     }
     tf = AbsAppl(h0);
   }
@@ -297,7 +326,6 @@ do_execute_n(Term t, Term mod, unsigned int n)
   if (IsExtensionFunctor(f)) {
     return CallError(TYPE_ERROR_CALLABLE, mod);
   }
-  arity = ArityOfFunctor(f);
     
   if (PRED_GOAL_EXPANSION_ALL) {
     LOCK(SignalLock);
@@ -337,7 +365,7 @@ do_execute_n(Term t, Term mod, unsigned int n)
   for (i = arity-n+1; i <= arity; i++,j++) {
     XREGS[i] = H[j];
   }
-  return (CallPredicate(pen, B, pen->CodeOfPred));
+  return CallPredicate(pen, B, pen->CodeOfPred);
 }
 
 static Int
@@ -355,23 +383,22 @@ EnterCreepMode(Term t, Term mod) {
   }
   PredCreep = RepPredProp(PredPropByFunc(FunctorCreep,1));
   if (mod) {
-    ARG1 = MkPairTerm(mod,ARG1);
+    ARG1 = MkPairTerm(mod,t);
   } else {
-    ARG1 = MkPairTerm(TermProlog,ARG1);
+    ARG1 = MkPairTerm(TermProlog,t);
   }
   LOCK(SignalLock);
   CreepFlag = CalculateStackGap();
   UNLOCK(SignalLock);
-  yap_flags[SPY_CREEP_FLAG] = TRUE;
   P_before_spy = P;
-  return (CallPredicate(PredCreep, B, PredCreep->CodeOfPred));
+  return CallPredicate(PredCreep, B, PredCreep->CodeOfPred);
 }
 
 static Int
 p_execute(void)
 {				/* '$execute'(Goal)	 */
   Term            t = Deref(ARG1);
-  return(do_execute(t, CurrentModule));
+  return do_execute(t, CurrentModule);
 }
 
 static void
@@ -395,7 +422,7 @@ p_execute2(void)
 {				/* '$execute'(Goal)	 */
   Term       t = Deref(ARG1);
   heap_store(Deref(ARG2));
-  return(do_execute_n(t, CurrentModule, 1));
+  return do_execute_n(t, CurrentModule, 1);
 }
 
 static Int
@@ -404,7 +431,7 @@ p_execute3(void)
   Term            t = Deref(ARG1);
   heap_store(Deref(ARG2));
   heap_store(Deref(ARG3));
-  return(do_execute_n(t, CurrentModule, 2));
+  return do_execute_n(t, CurrentModule, 2);
 }
 
 static Int
@@ -414,7 +441,123 @@ p_execute4(void)
   heap_store(Deref(ARG2));
   heap_store(Deref(ARG3));
   heap_store(Deref(ARG4));
-  return(do_execute_n(t, CurrentModule, 3));
+  return do_execute_n(t, CurrentModule, 3);
+}
+
+static Int
+p_execute5(void)
+{				/* '$execute'(Goal)	 */
+  Term            t = Deref(ARG1);
+  heap_store(Deref(ARG2));
+  heap_store(Deref(ARG3));
+  heap_store(Deref(ARG4));
+  heap_store(Deref(ARG5));
+  return do_execute_n(t, CurrentModule, 4);
+}
+
+static Int
+p_execute6(void)
+{				/* '$execute'(Goal)	 */
+  Term            t = Deref(ARG1);
+  heap_store(Deref(ARG2));
+  heap_store(Deref(ARG3));
+  heap_store(Deref(ARG4));
+  heap_store(Deref(ARG5));
+  heap_store(Deref(ARG6));
+  return do_execute_n(t, CurrentModule, 5);
+}
+
+static Int
+p_execute7(void)
+{				/* '$execute'(Goal)	 */
+  Term            t = Deref(ARG1);
+  heap_store(Deref(ARG2));
+  heap_store(Deref(ARG3));
+  heap_store(Deref(ARG4));
+  heap_store(Deref(ARG5));
+  heap_store(Deref(ARG6));
+  heap_store(Deref(ARG7));
+  return do_execute_n(t, CurrentModule, 6);
+}
+
+static Int
+p_execute8(void)
+{				/* '$execute'(Goal)	 */
+  Term            t = Deref(ARG1);
+  heap_store(Deref(ARG2));
+  heap_store(Deref(ARG3));
+  heap_store(Deref(ARG4));
+  heap_store(Deref(ARG5));
+  heap_store(Deref(ARG6));
+  heap_store(Deref(ARG7));
+  heap_store(Deref(ARG8));
+  return do_execute_n(t, CurrentModule, 7);
+}
+
+static Int
+p_execute9(void)
+{				/* '$execute'(Goal)	 */
+  Term            t = Deref(ARG1);
+  heap_store(Deref(ARG2));
+  heap_store(Deref(ARG3));
+  heap_store(Deref(ARG4));
+  heap_store(Deref(ARG5));
+  heap_store(Deref(ARG6));
+  heap_store(Deref(ARG7));
+  heap_store(Deref(ARG8));
+  heap_store(Deref(ARG9));
+  return do_execute_n(t, CurrentModule, 8);
+}
+
+static Int
+p_execute10(void)
+{				/* '$execute'(Goal)	 */
+  Term            t = Deref(ARG1);
+  heap_store(Deref(ARG2));
+  heap_store(Deref(ARG3));
+  heap_store(Deref(ARG4));
+  heap_store(Deref(ARG5));
+  heap_store(Deref(ARG6));
+  heap_store(Deref(ARG7));
+  heap_store(Deref(ARG8));
+  heap_store(Deref(ARG9));
+  heap_store(Deref(ARG10));
+  return(do_execute_n(t, CurrentModule, 9));
+}
+
+static Int
+p_execute11(void)
+{				/* '$execute'(Goal)	 */
+  Term            t = Deref(ARG1);
+  heap_store(Deref(ARG2));
+  heap_store(Deref(ARG3));
+  heap_store(Deref(ARG4));
+  heap_store(Deref(ARG5));
+  heap_store(Deref(ARG6));
+  heap_store(Deref(ARG7));
+  heap_store(Deref(ARG8));
+  heap_store(Deref(ARG9));
+  heap_store(Deref(ARG10));
+  heap_store(Deref(ARG11));
+  return(do_execute_n(t, CurrentModule, 10));
+}
+
+static Int
+p_execute12(void)
+{				/* '$execute'(Goal)	 */
+  Term            t = Deref(ARG1);
+  heap_store(Deref(ARG2));
+  heap_store(Deref(ARG3));
+  heap_store(Deref(ARG4));
+  heap_store(Deref(ARG5));
+  heap_store(Deref(ARG6));
+  heap_store(Deref(ARG7));
+  heap_store(Deref(ARG8));
+  heap_store(Deref(ARG9));
+  heap_store(Deref(ARG10));
+  heap_store(Deref(ARG11));
+  heap_store(Deref(ARG12));
+  return(do_execute_n(t, CurrentModule, 11));
 }
 
 static Int
@@ -422,7 +565,7 @@ p_execute_clause(void)
 {				/* '$execute_clause'(Goal)	 */
   Term            t = Deref(ARG1);
   Term            mod = Deref(ARG2);
-  choiceptr       cp = cp_from_integer(Deref(ARG4));
+  choiceptr       cut_cp = cp_from_integer(Deref(ARG4));
   unsigned int    arity;
   Prop            pe;
   yamop *code;
@@ -478,7 +621,7 @@ p_execute_clause(void)
   } else {
     code = Yap_ClauseFromTerm(clt)->ClCode;
   }
-  return CallPredicate(RepPredProp(pe), cp, code);
+  return CallPredicate(RepPredProp(pe), cut_cp, code);
 }
 
 static Int
@@ -601,6 +744,19 @@ p_execute_nonstop(void)
   /* call may not define new system predicates!! */
   if (RepPredProp(pe)->PredFlags & SpiedPredFlag) {
     return CallPredicate(RepPredProp(pe), B, RepPredProp(pe)->cs.p_code.TrueCodeOfPred);
+  }  else if ((RepPredProp(pe)->PredFlags & (AsmPredFlag|CPredFlag)) &&
+	      RepPredProp(pe)->OpcodeOfPred != Yap_opcode(_call_bfunc_xx)) {
+    /* USER C-Code may walk over registers */
+    if (RepPredProp(pe)->PredFlags & UserCPredFlag) {
+      save_machine_regs();
+    }
+    if (RepPredProp(pe)->PredFlags & UserCPredFlag) {
+      Int out = RepPredProp(pe)->cs.f_code();
+      restore_machine_regs();
+      return out;
+    } else {
+      return RepPredProp(pe)->cs.f_code();
+    }
   } else {
     return CallPredicate(RepPredProp(pe), B, RepPredProp(pe)->CodeOfPred);
   }
@@ -741,10 +897,6 @@ p_execute_3(void)
   Term            mod = Deref(ARG5);
   Prop            pe;
 
-  if (!IsAtomTerm(t)) {
-    Yap_Error(TYPE_ERROR_ATOM,ARG1,"call_with_args/4");
-    return(FALSE);
-  }
   if (IsAtomTerm(t)) {
     Atom            a;
     a = AtomOfTerm(t);
@@ -759,7 +911,7 @@ p_execute_3(void)
     CELL *ptr;
 
     if (IsExtensionFunctor(f)) {
-      Yap_Error(TYPE_ERROR_CALLABLE, t, "call_with_args/2");
+      Yap_Error(TYPE_ERROR_CALLABLE, t, "call_with_args/4");
       return(FALSE);
     }
     Arity = ArityOfFunctor(f);
@@ -1231,7 +1383,7 @@ p_pred_goal_expansion_on(void) {
 static int
 exec_absmi(int top)
 {
-  int lval;
+  int lval, out;
   if (top && (lval = sigsetjmp (Yap_RestartEnv, 1)) != 0) {
     switch(lval) {
     case 1:
@@ -1245,7 +1397,6 @@ exec_absmi(int top)
 	restore_B();
 	/* H is not so important, because we're gonna backtrack */
 	restore_H();
-	yap_flags[SPY_CREEP_FLAG] = 0;
 	LOCK(SignalLock);
 	CreepFlag = CalculateStackGap();
 	Yap_PrologMode = UserMode;
@@ -1275,15 +1426,15 @@ exec_absmi(int top)
   } else {
     Yap_PrologMode = UserMode;
   }
-  return(Yap_absmi(0));
+  out = Yap_absmi(0);
+  Yap_StartSlots();
+  return out;
 }
 
-static Term
-do_goal(Term t, yamop *CodeAdr, int arity, CELL *pt, int top)
-{
-  choiceptr saved_b = B;
-  Term out = 0L;
 
+static  void
+init_stack(int arity, CELL *pt, int top, choiceptr saved_b)
+{
   /* create an initial pseudo environment so that when garbage
      collection is going up in the environment chain it doesn't get
      confused */
@@ -1310,7 +1461,7 @@ do_goal(Term t, yamop *CodeAdr, int arity, CELL *pt, int top)
   B = (choiceptr)ASP;
   B--;
 #ifdef TABLING
-  if (top) {
+  if (top && GLOBAL_root_dep_fr) {
     DepFr_cons_cp(GLOBAL_root_dep_fr) = B;
   }
 #endif /* TABLING */
@@ -1329,8 +1480,19 @@ do_goal(Term t, yamop *CodeAdr, int arity, CELL *pt, int top)
   WPP = NULL;
 #endif
   YENV[E_CB] = Unsigned (B);
-  P = (yamop *) CodeAdr;
   CP = YESCODE;
+
+}
+
+static Term
+do_goal(Term t, yamop *CodeAdr, int arity, CELL *pt, int top)
+{
+  choiceptr saved_b = B;
+  Term out = 0L;
+
+  init_stack(arity, pt, top, saved_b);
+  P = (yamop *) CodeAdr;
+  S = CellPtr (RepPredProp (PredPropByFunc (Yap_MkFunctor(AtomCall, 1),0)));	/* A1 mishaps */
   S = CellPtr (RepPredProp (PredPropByFunc (Yap_MkFunctor(AtomCall, 1),0)));	/* A1 mishaps */
 
   out = exec_absmi(top);
@@ -1389,15 +1551,15 @@ Yap_execute_goal(Term t, int nargs, Term mod)
   if (pe == NIL) {
     return(CallMetaCall(mod));
   }
-  READ_LOCK(ppe->PRWLock);
+  LOCK(ppe->PELock);
   if (IsAtomTerm(t)) {
     CodeAdr = RepPredProp (pe)->CodeOfPred;
-    READ_UNLOCK(ppe->PRWLock);
+    UNLOCK(ppe->PELock);
     out = do_goal(t, CodeAdr, 0, pt, FALSE);
   } else {
     Functor f = FunctorOfTerm(t);
     CodeAdr = RepPredProp (pe)->CodeOfPred;
-    READ_UNLOCK(ppe->PRWLock);
+    UNLOCK(ppe->PELock);
     out = do_goal(t, CodeAdr, ArityOfFunctor(f), pt, FALSE);
   }
 
@@ -1534,9 +1696,9 @@ Yap_RunTopGoal(Term t)
     /* we must always start the emulator with Prolog code */
     return FALSE;
   }
-  READ_LOCK(ppe->PRWLock);
+  LOCK(ppe->PELock);
   CodeAdr = ppe->CodeOfPred;
-  READ_UNLOCK(ppe->PRWLock);
+  UNLOCK(ppe->PELock);
 #if !USE_SYSTEM_MALLOC
   if (Yap_TrailTop - HeapTop < 2048) {
     Yap_PrologMode = BootMode;
@@ -1627,25 +1789,72 @@ p_restore_regs2(void)
 #endif /* TABLING */
     B = pt0;
     HB = B->cp_h;
-    /*    trim_trail();*/
+    Yap_TrimTrail();
   }
   return(TRUE);
 }
 
 static Int
 p_clean_ifcp(void) {
+  Term t = Deref(ARG1);
+  if (IsVarTerm(t)) {
+    Yap_Error(INSTANTIATION_ERROR, t, "cut_at/1");
+    return FALSE;    
+  }
+  if (!IsIntegerTerm(t)) {
+    Yap_Error(TYPE_ERROR_INTEGER, t, "cut_at/1");
+    return FALSE;    
+  }
 #if SBA
-  choiceptr pt0 = (choiceptr)IntegerOfTerm(Deref(ARG1));
+  choiceptr pt0 = (choiceptr)IntegerOfTerm(t);
 #else
-  choiceptr pt0 = (choiceptr)(LCL0-IntOfTerm(Deref(ARG1)));
+  choiceptr pt0 = cp_from_integer(t);
 #endif
-  if (pt0 == B) {
+  if (pt0 < B) {
+    /* this should never happen */ 
+    return TRUE;
+  } else if (pt0 == B) {
     B = B->cp_b;
     HB = B->cp_h;
   } else {
     pt0->cp_ap = (yamop *)TRUSTFAILCODE;
   }
-  return(TRUE);
+  return TRUE;
+}
+
+
+
+static int disj_marker(yamop *apc) {
+  op_numbers opnum = Yap_op_from_opcode(apc->opc);
+
+  return opnum == _or_else || opnum == _or_last;
+}
+
+
+static Int
+p_cut_up_to_next_disjunction(void) {
+  choiceptr pt0 = B;
+  CELL *qenv = (CELL *)ENV[E_E];
+  
+  while (pt0 &&
+	 !( qenv == pt0->cp_env && disj_marker(pt0->cp_ap))) {
+    pt0 = pt0->cp_b;
+  }
+  if (!pt0)
+    return TRUE;
+#ifdef YAPOR
+  CUT_prune_to(pt0);
+#endif /* YAPOR */
+  /* find where to cut to */
+  if (SHOULD_CUT_UP_TO(B,pt0)) {
+    B = pt0;
+#ifdef TABLING
+    abolish_incomplete_subgoals(B);
+#endif /* TABLING */
+  }
+  HB = B->cp_h;
+  Yap_TrimTrail();
+  return TRUE;
 }
 
 
@@ -1745,6 +1954,7 @@ p_generate_pred_info(void) {
 void
 Yap_InitYaamRegs(void)
 {
+  Term h0var;
 
 #if PUSH_REGS
   /* Guarantee that after a longjmp we go back to the original abstract
@@ -1789,13 +1999,18 @@ Yap_InitYaamRegs(void)
   EX = 0L;
   /* for slots to work */
   Yap_StartSlots();
+  init_stack(0, NULL, TRUE, NULL);
+  /* the first real choice-point will also have AP=FAIL */ 
+  Yap_StartSlots();
+  GlobalArena = TermNil;
+  h0var = MkVarTerm();
 #if COROUTINING
-  RESET_VARIABLE((CELL *)Yap_GlobalBase);
-  DelayedVars = Yap_NewTimedVar(MkIntTerm(0));
+  DelayedVars = Yap_NewTimedVar(h0var);
   WokenGoals = Yap_NewTimedVar(TermNil);
-  AttsMutableList = Yap_NewTimedVar(MkIntTerm(0));
+  AttsMutableList = Yap_NewTimedVar(h0var);
+  GlobalDelayArena = TermNil;
 #endif
-  GcGeneration = Yap_NewTimedVar(MkIntTerm(0));
+  GcGeneration = Yap_NewTimedVar(h0var);
   GcCurrentPhase = 0L;
   GcPhase = Yap_NewTimedVar(MkIntTerm(GcCurrentPhase));
 #if defined(YAPOR) || defined(THREADS)
@@ -1803,11 +2018,12 @@ Yap_InitYaamRegs(void)
   WPP = NULL;
   PREG_ADDR = NULL;
 #endif
+  Yap_PreAllocCodeSpace();
 #ifdef CUT_C
   cut_c_initialize();
 #endif
-#if defined CUT_C && (defined MYDDAS_ODBC || defined MYDDAS_MYSQL)
-  Yap_REGS.MYDDAS_GLOBAL_POINTER = myddas_util_initialize_myddas(); 
+#if defined MYDDAS_MYSQL || defined MYDDAS_ODBC
+  Yap_REGS.MYDDAS_GLOBAL_POINTER = NULL;
 #endif
 }
 
@@ -1822,12 +2038,22 @@ p_uncaught_throw(void)
 void 
 Yap_InitExecFs(void)
 {
+  Term cm = CurrentModule;
   Yap_InitComma();
   Yap_InitCPred("$execute", 1, p_execute, HiddenPredFlag);
   Yap_InitCPred("$execute", 2, p_execute2, HiddenPredFlag);
   Yap_InitCPred("$execute", 3, p_execute3, HiddenPredFlag);
   Yap_InitCPred("$execute", 4, p_execute4, HiddenPredFlag);
+  Yap_InitCPred("$execute", 5, p_execute5, HiddenPredFlag);
+  Yap_InitCPred("$execute", 6, p_execute6, HiddenPredFlag);
+  Yap_InitCPred("$execute", 7, p_execute7, HiddenPredFlag);
+  Yap_InitCPred("$execute", 8, p_execute8, HiddenPredFlag);
+  Yap_InitCPred("$execute", 9, p_execute9, HiddenPredFlag);
+  Yap_InitCPred("$execute", 10, p_execute10, HiddenPredFlag);
+  Yap_InitCPred("$execute", 11, p_execute11, HiddenPredFlag);
+  Yap_InitCPred("$execute", 12, p_execute12, HiddenPredFlag);
   Yap_InitCPred("$execute_in_mod", 2, p_execute_in_mod, HiddenPredFlag);
+  Yap_InitCPred("$execute_wo_mod", 2, p_execute_in_mod, HiddenPredFlag);
   Yap_InitCPred("$call_with_args", 2, p_execute_0, HiddenPredFlag);
   Yap_InitCPred("$call_with_args", 3, p_execute_1, HiddenPredFlag);
   Yap_InitCPred("$call_with_args", 4, p_execute_2, HiddenPredFlag);
@@ -1845,11 +2071,18 @@ Yap_InitExecFs(void)
   Yap_InitCPred("$execute0", 2, p_execute0, HiddenPredFlag);
   Yap_InitCPred("$execute_nonstop", 2, p_execute_nonstop, HiddenPredFlag);
   Yap_InitCPred("$execute_clause", 4, p_execute_clause, HiddenPredFlag);
-  Yap_InitCPred("$save_current_choice_point", 1, p_save_cp, HiddenPredFlag);
+  CurrentModule = HACKS_MODULE;
+  Yap_InitCPred("current_choice_point", 1, p_save_cp, HiddenPredFlag);
+  Yap_InitCPred("current_choicepoint", 1, p_save_cp, HiddenPredFlag);
+  Yap_InitCPred("env_choice_point", 1, p_save_env_b, HiddenPredFlag);
+  Yap_InitCPred("trail_suspension_marker", 1, p_trail_suspension_marker, HiddenPredFlag);
+  Yap_InitCPred("cut_at", 1, p_clean_ifcp, SafePredFlag);
+  CurrentModule = cm;
   Yap_InitCPred("$pred_goal_expansion_on", 0, p_pred_goal_expansion_on, SafePredFlag|HiddenPredFlag);
   Yap_InitCPred("$restore_regs", 1, p_restore_regs, SafePredFlag|HiddenPredFlag);
   Yap_InitCPred("$restore_regs", 2, p_restore_regs2, SafePredFlag|HiddenPredFlag);
   Yap_InitCPred("$clean_ifcp", 1, p_clean_ifcp, SafePredFlag|HiddenPredFlag);
+  Yap_InitCPred("qpack_clean_up_to_disjunction", 0, p_cut_up_to_next_disjunction, SafePredFlag);
   Yap_InitCPred("$jump_env_and_store_ball", 1, p_jump_env, HiddenPredFlag);
   Yap_InitCPred("$generate_pred_info", 4, p_generate_pred_info, HiddenPredFlag);
   Yap_InitCPred("$uncaught_throw", 0, p_uncaught_throw, HiddenPredFlag);
