@@ -11,8 +11,27 @@
 * File:		gprof.c							 *
 * comments:	Interrupt Driven Profiler				 *
 *									 *
-* Last rev:     $Date: 2006/02/01 13:28:56 $,$Author: vsc $						 *
+* Last rev:     $Date: 2008/03/26 14:37:07 $,$Author: vsc $						 *
 * $Log: gprof.c,v $
+* Revision 1.10  2008/03/26 14:37:07  vsc
+* more icc fixes
+*
+* Revision 1.9  2007/10/08 23:02:15  vsc
+* minor fixes
+*
+* Revision 1.8  2007/04/10 22:13:20  vsc
+* fix max modules limitation
+*
+* Revision 1.7  2006/08/22 16:12:45  vsc
+* global variables
+*
+* Revision 1.6  2006/08/07 18:51:44  vsc
+* fix garbage collector not to try to garbage collect when we ask for large
+* chunks of stack in a single go.
+*
+* Revision 1.5  2006/04/27 20:58:59  rslopes
+* fix do profiler offline.
+*
 * Revision 1.4  2006/02/01 13:28:56  vsc
 * bignum support fixes
 *
@@ -102,6 +121,8 @@ RBTreeCreate(void) {
   return temp;
 }
 
+/* This is code originally written by Emin Martinian */
+
 /***********************************************************************/
 /*  FUNCTION:  LeftRotate */
 /**/
@@ -371,8 +392,11 @@ RBExactQuery(yamop* q) {
 
 static rb_red_blk_node*
 RBLookup(yamop *entry) {
-  rb_red_blk_node *current = ProfilerRoot->left;
+  rb_red_blk_node *current;
 
+  if (!ProfilerRoot)
+    return NULL;
+  current = ProfilerRoot->left;
   while (current != ProfilerNil) {
     if (current->key <= entry && current->lim >= entry) {
       return current;
@@ -714,8 +738,9 @@ static Int profend(void);
 static int
 showprofres(UInt type) { 
   clauseentry *pr, *t, *t2;
+  PredEntry *mype;
   UInt count=0, ProfCalls=0, InGrowHeap=0, InGrowStack=0, InGC=0, InError=0, InUnify=0, InCCall=0;
-  yamop *pc_ptr,*y; void *oldpc;
+  yamop *pc_ptr,*y; void *oldpc; 
 
   profend(); /* Make sure profiler has ended */
 
@@ -752,7 +777,7 @@ showprofres(UInt type) {
 
   t2=NULL;
   ProfCalls=0;
-  while(fscanf(FProf,"%p %p\n",&oldpc, &pc_ptr) >0){
+  while(fscanf(FProf,"%p %p %p\n",&oldpc, &pc_ptr,&mype) >0){
     if (type<10) ProfCalls++;
     
     if (oldpc!=0 && type<=2) {
@@ -837,7 +862,7 @@ showprofres(UInt type) {
       if (calls && myp->FunctorOfPred->KindOfPE==47872) {
         count+=calls;
 	printf("%p",myp);
-	printf(" %s",RepAtom(AtomOfTerm(myp->ModuleOfPred))->StrOfAE);
+	if (myp->ModuleOfPred) printf(" %s",RepAtom(AtomOfTerm(myp->ModuleOfPred))->StrOfAE);
 	printf(":%s",RepAtom(NameOfFunctor(myp->FunctorOfPred))->StrOfAE);
         if (myp->ArityOfPE) printf("/%d",myp->ArityOfPE);	
 	printf(" -> %lu (%3.1f%c)\n",(unsigned long int)calls,(float) calls*100/ProfCalls,'%');
@@ -862,7 +887,7 @@ showprofres(UInt type) {
 	}
         count+=calls;
 	//	printf("%p %p",t->pp, t->beg);
-        printf(" %s",RepAtom(AtomOfTerm(t->pp->ModuleOfPred))->StrOfAE);
+        if (t->pp->ModuleOfPred) printf(" %s",RepAtom(AtomOfTerm(t->pp->ModuleOfPred))->StrOfAE);
         printf(":%s",RepAtom(NameOfFunctor(t->pp->FunctorOfPred))->StrOfAE);
         if (t->pp->ArityOfPE) printf("/%d",t->pp->ArityOfPE);	
         printf(" -> %lu (%3.1f%c)\n",(unsigned long int)calls,(float) calls*100/ProfCalls,'%');
@@ -886,8 +911,6 @@ showprofres(UInt type) {
 
 #define TestMode (GCMode | GrowHeapMode | GrowStackMode | ErrorHandlingMode | InErrorMode | AbortMode | MallocMode)
 
-extern int Yap_absmiEND(void);
-
 static void
 prof_alrm(int signo, siginfo_t *si, void *scv)
 {
@@ -971,8 +994,9 @@ prof_alrm(int signo, siginfo_t *si, void *scv)
   }
 
   if (Yap_OffLineProfiler) {
+    fprintf(FProf,"%p %p ", oldpc, current_p);
     ProfOn = FALSE;
-    return;
+    //    return;
   }
 
   if (ProfOn) {
@@ -980,9 +1004,9 @@ prof_alrm(int signo, siginfo_t *si, void *scv)
     return;
   }
   ProfOn = TRUE;
-
   if ((node = RBLookup((yamop *)current_p))) {
     node->pcs++;
+    if (Yap_OffLineProfiler) fprintf(FProf,"%p\n", node->pe);
     ProfOn = FALSE;
     return;
   } else {
@@ -990,6 +1014,7 @@ prof_alrm(int signo, siginfo_t *si, void *scv)
     CODEADDR start, end;
 
     pp = Yap_PredEntryForCode(current_p, FIND_PRED_FROM_ANYWHERE, &start, &end);    
+    if (Yap_OffLineProfiler) fprintf(FProf,"%p\n", pp);
     if (!pp) {
 #if DEBUG
       fprintf(stderr,"lost %p, %d\n", P, Yap_op_from_opcode(P->opc));
@@ -1125,11 +1150,9 @@ profglobs(void) {
     Yap_unify(ARG6,MkIntegerTerm(ProfOns)) ;
 }
 
-static Int profinit(void)
+static Int
+do_profinit(void)
 {
-  if (ProfilerOn!=0) return (FALSE);
-  
-
   if (Yap_OffLineProfiler) {
     FPreds=fopen(profile_names(PROFPREDS_FILE),"w+"); 
     if (FPreds == NULL) return FALSE;
@@ -1147,6 +1170,16 @@ static Int profinit(void)
       }    
     }
   }
+  return TRUE;
+}
+
+static Int profinit(void)
+{
+  if (ProfilerOn!=0) return (FALSE);
+  
+  if (!do_profinit())
+    return FALSE;
+
   ProfilerOn = -1; /* Inited but not yet started */
   return(TRUE);
 }
@@ -1191,8 +1224,14 @@ static Int start_profilers(int msec)
   struct itimerval t;
   struct sigaction sa;
   
-  if (ProfilerOn!=-1) return (FALSE); /* have to go through profinit */
-
+  if (ProfilerOn!=-1) {
+    if (Yap_OffLineProfiler) {
+      return FALSE; /* have to go through profinit */
+    } else {
+      if (!do_profinit())
+	return FALSE;
+    }
+  }
   sa.sa_sigaction=prof_alrm;
   sigemptyset(&sa.sa_mask);
   sa.sa_flags=SA_SIGINFO;
@@ -1206,7 +1245,7 @@ static Int start_profilers(int msec)
   setitimer(ITIMER_PROF,&t,NULL);
 
   ProfilerOn = msec;
-  return(TRUE);
+  return TRUE;
 }
 
 
@@ -1257,7 +1296,7 @@ static Int getpredinfo(void)
   if (!pp)
     return FALSE;
   if (pp->ModuleOfPred == PROLOG_MODULE)
-    mod = ModuleName[0];
+    mod = TermProlog;
   else
     mod = pp->ModuleOfPred;
   if (pp->ModuleOfPred == IDB_MODULE) {
@@ -1305,7 +1344,7 @@ Yap_InitLowProf(void)
   ProfilerOn = FALSE;
   Yap_OffLineProfiler = FALSE;
   Yap_InitCPred("profinit",0, profinit, SafePredFlag);
-  Yap_InitCPred("profini1",1, profinit1, SafePredFlag);
+  Yap_InitCPred("profinit",1, profinit1, SafePredFlag);
   Yap_InitCPred("$proftype",1, proftype, SafePredFlag);
   Yap_InitCPred("profend" ,0, profend, SafePredFlag);
   Yap_InitCPred("profon" , 0, profon0, SafePredFlag);
