@@ -81,11 +81,11 @@ STD_PROTO(void  exit, (int));
 
 ADDR Yap_HeapBase;
 
-#ifdef THREADS
+#if defined(THREADS)
 
-struct restore_info rinfo[MAX_WORKERS];
+struct restore_info rinfo[MAX_THREADS];
 
-struct thread_globs Yap_thread_gl[MAX_WORKERS];
+struct thread_globs Yap_thread_gl[MAX_THREADS];
 
 #else
 
@@ -156,7 +156,7 @@ char            emacs_tmp[256], emacs_tmp2[256];
 
 /********* Prolog State ********************************************/
 
-prolog_exec_mode      Yap_PrologMode = BootMode;
+Int      Yap_PrologMode = BootMode;
 
 int      Yap_CritLocks = 0;
 
@@ -253,6 +253,11 @@ OpDec(int p, char *type, Atom a, Term m)
     info->KindOfPE = Ord(OpProperty);
     info->NextOfPE = RepAtom(a)->PropsOfAE;
     info->OpModule = m;
+    info->OpName = a;
+    LOCK(OpListLock);
+    info->OpNext = OpList;
+    OpList = info;
+    UNLOCK(OpListLock);
     RepAtom(a)->PropsOfAE = AbsOpProp(info);
     INIT_RWLOCK(info->OpRWLock);
     WRITE_LOCK(info->OpRWLock);
@@ -263,7 +268,8 @@ OpDec(int p, char *type, Atom a, Term m)
     WRITE_UNLOCK(ae->ARWLock);
   }
   if (i <= 3) {
-    if (info->Posfix != 0) /* there is a posfix operator */ {
+    if (yap_flags[STRICT_ISO_FLAG] && 
+	info->Posfix != 0) /* there is a posfix operator */ {
       /* ISO dictates */
       WRITE_UNLOCK(info->OpRWLock);
       Yap_Error(PERMISSION_ERROR_CREATE_OPERATOR,MkAtomTerm(a),"op/3");
@@ -271,7 +277,8 @@ OpDec(int p, char *type, Atom a, Term m)
     }
     info->Infix = p;
   } else if (i <= 5) {
-    if (info->Infix != 0) /* there is an infix operator */ {
+    if (yap_flags[STRICT_ISO_FLAG] && 
+	info->Infix != 0) /* there is an infix operator */ {
       /* ISO dictates */
       WRITE_UNLOCK(info->OpRWLock);
       Yap_Error(PERMISSION_ERROR_CREATE_OPERATOR,MkAtomTerm(a),"op/3");
@@ -331,7 +338,7 @@ Yap_GetOp(OpEntry *pp, int *prio, int fix)
     else
       n = 4, *prio = p;
   }
-  return (Yap_LookupAtom(optypes[n]));
+  return Yap_LookupAtom(optypes[n]);
 }
 
 typedef struct OPSTRUCT {
@@ -345,11 +352,13 @@ static Opdef    Ops[] = {
   {"?-", fx, 1200},
   {":-", fx, 1200},
   {"dynamic", fx, 1150},
+  {"thread_local", fx, 1150},
   {"initialization", fx, 1150},
   {"mode", fx, 1150},
   {"public", fx, 1150},
   {"multifile", fx, 1150},
   {"meta_predicate", fx, 1150},
+  {"module_transparent", fx, 1150},
   {"discontiguous", fx, 1150},
 #ifdef YAPOR
   {"sequential", fx, 1150},
@@ -364,6 +373,7 @@ static Opdef    Ops[] = {
   {"|", xfy, 1100},
   /*  {";", yf, 1100}, not allowed in ISO */
   {"->", xfy, 1050},
+  {"*->", xfy, 1050},
   {",", xfy, 1000},
   {".", xfy, 999},
   {"\\+", fy, 900},
@@ -378,12 +388,15 @@ static Opdef    Ops[] = {
   {"@>", xfx, 700},
   {"@=<", xfx, 700},
   {"@>=", xfx, 700},
+  {"=@=", xfx, 700},
+  {"\\=@=", xfx, 700},
   {"=:=", xfx, 700},
   {"=\\=", xfx, 700},
   {"<", xfx, 700},
   {">", xfx, 700},
   {"=<", xfx, 700},
   {">=", xfx, 700},
+  {"as", xfx, 600},
   {":", xfy, 600},
   {"+", yfx, 500},
   {"-", yfx, 500},
@@ -393,6 +406,7 @@ static Opdef    Ops[] = {
   {"+", fx, 500},
   {"-", fx, 500},
   {"\\", fx, 500},
+  {"rdiv", yfx, 400},
   {"*", yfx, 400},
   {"/", yfx, 400},
   {"<<", yfx, 400},
@@ -480,6 +494,7 @@ Yap_InitCPred(char *Name, unsigned long int Arity, CPredicate code, int flags)
     /* already exists */
     cl = ClauseCodeToStaticClause(pe->CodeOfPred);
     if ((flags | StandardPredFlag | CPredFlag) != pe->PredFlags) {
+      Yap_ClauseSpace -= cl->ClSize;
       Yap_FreeCodeSpace((ADDR)cl);
       cl = NULL;
     }
@@ -500,6 +515,7 @@ Yap_InitCPred(char *Name, unsigned long int Arity, CPredicate code, int flags)
 	return;
       }
     } else {
+      Yap_ClauseSpace += sz;
       cl->ClFlags = StaticMask;
       cl->ClNext = NULL;
       cl->ClSize = sz;
@@ -521,7 +537,9 @@ Yap_InitCPred(char *Name, unsigned long int Arity, CPredicate code, int flags)
     p_code->opc = Yap_opcode(_call_cpred);
   p_code->u.sla.bmap = NULL;
   p_code->u.sla.s = -Signed(RealEnvSize);
-  p_code->u.sla.sla_u.p = pe;
+  p_code->u.sla.sla_u.p =
+    p_code->u.sla.p0 =
+    pe;
   p_code = NEXTOP(p_code,sla);
   if (!(flags & SafePredFlag)) {
     p_code->opc = Yap_opcode(_deallocate);
@@ -533,7 +551,6 @@ Yap_InitCPred(char *Name, unsigned long int Arity, CPredicate code, int flags)
   p_code->opc = Yap_opcode(_Ystop);
   p_code->u.l.l = cl->ClCode;
   pe->OpcodeOfPred = pe->CodeOfPred->opc;
-  pe->ModuleOfPred = CurrentModule;
 }
 
 void 
@@ -562,6 +579,7 @@ Yap_InitCmpPred(char *Name, unsigned long int Arity, CmpPredicate cmp_code, int
 	  return;
 	}
       } else {
+	Yap_ClauseSpace += sz;
 	cl->ClFlags = StaticMask;
 	cl->ClNext = NULL;
 	cl->ClSize = sz;
@@ -604,11 +622,17 @@ Yap_InitAsmPred(char *Name,  unsigned long int Arity, int code, CPredicate def,
   pe->ModuleOfPred = CurrentModule;
   if (def != NULL) {
     yamop      *p_code = ((StaticClause *)NULL)->ClCode;
-    StaticClause     *cl = (StaticClause *)Yap_AllocCodeSpace((CELL)NEXTOP(NEXTOP(NEXTOP(((yamop *)p_code),sla),p),l)); 
+    StaticClause     *cl;
 
-    if (!cl) {
-      Yap_Error(OUT_OF_HEAP_ERROR,TermNil,"No Heap Space in InitAsmPred");
-      return;
+    if (pe->CodeOfPred == (yamop *)(&(pe->OpcodeOfPred))) {
+      cl = (StaticClause *)Yap_AllocCodeSpace((CELL)NEXTOP(NEXTOP(NEXTOP(((yamop *)p_code),sla),p),l)); 
+      if (!cl) {
+	Yap_Error(OUT_OF_HEAP_ERROR,TermNil,"No Heap Space in InitAsmPred");
+	return;
+      }
+      Yap_ClauseSpace += (CELL)NEXTOP(NEXTOP(NEXTOP(((yamop *)p_code),sla),p),l);
+    } else {
+      cl = ClauseCodeToStaticClause(pe->CodeOfPred);
     }
     cl->ClFlags = StaticMask;
     cl->ClNext = NULL;
@@ -619,7 +643,7 @@ Yap_InitAsmPred(char *Name,  unsigned long int Arity, int code, CPredicate def,
     p_code->opc = pe->OpcodeOfPred = Yap_opcode(_call_cpred);
     p_code->u.sla.bmap = NULL;
     p_code->u.sla.s = -Signed(RealEnvSize);
-    p_code->u.sla.sla_u.p = pe;
+    p_code->u.sla.sla_u.p = p_code->u.sla.p0 = pe;
     p_code = NEXTOP(p_code,sla);
     p_code->opc = Yap_opcode(_procceed);
     p_code->u.p.p = pe;
@@ -746,9 +770,11 @@ Yap_InitCPredBack(char *Name, unsigned long int Arity,
     cl->ClFlags = StaticMask;
     cl->ClNext = NULL;
 #ifdef CUT_C
+    Yap_ClauseSpace += (CELL)NEXTOP(NEXTOP(NEXTOP(NEXTOP(code,lds),lds),lds),l);
     cl->ClSize = 
       (CELL)NEXTOP(NEXTOP(NEXTOP(NEXTOP(code,lds),lds),lds),e);
 #else
+    Yap_ClauseSpace += (CELL)NEXTOP(NEXTOP(NEXTOP(code,lds),lds),l);
     cl->ClSize = 
       (CELL)NEXTOP(NEXTOP(NEXTOP(code,lds),lds),e);
 #endif
@@ -826,7 +852,6 @@ InitFlags(void)
   yap_flags[YAP_TO_CHARS_FLAG] = QUINTUS_TO_CHARS;
   yap_flags[LANGUAGE_MODE_FLAG] = 0;
   yap_flags[STRICT_ISO_FLAG] = FALSE;
-  yap_flags[SPY_CREEP_FLAG] = 0;
   yap_flags[SOURCE_MODE_FLAG] = FALSE;
   yap_flags[CHARACTER_ESCAPE_FLAG] = SICSTUS_CHARACTER_ESCAPES;
   yap_flags[WRITE_QUOTED_STRING_FLAG] = FALSE;
@@ -835,9 +860,11 @@ InitFlags(void)
 #else
   yap_flags[ALLOW_ASSERTING_STATIC_FLAG] = TRUE;
 #endif
+  yap_flags[GENERATE_DEBUG_INFO_FLAG] = TRUE;
   /* current default */
   yap_flags[INDEXING_MODE_FLAG] = INDEX_MODE_MULTI;
   yap_flags[TABLING_MODE_FLAG] = 0;
+  yap_flags[QUIET_MODE_FLAG] = FALSE;
 }
 
 static void 
@@ -868,7 +895,16 @@ InitCodes(void)
   Yap_heap_regs->attributes_module = MkAtomTerm(Yap_LookupAtom("attributes"));
   Yap_heap_regs->charsio_module = MkAtomTerm(Yap_LookupAtom("charsio"));
   Yap_heap_regs->terms_module = MkAtomTerm(Yap_LookupAtom("terms"));
+  Yap_heap_regs->system_module = MkAtomTerm(Yap_LookupAtom("system"));
+  Yap_heap_regs->readutil_module = MkAtomTerm(Yap_LookupAtom("readutil"));
+  Yap_heap_regs->hacks_module = MkAtomTerm(Yap_LookupAtom("yap_hacks"));
+  Yap_heap_regs->globals_module = MkAtomTerm(Yap_LookupAtom("nb"));
+  Yap_heap_regs->arg_module = MkAtomTerm(Yap_LookupAtom("arg"));
+  Yap_heap_regs->swi_module = MkAtomTerm(Yap_LookupAtom("swi"));
   Yap_InitModules();
+#ifdef BEAM
+  Yap_heap_regs->beam_retry_code.opc = Yap_opcode(_retry_eam);
+#endif
 #ifdef YAPOR
   Yap_heap_regs->seq_def = TRUE;
   Yap_heap_regs->getwork_code.opc = Yap_opcode(_getwork);
@@ -890,8 +926,13 @@ InitCodes(void)
 #endif /* YAPOR */
 #endif /* TABLING */
   Yap_heap_regs->expand_op_code = Yap_opcode(_expand_index);
+  INIT_LOCK(Yap_heap_regs->expand_clauses_list_lock);
+#ifdef LOW_LEVEL_TRACER
+  INIT_LOCK(Yap_heap_regs->low_level_trace_lock);
+#endif
   Yap_heap_regs->expand_clauses_first = NULL;
   Yap_heap_regs->expand_clauses_last = NULL;
+  Yap_heap_regs->expand_clauses = 0;
   Yap_heap_regs->failcode->opc = Yap_opcode(_op_fail);
   Yap_heap_regs->failcode_1 = Yap_opcode(_op_fail);
   Yap_heap_regs->failcode_2 = Yap_opcode(_op_fail);
@@ -908,16 +949,17 @@ InitCodes(void)
   Yap_heap_regs->env_for_yes_code.op = Yap_opcode(_call);
   Yap_heap_regs->env_for_yes_code.s = -Signed(RealEnvSize);
   Yap_heap_regs->env_for_yes_code.l2 = NULL;
-  Yap_heap_regs->yescode->opc = Yap_opcode(_Ystop);
+  Yap_heap_regs->yescode.opc = Yap_opcode(_Ystop);
   Yap_heap_regs->undef_op = Yap_opcode(_undef_p);
   Yap_heap_regs->index_op = Yap_opcode(_index_pred);
+  Yap_heap_regs->lockpred_op = Yap_opcode(_lock_pred);
   Yap_heap_regs->fail_op = Yap_opcode(_op_fail);
 
-  Yap_heap_regs->nocode->opc = Yap_opcode(_Nstop);
+  Yap_heap_regs->nocode.opc = Yap_opcode(_Nstop);
 
-  ((yamop *)(&Yap_heap_regs->rtrycode))->opc = Yap_opcode(_retry_and_mark);
-  ((yamop *)(&Yap_heap_regs->rtrycode))->u.ld.s = 0;
-  ((yamop *)(&Yap_heap_regs->rtrycode))->u.ld.d = NIL;
+  Yap_heap_regs->rtrycode.opc = Yap_opcode(_retry_and_mark);
+  Yap_heap_regs->rtrycode.u.ld.s = 0;
+  Yap_heap_regs->rtrycode.u.ld.d = NIL;
 #ifdef YAPOR
   INIT_YAMOP_LTT(&(Yap_heap_regs->rtrycode), 1);
 #endif /* YAPOR */
@@ -926,9 +968,14 @@ InitCodes(void)
   INIT_LOCK(Yap_heap_regs->thread_handles_lock);
   {
     int i;
-    for (i=0; i < MAX_WORKERS; i++) {
+    for (i=0; i < MAX_THREADS; i++) {
       Yap_heap_regs->thread_handle[i].in_use = FALSE;
+      Yap_heap_regs->thread_handle[i].zombie = FALSE;
       Yap_heap_regs->thread_handle[i].local_preds = NULL;
+#ifdef LOW_LEVEL_TRACER
+      Yap_heap_regs->thread_handle[i].thread_inst_count = 0LL;
+#endif
+      pthread_mutex_init(&Yap_heap_regs->thread_handle[i].tlock, NULL);
     }
   }
   Yap_heap_regs->thread_handle[0].id = 0;
@@ -942,6 +989,9 @@ InitCodes(void)
   Yap_heap_regs->n_of_threads_created = 1;
   Yap_heap_regs->threads_total_time = 0;
 #endif
+#ifdef  YAPOR
+  Yap_heap_regs->n_of_threads = 1;
+#endif
 #if defined(YAPOR) || defined(THREADS)
   INIT_LOCK(Yap_heap_regs->bgl);
   INIT_LOCK(Yap_heap_regs->free_blocks_lock);
@@ -950,20 +1000,60 @@ InitCodes(void)
   INIT_LOCK(Yap_heap_regs->dead_static_clauses_lock);
   INIT_LOCK(Yap_heap_regs->dead_mega_clauses_lock);
   INIT_LOCK(Yap_heap_regs->dead_static_indices_lock);
+  INIT_LOCK(Yap_heap_regs->op_list_lock);
   Yap_heap_regs->heap_top_owner = -1;
   {
     int i;
-    for (i=0; i < MAX_WORKERS; i++) {
+    for (i=0; i < MAX_AGENTS; i++) {
+      INIT_LOCK(Yap_heap_regs->wl[i].signal_lock);
+      Yap_heap_regs->wl[i].active_signals = 0;
       Yap_heap_regs->wl[i].scratchpad.ptr = NULL;
       Yap_heap_regs->wl[i].scratchpad.sz = SCRATCH_START_SIZE;
       Yap_heap_regs->wl[i].scratchpad.msz = SCRATCH_START_SIZE;
       Yap_heap_regs->wl[i].dynamic_arrays = NULL;
       Yap_heap_regs->wl[i].static_arrays = NULL;
+      Yap_heap_regs->wl[i].global_variables = NULL;
+      Yap_heap_regs->wl[i].global_arena = 0L;
+      Yap_heap_regs->wl[i].global_arena_overflows = 0;
+      Yap_heap_regs->wl[i].global_delay_arena = 0L;
+      Yap_heap_regs->wl[i].allow_restart = FALSE;
+      Yap_heap_regs->wl[i].tot_gc_time = 0;
+      Yap_heap_regs->wl[i].tot_gc_recovered = 0;
+      Yap_heap_regs->wl[i].gc_calls = 0;
+      Yap_heap_regs->wl[i].last_gc_time = 0;
+      Yap_heap_regs->wl[i].last_ss_time = 0;
+      Yap_heap_regs->wl[i].consultlow = (consult_obj *)Yap_AllocCodeSpace(sizeof(consult_obj)*InitialConsultCapacity);
+      if (Yap_heap_regs->wl[i].consultlow == NULL) {
+	Yap_Error(OUT_OF_HEAP_ERROR,TermNil,"No Heap Space in InitCodes");
+	return;
+      }
+      Yap_heap_regs->wl[i].consultcapacity = InitialConsultCapacity;
+      Yap_heap_regs->wl[i].consultbase = Yap_heap_regs->wl[i].consultsp =
+	Yap_heap_regs->wl[i].consultlow + Yap_heap_regs->wl[i].consultcapacity;
+      Yap_heap_regs->wl[i].Gc_timestamp = 0;
     }
   }
 #else
   Yap_heap_regs->wl.dynamic_arrays = NULL;
   Yap_heap_regs->wl.static_arrays = NULL;
+  Yap_heap_regs->wl.global_variables = NULL;
+  Yap_heap_regs->wl.global_arena = 0L;
+  Yap_heap_regs->wl.global_arena_overflows = 0;
+  Yap_heap_regs->wl.allow_restart = FALSE;
+  Yap_heap_regs->wl.global_delay_arena = 0L;
+  Yap_heap_regs->wl.tot_gc_time = 0;
+  Yap_heap_regs->wl.tot_gc_recovered = 0;
+  Yap_heap_regs->wl.gc_calls = 0;
+  Yap_heap_regs->wl.last_gc_time = 0;
+  Yap_heap_regs->wl.last_ss_time = 0;
+  Yap_heap_regs->wl.consultlow = (consult_obj *)Yap_AllocCodeSpace(sizeof(consult_obj)*InitialConsultCapacity);
+  if (Yap_heap_regs->wl.consultlow == NULL) {
+    Yap_Error(OUT_OF_HEAP_ERROR,TermNil,"No Heap Space in InitCodes");
+    return;
+  }
+  Yap_heap_regs->wl.consultcapacity = InitialConsultCapacity;
+  Yap_heap_regs->wl.consultbase = Yap_heap_regs->wl.consultsp =
+    Yap_heap_regs->wl.consultlow + Yap_heap_regs->wl.consultcapacity;
 #endif /* YAPOR */
   Yap_heap_regs->clausecode->arity = 0;
   Yap_heap_regs->clausecode->clause = NULL;
@@ -972,12 +1062,6 @@ InitCodes(void)
   Yap_heap_regs->invisiblechain.Entry = NIL;
   INIT_RWLOCK(Yap_heap_regs->invisiblechain.AERWLock);
   
-  Yap_heap_regs->consultlow = (consult_obj *)Yap_AllocCodeSpace(sizeof(consult_obj)*InitialConsultCapacity);
-  if (Yap_heap_regs->consultlow == NULL) {
-    Yap_Error(OUT_OF_HEAP_ERROR,TermNil,"No Heap Space in InitCodes");
-   return;
-  }
-  Yap_heap_regs->consultcapacity = InitialConsultCapacity;
   {
     Atom            at;
     PredEntry      *pred;
@@ -991,17 +1075,24 @@ InitCodes(void)
     at = Yap_FullLookupAtom("$spy");
     pred = RepPredProp(PredPropByFunc(Yap_MkFunctor(at, 1),0));
     Yap_heap_regs->spy_code = pred;
+    Yap_heap_regs->env_for_trustfail_code.p =
+      Yap_heap_regs->env_for_trustfail_code.p0 =
+      RepPredProp(PredPropByAtom(Yap_FullLookupAtom("fail"),PROLOG_MODULE));
   }
+#if DEBUG
+  Yap_heap_regs->new_cps = 0;
+  Yap_heap_regs->live_cps = 0;
+  Yap_heap_regs->dirty_cps = 0;
+  Yap_heap_regs->freed_cps = 0;
+#endif
   Yap_heap_regs->system_profiling = FALSE;
   Yap_heap_regs->system_call_counting = FALSE;
   Yap_heap_regs->system_pred_goal_expansion_all = FALSE;
-  Yap_heap_regs->system_pred_goal_expansion_func = FALSE;
   Yap_heap_regs->system_pred_goal_expansion_on = FALSE;
   Yap_heap_regs->update_mode = UPDATE_MODE_LOGICAL;
-  Yap_heap_regs->consultbase = Yap_heap_regs->consultsp =
-    Yap_heap_regs->consultlow + Yap_heap_regs->consultcapacity;
   Yap_heap_regs->compiler_compile_mode = 0; /* fast will be for native code */
   Yap_heap_regs->compiler_optimizer_on = TRUE;
+  Yap_heap_regs->compiler_compile_arrays = FALSE;
   Yap_heap_regs->maxdepth      = 0;
   Yap_heap_regs->maxlist       = 0;
   Yap_heap_regs->maxwriteargs  = 0;
@@ -1132,6 +1223,7 @@ InitCodes(void)
   Yap_heap_regs->functor_csult = Yap_MkFunctor(AtomCsult, 1);
   Yap_heap_regs->functor_eq = Yap_MkFunctor(AtomEq, 2);
   Yap_heap_regs->functor_execute_in_mod = Yap_MkFunctor(Yap_FullLookupAtom("$execute_in_mod"), 2);
+  Yap_heap_regs->functor_execute2_in_mod = Yap_MkFunctor(Yap_FullLookupAtom("$execute_wo_mod"), 2);
   Yap_heap_regs->functor_execute_within = Yap_MkFunctor(Yap_FullLookupAtom("$execute_within"), 1);
   Yap_heap_regs->functor_g_atom = Yap_MkFunctor(Yap_LookupAtom("atom"), 1);
   Yap_heap_regs->functor_g_atomic = Yap_MkFunctor(Yap_LookupAtom("atomic"), 1);
@@ -1151,6 +1243,7 @@ InitCodes(void)
   Yap_heap_regs->functor_mutable = Yap_MkFunctor(Yap_FullLookupAtom("$mutable_variable"),
 					 sizeof(timed_var)/sizeof(CELL));
 #endif
+  Yap_heap_regs->functor_nb_queue = Yap_MkFunctor(Yap_LookupAtom("queue"), 5);
   Yap_heap_regs->functor_not = Yap_MkFunctor(AtomNot, 1);
   Yap_heap_regs->functor_or = Yap_MkFunctor(AtomSemic, 2);
   Yap_heap_regs->functor_portray = Yap_MkFunctor(AtomPortray, 1);
@@ -1158,7 +1251,7 @@ InitCodes(void)
   Yap_heap_regs->functor_creep = Yap_MkFunctor(AtomCreep, 1);
   Yap_heap_regs->functor_static_clause = Yap_MkFunctor (Yap_FullLookupAtom("$static_clause"), 1);
   Yap_heap_regs->functor_stream = Yap_MkFunctor (AtomStream, 1);
-  Yap_heap_regs->functor_stream_pos = Yap_MkFunctor (AtomStreamPos, 3);
+  Yap_heap_regs->functor_stream_pos = Yap_MkFunctor (AtomStreamPos, 5);
   Yap_heap_regs->functor_stream_eOS = Yap_MkFunctor (Yap_LookupAtom("end_of_stream"), 1);
   Yap_heap_regs->functor_thread_run = Yap_MkFunctor (Yap_FullLookupAtom("$top_thread_goal"), 2);
   Yap_heap_regs->functor_change_module = Yap_MkFunctor (Yap_FullLookupAtom("$change_module"), 1);
@@ -1176,11 +1269,16 @@ InitCodes(void)
   Yap_heap_regs->file_aliases = NULL;
   Yap_heap_regs->foreign_code_loaded = NULL;
   Yap_heap_regs->yap_lib_dir = NULL;
+  Yap_heap_regs->agc_last_call = 0;
+  Yap_heap_regs->agc_threshold = 10000;
   Yap_heap_regs->agc_hook = NULL;
   Yap_heap_regs->parser_error_style = EXCEPTION_ON_PARSER_ERROR;
+  Yap_heap_regs->global_hold_entry = Yap_InitAtomHold();
   Yap_heap_regs->size_of_overflow  = 0;
   /* make sure no one else can use these two atoms */
   CurrentModule = 0;
+  OpList = NULL;
+  Yap_heap_regs->op_list = NULL;
   Yap_heap_regs->dead_static_clauses = NULL;
   Yap_heap_regs->dead_mega_clauses = NULL;
   Yap_heap_regs->dead_static_indices = NULL;
@@ -1193,8 +1291,9 @@ InitCodes(void)
   Yap_heap_regs->pred_meta_call = RepPredProp(PredPropByFunc(Yap_MkFunctor(Yap_heap_regs->atom_meta_call,4),PROLOG_MODULE));
   Yap_heap_regs->pred_dollar_catch = RepPredProp(PredPropByFunc(Yap_MkFunctor(Yap_FullLookupAtom("$catch"),3),PROLOG_MODULE));
   Yap_heap_regs->pred_recorded_with_key = RepPredProp(PredPropByFunc(Yap_MkFunctor(Yap_FullLookupAtom("$recorded_with_key"),3),PROLOG_MODULE));
-  Yap_heap_regs->pred_log_upd_clause = RepPredProp(PredPropByFunc(Yap_MkFunctor(Yap_FullLookupAtom("$do_log_upd_clause"),5),PROLOG_MODULE));
-  Yap_heap_regs->pred_log_upd_clause0 = RepPredProp(PredPropByFunc(Yap_MkFunctor(Yap_FullLookupAtom("$do_log_upd_clause"),4),PROLOG_MODULE));
+  Yap_heap_regs->pred_log_upd_clause = RepPredProp(PredPropByFunc(Yap_MkFunctor(Yap_FullLookupAtom("$do_log_upd_clause"),6),PROLOG_MODULE));
+  Yap_heap_regs->pred_log_upd_clause_erase = RepPredProp(PredPropByFunc(Yap_MkFunctor(Yap_FullLookupAtom("$do_log_upd_clause_erase"),6),PROLOG_MODULE));
+  Yap_heap_regs->pred_log_upd_clause0 = RepPredProp(PredPropByFunc(Yap_MkFunctor(Yap_FullLookupAtom("$do_log_upd_clause0"),6),PROLOG_MODULE));
   Yap_heap_regs->pred_static_clause = RepPredProp(PredPropByFunc(Yap_MkFunctor(Yap_FullLookupAtom("$do_static_clause"),5),PROLOG_MODULE));
   Yap_heap_regs->pred_throw = RepPredProp(PredPropByFunc(FunctorThrow,PROLOG_MODULE));
   Yap_heap_regs->pred_handle_throw = RepPredProp(PredPropByFunc(Yap_MkFunctor(Yap_FullLookupAtom("$handle_throw"),3),PROLOG_MODULE));
@@ -1213,6 +1312,9 @@ InitCodes(void)
 #endif /* YAPOR */
   Yap_heap_regs->db_erased_marker =
     (DBRef)Yap_AllocCodeSpace(sizeof(DBStruct));
+  Yap_LUClauseSpace += sizeof(DBStruct);
+  INIT_LOCK(Yap_heap_regs->dbterms_list_lock);
+  Yap_heap_regs->dbterms_list = NULL;
   Yap_heap_regs->db_erased_marker->id = FunctorDBRef;
   Yap_heap_regs->db_erased_marker->Flags = ErasedMask;
   Yap_heap_regs->db_erased_marker->Code = NULL;
@@ -1220,6 +1322,7 @@ InitCodes(void)
   Yap_heap_regs->db_erased_marker->Parent = NULL;
   Yap_heap_regs->logdb_erased_marker =
     (LogUpdClause *)Yap_AllocCodeSpace(sizeof(LogUpdClause)+(UInt)NEXTOP((yamop*)NULL,e));
+  Yap_LUClauseSpace += sizeof(LogUpdClause)+(UInt)NEXTOP((yamop*)NULL,e);
   Yap_heap_regs->logdb_erased_marker->Id = FunctorDBRef;
   Yap_heap_regs->logdb_erased_marker->ClFlags = ErasedMask|LogUpdMask;
   Yap_heap_regs->logdb_erased_marker->ClSource = NULL;
@@ -1243,7 +1346,11 @@ static void
 InitVersion(void)
 {
   Yap_PutValue(Yap_FullLookupAtom("$version_name"),
-	   MkAtomTerm(Yap_LookupAtom(YAP_VERSION)));
+	       MkAtomTerm(Yap_LookupAtom(YAP_VERSION)));
+#if defined MYDDAS_MYSQL || defined MYDDAS_ODBC
+  Yap_PutValue(Yap_FullLookupAtom("$myddas_version_name"),
+	       MkAtomTerm(Yap_LookupAtom(MYDDAS_VERSION)));
+#endif  
 }
 
 
@@ -1274,6 +1381,17 @@ Yap_InitWorkspace(int Heap, int Stack, int Trail, int max_table_size,
   /* also init memory page size, required by later functions */
   Yap_InitSysbits ();
 
+  if (Heap < MinHeapSpace)
+    Heap = MinHeapSpace;
+  Heap = AdjustPageSize(Heap * K);
+  /* sanity checking for data areas */
+  if (Trail < MinTrailSpace)
+    Trail = MinTrailSpace;
+  Trail = AdjustPageSize(Trail * K);
+  if (Stack < MinStackSpace)
+    Stack = MinStackSpace;
+  Stack = AdjustPageSize(Stack * K);
+
 #ifdef YAPOR
   worker_id = 0;
   if (n_workers > MAX_WORKERS)
@@ -1289,13 +1407,25 @@ Yap_InitWorkspace(int Heap, int Stack, int Trail, int max_table_size,
 #else
   Yap_InitMemory (Trail, Heap, Stack);
 #endif /* YAPOR */
-
+  Yap_ClauseSpace = 0;
+  Yap_IndexSpace_Tree = 0;
+  Yap_IndexSpace_EXT = 0;
+  Yap_IndexSpace_SW = 0;
+  Yap_LUClauseSpace = 0;
+  Yap_LUIndexSpace_Tree = 0;
+  Yap_LUIndexSpace_CP = 0;
+  Yap_LUIndexSpace_EXT = 0;
+  Yap_LUIndexSpace_SW = 0;
+#if USE_THREADED_CODE
+  Yap_heap_regs->op_rtable = NULL;
+#endif
 #if defined(YAPOR) || defined(TABLING)
   Yap_init_global(max_table_size, n_workers, sch_loop, delay_load);
 #endif /* YAPOR || TABLING */
 
   Yap_InitTime ();
   AtomHashTableSize = MaxHash;
+  WideAtomHashTableSize = MaxWideHash;
   HashChain = (AtomHashEntry *)Yap_AllocAtomSpace(sizeof(AtomHashEntry) * MaxHash);
   if (HashChain == NULL) {
     Yap_Error(FATAL_ERROR,MkIntTerm(0),"allocating initial atom table");
@@ -1304,7 +1434,26 @@ Yap_InitWorkspace(int Heap, int Stack, int Trail, int max_table_size,
     INIT_RWLOCK(HashChain[i].AERWLock);
     HashChain[i].Entry = NIL;
   }
+  WideHashChain = (AtomHashEntry *)Yap_AllocAtomSpace(sizeof(AtomHashEntry) * MaxWideHash);
+  if (WideHashChain == NULL) {
+    Yap_Error(FATAL_ERROR,MkIntTerm(0),"allocating initial atom table");
+  }
+  for (i = 0; i < MaxWideHash; ++i) {
+    INIT_RWLOCK(WideHashChain[i].AERWLock);
+    WideHashChain[i].Entry = NIL;
+  }
+  PredHash = (PredEntry **)Yap_AllocAtomSpace(sizeof(PredEntry **) * PredHashInitialSize);
+  PredHashTableSize = PredHashInitialSize;
+  if (PredHash == NULL) {
+    Yap_Error(FATAL_ERROR,MkIntTerm(0),"allocating initial predicate hash table");
+  }
+  for (i = 0; i < PredHashTableSize; ++i) {
+    PredHash[i] = NULL;
+  }
+  INIT_RWLOCK(PredHashRWLock);
   NOfAtoms = 0;
+  NOfWideAtoms = 0;
+  PredsInHashTable = 0;
 #if THREADS
   SF_STORE->AtFoundVar = Yap_LookupAtom(".");
   Yap_ReleaseAtom(AtomFoundVar);
@@ -1321,6 +1470,20 @@ Yap_InitWorkspace(int Heap, int Stack, int Trail, int max_table_size,
   Yap_LookupAtomWithAddress(".",&(SF_STORE->AtDot));
 #endif
   /* InitAbsmi must be done before InitCodes */
+  /* This must be done before initialising predicates */
+  Yap_heap_regs->system_pred_goal_expansion_func = FALSE;
+  for (i = 0; i <= LAST_FLAG; i++) {
+    yap_flags[i] = 0;
+  }
+#ifdef LOW_PROF
+  ProfilerOn = FALSE;
+  FPreds = NULL;
+#endif
+  ActiveSignals = 0;
+  DoingUndefp = FALSE;
+  DBErasedList = NULL;
+  DBErasedIList = NULL;
+  Yap_heap_regs->IntLUKeys = NULL;
 #ifdef MPW
   Yap_InitAbsmi(REGS, FunctorList);
 #else
@@ -1351,7 +1514,7 @@ Yap_exit (int value)
   remove("PROFILING");
 #endif
 #if defined MYDDAS_MYSQL || defined MYDDAS_ODBC
-  Yap_MyDDAS_delete_all_myddas_structs();
+  Yap_MYDDAS_delete_all_myddas_structs();
 #endif
   if (! (Yap_PrologMode & BootMode) )
     Yap_ShutdownLoadForeign();
