@@ -406,6 +406,7 @@ p_dif(void)
      */
     HBREG = H;
     B = (choiceptr) H;
+    B->cp_h = H;
     SET_BB(B);
     save_hb();
     d0 = Yap_IUnify(d0, d1);
@@ -711,9 +712,9 @@ p_functor(void)			/* functor(?,?,?) */
     *pt1++ = d0;
     d0 = AbsAppl(H);
     if (pt1+d1 > ENV - CreepFlag) {
-      if (!Yap_gc(3, ENV, P)) {
+      if (!Yap_gcl((1+d1)*sizeof(CELL), 3, ENV, P)) {
 	Yap_Error(OUT_OF_STACK_ERROR, TermNil, Yap_ErrorMessage);
-	return(FALSE);
+	return FALSE;
       }
       goto restart;
     }
@@ -800,7 +801,7 @@ p_cut_by( void)
 #endif /* TABLING */
     B = pt0;
     HB = B->cp_h;
-    /*    trim_trail();*/
+    Yap_TrimTrail();
   }
   ENDCHO(pt0);
   return(TRUE);
@@ -821,11 +822,86 @@ p_erroneous_call(void)
   return(FALSE);
 }
 
+static Int 
+init_genarg(void)
+{				/* getarg(?Atom)		 */
+  Term t0 = Deref(ARG1);
+  Term t1 = Deref(ARG2);
+  CELL *pt, *end;
+  int res;
+  UInt arity;
+
+  if (!IsVarTerm(t0)) {
+    res = p_arg();
+    if (res) {
+      cut_succeed();
+    } else {
+      cut_fail();
+    }
+  }
+  if (IsVarTerm(t1)) {
+    Yap_Error(INSTANTIATION_ERROR,t1,"genarg/3");
+    return FALSE;
+  }
+  if (IsPrimitiveTerm(t1)) {
+    Yap_Error(TYPE_ERROR_COMPOUND,t1,"genarg/3");
+    return FALSE;
+  }
+  if (IsPairTerm(t1)) {
+    pt = RepPair(t1);
+    end = RepPair(t1)+1;
+    arity = 2;
+  } else {
+    arity = ArityOfFunctor(FunctorOfTerm(t1));
+    pt = RepAppl(t1);
+    end = pt+arity;
+    pt += 1;
+  }
+  res = Yap_unify(ARG1,MkIntTerm(1)) &&
+    Yap_unify(ARG3,pt[0]);
+  if (arity == 1) {
+    if (res) {
+      cut_succeed();
+    } else {
+      cut_fail();
+    }
+  }
+  EXTRA_CBACK_ARG(3,1) = (Term)(pt+1);
+  EXTRA_CBACK_ARG(3,2) = (Term)(end);
+  EXTRA_CBACK_ARG(3,3) = MkIntegerTerm(arity);
+  return res;
+}
+
+static Int
+cont_genarg(void)
+{				/* genarg(?Atom)		 */
+  CELL *pt, *end;
+  int res;
+  UInt arity;
+
+  pt = (CELL *)EXTRA_CBACK_ARG(3,1);
+  end = (CELL *)EXTRA_CBACK_ARG(3,2);
+  arity = IntegerOfTerm(EXTRA_CBACK_ARG(3,3));
+  if (pt == end) {
+    res = Yap_unify(ARG1,MkIntegerTerm(arity)) &&
+      Yap_unify(ARG3,pt[0]);
+    if (res) {
+      cut_succeed();
+    } else {
+      cut_fail();
+    }
+  }
+  EXTRA_CBACK_ARG(3,1) = (Term)(pt+1);
+  return Yap_unify(ARG1,MkIntegerTerm(arity-(end-pt))) &&
+      Yap_unify(ARG3,pt[0]);
+}
+
+
 void 
 Yap_InitInlines(void)
 {
+  Term cm = CurrentModule;
   Yap_InitAsmPred("$$cut_by", 1, _cut_by, p_cut_by, SafePredFlag);
-
   Yap_InitAsmPred("atom", 1, _atom, p_atom, SafePredFlag);
   Yap_InitAsmPred("atomic", 1, _atomic, p_atomic, SafePredFlag);
   Yap_InitAsmPred("integer", 1, _integer, p_integer, SafePredFlag);
@@ -849,5 +925,8 @@ Yap_InitInlines(void)
   Yap_InitAsmPred("$or", 3, _or, p_erroneous_call, SafePredFlag);
   Yap_InitAsmPred("$sll", 3, _sll, p_erroneous_call, SafePredFlag);
   Yap_InitAsmPred("$slr", 3, _slr, p_erroneous_call, SafePredFlag);
+  CurrentModule = GLOBALS_MODULE;
+  Yap_InitCPredBack("genarg", 3, 3, init_genarg, cont_genarg,SafePredFlag);
+  CurrentModule = cm;
 }
 
