@@ -310,6 +310,7 @@ ParseArgs(Atom a, JMPBUFF *FailBuff)
 {
   int nargs = 0;
   Term *p, t;
+  Functor func;
 #ifdef SFUNC
   SFEntry *pe = (SFEntry *) Yap_GetAProp(a, SFProperty);
 #endif
@@ -340,6 +341,11 @@ ParseArgs(Atom a, JMPBUFF *FailBuff)
     Yap_ErrorMessage = "Stack Overflow";
     FAIL;
   }  
+  func = Yap_MkFunctor(a, nargs);
+  if (func == NULL) {
+    Yap_ErrorMessage = "Heap Overflow";
+    FAIL;
+  }
 #ifdef SFUNC
   if (pe)
     t = MkSFTerm(Yap_MkFunctor(a, SFArity), nargs, p, pe->NilValue);
@@ -349,7 +355,7 @@ ParseArgs(Atom a, JMPBUFF *FailBuff)
   if (a == AtomDBRef && nargs == 2) 
     t = MkDBRefTerm((DBRef)IntegerOfTerm(p[0]));
   else
-    t = Yap_MkApplTerm(Yap_MkFunctor(a, nargs), nargs, p);
+    t = Yap_MkApplTerm(func, nargs, p);
 #endif
   if (H > ASP-4096) {
     Yap_ErrorMessage = "Stack Overflow";
@@ -485,6 +491,10 @@ ParseTerm(int prio, JMPBUFF *FailBuff)
 	TRY(
 	  /* build appl on the heap */
 	  func = Yap_MkFunctor((Atom) t, 1);
+	  if (func == NULL) {
+	    Yap_ErrorMessage = "Heap Overflow";
+	    FAIL;
+	  }	  
 	  t = ParseTerm(oprprio, FailBuff);
 	  t = Yap_MkApplTerm(func, 1, &t);
 	  /* check for possible overflow against local stack */
@@ -517,14 +527,34 @@ ParseTerm(int prio, JMPBUFF *FailBuff)
 	t = MkAtomTerm(AtomNil);
       else if (yap_flags[YAP_DOUBLE_QUOTES_FLAG] == STRING_AS_CHARS)
 	t = Yap_StringToListOfAtoms(p);
-      else if (yap_flags[YAP_DOUBLE_QUOTES_FLAG] == STRING_AS_ATOM)
-	t = MkAtomTerm(Yap_LookupAtom(p));
-      else
+      else if (yap_flags[YAP_DOUBLE_QUOTES_FLAG] == STRING_AS_ATOM) {
+	Atom at = Yap_LookupAtom(p);
+	if (at == NIL) {
+	  Yap_ErrorMessage = "Heap Overflow";
+	  FAIL;	  
+	}
+	t = MkAtomTerm(at);
+      } else
 	t = Yap_StringToList(p);
       NextToken;
     }
   break;
 
+  case WString_tok:	/* build list on the heap */
+    {
+      Volatile wchar_t *p = (wchar_t *) Yap_tokptr->TokInfo;
+      if (*p == 0)
+	t = MkAtomTerm(AtomNil);
+      else if (yap_flags[YAP_DOUBLE_QUOTES_FLAG] == STRING_AS_CHARS)
+	t = Yap_WStringToListOfAtoms(p);
+      else if (yap_flags[YAP_DOUBLE_QUOTES_FLAG] == STRING_AS_ATOM)
+	t = MkAtomTerm(Yap_LookupWideAtom(p));
+      else
+	t = Yap_WStringToList(p);
+      NextToken;
+    }
+  break;
+
   case Var_tok:
     varinfo = (VarEntry *) (Yap_tokptr->TokInfo);
     if ((t = varinfo->VarAdr) == TermNil) {
@@ -546,8 +576,14 @@ ParseTerm(int prio, JMPBUFF *FailBuff)
       break;
     case '[':
       NextToken;
-      t = ParseList(FailBuff);
-      checkfor((Term) ']', FailBuff);
+      if (Yap_tokptr->Tok == Ord(Ponctuation_tok) &&
+	  Unsigned(Yap_tokptr->TokInfo) == ']') {
+	t = TermNil;
+	NextToken;
+      } else {
+	t = ParseList(FailBuff);
+	checkfor((Term) ']', FailBuff);
+      }
       break;
     case '{':
       NextToken;
@@ -581,12 +617,16 @@ ParseTerm(int prio, JMPBUFF *FailBuff)
     if (Yap_tokptr->Tok == Ord(Name_tok)
 	&& (opinfo = Yap_GetOpProp((Atom)(Yap_tokptr->TokInfo)))) {
       OpEntry *save_opinfo = opinfo;
-      if (IsInfixOp(opinfo, &opprio, &oplprio, &oprprio)
+      if (IsInfixOp(opinfo, &opprio, &oplprio, &oprprio) 
 	  && opprio <= prio && oplprio >= curprio) {
 	/* try parsing as infix operator */
 	Volatile int oldprio = curprio;
 	TRY3(
 	     func = Yap_MkFunctor((Atom) Yap_tokptr->TokInfo, 2);
+	     if (func == NULL) {
+	       Yap_ErrorMessage = "Heap Overflow";
+	       FAIL;
+	     }
 	     NextToken;
 	     {
 	       Term args[2];
@@ -610,7 +650,12 @@ ParseTerm(int prio, JMPBUFF *FailBuff)
       if (IsPosfixOp(opinfo, &opprio, &oplprio)
 	  && opprio <= prio && oplprio >= curprio) {
 	/* parse as posfix operator */
-	t = Yap_MkApplTerm(Yap_MkFunctor((Atom) Yap_tokptr->TokInfo, 1), 1, &t);
+	Functor func = Yap_MkFunctor((Atom) Yap_tokptr->TokInfo, 1);
+	if (func == NULL) {
+	  Yap_ErrorMessage = "Heap Overflow";
+	  FAIL;
+	}
+	t = Yap_MkApplTerm(func, 1, &t);
 	/* check for possible overflow against local stack */
 	if (H > ASP-4096) {
 	  Yap_ErrorMessage = "Stack Overflow";
@@ -653,7 +698,7 @@ ParseTerm(int prio, JMPBUFF *FailBuff)
 	continue;
       }
     }
-    if (Yap_tokptr->Tok <= Ord(String_tok))
+    if (Yap_tokptr->Tok <= Ord(WString_tok))
       FAIL;
     break;
   }
