@@ -20,6 +20,7 @@ static char     SccsId[] = "@(#)save.c	1.3 3/15/90";
 
 #if _MSC_VER || defined(__MINGW32__)
 #include <windows.h>
+#include <psapi.h>
 #endif
 #include "absmi.h"
 #include "alloc.h"
@@ -112,8 +113,7 @@ STATIC_PROTO(void  RestoreDB, (DBEntry *));
 STATIC_PROTO(void  CleanClauses, (yamop *, yamop *,PredEntry *));
 STATIC_PROTO(void  rehash, (CELL *, int, int));
 STATIC_PROTO(void  CleanCode, (PredEntry *));
-STATIC_PROTO(void  RestoreEntries, (PropEntry *));
-STATIC_PROTO(void  RestoreInvisibleAtoms, (void));
+STATIC_PROTO(void  RestoreEntries, (PropEntry *, int));
 STATIC_PROTO(void  RestoreFreeSpace, (void));
 STATIC_PROTO(void  restore_heap, (void));
 #ifdef DEBUG_RESTORE3
@@ -132,9 +132,11 @@ STATIC_PROTO(void NewFileInfo, (long, long));
 extern int      DefVol;
 #endif
 
+#ifdef _WIN32
 #if HAVE_IO_H
 #include <io.h>
 #endif
+#endif
 
 #ifdef LIGHT
 
@@ -184,7 +186,7 @@ myread(int fd, char *buff, Int len)
 {
   while (len > 0) {
     int nchars = read(fd, buff, len);
-    if (nchars < 0) {
+    if (nchars <= 0) {
       return do_system_error(PERMISSION_ERROR_INPUT_PAST_END_OF_STREAM, "bad read on saved state");
     }
     len -= nchars;
@@ -434,7 +436,7 @@ save_regs(int mode)
   putcellptr(CellPtr(Yap_HeapBase));
   putcellptr(CellPtr(HeapTop));
   /* and the space it ocuppies */
-  putout(Unsigned(HeapUsed));
+  putout(Unsigned(Yap_heap_regs->heap_used));
   /* Then the start of the free code */
   putcellptr(CellPtr(FreeBlocks));
   putcellptr(AuxSp);
@@ -471,19 +473,19 @@ save_heap(void)
 {
   int j;
   /* Then save the whole heap */
-#if defined(YAPOR) || defined(TABLING) && !defined(YAP_MEMORY_ALLOC_SCHEME)
+#if defined(YAPOR) || (defined(TABLING) && !defined(YAP_MEMORY_ALLOC_SCHEME))
   /* skip the local and global data structures */
   j = Unsigned(&GLOBAL) - Unsigned(Yap_HeapBase);
   putout(j);
   mywrite(splfild, (char *) Yap_HeapBase, j);
-  j = Unsigned(HeapTop) - Unsigned(&HashChain);
+  j = Unsigned(HeapTop) - Unsigned(REMOTE+MAX_WORKERS);
   putout(j);
-  mywrite(splfild, (char *) &HashChain, j);
+  mywrite(splfild, (char *) &(REMOTE[MAX_WORKERS]), j);
 #else
   j = Unsigned(HeapTop) - Unsigned(Yap_HeapBase);
   /* store 10 more cells because of the memory manager */
   mywrite(splfild, (char *) Yap_HeapBase, j);
-#endif /* YAPOR || TABLING */
+#endif /* YAPOR || (TABLING && !YAP_MEMORY_ALLOC_SCHEME) */
 }
 
 static void
@@ -545,6 +547,12 @@ do_save(int mode) {
   NewFileInfo('YAPS', 'MYap');
 #endif
   Term t1 = Deref(ARG1);
+
+  if (Yap_HoleSize) {
+    Yap_Error(SYSTEM_ERROR,MkAtomTerm(Yap_LookupAtom(Yap_FileNameBuf)),
+	      "restore/1: address space has holes of size %ld, cannot save", (long int)Yap_HoleSize);
+    return FALSE;
+  }
   if (!Yap_GetName(Yap_FileNameBuf, YAP_FILENAME_MAX, t1)) {
     Yap_Error(TYPE_ERROR_LIST,t1,"save/1");
     return FALSE;
@@ -611,11 +619,13 @@ check_header(CELL *info, CELL *ATrail, CELL *AStack, CELL *AHeap)
   char            pp[80];
   char msg[256];
   CELL hp_size, gb_size, lc_size, tr_size, mode;
+  int n;
 
   /* make sure we always check if there are enough bytes */
   /* skip the first line */
+  pp[0] = '\0';
   do {
-    if (read(splfild, pp, 1) < 0) {
+    if ((n = read(splfild, pp, 1)) <= 0) {
       do_system_error(PERMISSION_ERROR_INPUT_PAST_END_OF_STREAM,"failed to scan first line from saved state");
       return FAIL_RESTORE;
     }
@@ -625,7 +635,7 @@ check_header(CELL *info, CELL *ATrail, CELL *AStack, CELL *AHeap)
   {
     int count = 0, n, to_read = Unsigned(strlen(msg) + 1);
     while (count < to_read) {
-      if ((n = read(splfild, pp, to_read-count)) < 0) {
+      if ((n = read(splfild, pp, to_read-count)) <= 0) {
 	do_system_error(PERMISSION_ERROR_INPUT_PAST_END_OF_STREAM,"failed to scan version info from saved state");
 	return FAIL_RESTORE;
       }
@@ -873,6 +883,7 @@ get_regs(int flag)
     OldH = H;
     OldTR = TR;
     GDiff = Unsigned(NewGlobalBase) - Unsigned(Yap_GlobalBase);
+    GDiff0 = 0;
     LDiff = Unsigned(NewLCL0) - Unsigned(LCL0);
     TrDiff = LDiff;
     Yap_GlobalBase = (ADDR)NewGlobalBase;
@@ -899,11 +910,11 @@ get_hash(void)
 static int 
 CopyCode(void)
 {
-#if (defined(YAPOR) || defined(TABLING)) && !defined(YAP_MEMORY_ALLOC_SCHEME)
+#if defined(YAPOR) || (defined(TABLING) && !defined(YAP_MEMORY_ALLOC_SCHEME))
   /* skip the local and global data structures */
   CELL j = get_cell();
   if (Yap_ErrorMessage)
-      return -1;
+    return -1;
   if (j != Unsigned(&GLOBAL) - Unsigned(Yap_HeapBase)) {
     Yap_ErrorMessage = "code space size does not match saved state";
     return -1;
@@ -913,13 +924,13 @@ CopyCode(void)
   j = get_cell();
   if (Yap_ErrorMessage)
       return -1;
-  if (myread(splfild, (char *) &HashChain, j) < 0)
+  if (myread(splfild, (char *) &(REMOTE[MAX_WORKERS]), j) < 0)
       return -1;
 #else
   if (myread(splfild, (char *) Yap_HeapBase,
 	     (Unsigned(OldHeapTop) - Unsigned(OldHeapBase))) < 0)
     return -1;
-#endif /* YAPOR || TABLING */
+#endif /* YAPOR || (TABLING && !YAP_MEMORY_ALLOC_SCHEME) */
   return 1;
 }
 
@@ -998,9 +1009,12 @@ get_coded(int flag, OPCODE old_ops[])
 static void 
 restore_heap_regs(void)
 {
-  HeapTop = AddrAdjust(HeapTop);
-  *((YAP_SEG_SIZE *) HeapTop) = InUseFlag;
-  HeapMax = HeapUsed = OldHeapUsed;
+  if (HeapTop) {
+    HeapTop = AddrAdjust(HeapTop);
+    *((YAP_SEG_SIZE *) HeapTop) = InUseFlag;
+  }
+  HeapMax = Yap_heap_regs->heap_used = OldHeapUsed;
+  HeapLim = Yap_GlobalBase;
   restore_codes();
 }
 
@@ -1250,30 +1264,70 @@ RestoreFreeSpace(void)
 #endif
 }
 
-/* restore the atom entries which are invisible for the user */
-static void 
-RestoreInvisibleAtoms(void)
+static void
+RestoreAtomList(Atom atm)
 {
   AtomEntry      *at;
-  Atom            atm = INVISIBLECHAIN.Entry;
 
-  INVISIBLECHAIN.Entry = atm = AtomAdjust(atm);
   at = RepAtom(atm);
   if (EndOfPAEntr(at))
     return;
   do {
-#ifdef DEBUG_RESTORE2		/* useful during debug */
-    fprintf(errout, "Restoring %s\n", at->StrOfAE);
-#endif
-    at->PropsOfAE = PropAdjust(at->PropsOfAE);
-    RestoreEntries(RepProp(at->PropsOfAE));
-    atm = at->NextOfAE;
-    at->NextOfAE = atm = AtomAdjust(atm);
-    at = RepAtom(atm);
-  }
-  while (!EndOfPAEntr(at));
+    RestoreAtom(at);
+    at = RepAtom(at->NextOfAE);
+  } while (!EndOfPAEntr(at));
 }
 
+
+static void
+RestoreHashPreds(void)
+{
+  UInt size = PredHashTableSize;
+  int malloced = FALSE;
+  PredEntry **np;
+  UInt i;
+  PredEntry **oldp = PredHash = (PredEntry **)AddrAdjust((ADDR)PredHash);
+
+  np = (PredEntry **) Yap_AllocAtomSpace(sizeof(PredEntry **)*size);
+  if (!np) {
+    if (!(np = (PredEntry **) malloc(sizeof(PredEntry **)*size))) {
+	Yap_Error(FATAL_ERROR,TermNil,"Could not allocate space for pred table");
+	return;
+      }
+    malloced = TRUE;
+  }
+  for (i = 0; i < size; i++) {
+    np[i] = NULL;
+  }
+  for (i = 0; i < PredHashTableSize; i++) {
+    PredEntry *p = oldp[i];
+
+    if (p)
+      p = PredEntryAdjust(p);
+    while (p) {
+      Prop nextp;
+      UInt hsh;
+      
+      if (p->NextOfPE)
+	p->NextOfPE = PropAdjust(p->NextOfPE);
+      nextp = p->NextOfPE;
+      CleanCode(p);
+      hsh = PRED_HASH(p->FunctorOfPred, p->ModuleOfPred, size);
+      p->NextOfPE = AbsPredProp(np[hsh]);
+      np[hsh] = p;
+      p = RepPredProp(nextp);
+    }
+  }
+  for (i = 0; i < size; i++) {
+    PredHash[i] = np[i];
+  }
+  if (malloced)
+    free((ADDR)np);
+  else
+    Yap_FreeAtomSpace((ADDR)np);
+}
+
+
 /*
  * This is the really tough part, to restore the whole of the heap 
  */
@@ -1283,24 +1337,21 @@ restore_heap(void)
   AtomHashEntry *HashPtr = HashChain;
   register int    i;
   for (i = 0; i < AtomHashTableSize; ++i) {
-    Atom atm = HashPtr->Entry;
-    if (atm) {
-      AtomEntry      *at;
-      HashPtr->Entry = atm = AtomAdjust(atm);
-      at =  RepAtom(atm);
-      do {
-#ifdef DEBUG_RESTORE2			/* useful during debug */
-	fprintf(errout, "Restoring %s\n", at->StrOfAE);
-#endif
-	at->PropsOfAE = PropAdjust(at->PropsOfAE);
-	RestoreEntries(RepProp(at->PropsOfAE));
-	atm = at->NextOfAE = AtomAdjust(at->NextOfAE);
-	at = RepAtom(atm);
-      } while (!EndOfPAEntr(at));
-    }
+    HashPtr->Entry = AtomAdjust(HashPtr->Entry);
+    RestoreAtomList(HashPtr->Entry);
     HashPtr++;
   }
-  RestoreInvisibleAtoms();
+  HashPtr = WideHashChain;
+  for (i = 0; i < WideAtomHashTableSize; ++i) {
+    HashPtr->Entry = AtomAdjust(HashPtr->Entry);
+    RestoreAtomList(HashPtr->Entry);
+    HashPtr++;
+  }
+  INVISIBLECHAIN.Entry = AtomAdjust(INVISIBLECHAIN.Entry);
+  RestoreAtomList(INVISIBLECHAIN.Entry);
+  RestoreAtom(RepAtom(AtomFoundVar));
+  RestoreAtom(RepAtom(AtomFreeTerm));
+  RestoreHashPreds();
   RestoreForeignCodeStructure();
   RestoreIOStructures();
 }
@@ -1333,6 +1384,18 @@ ShowAtoms()
     }
     HashPtr++;
   }
+  HashPtr = WideHashChain;
+  for (i = 0; i < WideAtomHashTableSize; ++i) {
+    if (HashPtr->Entry != NIL) {
+      AtomEntry      *at;
+      at = RepAtom(HashPtr->Entry);
+      do {
+	fprintf(Yap_stderr,"Passei ao %s em %x\n", at->StrOfAE, at);
+	ShowEntries(RepProp(at->PropsOfAE));
+      } while (!EndOfPAEntr(at = RepAtom(at->NextOfAE)));
+    }
+    HashPtr++;
+  }
 }
 
 #endif /* DEBUG_RESTORE3 */
@@ -1347,7 +1410,7 @@ commit_to_saved_state(char *s, CELL *Astate, CELL *ATrail, CELL *AStack, CELL *A
     return(FAIL_RESTORE);
   Yap_PrologMode = BootMode;
   if (Yap_HeapBase) {
-    if (!yap_flags[HALT_AFTER_CONSULT_FLAG]) {
+    if (!yap_flags[HALT_AFTER_CONSULT_FLAG] && !yap_flags[QUIET_MODE_FLAG]) {
       Yap_TrueFileName(s,Yap_FileNameBuf2, YAP_FILENAME_MAX);
       fprintf(stderr, "%% Restoring file %s\n", Yap_FileNameBuf2);
     }
@@ -1379,9 +1442,13 @@ OpenRestore(char *inpf, char *YapLibDir, CELL *Astate, CELL *ATrail, CELL *AStac
 {
   int mode = FAIL_RESTORE;
 
-  Yap_ErrorMessage = NULL;
-  if (inpf == NULL)
-    inpf = StartUpFile;
+  //  Yap_ErrorMessage = NULL;
+  if (inpf == NULL) {
+#if _MSC_VER || defined(__MINGW32__)
+    if (!(inpf = Yap_RegistryGetString("startup")))
+#endif
+      inpf = StartUpFile;
+  }
 #if __simplescalar__
   /* does not implement getcwd */
   strncpy(Yap_FileNameBuf,yap_pwd,YAP_FILENAME_MAX);
@@ -1438,6 +1505,50 @@ OpenRestore(char *inpf, char *YapLibDir, CELL *Astate, CELL *ATrail, CELL *AStac
       }
     }
   }
+#if _MSC_VER || defined(__MINGW32__)
+  {
+    DWORD fatts;
+    int buflen;
+    char *pt;
+
+    /* try to get it from current executable */
+    if ((fatts = GetFileAttributes(Yap_FileNameBuf)) == 0xFFFFFFFFL ||
+	!(fatts & FILE_ATTRIBUTE_DIRECTORY)) {
+      /* couldn't find it where it was supposed to be,
+	 let's try using the executable */
+      if (!GetModuleFileNameEx( GetCurrentProcess(), NULL, Yap_FileNameBuf, YAP_FILENAME_MAX)) {
+	/* do nothing */
+	goto end;
+      }
+      buflen = strlen(Yap_FileNameBuf);
+      pt = Yap_FileNameBuf+strlen(Yap_FileNameBuf);
+      while (*--pt != '\\') {
+	/* skip executable */
+	if (pt == Yap_FileNameBuf) {
+	  /* do nothing */
+	  goto end;
+	}
+      }
+      while (*--pt != '\\') {
+	/* skip parent directory "bin\\" */
+	if (pt == Yap_FileNameBuf) {
+	  goto end;
+	}
+      }
+      /* now, this is a possible location for the ROOT_DIR, let's look for a share directory here */
+      pt[1] = '\0';
+      strncat(Yap_FileNameBuf,"lib/Yap/startup",YAP_FILENAME_MAX);
+    }
+    if ((splfild = open_file(Yap_FileNameBuf, O_RDONLY)) > 0) {
+      if ((mode = commit_to_saved_state(Yap_FileNameBuf,Astate,ATrail,AStack,AHeap)) != FAIL_RESTORE) {
+	Yap_ErrorMessage = NULL;
+	return(mode);
+      }
+    }
+  }
+ end:
+#endif
+  /* try to open from current directory */
   /* could not open file */
   if (Yap_ErrorMessage == NULL) {
 #if __simplescalar__
@@ -1467,6 +1578,7 @@ CloseRestore(void)
   Yap_PrologMode = UserMode;
 }
 
+#if !defined(_WIN32)
 static int 
 check_opcodes(OPCODE old_ops[])
 {
@@ -1479,11 +1591,13 @@ check_opcodes(OPCODE old_ops[])
       break;
     }
   }
-  return(have_shifted);
+  return have_shifted;
 #else
-  return(FALSE);
+  /* be conservative */
+  return TRUE;
 #endif
 }
+#endif
 
 static void 
 RestoreHeap(OPCODE old_ops[])
@@ -1492,9 +1606,16 @@ RestoreHeap(OPCODE old_ops[])
   Term mod = CurrentModule;
 
   CurrentModule = PROLOG_MODULE;
+#if defined(_WIN32)
+  /* It seems that under WIN32 opcodes may not have moved but the
+     remaining code may have bmoved */ 
+  opcodes_moved = TRUE;
+#else
   opcodes_moved = check_opcodes(old_ops);
+#endif
   /* opcodes_moved has side-effects and should be tried first */
   if (heap_moved) {
+    opcodes_moved = TRUE;
     RestoreFreeSpace();
   }
   if (heap_moved || opcodes_moved) {
@@ -1552,34 +1673,37 @@ UnmarkTrEntries(void)
   B->cp_ap = NOCODE;
   Entries = (CELL *)Yap_TrailBase;
   while ((entry = *Entries++) != (CELL)NULL) {
-    if (IsVarTerm(entry)) {
-      RESET_VARIABLE((CELL *)entry);
-    } else if (IsPairTerm(entry)) {
-      CELL *ent = CellPtoHeapAdjust(RepPair(entry));
-      register CELL flags;
+    if (!IsVarTerm(entry)) {
+      if(IsPairTerm(entry)) {
+	CELL *ent = CellPtoHeapAdjust(RepPair(entry));
+	register CELL flags;
 
-      flags = *ent;
-      ResetFlag(InUseMask, flags);
-      *ent = flags;
-      if (FlagOn(ErasedMask, flags)) {
-	if (FlagOn(DBClMask, flags)) {
-	  Yap_ErDBE(DBStructFlagsToDBStruct(ent));
-	} else {
-	  if (flags & LogUpdMask) {
-	    if (flags & IndexMask) {
-	      Yap_ErLogUpdIndex(ClauseFlagsToLogUpdIndex(ent), NULL);
-	    } else {
-	      Yap_ErLogUpdCl(ClauseFlagsToLogUpdClause(ent));
-	    }
+	flags = *ent;
+	ResetFlag(InUseMask, flags);
+	*ent = flags;
+	if (FlagOn((DirtyMask|ErasedMask), flags)) {
+	  if (FlagOn(DBClMask, flags)) {
+	    Yap_ErDBE(DBStructFlagsToDBStruct(ent));
 	  } else {
-	    Yap_ErCl(ClauseFlagsToDynamicClause(ent));
+	    if (flags & LogUpdMask) {
+	      if (flags & IndexMask) {
+		if (FlagOn(ErasedMask, flags))
+		  Yap_ErLogUpdIndex(ClauseFlagsToLogUpdIndex(ent));
+		else
+		  Yap_CleanUpIndex(ClauseFlagsToLogUpdIndex(ent));
+	      } else {
+		Yap_ErLogUpdCl(ClauseFlagsToLogUpdClause(ent));
+	      }
+	    } else {
+	      Yap_ErCl(ClauseFlagsToDynamicClause(ent));
+	    }
 	  }
 	}
-      }
 #ifdef MULTI_ASSIGNMENT_VARIABLES
-    } else /* if (IsApplTerm(d1)) */ {
-      Entries += 2;
+      } else /* if (IsApplTerm(d1)) */ {
+	Entries += 2;
 #endif
+      }
     }
   }
   B = NULL;
@@ -1599,8 +1723,8 @@ Restore(char *s, char *lib_dir)
   int restore_mode;
 
   OPCODE old_ops[_std_top+1];
-
   CELL MyTrail, MyStack, MyHeap, MyState;
+
   if ((restore_mode = OpenRestore(s, lib_dir, &MyState, &MyTrail, &MyStack, &MyHeap)) == FAIL_RESTORE)
     return(FALSE);
   Yap_ShutdownLoadForeign();
@@ -1639,6 +1763,7 @@ Restore(char *s, char *lib_dir)
   Yap_InitPlIO();
   /* reset time */
   Yap_ReInitWallTime();
+  Yap_InitSysPath();
   CloseRestore();
   if (which_save == 2) {
     Yap_unify(ARG2, MkIntTerm(0));
