@@ -50,7 +50,7 @@
 #endif
 
 /* You just can't trust some machines */
-#define my_isxdigit(C,SU,SL)	(chtype[C] == NU || (C >= 'A' && \
+#define my_isxdigit(C,SU,SL)	(chtype(C) == NU || (C >= 'A' &&	\
 				 C <= (SU)) || (C >= 'a' && C <= (SL)))
 #define my_isupper(C)	( C >= 'A' && C <= 'Z' )
 #define my_islower(C)	( C >= 'a' && C <= 'z' )
@@ -119,7 +119,6 @@ EF,
 #endif
 };
 
-#define chtype (chtype0+1)
 char *Yap_chtype = chtype0+1;
 
 /* in case there is an overflow */
@@ -128,12 +127,6 @@ typedef struct scanner_extra_alloc {
   void *filler;
 } ScannerExtraBlock;
 
-#if USE_SYSTEM_MALLOC
-#define EXPAND_TRAIL TRUE
-#else
-#define EXPAND_TRAIL FALSE
-#endif
-
 static char *
 AllocScannerMemory(unsigned int size)
 {
@@ -155,7 +148,7 @@ AllocScannerMemory(unsigned int size)
  
     if (size > alloc_size)
       alloc_size = size;
-    if(!EXPAND_TRAIL || !Yap_growtrail (alloc_size, TRUE)) {
+    if(!Yap_growtrail(alloc_size, TRUE)) {
       struct scanner_extra_alloc *ptr;
 
       if (!(ptr = (struct scanner_extra_alloc *)malloc(size+sizeof(ScannerExtraBlock)))) {
@@ -237,7 +230,7 @@ read_int_overflow(const char *s, Int base, Int val)
 #endif    
 }
 
-static unsigned int
+static wchar_t
 read_quoted_char(int *scan_nextp, int inp_stream, int (*QuotedNxtch)(int))
 {
   int ch;
@@ -261,7 +254,7 @@ read_quoted_char(int *scan_nextp, int inp_stream, int (*QuotedNxtch)(int))
     } else {
       /* sicstus */
       ch = QuotedNxtch(inp_stream);
-      if (chtype[ch] == SL) {
+      if (chtype(ch) == SL) {
 	goto restart;
       } else {
 	return 'c';
@@ -279,6 +272,46 @@ read_quoted_char(int *scan_nextp, int inp_stream, int (*QuotedNxtch)(int))
     return '\r';
   case 't':
     return '\t';
+  case 'u':
+    {
+      int i;
+      wchar_t wc='\0';
+
+      for (i=0; i< 4; i++) {
+	ch = QuotedNxtch(inp_stream);
+	if (ch>='0' && ch <= '9') {
+	  wc += (ch-'0')<<((3-i)*4);
+	} else if (ch>='a' && ch <= 'f') {
+	  wc += ((ch-'a')+10)<<((3-i)*4);
+	} else if (ch>='A' && ch <= 'F') {
+	  wc += ((ch-'A')+10)<<((3-i)*4);
+	} else {
+	  Yap_ErrorMessage = "invalid escape sequence";
+	  return 0;
+	}
+      }
+      return wc;
+    }
+  case 'U':
+    {
+      int i;
+      wchar_t wc='\0';
+
+      for (i=0; i< 8; i++) {
+	ch = QuotedNxtch(inp_stream);
+	if (ch>='0' && ch <= '9') {
+	  wc += (ch-'0')<<((7-i)*4);
+	} else if (ch>='a' && ch <= 'f') {
+	  wc += ((ch-'a')+10)<<((7-i)*4);
+	} else if (ch>='A' && ch <= 'F') {
+	  wc += ((ch-'A')+10)<<((7-i)*4);
+	} else {
+	  Yap_ErrorMessage = "invalid escape sequence";
+	  return 0;
+	}
+      }
+      return wc;
+    }
   case 'v':
     return '\v';
   case '\\':
@@ -363,11 +396,11 @@ read_quoted_char(int *scan_nextp, int inp_stream, int (*QuotedNxtch)(int))
       unsigned char so_far = 0; 
       ch = QuotedNxtch(inp_stream);
       if (my_isxdigit(ch,'f','F')) {/* hexa */
-	so_far = so_far * 16 + (chtype[ch] == NU ? ch - '0' :
+	so_far = so_far * 16 + (chtype(ch) == NU ? ch - '0' :
 				(my_isupper(ch) ? ch - 'A' : ch - 'a') + 10);
 	ch = QuotedNxtch(inp_stream);
 	if (my_isxdigit(ch,'f','F')) { /* hexa */
-	  so_far = so_far * 16 + (chtype[ch] == NU ? ch - '0' :
+	  so_far = so_far * 16 + (chtype(ch) == NU ? ch - '0' :
 				  (my_isupper(ch) ? ch - 'A' : ch - 'a') + 10);
 	  ch = QuotedNxtch(inp_stream);
 	  if (ch == '\\') {
@@ -392,11 +425,11 @@ read_quoted_char(int *scan_nextp, int inp_stream, int (*QuotedNxtch)(int))
       /* sicstus mode */
       unsigned char so_far = 0;
       ch = QuotedNxtch(inp_stream);
-      so_far = (chtype[ch] == NU ? ch - '0' :
+      so_far = (chtype(ch) == NU ? ch - '0' :
 		my_isupper(ch) ? ch - 'A' + 10 : 
 		my_islower(ch) ? ch - 'a' +10 : 0);
       ch = QuotedNxtch(inp_stream);
-      return so_far*16 + (chtype[ch] == NU ? ch - '0' :
+      return so_far*16 + (chtype(ch) == NU ? ch - '0' :
 		       my_isupper(ch) ? ch - 'A' +10 :
 		       my_islower(ch) ? ch - 'a' + 10 : 0);
     }
@@ -409,7 +442,7 @@ read_quoted_char(int *scan_nextp, int inp_stream, int (*QuotedNxtch)(int))
       return 0;
     } else {
       /* sicstus */
-      if (chtype[ch] == SL) {
+      if (chtype(ch) == SL) {
 	goto restart;
       } else {
 	return ch;
@@ -434,7 +467,7 @@ get_num(int *chp, int *chbuffp, int inp_stream, int (*Nxtch) (int), int (*Quoted
    * because of things like 00'2, 03'2 and even better 12'2, I need to
    * do this (have mercy) 
    */
-  if (chtype[ch] == NU) {
+  if (chtype(ch) == NU) {
     *sp++ = ch;
     if (--max_size == 0) {
       Yap_ErrorMessage = "Number Too Long";
@@ -456,7 +489,7 @@ get_num(int *chp, int *chbuffp, int inp_stream, int (*Nxtch) (int), int (*Quoted
     *sp++ = ch;
     ch = Nxtch(inp_stream);
     if (base == 0) {
-      Int ascii = ch;
+      wchar_t ascii = ch;
       int scan_extra = TRUE;
 
       if (ch == '\\' &&
@@ -466,21 +499,22 @@ get_num(int *chp, int *chbuffp, int inp_stream, int (*Nxtch) (int), int (*Quoted
       /* a quick way to represent ASCII */
       if (scan_extra)
 	*chp = Nxtch(inp_stream);
-      return MkIntTerm(ascii);
+      return MkIntegerTerm(ascii);
     } else if (base >= 10 && base <= 36) {
       int upper_case = 'A' - 11 + base;
       int lower_case = 'a' - 11 + base;
 
       while (my_isxdigit(ch, upper_case, lower_case)) {
 	Int oval = val;
+	int chval = (chtype(ch) == NU ? ch - '0' :
+		     (my_isupper(ch) ? ch - 'A' : ch - 'a') + 10);
 	if (--max_size == 0) {
 	  Yap_ErrorMessage = "Number Too Long";
 	  return TermNil;
 	}
 	*sp++ = ch;
-	val = val * base + (chtype[ch] == NU ? ch - '0' :
-			    (my_isupper(ch) ? ch - 'A' : ch - 'a') + 10);
-	if (oval >= val && oval != 0) /* overflow */
+	val = oval * base + chval;
+	if (oval != (val-chval)/base) /* overflow */
 	  has_overflow = (has_overflow || TRUE);
 	ch = Nxtch(inp_stream);
       }
@@ -495,15 +529,16 @@ get_num(int *chp, int *chbuffp, int inp_stream, int (*Nxtch) (int), int (*Quoted
     ch = Nxtch(inp_stream);
     while (my_isxdigit(ch, 'F', 'f')) {
       Int oval = val;
+      int chval = (chtype(ch) == NU ? ch - '0' :
+		   (my_isupper(ch) ? ch - 'A' : ch - 'a') + 10);
       if (--max_size == 0) {
 	Yap_ErrorMessage = "Number Too Long";
 	return TermNil;
       }
       *sp++ = ch;
-      val = val * 16 + (chtype[ch] == NU ? ch - '0' :
-			(my_isupper(ch) ? ch - 'A' : ch - 'a') + 10);
-      if (oval >= val && oval != 0) /* overflow */
-	has_overflow = (has_overflow || TRUE);
+      val = val * 16 + chval;
+      if (oval != (val-chval)/16) /* overflow */
+	has_overflow = TRUE;
       ch = Nxtch(inp_stream);
     }
     *chp = ch;
@@ -522,9 +557,9 @@ get_num(int *chp, int *chbuffp, int inp_stream, int (*Nxtch) (int), int (*Quoted
     val = base;
     base = 10;
   }
-  while (chtype[ch] == NU) {
+  while (chtype(ch) == NU) {
     Int oval = val;
-    if (!(val == 0 && ch == '0')) {
+    if (!(val == 0 && ch == '0') || has_overflow) {
       if (--max_size == 0) {
 	Yap_ErrorMessage = "Number Too Long";
 	return (TermNil);
@@ -535,7 +570,7 @@ get_num(int *chp, int *chbuffp, int inp_stream, int (*Nxtch) (int), int (*Quoted
       return MkIntegerTerm(val);
     val = val * base + ch - '0';
     if (val/base != oval || val -oval*base != ch-'0') /* overflow */
-      has_overflow = (has_overflow || TRUE);
+      has_overflow = TRUE;
     ch = Nxtch(inp_stream);
   }
   if (might_be_float && (ch == '.' || ch == 'e' || ch == 'E')) {
@@ -545,7 +580,7 @@ get_num(int *chp, int *chbuffp, int inp_stream, int (*Nxtch) (int), int (*Quoted
 	return TermNil;
       }
       *sp++ = '.';
-      if (chtype[ch = Nxtch(inp_stream)] != NU) {
+      if (chtype(ch = Nxtch(inp_stream)) != NU) {
 	*chbuffp = '.';
 	*chp = ch;
 	*--sp = '\0';
@@ -560,7 +595,7 @@ get_num(int *chp, int *chbuffp, int inp_stream, int (*Nxtch) (int), int (*Quoted
 	}
 	*sp++ = ch;
       }
-      while (chtype[ch = Nxtch(inp_stream)] == NU);
+      while (chtype(ch = Nxtch(inp_stream)) == NU);
     }
     if (ch == 'e' || ch == 'E') {
       char *sp0 = sp;
@@ -584,7 +619,7 @@ get_num(int *chp, int *chbuffp, int inp_stream, int (*Nxtch) (int), int (*Quoted
 	cbuff = '+';
 	ch = Nxtch(inp_stream);
       }
-      if (chtype[ch] != NU) {
+      if (chtype(ch) != NU) {
 	/* error */
 	char *sp;
 	*chp = ch;
@@ -610,7 +645,7 @@ get_num(int *chp, int *chbuffp, int inp_stream, int (*Nxtch) (int), int (*Quoted
 	  return TermNil;
 	}
 	*sp++ = ch;
-      } while (chtype[ch = Nxtch(inp_stream)] == NU);
+      } while (chtype(ch = Nxtch(inp_stream)) == NU);
     }
     *sp = '\0';
     *chp = ch;
@@ -657,14 +692,14 @@ Yap_scan_num(int (*Nxtch) (int))
   } else if (ch == '+') {
     ch = Nxtch(-1);
   }
-  if (chtype[ch] != NU) {
+  if (chtype(ch) != NU) {
     Yap_clean_tokenizer(NULL, NULL, NULL);
     return TermNil;
   }
-  cherr = 0;
+  cherr = '\0';
   if (ASP-H < 1024)
     return TermNil;
-  out = get_num(&ch, &cherr, -1, Nxtch, Nxtch, ptr, 4096);
+  out = get_num(&ch, &cherr, -1, Nxtch, Nxtch, ptr, 4096); /*  */
   PopScannerMemory(ptr, 4096);
   if (sign == -1) {
     if (IsIntegerTerm(out))
@@ -675,7 +710,25 @@ Yap_scan_num(int (*Nxtch) (int))
   Yap_clean_tokenizer(NULL, NULL, NULL);
   if (Yap_ErrorMessage != NULL || ch != -1 || cherr)
     return TermNil;
-  return(out);
+  return out;
+}
+
+
+static wchar_t *
+ch_to_wide(char *base, char *charp)
+{
+  int n = charp-base, i;
+  wchar_t *nb = (wchar_t *)base;
+
+  if ((nb+n) + 1024 > (wchar_t *)AuxSp) {
+    Yap_Error_TYPE = OUT_OF_AUXSPACE_ERROR;	  
+    Yap_ErrorMessage = "Heap Overflow While Scanning: please increase code space (-h)";
+    return NULL;
+  }
+  for (i=n; i > 0; i--) {
+    nb[i-1] = base[i-1];
+  }
+  return nb+n;
 }
 
 TokEntry *
@@ -685,10 +738,12 @@ Yap_tokenizer(int inp_stream)
   enum TokenKinds kind;
   int solo_flag = TRUE;
   int ch;
-  int (*Nxtch) (int) = Stream[inp_stream].stream_getc_for_read;
-  int (*QuotedNxtch) (int) = Stream[inp_stream].stream_getc;
+  wchar_t *wcharp;
+  int (*Nxtch) (int) = Stream[inp_stream].stream_wgetc_for_read;
+  int (*QuotedNxtch) (int) = Stream[inp_stream].stream_wgetc;
 
   Yap_ErrorMessage = NULL;
+  Yap_Error_Size = 0;
   Yap_VarTable = NULL;
   Yap_AnonVarTable = NULL;
   Yap_eot_before_eof = FALSE;
@@ -696,9 +751,11 @@ Yap_tokenizer(int inp_stream)
   ScannerExtraBlocks = NULL;
   l = NULL;
   p = NULL;			/* Just to make lint happy */
+  LOCK(Stream[inp_stream].streamlock);
   ch = Nxtch(inp_stream);
   do {
-    int och, quote, isvar;
+    wchar_t och;
+    int quote, isvar;
     char *charp, *mp;
     unsigned int len;
     char *TokImage = NULL;
@@ -712,6 +769,7 @@ Yap_tokenizer(int inp_stream)
       if (p)
 	p->Tok = Ord(kind = eot_tok);
       /* serious error now */
+      UNLOCK(Stream[inp_stream].streamlock);
       return l;
     }
     if (!l)
@@ -720,16 +778,16 @@ Yap_tokenizer(int inp_stream)
       p->TokNext = t;
     p = t;
   restart:
-    while (chtype[ch] == BS) {
+    while (chtype(ch) == BS) {
       ch = Nxtch(inp_stream);
     }
     t->TokPos = GetCurInpPos(inp_stream);
 
-    switch (chtype[ch]) {
+    switch (chtype(ch)) {
 
     case CC:
-      while ((ch = Nxtch(inp_stream)) != 10 && chtype[ch] != EF);
-      if (chtype[ch] != EF) {
+      while ((ch = Nxtch(inp_stream)) != 10 && chtype(ch) != EF);
+      if (chtype(ch) != EF) {
 	/* blank space */
 	goto restart;
       } else {
@@ -745,10 +803,11 @@ Yap_tokenizer(int inp_stream)
     scan_name:
       TokImage = ((AtomEntry *) ( Yap_PreAllocCodeSpace()))->StrOfAE;
       charp = TokImage;
-      isvar = (chtype[och] != LC);
+      isvar = (chtype(och) != LC);
       *charp++ = och;
-      for (; chtype[ch] <= NU; ch = Nxtch(inp_stream)) {
+      for (; chtype(ch) <= NU; ch = Nxtch(inp_stream)) {
 	if (charp == (char *)AuxSp-1024) {
+	huge_var_error:
 	  /* huge atom or variable, we are in trouble */
 	  Yap_ErrorMessage = "Code Space Overflow due to huge atom";
 	  Yap_Error_TYPE = OUT_OF_AUXSPACE_ERROR;	  
@@ -756,10 +815,18 @@ Yap_tokenizer(int inp_stream)
 	  if (p)
 	    p->Tok = Ord(kind = eot_tok);
 	  /* serious error now */
+	  UNLOCK(Stream[inp_stream].streamlock);
 	  return l;
 	}
 	*charp++ = ch;
       }
+      while (ch == '\'' && isvar && yap_flags[VARS_CAN_HAVE_QUOTE_FLAG]) {
+	if (charp == (char *)AuxSp-1024) {
+	  goto huge_var_error;
+	}
+	*charp++ = ch;
+	ch = Nxtch(inp_stream);
+      }
       *charp++ = '\0';
       if (!isvar) {
 	/* don't do this in iso */
@@ -770,6 +837,7 @@ Yap_tokenizer(int inp_stream)
 	  if (p)
 	    t->Tok = Ord(kind = eot_tok);
 	  /* serious error now */
+	  UNLOCK(Stream[inp_stream].streamlock);
 	  return l;
 	}
 	t->TokInfo = Unsigned(ae);
@@ -786,11 +854,13 @@ Yap_tokenizer(int inp_stream)
 
     case NU:
       {
-	int cherr, cha = ch;
+	int cherr;
+	int cha = ch;
 	char *ptr;
 
 	cherr = 0;
 	if (!(ptr = AllocScannerMemory(4096))) {
+	  UNLOCK(Stream[inp_stream].streamlock);
 	  Yap_ErrorMessage = "Trail Overflow";
 	  Yap_Error_TYPE = OUT_OF_TRAIL_ERROR;	            
 	  if (p)
@@ -799,14 +869,17 @@ Yap_tokenizer(int inp_stream)
 	  return l;
 	}
 	if (ASP-H < 1024) {
+	  UNLOCK(Stream[inp_stream].streamlock);
 	  Yap_ErrorMessage = "Stack Overflow";
 	  Yap_Error_TYPE = OUT_OF_STACK_ERROR;	            
+	  Yap_Error_Size = 0L;	            
 	  if (p)
 	    p->Tok = Ord(kind = eot_tok);
 	  /* serious error now */
 	  return l;
 	}
 	if ((t->TokInfo = get_num(&cha,&cherr,inp_stream,Nxtch,QuotedNxtch,ptr,4096)) == 0L) {
+	  UNLOCK(Stream[inp_stream].streamlock);
 	  if (p)
 	    p->Tok = Ord(kind = eot_tok);
 	  /* serious error now */
@@ -820,6 +893,7 @@ Yap_tokenizer(int inp_stream)
 	  t->TokPos = GetCurInpPos(inp_stream);
 	  e = (TokEntry *) AllocScannerMemory(sizeof(TokEntry));
 	  if (e == NULL) {
+	    UNLOCK(Stream[inp_stream].streamlock);
 	    Yap_ErrorMessage = "Trail Overflow";
 	    Yap_Error_TYPE = OUT_OF_TRAIL_ERROR;	            
 	    if (p)
@@ -849,6 +923,7 @@ Yap_tokenizer(int inp_stream)
 	      t->TokPos = GetCurInpPos(inp_stream);
 	      e2 = (TokEntry *) AllocScannerMemory(sizeof(TokEntry));
 	      if (e2 == NULL) {
+		UNLOCK(Stream[inp_stream].streamlock);
 		Yap_ErrorMessage = "Trail Overflow";
 		Yap_Error_TYPE = OUT_OF_TRAIL_ERROR;	            
 		if (p)
@@ -880,6 +955,7 @@ Yap_tokenizer(int inp_stream)
 	      t->TokPos = GetCurInpPos(inp_stream);
 	      e2 = (TokEntry *) AllocScannerMemory(sizeof(TokEntry));
 	      if (e2 == NULL) {
+		UNLOCK(Stream[inp_stream].streamlock);
 		Yap_ErrorMessage = "Trail Overflow";
 		Yap_Error_TYPE = OUT_OF_TRAIL_ERROR;	            
 		t->Tok = Ord(kind = eot_tok);
@@ -909,12 +985,18 @@ Yap_tokenizer(int inp_stream)
       quote = ch;
       len = 0;
       ch = QuotedNxtch(inp_stream);
-      while (1) {
+      wcharp = NULL;
+
+      while (TRUE) {
 	if (charp + 1024 > (char *)AuxSp) {
 	  Yap_Error_TYPE = OUT_OF_AUXSPACE_ERROR;	  
 	  Yap_ErrorMessage = "Heap Overflow While Scanning: please increase code space (-h)";
 	  break;
 	}
+	if (!wcharp && ch > MAX_ISO_LATIN1){
+	  /* does not fit in ISO-LATIN */
+	  wcharp = ch_to_wide(TokImage, charp);
+	}
 	if (ch == 10  &&  yap_flags[CHARACTER_ESCAPE_FLAG] == ISO_CHARACTER_ESCAPES) {
 	  /* in ISO a new line terminates a string */
 	  Yap_ErrorMessage = "layout character \n inside quotes";
@@ -924,24 +1006,42 @@ Yap_tokenizer(int inp_stream)
 	  ch = QuotedNxtch(inp_stream);
 	  if (ch != quote)
 	    break;
-	  *charp++ = ch;
+	  if (wcharp) 
+	    *wcharp++ = ch;
+	  else
+	    *charp++ = ch;
 	  ch = QuotedNxtch(inp_stream);
 	} else if (ch == '\\' && yap_flags[CHARACTER_ESCAPE_FLAG] != CPROLOG_CHARACTER_ESCAPES) {
 	  int scan_next = TRUE;
-	  *charp++ = read_quoted_char(&scan_next, inp_stream, QuotedNxtch);
+	  if (wcharp) 
+	    *wcharp++ = read_quoted_char(&scan_next, inp_stream, QuotedNxtch);
+	  else {
+	    wchar_t next = read_quoted_char(&scan_next, inp_stream, QuotedNxtch);
+	    if (next > MAX_ISO_LATIN1){
+	      /* does not fit in ISO-LATIN */
+	      wcharp = ch_to_wide(TokImage, charp);
+	      *wcharp++ = next;
+	    } else {
+	      *charp++ = next;
+	    }
+	  }
 	  if (scan_next) {
 	    ch = QuotedNxtch(inp_stream);
 	  }
-	} else if (chtype[ch] == EF) {
+	} else if (chtype(ch) == EF && ch <= MAX_ISO_LATIN1) {
 	  Yap_ReleasePreAllocCodeSpace((CODEADDR)TokImage);
 	  t->Tok = Ord(kind = eot_tok);
 	  break;
 	} else {
-	  *charp++ = ch;
+	  if (wcharp) 
+	    *wcharp++ = ch;
+	  else
+	    *charp++ = ch;
 	  ch = QuotedNxtch(inp_stream);
 	}
 	++len;
 	if (charp > (char *)AuxSp - 1024) {
+	  UNLOCK(Stream[inp_stream].streamlock);
 	  /* Not enough space to read in the string. */
 	  Yap_Error_TYPE = OUT_OF_AUXSPACE_ERROR;	  
 	  Yap_ErrorMessage = "not enough space to read in string or quoted atom";
@@ -951,21 +1051,50 @@ Yap_tokenizer(int inp_stream)
 	  return l;
 	}
       }
-      *charp = '\0';
+      if (wcharp) {
+	*wcharp = '\0';
+      }  else  {
+	*charp = '\0';
+      }
       if (quote == '"') {
-	mp = AllocScannerMemory(len + 1);
+	if (wcharp) {
+	  mp = AllocScannerMemory(sizeof(wchar_t)*(len+1));
+	} else {
+	  mp = AllocScannerMemory(len + 1);
+	}
 	if (mp == NULL) {
+	  UNLOCK(Stream[inp_stream].streamlock);
 	  Yap_ErrorMessage = "not enough heap space to read in string or quoted atom";
 	  Yap_ReleasePreAllocCodeSpace((CODEADDR)TokImage);
 	  t->Tok = Ord(kind = eot_tok);
 	  return l;
 	}
-	strcpy(mp, TokImage);
+	if (wcharp) 
+	  wcscpy((wchar_t *)mp,(wchar_t *)TokImage);
+	else
+	  strcpy(mp, TokImage);
 	t->TokInfo = Unsigned(mp);
 	Yap_ReleasePreAllocCodeSpace((CODEADDR)TokImage);
-	t->Tok = Ord(kind = String_tok);
+	if (wcharp) {
+	  t->Tok = Ord(kind = WString_tok);
+	} else {
+	  t->Tok = Ord(kind = String_tok);
+	}
       } else {
-	t->TokInfo = Unsigned(Yap_LookupAtom(TokImage));
+	if (wcharp) {
+	  t->TokInfo = Unsigned(Yap_LookupWideAtom((wchar_t *)TokImage));
+	} else {
+	  t->TokInfo = Unsigned(Yap_LookupAtom(TokImage));
+	  if (t->TokInfo == (CELL)NIL) {
+	    Yap_Error_TYPE = OUT_OF_HEAP_ERROR;	  
+	    Yap_ErrorMessage = "Code Space Overflow";
+	    if (p)
+	      t->Tok = Ord(kind = eot_tok);
+	    /* serious error now */
+	    UNLOCK(Stream[inp_stream].streamlock);
+	    return l;
+	  }
+	}
 	Yap_ReleasePreAllocCodeSpace((CODEADDR)TokImage);
 	t->Tok = Ord(kind = Name_tok);
 	if (ch == '(')
@@ -977,32 +1106,41 @@ Yap_tokenizer(int inp_stream)
       och = ch;
       ch = Nxtch(inp_stream);
       if (och == '/' && ch == '*') {
-	while ((och != '*' || ch != '/') && chtype[ch] != EF) {
+	while ((och != '*' || ch != '/') && chtype(ch) != EF) {
 	  och = ch;
 	  ch = Nxtch(inp_stream);
 	}
-	if (chtype[ch] == EF) {
+	if (chtype(ch) == EF) {
 	  t->Tok = Ord(kind = eot_tok);
 	}
 	ch = Nxtch(inp_stream);
 	goto restart;
       }
     enter_symbol:
-      if (och == '.' && (chtype[ch] == BS || chtype[ch] == EF
-			 || chtype[ch] == CC)) {
+      if (och == '.' && (chtype(ch) == BS || chtype(ch) == EF
+			 || chtype(ch) == CC)) {
 	Yap_eot_before_eof = TRUE;
-	if (chtype[ch] == CC)
-	  while ((ch = Nxtch(inp_stream)) != 10 && chtype[ch] != EF);
+	if (chtype(ch) == CC)
+	  while ((ch = Nxtch(inp_stream)) != 10 && chtype(ch) != EF);
 	t->Tok = Ord(kind = eot_tok);
       }
       else {
 	TokImage = ((AtomEntry *) ( Yap_PreAllocCodeSpace()))->StrOfAE;
 	charp = TokImage;
 	*charp++ = och;
-	for (; chtype[ch] == SY; ch = Nxtch(inp_stream))
+	for (; chtype(ch) == SY; ch = Nxtch(inp_stream))
 	  *charp++ = ch;
 	*charp = '\0';
 	t->TokInfo = Unsigned(Yap_LookupAtom(TokImage));
+	if (t->TokInfo == (CELL)NIL) {
+	  Yap_Error_TYPE = OUT_OF_HEAP_ERROR;	  
+	  Yap_ErrorMessage = "Code Space Overflow";
+	  if (p)
+	    t->Tok = Ord(kind = eot_tok);
+	  /* serious error now */
+	  UNLOCK(Stream[inp_stream].streamlock);
+	  return l;
+	}
 	Yap_ReleasePreAllocCodeSpace((CODEADDR)TokImage);
 	t->Tok = Ord(kind = Name_tok);
 	if (ch == '(')
@@ -1025,23 +1163,13 @@ Yap_tokenizer(int inp_stream)
 
     case BK:
       och = ch;
-      do {
-	ch = Nxtch(inp_stream);
-      } while (chtype[ch] == BS);
-      if (och == '[' && ch == ']') {
-	t->TokInfo = Unsigned(AtomNil);
-	ch = Nxtch(inp_stream);
-	t->Tok = Ord(kind = Name_tok);
-	if (ch == '(')
-	  solo_flag = FALSE;
-      } else {
-	t->TokInfo = och;
-	if (t->TokInfo == '(' && !solo_flag) {
-	  t->TokInfo = 'l';
-	  solo_flag = TRUE;
-	}
-	t->Tok = Ord(kind = Ponctuation_tok);
+      ch = Nxtch(inp_stream);
+      t->TokInfo = och;
+      if (t->TokInfo == '(' && !solo_flag) {
+	t->TokInfo = 'l';
+	solo_flag = TRUE;
       }
+      t->Tok = Ord(kind = Ponctuation_tok);
       break;
 
     case EF:
@@ -1050,7 +1178,7 @@ Yap_tokenizer(int inp_stream)
 
     default:
 #ifdef DEBUG
-      fprintf(Yap_stderr, "\n++++ token: wrong char type %c %d\n", ch, chtype[ch]);
+      fprintf(Yap_stderr, "\n++++ token: wrong char type %c %d\n", ch, chtype(ch));
 #endif
       t->Tok = Ord(kind = eot_tok);
     }
@@ -1061,6 +1189,7 @@ Yap_tokenizer(int inp_stream)
       /* insert an error token to inform the system of what happened */
       TokEntry *e = (TokEntry *) AllocScannerMemory(sizeof(TokEntry));
       if (e == NULL) {
+	UNLOCK(Stream[inp_stream].streamlock);
 	Yap_ErrorMessage = "Trail Overflow";
 	Yap_Error_TYPE = OUT_OF_TRAIL_ERROR;	            
 	p->Tok = Ord(kind = eot_tok);
@@ -1076,6 +1205,7 @@ Yap_tokenizer(int inp_stream)
       p = e;
     }
   } while (kind != eot_tok);
+  UNLOCK(Stream[inp_stream].streamlock);
   return (l);
 }
 
