@@ -39,7 +39,7 @@ static char SccsId[] = "%W% %G%";
 #if HAVE_WINDOWS_H
 #include <windows.h>
 #endif
-#if HAVE_SYS_TIME_H && !defined(__MINGW32__) && !_MSC_VER
+#if HAVE_SYS_TIME_H && !_MSC_VER
 #include <sys/time.h>
 #endif
 #if HAVE_UNISTD_H
@@ -137,6 +137,27 @@ Yap_WinError(char *yap_error)
 			       (C) <= 'Z') || (C) == '_' )
 
 
+static int
+is_directory(char *FileName)
+{
+#ifdef _WIN32
+  DWORD dwAtts = GetFileAttributes(FileName);
+  if (dwAtts == INVALID_FILE_ATTRIBUTES)
+    return FALSE;
+  return (dwAtts & FILE_ATTRIBUTE_DIRECTORY);
+#elif HAVE_LSTAT 
+  struct stat buf;
+
+  if (lstat(FileName, &buf) == -1) {
+    /* return an error number */
+    return FALSE;
+  }
+  return S_ISDIR(buf.st_mode);
+#else
+  return FALSE;
+#endif
+}
+
 static int
 dir_separator (int ch)
 {
@@ -157,12 +178,67 @@ Yap_dir_separator (int ch)
   return dir_separator (ch);
 }
 
+#if _MSC_VER || defined(__MINGW32__)
+#include <psapi.h>
+
+char *libdir = NULL;
+#endif
+
 void
 Yap_InitSysPath(void) {
+#if _MSC_VER || defined(__MINGW32__)
+  {
+    char *dir;
+    if ((dir = Yap_RegistryGetString("library"))) {
+      Yap_PutValue(Yap_LookupAtom("system_library_directory"),
+		   MkAtomTerm(Yap_LookupAtom(dir)));
+      return;
+    }
+  }
+#endif
   strncpy(Yap_FileNameBuf, SHARE_DIR, YAP_FILENAME_MAX);
-#ifdef MAC
-  strncat(Yap_FileNameBuf,":", YAP_FILENAME_MAX);
-#elif ATARI || _MSC_VER || defined(__MINGW32__)
+#if _MSC_VER || defined(__MINGW32__)
+  {
+    DWORD fatts;
+    int buflen;
+    char *pt;
+
+    if ((fatts = GetFileAttributes(Yap_FileNameBuf)) == 0xFFFFFFFFL ||
+	!(fatts & FILE_ATTRIBUTE_DIRECTORY)) {
+      /* couldn't find it where it was supposed to be,
+	 let's try using the executable */
+      if (!GetModuleFileNameEx( GetCurrentProcess(), NULL, Yap_FileNameBuf, YAP_FILENAME_MAX)) {
+	Yap_Error(OPERATING_SYSTEM_ERROR, TermNil, "could not find executable name"); 
+	/* do nothing */
+	return;
+      }
+      buflen = strlen(Yap_FileNameBuf);
+      pt = Yap_FileNameBuf+strlen(Yap_FileNameBuf);
+      while (*--pt != '\\') {
+	/* skip executable */
+	if (pt == Yap_FileNameBuf) {
+	  Yap_Error(OPERATING_SYSTEM_ERROR, TermNil, "could not find executable name");
+	  /* do nothing */
+	  return;
+	}
+      }
+      while (*--pt != '\\') {
+	/* skip parent directory "bin\\" */
+	if (pt == Yap_FileNameBuf) {
+	  Yap_Error(OPERATING_SYSTEM_ERROR, TermNil, "could not find executable name");
+	  /* do nothing */
+	}
+      }
+      /* now, this is a possible location for the ROOT_DIR, let's look for a share directory here */
+      pt[1] = '\0';
+      /* grosse */
+      strncat(Yap_FileNameBuf,"lib\\Yap",YAP_FILENAME_MAX);
+      libdir = Yap_AllocCodeSpace(strlen(Yap_FileNameBuf)+1);
+      strncpy(libdir, Yap_FileNameBuf, strlen(Yap_FileNameBuf)+1);
+      pt[1] = '\0';
+      strncat(Yap_FileNameBuf,"share",YAP_FILENAME_MAX);
+    }
+  }
   strncat(Yap_FileNameBuf,"\\", YAP_FILENAME_MAX);
 #else
   strncat(Yap_FileNameBuf,"/", YAP_FILENAME_MAX);
@@ -236,6 +312,7 @@ bla bla
 #if THREADS
 #define StartOfTimes (*(ThreadHandle[worker_id].start_of_timesp))
 #define last_time    (*(ThreadHandle[worker_id].last_timep))
+
 #else
 /* since the point YAP was started */
 static struct timeval StartOfTimes;
@@ -333,6 +410,18 @@ InitTime (void)
   }
 }
 
+#ifdef __GNUC__
+static unsigned long long int 
+sub_utime(FILETIME t1, FILETIME t2)
+{
+  ULARGE_INTEGER u[2];
+  memcpy((void *)u,(void *)&t1,sizeof(FILETIME));
+  memcpy((void *)(u+1),(void *)&t2,sizeof(FILETIME));
+  return 
+    u[0].QuadPart - u[1].QuadPart;
+}
+#endif
+
 UInt
 Yap_cputime (void)
 {
@@ -345,8 +434,7 @@ Yap_cputime (void)
   } else {
 #ifdef __GNUC__
     unsigned long long int t =
-      *(unsigned long long int *)&UserTime - 
-      *(unsigned long long int *)&StartOfTimes;
+      sub_utime(UserTime,StartOfTimes);
     do_div(t,10000);
     return((Int)t);
 #endif
@@ -370,11 +458,9 @@ void Yap_cputime_interval(Int *now,Int *interval)
   } else {
 #ifdef __GNUC__
     unsigned long long int t1 =
-      *(unsigned long long int *)&UserTime -
-      *(unsigned long long int *)&StartOfTimes;
+      sub_utime(UserTime, StartOfTimes);
     unsigned long long int t2 =
-      *(unsigned long long int *)&UserTime -
-      *(unsigned long long int *)&last_time;
+      sub_utime(UserTime, last_time);
     do_div(t1,10000);
     *now = (Int)t1;
     do_div(t2,10000);
@@ -1239,7 +1325,7 @@ InteractSIGINT(int ch) {
   switch (ch) {
   case 'a':
     /* abort computation */
-    if (Yap_PrologMode & ConsoleGetcMode) {
+    if (Yap_PrologMode & (GCMode|ConsoleGetcMode|GrowStackMode|GrowHeapMode)) {
       Yap_PrologMode |= AbortMode;
     } else {
       Yap_Error(PURE_ABORT, TermNil, "");
@@ -1406,19 +1492,25 @@ ReceiveSignal (int s)
     case SIGKILL:
       Yap_Error(INTERRUPT_ERROR,MkIntTerm(s),NULL);
 #endif
-#if defined(SIGUSR1)
+#ifdef SIGUSR1
     case SIGUSR1:
       /* force the system to creep */
       Yap_signal (YAP_USR1_SIGNAL);
       break;
 #endif /* defined(SIGUSR1) */
-#if defined(SIGUSR2)
+#ifdef SIGUSR2
     case SIGUSR2:
       /* force the system to creep */
       Yap_signal (YAP_USR2_SIGNAL);
       break;
 #endif /* defined(SIGUSR2) */
-#if defined(SIGHUP)
+#ifdef SIGPIPE
+    case SIGPIPE:
+      /* force the system to creep */
+      Yap_signal (YAP_PIPE_SIGNAL);
+      break;
+#endif /* defined(SIGPIPE) */
+#ifdef SIGHUP
     case SIGHUP:
       /* force the system to creep */
       Yap_signal (YAP_HUP_SIGNAL);
@@ -1459,6 +1551,9 @@ InitSignals (void)
     my_signal (SIGHUP,  ReceiveSignal);
     my_signal (SIGALRM, HandleALRM);
 #endif
+#ifdef SIGPIPE
+    my_signal (SIGPIPE, ReceiveSignal);
+#endif
 #if _MSC_VER || defined(__MINGW32__)
     signal (SIGINT, SIG_IGN);
     SetConsoleCtrlHandler(MSCHandleSignal,TRUE);
@@ -1513,13 +1608,14 @@ Yap_volume_header(char *file)
   return volume_header(file);
 }
 
+/******
+      TODO: rewrite to use wordexp
+ ****/
 static int
-TrueFileName (char *source, char *result, int in_lib)
+TrueFileName (char *source, char *root, char *result, int in_lib)
 {
-  register int ch;
-  register char *res0 = result, *work;
-  char ares1[YAP_FILENAME_MAX], *res1 = ares1;
-  char *var_name;
+  char *res0 = result, *work;
+  char ares1[YAP_FILENAME_MAX];
 
   result[0] = '\0';
 #if defined(__MINGW32__) || _MSC_VER
@@ -1534,138 +1630,143 @@ TrueFileName (char *source, char *result, int in_lib)
   }
 #endif
   /* step 1: eating home information */
-  if (source[0] == '~')
-    {
-      if (dir_separator(source[1]) || source[1] == '\0')
-	{
-	  char *s;
-	  source++;
+  if (source[0] == '~') {
+    if (dir_separator(source[1]) || source[1] == '\0')
+      {
+	char *s;
+	source++;
 #if defined(_WIN32)
-	  s = getenv("HOMEDRIVE");
-	  if (s != NULL)
-	    strncpy (result, getenv ("HOMEDRIVE"), YAP_FILENAME_MAX);
-	  s = getenv("HOMEPATH");
-	  if (s != NULL)
-	    strncpy (result, s, YAP_FILENAME_MAX);
+	s = getenv("HOMEDRIVE");
+	if (s != NULL)
+	  strncpy (result, getenv ("HOMEDRIVE"), YAP_FILENAME_MAX);
+	s = getenv("HOMEPATH");
+	if (s != NULL)
+	  strncpy (result, s, YAP_FILENAME_MAX);
 #else
-	  s = getenv ("HOME");
-	  if (s != NULL)
-	    strncpy (result, s, YAP_FILENAME_MAX);
+	s = getenv ("HOME");
+	if (s != NULL)
+	  strncpy (result, s, YAP_FILENAME_MAX);
 #endif
-	}
+      } else {
 #if HAVE_GETPWNAM
-      else
-	{
-	  struct passwd *user_passwd;
+      struct passwd *user_passwd;
 
-	  source++;
-	  while (!dir_separator((*res0 = *source)) && *res0 != '\0')
-	    res0++, source++;
-	  *res0++ = '\0';
-	  if ((user_passwd = getpwnam (result)) == NULL)
-	    {
-	      return(FALSE);
-	    }
-	  strncpy (result, user_passwd->pw_dir, YAP_FILENAME_MAX);
-	}
-#endif
-      strncat (result, source, YAP_FILENAME_MAX);
-    }
-  else
-    strncpy (result, source, YAP_FILENAME_MAX);
-  /* step 2: handling environment variables in file names */
-  strncpy (ares1, result, YAP_FILENAME_MAX);
-  res0 = result;
-  while ((ch = *res1++)!=0)
-    {
-      if (ch == '\\' && !dir_separator('\\'))
-	{
-	  ch = *res1++;
-	  if (ch == '\0')
-	    {
-	      *res0 = '\0';
-	      break;
-	    }
-	  else
-	    *res0++ = ch;
-	}
-      if (ch != '$')
-	*res0++ = ch;
-      else
-	{
-	  char env_var[256], *sptr = env_var;
-	  while (((ch = *res1)!=0) && is_valid_env_char (ch))
-	    {
-	      res1++;
-	      *sptr++ = ch;
-	    }
-	  *sptr = '\0';
-	  if ((var_name = (char *) getenv (env_var)) == NULL)
-	    {
-	      return(FALSE);
-	    }
-	  else
-	    while ((*res0 = *var_name++)!=0)
-	      res0++;
-	}
-    }
-  *res0 = '\0';
-  /* step 3: get the full file name */
-  if (!dir_separator(result[0]) && !volume_header(result))
-    {
-
-#if __simplescalar__
-      /* does not implement getcwd */
-      strncpy(ares1,yap_pwd,YAP_FILENAME_MAX);
-#elif HAVE_GETCWD
-      if (getcwd (ares1, YAP_FILENAME_MAX) == NULL)
+      source++;
+      while (!dir_separator((*res0 = *source)) && *res0 != '\0')
+	res0++, source++;
+      *res0++ = '\0';
+      if ((user_passwd = getpwnam (result)) == NULL) {
 	return FALSE;
+      }
+      strncpy (result, user_passwd->pw_dir, YAP_FILENAME_MAX);
 #else
-      if (getwd (ares1) == NULL)
-	return FALSE;
+      return FALSE;
 #endif
+    }
+    strncat (result, source, YAP_FILENAME_MAX);
+  } else if (source[0] == '$') {
+    /* follow SICStus expansion rules */
+    int ch;
+    char *s;
+    char *res0 = source+1;
+
+    while ((ch = *res0) && is_valid_env_char (ch)) {
+      res0++;
+    }
+    *res0 = '\0';
+    if (!(s = (char *) getenv (source+1))) {
+      return FALSE;
+    }
+    *res0 = ch;
+    strncpy (result, s, YAP_FILENAME_MAX);
+    strncat (result, res0, YAP_FILENAME_MAX);
+  } else {
+    strncpy (result, source, YAP_FILENAME_MAX);
+  }
+#if defined(_WIN32)
+ {
+   int ch;
+
+   res0 = result;
+   /* step 2 WINDOWS: replacing \ by / */
+   while ((ch = *res0++)) {
+     if (ch == '\\' && dir_separator('\\')) {
+       res0[-1] = '/';
+     }
+   }
+ }
+#endif
+  /* step 3: get the full file name */
+  if (!dir_separator(result[0]) && !volume_header(result)) {
+#if __simplescalar__
+    /* does not implement getcwd */
+    strncpy(ares1,yap_pwd,YAP_FILENAME_MAX);
+#elif HAVE_GETCWD
+    if (getcwd (ares1, YAP_FILENAME_MAX) == NULL)
+      return FALSE;
+#else
+    if (getwd (ares1) == NULL)
+      return FALSE;
+#endif
+#if _MSC_VER || defined(__MINGW32__)
+    strncat (ares1, "\\", YAP_FILENAME_MAX);
+#else
+    strncat (ares1, "/", YAP_FILENAME_MAX);
+#endif
+    if (root) {
+      if (!dir_separator(root[0]) && !volume_header(root)) {
+	strncat(ares1, root, YAP_FILENAME_MAX);
+      } else {
+	strncpy(ares1, root, YAP_FILENAME_MAX);
+      }
 #if _MSC_VER || defined(__MINGW32__)
       strncat (ares1, "\\", YAP_FILENAME_MAX);
 #else
       strncat (ares1, "/", YAP_FILENAME_MAX);
 #endif
-      strncat (ares1, result, YAP_FILENAME_MAX);
-      if (in_lib) {
-	int tmpf;
-	if ((tmpf = open(ares1, O_RDONLY)) < 0) {
-	  /* not in current directory, let us try the library */
-	  if  (Yap_LibDir != NULL) {
-	    strncpy(Yap_FileNameBuf, Yap_LibDir, YAP_FILENAME_MAX);
+    }
+    strncat (ares1, result, YAP_FILENAME_MAX);
+    if (in_lib) {
+      int tmpf;
+      if ((tmpf = open(ares1, O_RDONLY)) < 0) {
+	/* not in current directory, let us try the library */
+	if  (Yap_LibDir != NULL) {
+	  strncpy(Yap_FileNameBuf, Yap_LibDir, YAP_FILENAME_MAX);
 #if HAVE_GETENV
+	} else {
+	  char *yap_env = getenv("YAPLIBDIR");
+	  if (yap_env != NULL) {
+	    strncpy(ares1, yap_env, YAP_FILENAME_MAX);
+#endif
 	  } else {
-	    char *yap_env = getenv("YAPLIBDIR");
-	    if (yap_env != NULL) {
-	      strncpy(ares1, yap_env, YAP_FILENAME_MAX);
+#if _MSC_VER || defined(__MINGW32__)
+	    if (libdir)
+	      strncpy(ares1, libdir, YAP_FILENAME_MAX);
+	    else
 #endif
-	    } else {
 	      strncpy(ares1, LIB_DIR, YAP_FILENAME_MAX);
-	    }
-#if HAVE_GETENV
 	  }
+#if HAVE_GETENV
+	}
 #endif
 #if _MSC_VER || defined(__MINGW32__)
-	  strncat(ares1,"\\", YAP_FILENAME_MAX);
+	strncat(ares1,"\\", YAP_FILENAME_MAX);
 #else
-	  strncat(ares1,"/", YAP_FILENAME_MAX);
+	strncat(ares1,"/", YAP_FILENAME_MAX);
 #endif
-	  strncat(ares1,result, YAP_FILENAME_MAX);
-	  if ((tmpf = open(ares1, O_RDONLY)) >= 0) {
-	    close(tmpf);
-	    strncpy (result, ares1, YAP_FILENAME_MAX);
-	  }
-	} else {
-	  strncpy (result, ares1, YAP_FILENAME_MAX);
+	strncat(ares1,result, YAP_FILENAME_MAX);
+	if ((tmpf = open(ares1, O_RDONLY)) >= 0) {
 	  close(tmpf);
+	  strncpy (result, ares1, YAP_FILENAME_MAX);
 	}
       } else {
 	strncpy (result, ares1, YAP_FILENAME_MAX);
+	close(tmpf);
       }
+    } else {
+      strncpy (result, ares1, YAP_FILENAME_MAX);
     }
+  }
   /* step 4: simplifying the file name */
   work = result;
   while (*work != '\0')
@@ -1695,15 +1796,69 @@ TrueFileName (char *source, char *result, int in_lib)
 	}
       while ((*new_work++ = *next_work++)!=0);
     }
-  return (TRUE);
+  if (work != result && dir_separator(work[-1])) {
+    /* should only do this on result being a directory */
+    int ch0 = work[-1];
+    work--;
+    work[0] = '\0';
+    if (!is_directory(result)) {
+      /* put it back: */
+      work[0] = ch0;
+      work++;
+    }
+  }
+  return TRUE;
 }
 
 int
 Yap_TrueFileName (char *source, char *result, int in_lib)
 {
-  return TrueFileName (source, result, in_lib);
+  return TrueFileName (source, NULL, result, in_lib);
 }
 
+static Int
+p_true_file_name (void)
+{
+  Term t = Deref(ARG1);
+  
+  if (IsVarTerm(t)) {
+    Yap_Error(INSTANTIATION_ERROR,t,"argument to true_file_name unbound");
+    return FALSE;
+  }
+  if (!IsAtomTerm(t)) {
+    Yap_Error(TYPE_ERROR_ATOM,t,"argument to true_file_name");
+    return FALSE;
+  }
+  TrueFileName (RepAtom(AtomOfTerm(t))->StrOfAE, NULL, Yap_FileNameBuf, FALSE);
+  return Yap_unify(ARG2, MkAtomTerm(Yap_LookupAtom(Yap_FileNameBuf)));
+}
+
+static Int
+p_true_file_name3 (void)
+{
+  Term t = Deref(ARG1), t2 = Deref(ARG2);
+  char *root = NULL;
+  
+  if (IsVarTerm(t)) {
+    Yap_Error(INSTANTIATION_ERROR,t,"argument to true_file_name unbound");
+    return FALSE;
+  }
+  if (!IsAtomTerm(t)) {
+    Yap_Error(TYPE_ERROR_ATOM,t,"argument to true_file_name");
+    return FALSE;
+  }
+  if (!IsVarTerm(t2)) {
+    if (!IsAtomTerm(t)) {
+      Yap_Error(TYPE_ERROR_ATOM,t2,"argument to true_file_name");
+      return FALSE;
+    }
+    root = RepAtom(AtomOfTerm(t2))->StrOfAE;
+  }
+  TrueFileName (RepAtom(AtomOfTerm(t))->StrOfAE, root, Yap_FileNameBuf, FALSE);
+  return Yap_unify(ARG3, MkAtomTerm(Yap_LookupAtom(Yap_FileNameBuf)));
+}
+
+
 static Int
 p_getcwd(void)
 {
@@ -1715,7 +1870,7 @@ p_getcwd(void)
 #if HAVE_STRERROR
     Yap_Error(OPERATING_SYSTEM_ERROR, ARG1, "%s in getcwd/1", strerror(errno));
 #else
-    Yap_Error(OPERATING_SYSTEM_ERROR, ARG1, "in getcwd/1");
+    Yap_Error(OPERATING_SYSTEM_ERROR, ARG1, "error %d in getcwd/1", errno);
 #endif
     return FALSE;
   }
@@ -1783,19 +1938,19 @@ p_shell (void)
     bourne = TRUE;
   /* Yap_CloseStreams(TRUE); */
   if (bourne)
-    return (system (RepAtom(AtomOfTerm(t1))->StrOfAE) == 0);
+    return system(RepAtom(AtomOfTerm(t1))->StrOfAE) == 0;
   else {
-      int status = -1;
-      int child = fork ();
-      if (child == 0)
-	{			/* let the children go */
-	  if (!execl (shell, shell, "-c", RepAtom(AtomOfTerm(t1))->StrOfAE , NULL)) {
-	    exit(-1);
-	  }
-	  exit(TRUE);
-	}
-      {				/* put the father on wait */
-	int result = child < 0 ||
+    int status = -1;
+    int child = fork ();
+
+    if (child == 0) {			/* let the children go */
+      if (!execl (shell, shell, "-c", RepAtom(AtomOfTerm(t1))->StrOfAE , NULL)) {
+	exit(-1);
+      }
+      exit(TRUE);
+    }
+    {				/* put the father on wait */
+      int result = child < 0 ||
 /* vsc:I am not sure this is used, Stevens say wait returns an integer.
 #if NO_UNION_WAIT
 */
@@ -1815,10 +1970,10 @@ p_shell (void)
   shell = "msh -i";
   /* Yap_CloseStreams(); */
   system (shell);
-  return (TRUE);
+  return TRUE;
 #else
   Yap_Error (SYSTEM_ERROR,TermNil,"shell not available in this configuration");
-  return(FALSE);
+  return FALSE;
 #endif
 #endif /* HAVE_SYSTEM */
 #endif /* _MSC_VER */
@@ -1892,8 +2047,8 @@ p_mv (void)
   } else if (!IsAtomTerm(t2)) {
     Yap_Error(TYPE_ERROR_ATOM, t2, "second argument to rename/2 not atom");
   }
-  TrueFileName (RepAtom(AtomOfTerm(t1))->StrOfAE, oldname, FALSE);
-  TrueFileName (RepAtom(AtomOfTerm(t2))->StrOfAE, newname, FALSE);
+  TrueFileName (RepAtom(AtomOfTerm(t1))->StrOfAE, NULL, oldname, FALSE);
+  TrueFileName (RepAtom(AtomOfTerm(t2))->StrOfAE, NULL, newname, FALSE);
   if ((r = link (oldname, newname)) == 0 && (r = unlink (oldname)) != 0)
     unlink (newname);
   if (r != 0) {
@@ -1927,7 +2082,7 @@ p_file_directory_name (void)
     Yap_Error(TYPE_ERROR_ATOM, t1,  "first arg of file_directory_name/2");
     return FALSE;
   }
-  TrueFileName (RepAtom(AtomOfTerm(t1))->StrOfAE,  Yap_FileNameBuf, FALSE);
+  TrueFileName (RepAtom(AtomOfTerm(t1))->StrOfAE,  NULL, Yap_FileNameBuf, FALSE);
   chp = Yap_FileNameBuf+strlen(Yap_FileNameBuf);
   while (!dir_separator(*--chp) && chp != Yap_FileNameBuf);
   if (chp == Yap_FileNameBuf) {
@@ -1947,7 +2102,7 @@ p_cd (void)
     Yap_Error(INSTANTIATION_ERROR,t1,"argument to cd/1 is not valid");
     return FALSE;
   } else if (IsAtomTerm(t1)) {
-    TrueFileName (RepAtom(AtomOfTerm(t1))->StrOfAE, Yap_FileNameBuf2, FALSE);
+    TrueFileName (RepAtom(AtomOfTerm(t1))->StrOfAE, NULL, Yap_FileNameBuf2, FALSE);
   } else {
     if (t1 == TermNil)
       return TRUE;
@@ -1955,7 +2110,7 @@ p_cd (void)
       Yap_Error(TYPE_ERROR_ATOM,t1,"argument to cd/1 is not valid");
       return FALSE;
     }
-    TrueFileName (Yap_FileNameBuf, Yap_FileNameBuf2, FALSE);
+    TrueFileName (Yap_FileNameBuf, NULL, Yap_FileNameBuf2, FALSE);
   }
 #if HAVE_CHDIR
 #if  __simplescalar__
@@ -2062,12 +2217,17 @@ static Int p_putenv(void)
 	  "second arg to putenv/2");
     return(FALSE);
   } else s2 = RepAtom(AtomOfTerm(t2))->StrOfAE;
-  p0 = p = Yap_AllocAtomSpace(strlen(s)+strlen(s2)+3);
+  while (!(p0 = p = Yap_AllocAtomSpace(strlen(s)+strlen(s2)+3))) {
+    if (!Yap_growheap(FALSE, MinHeapGap, NULL)) {
+      Yap_Error(OUT_OF_HEAP_ERROR, TermNil, Yap_ErrorMessage);
+      return FALSE;
+    }
+  }
   while ((*p++ = *s++) != '\0');
   p[-1] = '=';
   while ((*p++ = *s2++) != '\0');
   if (putenv(p0) == 0)
-    return(TRUE);
+    return TRUE;
 #if HAVE_STRERROR
   Yap_Error(OPERATING_SYSTEM_ERROR, TermNil,
 	"in putenv(%s)", strerror(errno), p0);
@@ -2075,11 +2235,11 @@ static Int p_putenv(void)
   Yap_Error(OPERATING_SYSTEM_ERROR, TermNil,
 	"in putenv(%s)", p0);
 #endif
-  return (FALSE);
+  return FALSE;
 #else
     Yap_Error(SYSTEM_ERROR, TermNil,
 	  "putenv not available in this configuration");
-    return (FALSE);
+    return FALSE;
 #endif
 }
 
@@ -2121,13 +2281,15 @@ static Int p_file_age(void)
 static DWORD WINAPI
 DoTimerThread(LPVOID targ)
 {
-  Int time = *(Int *)targ;
+  Int *time = (Int *)targ;
   HANDLE htimer;
   LARGE_INTEGER liDueTime;
 
-  htimer = CreateWaitableTimer(NULL,FALSE,NULL);
+  htimer = CreateWaitableTimer(NULL, FALSE, NULL);
   liDueTime.QuadPart =  -10000000;
-  liDueTime.QuadPart *=  time;
+  liDueTime.QuadPart *=  time[0];
+  /* add time in usecs */
+  liDueTime.QuadPart -=  time[1]*10;
   /* Copy the relative time into a LARGE_INTEGER. */
   if (SetWaitableTimer(htimer, &liDueTime,0,NULL,NULL,0) == 0) {
     return(FALSE);
@@ -2149,6 +2311,7 @@ static Int
 p_alarm(void)
 {
   Term t = Deref(ARG1);
+  Term t2 = Deref(ARG2);
   if (IsVarTerm(t)) {
     Yap_Error(INSTANTIATION_ERROR, t, "alarm/2");
     return(FALSE);
@@ -2157,12 +2320,23 @@ p_alarm(void)
     Yap_Error(TYPE_ERROR_INTEGER, t, "alarm/2");
     return(FALSE);
   }
+  if (IsVarTerm(t2)) {
+    Yap_Error(INSTANTIATION_ERROR, t2, "alarm/2");
+    return(FALSE);
+  }
+  if (!IsIntegerTerm(t2)) {
+    Yap_Error(TYPE_ERROR_INTEGER, t2, "alarm/2");
+    return(FALSE);
+  }
 #if _MSC_VER || defined(__MINGW32__)
   {
     Term tout;
-    Int time = IntegerOfTerm(t);
+    Int time[2];
+
+    time[0] = IntegerOfTerm(t);
+    time[1] = IntegerOfTerm(t2);
     
-    if (time != 0) {
+    if (time[0] != 0 && time[1] != 0) {
       DWORD dwThreadId; 
       HANDLE hThread; 
 
@@ -2170,7 +2344,7 @@ p_alarm(void)
 			     NULL,     /* no security attributes */
 			     0,        /* use default stack size */ 
 			     DoTimerThread, /* thread function */
-			     (LPVOID)&time,  /* argument to thread function */
+			     (LPVOID)time,  /* argument to thread function */
 			     0,        /* use default creation flags  */
 			     &dwThreadId);  /* returns the thread identifier */
  
@@ -2180,16 +2354,35 @@ p_alarm(void)
       }
     }
     tout = MkIntegerTerm(0);
-    return(Yap_unify(ARG2,tout));
+    return Yap_unify(ARG3,tout) && Yap_unify(ARG4,MkIntTerm(0));
   }
-#elif HAVE_ALARM
+#elif HAVE_SETITIMER && !SUPPORT_CONDOR
+  {
+    struct itimerval new, old;
+
+    new.it_interval.tv_sec = 0;
+    new.it_interval.tv_usec = 0;
+    new.it_value.tv_sec = IntegerOfTerm(t);
+    new.it_value.tv_usec = IntegerOfTerm(t2);
+    if (setitimer(ITIMER_REAL, &new, &old) < 0) {
+#if HAVE_STRERROR
+      Yap_Error(OPERATING_SYSTEM_ERROR, ARG1, "setitimer: %s", strerror(errno));
+#else
+      Yap_Error(OPERATING_SYSTEM_ERROR, ARG1, "setitimer %d", errno);
+#endif
+      return FALSE;
+    }
+    return Yap_unify(ARG3,MkIntegerTerm(old.it_value.tv_sec)) &&
+      Yap_unify(ARG4,MkIntegerTerm(old.it_value.tv_usec));
+  }
+#elif HAVE_ALARM && !SUPPORT_CONDOR
   {
     Int left;
     Term tout;
 
     left = alarm(IntegerOfTerm(t));
     tout = MkIntegerTerm(left);
-    return Yap_unify(ARG2,tout);
+    return Yap_unify(ARG3,tout) && Yap_unify(ARG4,MkIntTerm(0)) ;
   }
 #else
   /* not actually trying to set the alarm */
@@ -2197,7 +2390,7 @@ p_alarm(void)
     return TRUE;
   Yap_Error(SYSTEM_ERROR, TermNil,
 	"alarm not available in this configuration");
-  return(FALSE);
+  return FALSE;
 #endif
 }
 
@@ -2329,6 +2522,9 @@ static Int
 p_first_signal(void)
 {
   LOCK(SignalLock);
+#ifdef THREADS
+  pthread_mutex_lock(&(ThreadHandle[worker_id].tlock));
+#endif  
   /* always do wakeups first, because you don't want to keep the
      non-backtrackable variable bad */
   if (ActiveSignals & YAP_WAKEUP_SIGNAL) {
@@ -2356,6 +2552,11 @@ p_first_signal(void)
     UNLOCK(SignalLock);
     return Yap_unify(ARG1, MkAtomTerm(Yap_LookupAtom("sig_usr1")));
   }
+  if (ActiveSignals & YAP_PIPE_SIGNAL) {
+    ActiveSignals &= ~YAP_PIPE_SIGNAL;
+    UNLOCK(SignalLock);
+    return Yap_unify(ARG1, MkAtomTerm(Yap_LookupAtom("sig_pipe")));
+  }
   if (ActiveSignals & YAP_HUP_SIGNAL) {
     ActiveSignals &= ~YAP_HUP_SIGNAL;
     UNLOCK(SignalLock);
@@ -2402,6 +2603,9 @@ p_first_signal(void)
     return Yap_unify(ARG1, MkAtomTerm(Yap_LookupAtom("sig_statistics")));
   }
   UNLOCK(SignalLock);
+#ifdef THREADS
+  pthread_mutex_unlock(&(ThreadHandle[worker_id].tlock));
+#endif  
   return FALSE;
 }
 
@@ -2448,12 +2652,244 @@ p_continue_signals(void)
   if (ActiveSignals & YAP_STATISTICS_SIGNAL) {
     Yap_signal(YAP_STATISTICS_SIGNAL);
   }
+#ifdef THREADS
+  pthread_mutex_unlock(&(ThreadHandle[worker_id].tlock));
+#endif  
   return TRUE;
 }
 
+static Int
+p_unix(void)
+{
+#ifdef unix
+  return TRUE;
+#else
+#ifdef __unix__
+  return TRUE;
+#else
+#ifdef __APPLE__
+  return TRUE;
+#else
+  return FALSE;
+#endif
+#endif
+#endif
+}
+
+static Int
+p_win32(void)
+{
+#ifdef _WIN32
+  return TRUE;
+#else
+#ifdef __CYGWIN__
+  return TRUE;
+#else
+  return FALSE;
+#endif
+#endif
+}
+
+static Int
+p_ld_path(void)
+{
+  return Yap_unify(ARG1,MkAtomTerm(Yap_LookupAtom("LD_LIBRARY_PATH")));
+}
+
+
+
+#ifdef _WIN32
+
+/* This code is from SWI-Prolog by Jan Wielemaker */
+
+#define wstreq(s,q) (wcscmp((s), (q)) == 0)
+
+static HKEY
+reg_open_key(const wchar_t *which, int create)
+{ HKEY key = HKEY_CURRENT_USER;
+  DWORD disp;
+  LONG rval;
+
+  while(*which)
+  { wchar_t buf[256];
+    wchar_t *s;
+    HKEY tmp;
+
+    for(s=buf; *which && !(*which == '/' || *which == '\\'); )
+      *s++ = *which++;
+    *s = '\0';
+    if ( *which )
+      which++;
+
+    if ( wstreq(buf, L"HKEY_CLASSES_ROOT") )
+    { key = HKEY_CLASSES_ROOT;
+      continue;
+    } else if ( wstreq(buf, L"HKEY_CURRENT_USER") )
+    { key = HKEY_CURRENT_USER;
+      continue;
+    } else if ( wstreq(buf, L"HKEY_LOCAL_MACHINE") )
+    { key = HKEY_LOCAL_MACHINE;
+      continue;
+    } else if ( wstreq(buf, L"HKEY_USERS") )
+    { key = HKEY_USERS;
+      continue;
+    }
+
+    if ( RegOpenKeyExW(key, buf, 0L, KEY_READ, &tmp) == ERROR_SUCCESS )
+    { RegCloseKey(key);
+      key = tmp;
+      continue;
+    }
+
+    if ( !create )
+      return NULL;
+
+    rval = RegCreateKeyExW(key, buf, 0, L"", 0,
+			  KEY_ALL_ACCESS, NULL, &tmp, &disp);
+    RegCloseKey(key);
+    if ( rval == ERROR_SUCCESS )
+      key = tmp;
+    else
+      return NULL;
+  }
+
+  return key;
+}
+
+#define MAXREGSTRLEN 1024
+
+static void
+recover_space(wchar_t *k, Atom At)
+{
+  if (At->WStrOfAE != k)
+    Yap_FreeCodeSpace((char *)k);
+}
+
+static wchar_t *
+WideStringFromAtom(Atom KeyAt)
+{
+  if (IsWideAtom(KeyAt)) {
+    return KeyAt->WStrOfAE;
+  } else {
+    int len = strlen(KeyAt->StrOfAE);
+    int sz = sizeof(wchar_t)*(len+1);
+    char *chp = KeyAt->StrOfAE;
+    wchar_t *kptr, *k;
+
+    k = (wchar_t *)Yap_AllocCodeSpace(sz);
+    while (k == NULL) {
+      if (!Yap_growheap(FALSE, sz, NULL)) {
+	Yap_Error(OUT_OF_HEAP_ERROR, MkIntegerTerm(sz), "generating key in win_registry_get_value/3");
+	return FALSE;
+      }      
+    }
+    kptr = k;
+    while ((*kptr++ = *chp++));
+    return k;
+  }
+}
+
+static Int
+p_win_registry_get_value(void)
+{
+  DWORD type;
+  BYTE  data[MAXREGSTRLEN];
+  DWORD len = sizeof(data);
+  wchar_t *k, *name;
+  HKEY key;
+  Term Key = Deref(ARG1);
+  Term Name = Deref(ARG2);
+  Atom KeyAt, NameAt;
+
+  if (IsVarTerm(Key)) {
+    Yap_Error(INSTANTIATION_ERROR,Key,"argument to win_registry_get_value unbound");
+    return FALSE;
+  }
+  if (!IsAtomTerm(Key)) {
+    Yap_Error(TYPE_ERROR_ATOM,Key,"argument to win_registry_get_value");
+    return FALSE;
+  }
+  KeyAt = AtomOfTerm(Key);
+  if (IsVarTerm(Name)) {
+    Yap_Error(INSTANTIATION_ERROR,Key,"argument to win_registry_get_value unbound");
+    return FALSE;
+  }
+  if (!IsAtomTerm(Name)) {
+    Yap_Error(TYPE_ERROR_ATOM,Key,"argument to win_registry_get_value");
+    return FALSE;
+  }
+  NameAt = AtomOfTerm(Name);
+
+  k = WideStringFromAtom(KeyAt);
+  if ( !(key=reg_open_key(k, FALSE)) ) {
+    Yap_Error(EXISTENCE_ERROR_KEY, Key, "argument to win_registry_get_value");
+    recover_space(k, KeyAt);
+    return FALSE;
+  }
+  name = WideStringFromAtom(NameAt);
+
+  if ( RegQueryValueExW(key, name, NULL, &type, data, &len) == ERROR_SUCCESS ) {
+    RegCloseKey(key);
+    switch(type) {
+    case REG_SZ:
+      recover_space(k, KeyAt);
+      recover_space(name, NameAt);
+      ((wchar_t *)data)[len] = '\0';
+      return Yap_unify(MkAtomTerm(Yap_LookupMaybeWideAtom((wchar_t *)data)),ARG3);
+    case REG_DWORD:
+      recover_space(k, KeyAt);
+      recover_space(name, NameAt);
+      return Yap_unify(MkIntegerTerm(*((DWORD *)data)),ARG3);
+    default:
+      recover_space(k, KeyAt);
+      recover_space(name, NameAt);
+      return FALSE;
+    }
+  }
+  recover_space(k, KeyAt);
+  recover_space(name, NameAt);
+  return FALSE;
+}
+
+char *
+Yap_RegistryGetString(char *name)
+{
+  DWORD type;
+  BYTE  data[MAXREGSTRLEN];
+  DWORD len = sizeof(data);
+  HKEY key;
+  char *ptr;
+  int i;
+
+  if ( !(key=reg_open_key(L"HKEY_LOCAL_MACHINE/SOFTWARE/YAP/Prolog", FALSE)) ) {
+    return NULL;
+  }
+  if ( RegQueryValueEx(key, name, NULL, &type, data, &len) == ERROR_SUCCESS ) {
+    RegCloseKey(key);
+    switch(type) {
+    case REG_SZ:
+      ptr = malloc(len+2);
+      if (!ptr)
+	return NULL;
+      for (i=0; i<= len; i++)
+	ptr[i] = data[i];
+      ptr[len+1] = '\0';
+      return ptr;
+    default:
+      return NULL;
+    }
+  }
+  return NULL;
+}
+
+
+#endif
+
 void
 Yap_InitSysPreds(void)
 {
+  Term cm = CurrentModule;
+
   /* can only do after heap is initialised */
   InitLastWtime();
   Yap_InitCPred ("srandom", 1, p_srandom, SafePredFlag);
@@ -2465,7 +2901,7 @@ Yap_InitSysPreds(void)
   Yap_InitCPred ("$yap_home", 1, p_yap_home, SafePredFlag|SyncPredFlag|HiddenPredFlag);
   Yap_InitCPred ("getcwd", 1, p_getcwd, SafePredFlag|SyncPredFlag);
   Yap_InitCPred ("$dir_separator", 1, p_dir_sp, SafePredFlag|HiddenPredFlag);
-  Yap_InitCPred ("$alarm", 2, p_alarm, SafePredFlag|SyncPredFlag|HiddenPredFlag);
+  Yap_InitCPred ("$alarm", 4, p_alarm, SafePredFlag|SyncPredFlag|HiddenPredFlag);
   Yap_InitCPred ("$getenv", 2, p_getenv, SafePredFlag|HiddenPredFlag);
   Yap_InitCPred ("$putenv", 2, p_putenv, SafePredFlag|SyncPredFlag|HiddenPredFlag);
   Yap_InitCPred ("$file_age", 2, p_file_age, SafePredFlag|SyncPredFlag|HiddenPredFlag);
@@ -2475,6 +2911,16 @@ Yap_InitSysPreds(void)
   Yap_InitCPred ("$continue_signals", 0, p_continue_signals, SafePredFlag|SyncPredFlag|HiddenPredFlag);
   Yap_InitCPred ("file_directory_name", 2, p_file_directory_name, SafePredFlag);
   Yap_InitCPred ("$env_separator", 1, p_env_separator, SafePredFlag);
+  Yap_InitCPred ("$unix", 0, p_unix, SafePredFlag);
+  Yap_InitCPred ("$win32", 0, p_win32, SafePredFlag);
+  Yap_InitCPred ("$ld_path", 1, p_ld_path, SafePredFlag);
+#ifdef _WIN32
+  Yap_InitCPred ("win_registry_get_value", 3, p_win_registry_get_value,0);
+#endif
+  CurrentModule = SYSTEM_MODULE;
+  Yap_InitCPred ("true_file_name", 2, p_true_file_name, SyncPredFlag);
+  Yap_InitCPred ("true_file_name", 3, p_true_file_name3, SyncPredFlag);
+  CurrentModule = cm;
 }
 
 
@@ -2610,7 +3056,7 @@ rw_lock_voodoo(void) {
 #endif /* sparc */
 
 
-#ifdef i386
+#if defined(i386) || defined(__x86_64__)
 asm(
 
 ".align	4\n"
