@@ -19,7 +19,7 @@ static char     SccsId[] = "@(#)utilpreds.c	1.3";
 #endif
 
 #include "Yap.h"
-#include "Yatom.h"
+#include "clause.h"
 #include "Heap.h"
 #include "yapio.h"
 #include "eval.h"
@@ -31,7 +31,7 @@ typedef struct {
 }              *vcell;
 
 
-STATIC_PROTO(int   copy_complex_term, (CELL *, CELL *, CELL *, CELL *));
+STATIC_PROTO(int   copy_complex_term, (CELL *, CELL *, int, int, CELL *, CELL *));
 STATIC_PROTO(CELL  vars_in_complex_term, (CELL *, CELL *, Term));
 STATIC_PROTO(Int   p_non_singletons_in_term, (void));
 STATIC_PROTO(CELL  non_singletons_in_complex_term, (CELL *, CELL *));
@@ -76,16 +76,19 @@ clean_dirty_tr(tr_fr_ptr TR0) {
   }
 }
 
+
 static int
-copy_complex_term(register CELL *pt0, register CELL *pt0_end, CELL *ptf, CELL *HLow)
+copy_complex_term(CELL *pt0, CELL *pt0_end, int share, int newattvs, CELL *ptf, CELL *HLow)
 {
 
-  CELL **to_visit0, **to_visit = (CELL **)Yap_PreAllocCodeSpace();
+  struct cp_frame *to_visit0, *to_visit = (struct cp_frame *)Yap_PreAllocCodeSpace();
   CELL *HB0 = HB;
   tr_fr_ptr TR0 = TR;
+  int ground = TRUE;
 #ifdef COROUTINING
   CELL *dvars = NULL;
 #endif
+
   HB = HLow;
   to_visit0 = to_visit;
  loop:
@@ -108,27 +111,30 @@ copy_complex_term(register CELL *pt0, register CELL *pt0_end, CELL *ptf, CELL *H
 	*ptf = AbsPair(H);
 	ptf++;
 #ifdef RATIONAL_TREES
-	if (to_visit + 4 >= (CELL **)AuxSp) {
+	if (to_visit+1 >= (struct cp_frame *)AuxSp) {
 	  goto heap_overflow;
 	}
-	to_visit[0] = pt0;
-	to_visit[1] = pt0_end;
-	to_visit[2] = ptf;
-	to_visit[3] = (CELL *)*pt0;
+	to_visit->start_cp = pt0;
+	to_visit->end_cp = pt0_end;
+	to_visit->to = ptf;
+	to_visit->oldv = *pt0;
+	to_visit->ground = ground;
 	/* fool the system into thinking we had a variable there */
 	*pt0 = AbsPair(H);
-	to_visit += 4;
+	to_visit ++;
 #else
 	if (pt0 < pt0_end) {
-	  if (to_visit + 3 >= (CELL **)AuxSp) {
+	  if (to_visit+1 >= (struct cp_frame *)AuxSp) {
 	    goto heap_overflow;
 	  }
-	  to_visit[0] = pt0;
-	  to_visit[1] = pt0_end;
-	  to_visit[2] = ptf;
-	  to_visit += 3;
+	  to_visit->start_cp = pt0;
+	  to_visit->end_cp = pt0_end;
+	  to_visit->to = ptf;
+	  to_visit->ground = ground;
+	  to_visit ++;
 	}
 #endif
+	ground = TRUE;
 	pt0 = ap2 - 1;
 	pt0_end = ap2 + 1;
 	ptf = H;
@@ -149,36 +155,52 @@ copy_complex_term(register CELL *pt0, register CELL *pt0_end, CELL *ptf, CELL *H
 	f = (Functor)(*ap2);
 
 	if (IsExtensionFunctor(f)) {
-	    {
-	      *ptf++ = d0;  /* you can just copy other extensions. */
+#if defined(YAPOR) || defined(THREADS)
+	  if (f == FunctorDBRef) {
+	    DBRef  entryref = DBRefOfTerm(d0);
+	    if (entryref->Flags & LogUpdMask) {
+	      LogUpdClause *luclause = (LogUpdClause *)entryref;
+	      LOCK(luclause->ClPred->PELock);
+	      UNLOCK(luclause->ClPred->PELock);
+	    } else {
+	      LOCK(entryref->lock);
+	      TRAIL_REF(entryref);	/* So that fail will erase it */
+	      INC_DBREF_COUNT(entryref);
+	      UNLOCK(entryref->lock);
 	    }
+	  }
+#endif
+	  *ptf++ = d0;  /* you can just copy other extensions. */
 	  continue;
 	}
 	*ptf = AbsAppl(H);
 	ptf++;
 	/* store the terms to visit */
 #ifdef RATIONAL_TREES
-	if (to_visit + 4 >= (CELL **)AuxSp) {
+	if (to_visit+1 >= (struct cp_frame *)AuxSp) {
 	  goto heap_overflow;
 	}
-	to_visit[0] = pt0;
-	to_visit[1] = pt0_end;
-	to_visit[2] = ptf;
-	to_visit[3] = (CELL *)*pt0;
+	to_visit->start_cp = pt0;
+	to_visit->end_cp = pt0_end;
+	to_visit->to = ptf;
+	to_visit->oldv = *pt0;
+	to_visit->ground = ground;
 	/* fool the system into thinking we had a variable there */
 	*pt0 = AbsAppl(H);
-	to_visit += 4;
+	to_visit ++;
 #else
 	if (pt0 < pt0_end) {
-	  if (to_visit + 3 >= (CELL **)AuxSp) {
+	  if (to_visit+1 >= (struct cp_frame *)AuxSp) {
 	    goto heap_overflow;
 	  }
-	  to_visit[0] = pt0;
-	  to_visit[1] = pt0_end;
-	  to_visit[2] = ptf;
-	  to_visit += 3;
+	  to_visit->start_cp = pt0;
+	  to_visit->end_cp = pt0_end;
+	  to_visit->to = ptf;
+	  to_visit->ground = ground;
+	  to_visit ++;
 	}
 #endif
+	ground = (f != FunctorMutable);
 	d0 = ArityOfFunctor(f);
 	pt0 = ap2;
 	pt0_end = ap2 + d0;
@@ -197,19 +219,20 @@ copy_complex_term(register CELL *pt0, register CELL *pt0_end, CELL *ptf, CELL *H
     }
 
     derefa_body(d0, ptd0, copy_term_unk, copy_term_nvar);
+    ground = FALSE;
     if (ptd0 >= HLow && ptd0 < H) { 
       /* we have already found this cell */
       *ptf++ = (CELL) ptd0;
     } else {
 #if COROUTINING
-      if (IsAttachedTerm((CELL)ptd0)) {
+      if (newattvs && IsAttachedTerm((CELL)ptd0)) {
 	/* if unbound, call the standard copy term routine */
-	CELL **bp[1];
+	struct cp_frame *bp[1];
 
 	if (dvars == NULL) {
 	  dvars = (CELL *)DelayTop();
 	} 	
-	if (ptd0 >= dvars) {
+	if (ptd0 < dvars) {
 	  *ptf++ = (CELL) ptd0;
 	} else {
 	  tr_fr_ptr CurTR;
@@ -223,12 +246,24 @@ copy_complex_term(register CELL *pt0, register CELL *pt0_end, CELL *ptf, CELL *H
 	  to_visit = bp[0];
 	  HB = HLow;
 	  ptf++;
+	  if (TR > (tr_fr_ptr)Yap_TrailTop - 256) {
+	    /* Trail overflow */
+	    if (!Yap_growtrail((TR-TR0)*sizeof(tr_fr_ptr *), TRUE)) {
+	      goto trail_overflow;
+	    }
+	  }
 	  Bind_Global(ptd0, ptf[-1]);
 	}
       } else {
 #endif
 	/* first time we met this term */
 	RESET_VARIABLE(ptf);
+	if (TR > (tr_fr_ptr)Yap_TrailTop - 256) {
+	  /* Trail overflow */
+	  if (!Yap_growtrail((TR-TR0)*sizeof(tr_fr_ptr *), TRUE)) {
+	    goto trail_overflow;
+	  }
+	}
 	Bind_Global(ptd0, (CELL)ptf);
 	ptf++;
 #ifdef COROUTINING
@@ -238,25 +273,32 @@ copy_complex_term(register CELL *pt0, register CELL *pt0_end, CELL *ptf, CELL *H
   }
   /* Do we still have compound terms to visit */
   if (to_visit > to_visit0) {
+    to_visit --;
+    if (ground && share) {
+      CELL old = to_visit->oldv;
+      CELL *newp = to_visit->to-1;
+      CELL new = *newp;
+
+      *newp = old;
+      if (IsApplTerm(new))
+	H = RepAppl(new);
+      else
+	H = RepPair(new);
+    }
+    pt0 = to_visit->start_cp;
+    pt0_end = to_visit->end_cp;
+    ptf = to_visit->to;
 #ifdef RATIONAL_TREES
-    to_visit -= 4;
-    pt0 = to_visit[0];
-    pt0_end = to_visit[1];
-    ptf = to_visit[2];
-    *pt0 = (CELL)to_visit[3];
-#else
-    to_visit -= 3;
-    pt0 = to_visit[0];
-    pt0_end = to_visit[1];
-    ptf = to_visit[2];
+    *pt0 = to_visit->oldv;
 #endif
+    ground = (ground && to_visit->ground);
     goto loop;
   }
 
   /* restore our nice, friendly, term to its original state */
   HB = HB0;
   clean_dirty_tr(TR0);
-  return 0;
+  return ground;
 
  overflow:
   /* oops, we're in trouble */
@@ -266,16 +308,40 @@ copy_complex_term(register CELL *pt0, register CELL *pt0_end, CELL *ptf, CELL *H
   HB = HB0;
 #ifdef RATIONAL_TREES
   while (to_visit > to_visit0) {
-    to_visit -= 4;
-    pt0 = to_visit[0];
-    pt0_end = to_visit[1];
-    ptf = to_visit[2];
-    *pt0 = (CELL)to_visit[3];
+    to_visit --;
+    pt0 = to_visit->start_cp;
+    pt0_end = to_visit->end_cp;
+    ptf = to_visit->to;
+    *pt0 = to_visit->oldv;
   }
 #endif
   reset_trail(TR0);
   return -1;
 
+trail_overflow:
+  /* oops, we're in trouble */
+  H = HLow;
+  /* we've done it */
+  /* restore our nice, friendly, term to its original state */
+  HB = HB0;
+#ifdef RATIONAL_TREES
+  while (to_visit > to_visit0) {
+    to_visit --;
+    pt0 = to_visit->start_cp;
+    pt0_end = to_visit->end_cp;
+    ptf = to_visit->to;
+    *pt0 = to_visit->oldv;
+  }
+#endif
+  {
+    tr_fr_ptr oTR =  TR;
+    reset_trail(TR0);
+    if (!Yap_growtrail((oTR-TR0)*sizeof(tr_fr_ptr *), FALSE)) {
+      return -4;
+    }
+    return -2;
+  }
+
  heap_overflow:
   /* oops, we're in trouble */
   H = HLow;
@@ -284,24 +350,56 @@ copy_complex_term(register CELL *pt0, register CELL *pt0_end, CELL *ptf, CELL *H
   HB = HB0;
 #ifdef RATIONAL_TREES
   while (to_visit > to_visit0) {
-    to_visit -= 4;
-    pt0 = to_visit[0];
-    pt0_end = to_visit[1];
-    ptf = to_visit[2];
-    *pt0 = (CELL)to_visit[3];
+    to_visit --;
+    pt0 = to_visit->start_cp;
+    pt0_end = to_visit->end_cp;
+    ptf = to_visit->to;
+    *pt0 = to_visit->oldv;
   }
 #endif
   reset_trail(TR0);
-  return -2;
+  Yap_Error_Size = (ADDR)AuxSp-(ADDR)to_visit0;
+  return -3;
+}
+
+
+static Term
+handle_cp_overflow(int res, UInt arity, Term t)
+{
+  XREGS[arity+1] = t;
+  switch(res) {
+  case -1:
+    if (!Yap_gcl((ASP-H)*sizeof(CELL), arity+1, ENV, P)) {
+      Yap_Error(OUT_OF_STACK_ERROR, TermNil, Yap_ErrorMessage);
+      return 0L;
+    }
+    return Deref(XREGS[arity+1]);
+  case -2:
+    return Deref(XREGS[arity+1]);
+  case -3:
+    {
+      UInt size = Yap_Error_Size;
+      Yap_Error_Size = 0L;
+      if (size > 4*1024*1024)
+	size = 4*1024*1024;
+      if (!Yap_ExpandPreAllocCodeSpace(size,NULL)) {
+	Yap_Error(OUT_OF_AUXSPACE_ERROR, TermNil, Yap_ErrorMessage);
+	return 0L;
+      }
+    }
+    return Deref(XREGS[arity+1]);
+  default:
+    return 0L;
+  }
 }
 
 static Term
-CopyTerm(Term inp, UInt arity) {
+CopyTerm(Term inp, UInt arity, int share, int newattvs) {
   Term t = Deref(inp);
 
   if (IsVarTerm(t)) {
 #if COROUTINING
-    if (IsAttachedTerm(t)) {
+    if (newattvs && IsAttachedTerm(t)) {
       CELL *Hi;
       int res;
     restart_attached:
@@ -309,24 +407,11 @@ CopyTerm(Term inp, UInt arity) {
       *H = t;
       Hi = H+1;
       H += 2;
-      if ((res = copy_complex_term(Hi-2, Hi-1, Hi, Hi)) < 0) {
-	XREGS[arity+1] = t;
+      if ((res = copy_complex_term(Hi-2, Hi-1, share, newattvs, Hi, Hi)) < 0) {
 	H = Hi-1;
-	if (res == -1) { /* handle overflow */
-	  if (!Yap_gc(arity+1, ENV, P)) {
-	    Yap_Error(OUT_OF_STACK_ERROR, TermNil, Yap_ErrorMessage);
-	    return FALSE;
-	  }
-	  t = Deref(XREGS[arity+1]);
-	  goto restart_attached;
-	} else { /* handle overflow */
-	  if (!Yap_ExpandPreAllocCodeSpace(0,NULL)) {
-	    Yap_Error(OUT_OF_AUXSPACE_ERROR, TermNil, Yap_ErrorMessage);
-	    return FALSE;
-	  }
-	  t = Deref(XREGS[arity+1]);
-	  goto restart_attached;
-	}
+	if ((t = handle_cp_overflow(res,arity,t))== 0L)
+	  return FALSE;
+	goto restart_attached;
       }
       return Hi[0];
     }
@@ -346,24 +431,14 @@ CopyTerm(Term inp, UInt arity) {
     H += 2;
     {
       int res;
-      if ((res = copy_complex_term(ap-1, ap+1, Hi, Hi)) < 0) {
+      if ((res = copy_complex_term(ap-1, ap+1, share, newattvs, Hi, Hi)) < 0) {
 	H = Hi;
-	XREGS[arity+1] = t;
-	if (res == -1) { /* handle overflow */
-	  if (!Yap_gc(arity+1, ENV, P)) {
-	    Yap_Error(OUT_OF_STACK_ERROR, TermNil, Yap_ErrorMessage);
-	    return FALSE;
-	  }
-	  t = Deref(XREGS[arity+1]);
-	  goto restart_list;
-	} else { /* handle overflow */
-	  if (!Yap_ExpandPreAllocCodeSpace(0,NULL)) {
-	    Yap_Error(OUT_OF_AUXSPACE_ERROR, TermNil, Yap_ErrorMessage);
-	    return FALSE;
-	  }
-	  t = Deref(XREGS[arity+1]);
-	  goto restart_list;
-	}
+	if ((t = handle_cp_overflow(res,arity,t))== 0L)
+	  return FALSE;
+	goto restart_list;
+      } else if (res && share && FunctorOfTerm(t) != FunctorMutable) {
+	H = Hi;
+	return t;
       }
     }
     return tf;
@@ -380,27 +455,22 @@ CopyTerm(Term inp, UInt arity) {
     tf = AbsAppl(H);
     H[0] = (CELL)f;
     H += 1+ArityOfFunctor(f);
-    {
+    if (H > ASP-128) {
+      H = HB0;
+      if ((t = handle_cp_overflow(-1,arity,t))== 0L)
+	return FALSE;
+      goto restart_appl;
+    } else {
       int res;
 
-      if ((res = copy_complex_term(ap, ap+ArityOfFunctor(f), HB0+1, HB0)) < 0) {
+      if ((res = copy_complex_term(ap, ap+ArityOfFunctor(f), share, newattvs, HB0+1, HB0)) < 0) {
 	H = HB0;
-	XREGS[arity+1] = t;
-	if (res == -1) {
-	  if (!Yap_gc(arity+1, ENV, P)) {
-	    Yap_Error(OUT_OF_STACK_ERROR, TermNil, Yap_ErrorMessage);
-	    return FALSE;
-	  }
-	  t = Deref(XREGS[arity+1]);
-	  goto restart_appl;
-	} else { /* handle overflow */
-	  if (!Yap_ExpandPreAllocCodeSpace(0,NULL)) {
-	    Yap_Error(OUT_OF_AUXSPACE_ERROR, TermNil, Yap_ErrorMessage);
-	    return FALSE;
-	  }
-	  t = Deref(XREGS[arity+1]);
-	  goto restart_appl;
-	}
+	if ((t = handle_cp_overflow(res,arity,t))== 0L)
+	  return FALSE;
+	goto restart_appl;
+      } else if (res && share) {
+	H = HB0;
+	return t;
       }
     }
     return tf;
@@ -409,284 +479,33 @@ CopyTerm(Term inp, UInt arity) {
  
 Term
 Yap_CopyTerm(Term inp) {
-  return CopyTerm(inp, 0);
+  return CopyTerm(inp, 0, TRUE, TRUE);
 }
 
 static Int 
 p_copy_term(void)		/* copy term t to a new instance  */
 {
-  Term t = CopyTerm(ARG1, 2); 
+  Term t = CopyTerm(ARG1, 2, TRUE, TRUE); 
   if (t == 0L)
     return FALSE;
   /* be careful, there may be a stack shift here */
   return Yap_unify(ARG2,t);
 }
 
-static int copy_complex_term_no_delays(register CELL *pt0, register CELL *pt0_end, CELL *ptf, CELL *HLow)
+static Int 
+p_duplicate_term(void)		/* copy term t to a new instance  */
 {
-
-  CELL **to_visit0, **to_visit = (CELL **)Yap_PreAllocCodeSpace();
-  tr_fr_ptr TR0 = TR;
-  CELL *HB0 = HB;
-  HB = HLow;
-
-  to_visit0 = to_visit;
- loop:
-  while (pt0 < pt0_end) {
-    register CELL d0;
-    register CELL *ptd0;
-    ++ pt0;
-    ptd0 = pt0;
-    d0 = *ptd0;
-    deref_head(d0, copy_term_unk);
-  copy_term_nvar:
-    {
-      if (IsPairTerm(d0)) {
-	CELL *ap2 = RepPair(d0);
-	if (ap2 >= HB && ap2 < H) {
-	  /* If this is newer than the current term, just reuse */
-	  *ptf++ = d0;
-	  continue;
-	} 
-	*ptf = AbsPair(H);
-	ptf++;
-#ifdef RATIONAL_TREES
-	if (to_visit + 4 >= (CELL **)AuxSp) {
-	  goto heap_overflow;
-	}
-	to_visit[0] = pt0;
-	to_visit[1] = pt0_end;
-	to_visit[2] = ptf;
-	to_visit[3] = (CELL *)*pt0;
-	/* fool the system into thinking we had a variable there */
-	*pt0 = AbsPair(H);
-	to_visit += 4;
-#else
-	if (pt0 < pt0_end) {
-	  if (to_visit + 3 >= (CELL **)AuxSp) {
-	    goto heap_overflow;
-	  }
-	  to_visit[0] = pt0;
-	  to_visit[1] = pt0_end;
-	  to_visit[2] = ptf;
-	  to_visit += 3;
-	}
-#endif
-	pt0 = ap2 - 1;
-	pt0_end = ap2 + 1;
-	ptf = H;
-	H += 2;
-	if (H > ENV - 2048) {
-	  goto overflow;
-	}
-      } else if (IsApplTerm(d0)) {
-	register Functor f;
-	register CELL *ap2;
-	/* store the terms to visit */
-	ap2 = RepAppl(d0);
-	if (ap2 >= HB && ap2 < H) {
-	  /* If this is newer than the current term, just reuse */
-	  *ptf++ = d0;
-	  continue;
-	} 
-	f = (Functor)(*ap2);
-
-	if (IsExtensionFunctor(f)) {
-	  *ptf++ = d0;  /* you can just copy other extensions. */
-	  continue;
-	}
-	*ptf = AbsAppl(H);
-	ptf++;
-	/* store the terms to visit */
-#ifdef RATIONAL_TREES
-	if (to_visit + 4 >= (CELL **)AuxSp) {
-	  goto heap_overflow;
-	}
-	to_visit[0] = pt0;
-	to_visit[1] = pt0_end;
-	to_visit[2] = ptf;
-	to_visit[3] = (CELL *)*pt0;
-	/* fool the system into thinking we had a variable there */
-	*pt0 = AbsAppl(H);
-	to_visit += 4;
-#else
-	if (to_visit + 3 >= (CELL **)AuxSp) {
-	  goto heap_overflow;
-	}
-	if (pt0 < pt0_end) {
-	  to_visit[0] = pt0;
-	  to_visit[1] = pt0_end;
-	  to_visit[2] = ptf;
-	  to_visit += 3;
-	}
-#endif
-	d0 = ArityOfFunctor(f);
-	pt0 = ap2;
-	pt0_end = ap2 + d0;
-	/* store the functor for the new term */
-	H[0] = (CELL)f;
-	ptf = H+1;
-	H += 1+d0;
-	if (H > ENV - 2048) {
-	  goto overflow;
-	}
-      } else {
-	/* just copy atoms or integers */
-	*ptf++ = d0;
-      }
-      continue;
-    }
-    
-
-    derefa_body(d0, ptd0, copy_term_unk, copy_term_nvar);
-    if (ptd0 >= HLow && ptd0 < H) { 
-      /* we have already found this cell */
-      *ptf++ = (CELL) ptd0;
-    } else {
-      /* first time we met this term */
-      RESET_VARIABLE(ptf);
-      Bind_Global(ptd0, (CELL)ptf);
-      ptf++;
-    }
-  }
-  /* Do we still have compound terms to visit */
-  if (to_visit > to_visit0) {
-#ifdef RATIONAL_TREES
-    to_visit -= 4;
-    pt0 = to_visit[0];
-    pt0_end = to_visit[1];
-    ptf = to_visit[2];
-    *pt0 = (CELL)to_visit[3];
-#else
-    to_visit -= 3;
-    pt0 = to_visit[0];
-    pt0_end = to_visit[1];
-    ptf = to_visit[2];
-#endif
-    goto loop;
-  }
-
-  /* we've done it */
-  /* restore our nice, friendly, term to its original state */
-  HB = HB0;
-  clean_tr(TR0);
-  return(0); 
-
- overflow:
-  /* oops, we're in trouble */
-  H = HLow;
-  /* we've done it */
-  /* restore our nice, friendly, term to its original state */
-  HB = HB0;
-#ifdef RATIONAL_TREES
-  while (to_visit > to_visit0) {
-    to_visit -= 4;
-    pt0 = to_visit[0];
-    pt0_end = to_visit[1];
-    ptf = to_visit[2];
-    *pt0 = (CELL)to_visit[3];
-  }
-#endif
-  clean_tr(TR0);
-  return(-1);
-
- heap_overflow:
-  /* oops, we're in trouble */
-  H = HLow;
-  /* we've done it */
-  /* restore our nice, friendly, term to its original state */
-  HB = HB0;
-#ifdef RATIONAL_TREES
-  while (to_visit > to_visit0) {
-    to_visit -= 4;
-    pt0 = to_visit[0];
-    pt0_end = to_visit[1];
-    ptf = to_visit[2];
-    *pt0 = (CELL)to_visit[3];
-  }
-#endif
-  clean_tr(TR0);
-  return(-2);
+  Term t = CopyTerm(ARG1, 2, FALSE, TRUE); 
+  if (t == 0L)
+    return FALSE;
+  /* be careful, there may be a stack shift here */
+  return Yap_unify(ARG2,t);
 }
 
-static Term
-CopyTermNoDelays(Term inp) {
-  Term t = Deref(inp);
-  int res;
-
-  if (IsVarTerm(t)) {
-    return(MkVarTerm());
-  } else if (IsPrimitiveTerm(t)) {
-    return(t);
-  } else if (IsPairTerm(t)) {
-    Term tf;
-    CELL *ap, *Hi;
-
-  restart_list:
-    Hi = H;
-    ap = RepPair(t);
-    tf = AbsPair(H);
-    H += 2;
-    res = copy_complex_term_no_delays(ap-1, ap+1, H-2, H-2);
-    if (res) {
-      H = Hi;
-      if (res == -1) { /* handle overflow */
-	if (!Yap_gc(2, ENV, P)) {
-	  Yap_Error(OUT_OF_STACK_ERROR, TermNil, Yap_ErrorMessage);
-	  return(FALSE);
-	}
-	t = Deref(ARG1);
-	goto restart_list;
-      } else { /* handle overflow */
-	if (!Yap_ExpandPreAllocCodeSpace(0,NULL)) {
-	  Yap_Error(OUT_OF_AUXSPACE_ERROR, TermNil, Yap_ErrorMessage);
-	  return(FALSE);
-	}
-	t = Deref(ARG1);
-	goto restart_list;
-      }
-    }
-    return(tf);
-  } else {
-    Functor f;
-    Term tf;
-    CELL *HB0;
-    CELL *ap;
-
-  restart_appl:
-    f = FunctorOfTerm(t);
-    HB0 = H;
-    ap = RepAppl(t);
-    tf = AbsAppl(H);
-    H[0] = (CELL)f;
-    H += 1+ArityOfFunctor(f);
-    res = copy_complex_term_no_delays(ap, ap+ArityOfFunctor(f), HB0+1, HB0);
-    if (res) {
-      H = HB0;
-      if (res == -1) {
-	if (!Yap_gc(2, ENV, P)) {
-	  Yap_Error(OUT_OF_STACK_ERROR, TermNil, Yap_ErrorMessage);
-	  return(FALSE);
-	}
-	t = Deref(ARG1);
-	goto restart_appl;
-      } else { /* handle overflow */
-	if (!Yap_ExpandPreAllocCodeSpace(0,NULL)) {
-	  Yap_Error(OUT_OF_AUXSPACE_ERROR, TermNil, Yap_ErrorMessage);
-	  return(FALSE);
-	}
-	t = Deref(ARG1);
-	goto restart_appl;
-      }
-    }
-    return(tf);
-  }
-}
- 
 static Int 
 p_copy_term_no_delays(void)		/* copy term t to a new instance  */
 {
-  Term t = CopyTermNoDelays(ARG1);
+  Term t = CopyTerm(ARG1, 2, TRUE, FALSE);
   if (t == 0L) {
     return FALSE;
   }
@@ -715,6 +534,9 @@ static Term vars_in_complex_term(register CELL *pt0, register CELL *pt0_end, Ter
   vars_in_term_nvar:
     {
       if (IsPairTerm(d0)) {
+	if (to_visit + 1024 >= (CELL **)AuxSp) {
+	  goto aux_overflow;
+	}
 #ifdef RATIONAL_TREES
 	to_visit[0] = pt0;
 	to_visit[1] = pt0_end;
@@ -736,11 +558,13 @@ static Term vars_in_complex_term(register CELL *pt0, register CELL *pt0_end, Ter
 	/* store the terms to visit */
 	ap2 = RepAppl(d0);
 	f = (Functor)(*ap2);
-
 	if (IsExtensionFunctor(f)) {
 	  continue;
 	}
 	/* store the terms to visit */
+	if (to_visit + 1024 >= (CELL **)AuxSp) {
+	  goto aux_overflow;
+	}
 #ifdef RATIONAL_TREES
 	to_visit[0] = pt0;
 	to_visit[1] = pt0_end;
@@ -773,6 +597,12 @@ static Term vars_in_complex_term(register CELL *pt0, register CELL *pt0_end, Ter
     H += 2;
     H[-2] = (CELL)ptd0;
     /* next make sure noone will see this as a variable again */ 
+    if (TR > (tr_fr_ptr)Yap_TrailTop - 256) {
+      /* Trail overflow */
+      if (!Yap_growtrail((TR-TR0)*sizeof(tr_fr_ptr *), TRUE)) {
+	goto trail_overflow;
+      }
+    }
     TrailTerm(TR++) = (CELL)ptd0;
   }
   /* Do we still have compound terms to visit */
@@ -805,12 +635,81 @@ static Term vars_in_complex_term(register CELL *pt0, register CELL *pt0_end, Ter
   } else {
     return(inp);
   }
- global_overflow:
+
+ trail_overflow:
+#ifdef RATIONAL_TREES
+  while (to_visit > to_visit0) {
+    to_visit -= 3;
+    pt0 = to_visit[0];
+    *pt0 = (CELL)to_visit[2];
+  }
+#endif
+  Yap_Error_TYPE = OUT_OF_TRAIL_ERROR;
+  Yap_Error_Size = (TR-TR0)*sizeof(tr_fr_ptr *);
   clean_tr(TR0);
   Yap_ReleasePreAllocCodeSpace((ADDR)to_visit0);
   H = InitialH;
   return 0L;
   
+ aux_overflow:
+  Yap_Error_Size = (to_visit-to_visit0)*sizeof(CELL **);
+#ifdef RATIONAL_TREES
+  while (to_visit > to_visit0) {
+    to_visit -= 3;
+    pt0 = to_visit[0];
+    *pt0 = (CELL)to_visit[2];
+  }
+#endif
+  Yap_Error_TYPE = OUT_OF_AUXSPACE_ERROR;
+  clean_tr(TR0);
+  Yap_ReleasePreAllocCodeSpace((ADDR)to_visit0);
+  H = InitialH;
+  return 0L;
+  
+ global_overflow:
+#ifdef RATIONAL_TREES
+  while (to_visit > to_visit0) {
+    to_visit -= 3;
+    pt0 = to_visit[0];
+    *pt0 = (CELL)to_visit[2];
+  }
+#endif
+  clean_tr(TR0);
+  Yap_ReleasePreAllocCodeSpace((ADDR)to_visit0);
+  H = InitialH;
+  Yap_Error_TYPE = OUT_OF_STACK_ERROR;
+  Yap_Error_Size = (ASP-H)*sizeof(CELL);
+  return 0L;
+  
+}
+
+static int
+expand_vts(void)
+{
+  UInt expand = Yap_Error_Size;
+  yap_error_number yap_errno = Yap_Error_TYPE;
+  
+  Yap_Error_Size = 0;
+  Yap_Error_TYPE = YAP_NO_ERROR;
+  if (yap_errno == OUT_OF_TRAIL_ERROR) {
+    /* Trail overflow */
+    if (!Yap_growtrail(expand, FALSE)) {
+      return FALSE;
+    }
+  } else if (yap_errno == OUT_OF_AUXSPACE_ERROR) {
+    /* Aux space overflow */
+    if (expand > 4*1024*1024)
+      expand = 4*1024*1024;
+    if (!Yap_ExpandPreAllocCodeSpace(expand, NULL)) {
+      return FALSE;
+    }
+  } else {
+    if (!Yap_gcl(expand, 3, ENV, P)) {
+      Yap_Error(OUT_OF_STACK_ERROR, TermNil, "in term_variables");
+      return FALSE;
+    }
+  }
+  return TRUE;
 }
  
 static Int 
@@ -840,10 +739,8 @@ p_variables_in_term(void)	/* variables in term t		 */
 				 ArityOfFunctor(f), ARG2);
     }
     if (out == 0L) {
-      if (!Yap_gc(3, ENV, P)) {
-	Yap_Error(OUT_OF_STACK_ERROR, TermNil, "in term_variables");
+      if (!expand_vts())
 	return FALSE;
-      }
     }
   } while (out == 0L);
   return(Yap_unify(ARG3,out));
@@ -857,7 +754,11 @@ p_term_variables(void)	/* variables in term t		 */
   do {
     Term t = Deref(ARG1);
     if (IsVarTerm(t)) {
-      return Yap_unify(MkPairTerm(t,TermNil), ARG2);
+      Term out = Yap_MkNewPairTerm();
+      return
+	Yap_unify(t,HeadOfTerm(out)) &&
+	Yap_unify(TermNil, TailOfTerm(out)) &&
+	Yap_unify(out, ARG2);
     }  else if (IsPrimitiveTerm(t)) {
       return Yap_unify(TermNil, ARG2);
     } else if (IsPairTerm(t)) {
@@ -871,10 +772,8 @@ p_term_variables(void)	/* variables in term t		 */
 				 ArityOfFunctor(f), TermNil);
     }
     if (out == 0L) {
-      if (!Yap_gc(2, ENV, P)) {
-	Yap_Error(OUT_OF_STACK_ERROR, TermNil, "in term_variables");
+      if (!expand_vts())
 	return FALSE;
-      }
     }
   } while (out == 0L);
   return Yap_unify(ARG2,out);
@@ -888,7 +787,11 @@ p_term_variables3(void)	/* variables in term t		 */
   do {
     Term t = Deref(ARG1);
     if (IsVarTerm(t)) {
-      return Yap_unify(MkPairTerm(t,ARG3), ARG2);
+      Term out = Yap_MkNewPairTerm();
+      return
+	Yap_unify(t,HeadOfTerm(out)) &&
+	Yap_unify(ARG3, TailOfTerm(out)) &&
+	Yap_unify(out, ARG2);
     }  else if (IsPrimitiveTerm(t)) {
       return Yap_unify(ARG2, ARG3);
     } else if (IsPairTerm(t)) {
@@ -902,10 +805,8 @@ p_term_variables3(void)	/* variables in term t		 */
 				 ArityOfFunctor(f), ARG3);
     }
     if (out == 0L) {
-      if (!Yap_gc(3, ENV, P)) {
-	Yap_Error(OUT_OF_STACK_ERROR, TermNil, "in term_variables");
+      if (!expand_vts())
 	return FALSE;
-      }
     }
   } while (out == 0L);
 
@@ -1026,6 +927,13 @@ static Term non_singletons_in_complex_term(register CELL *pt0, register CELL *pt
   }
 
  aux_overflow:
+#ifdef RATIONAL_TREES
+  while (to_visit > to_visit0) {
+    to_visit -= 3;
+    pt0 = to_visit[0];
+    *pt0 = (CELL)to_visit[2];
+  }
+#endif
   clean_tr(TR0);
   if (H != InitialH) {
     /* close the list */
@@ -1169,7 +1077,6 @@ static Int ground_complex_term(register CELL *pt0, register CELL *pt0_end)
   while (to_visit > to_visit0) {
     to_visit -= 3;
     pt0 = to_visit[0];
-    pt0_end = to_visit[1];
     *pt0 = (CELL)to_visit[2];
   }
 #endif
@@ -1198,7 +1105,7 @@ p_ground(void)			/* ground(+T)		 */
       Functor fun = FunctorOfTerm(t);
       
       if (IsExtensionFunctor(fun))
-	return(TRUE);
+	return TRUE;
       else if ((out = ground_complex_term(RepAppl(t),
 					     RepAppl(t)+
 					     ArityOfFunctor(fun))) >= 0) {
@@ -1214,6 +1121,174 @@ p_ground(void)			/* ground(+T)		 */
   }
 }
 
+static int
+SizeOfExtension(Term t)
+{
+  Functor f = FunctorOfTerm(t);
+  if (f== FunctorDouble) {
+    return 2 + sizeof(Float)/sizeof(CELL);
+  }
+  if (f== FunctorLongInt) {
+    return 2 + sizeof(Float)/sizeof(CELL);
+  }
+  if (f== FunctorDBRef) {
+    return 0;
+  }
+  if (f== FunctorBigInt) {
+    CELL *pt = RepAppl(t)+1;
+    return 2+sizeof(MP_INT)+(((MP_INT *)(pt+1))->_mp_alloc*sizeof(mp_limb_t));
+  }
+  return 0;
+}
+
+
+static Int sz_ground_complex_term(register CELL *pt0, register CELL *pt0_end, int ground)
+{
+
+  register CELL **to_visit0, **to_visit = (CELL **)Yap_PreAllocCodeSpace();
+  Int sz = 0;
+
+  to_visit0 = to_visit;
+ loop:
+  while (pt0 < pt0_end) {
+    register CELL d0;
+    register CELL *ptd0;
+
+    ++pt0;
+    ptd0 = pt0;
+    d0 = *ptd0;
+    deref_head(d0, vars_in_term_unk);
+  vars_in_term_nvar:
+    {
+      if (IsPairTerm(d0)) {
+	sz += 2;
+	if (to_visit + 1024 >= (CELL **)AuxSp) {
+	  goto aux_overflow;
+	}
+#ifdef RATIONAL_TREES
+	to_visit[0] = pt0;
+	to_visit[1] = pt0_end;
+	to_visit[2] = (CELL *)*pt0;
+	to_visit += 3;
+	*pt0 = TermNil;
+#else
+	if (pt0 < pt0_end) {
+	  to_visit[0] = pt0;
+	  to_visit[1] = pt0_end;
+	  to_visit += 2;
+	}
+#endif
+	pt0 = RepPair(d0) - 1;
+	pt0_end = RepPair(d0) + 1;
+      } else if (IsApplTerm(d0)) {
+	register Functor f;
+	register CELL *ap2;
+	/* store the terms to visit */
+	ap2 = RepAppl(d0);
+	f = (Functor)(*ap2);
+
+	if (IsExtensionFunctor(f)) {
+	  sz += SizeOfExtension(d0);
+	  continue;
+	}
+	if (to_visit + 1024 >= (CELL **)AuxSp) {
+	  goto aux_overflow;
+	}
+#ifdef RATIONAL_TREES
+	to_visit[0] = pt0;
+	to_visit[1] = pt0_end;
+	to_visit[2] = (CELL *)*pt0;
+	to_visit += 3;
+	*pt0 = TermNil;
+#else
+	/* store the terms to visit */
+	if (pt0 < pt0_end) {
+	  to_visit[0] = pt0;
+	  to_visit[1] = pt0_end;
+	  to_visit += 2;
+	}
+#endif
+	d0 = ArityOfFunctor(f);
+	sz += (1+d0);
+	pt0 = ap2;
+	pt0_end = ap2 + d0;
+      }
+      continue;
+    }
+    
+
+    derefa_body(d0, ptd0, vars_in_term_unk, vars_in_term_nvar);
+    if (!ground)
+      continue;
+#ifdef RATIONAL_TREES
+    while (to_visit > to_visit0) {
+      to_visit -= 3;
+      pt0 = to_visit[0];
+      pt0_end = to_visit[1];
+      *pt0 = (CELL)to_visit[2];
+    }
+#endif
+    return 0;
+  }
+  /* Do we still have compound terms to visit */
+  if (to_visit > to_visit0) {
+#ifdef RATIONAL_TREES
+    to_visit -= 3;
+    pt0 = to_visit[0];
+    pt0_end = to_visit[1];
+    *pt0 = (CELL)to_visit[2];
+#else
+    to_visit -= 2;
+    pt0 = to_visit[0];
+    pt0_end = to_visit[1];
+#endif
+    goto loop;
+  }
+  return sz;
+
+ aux_overflow:
+  /* unwind stack */
+#ifdef RATIONAL_TREES
+  while (to_visit > to_visit0) {
+    to_visit -= 3;
+    pt0 = to_visit[0];
+    *pt0 = (CELL)to_visit[2];
+  }
+#endif
+  return -1;
+}
+
+int
+Yap_SizeGroundTerm(Term t, int ground)
+{
+  if (IsVarTerm(t)) {
+    if (!ground)
+      return 1;
+    return 0;
+  }  else if (IsPrimitiveTerm(t)) {
+    return 1;
+  } else if (IsPairTerm(t)) {
+    int sz = sz_ground_complex_term(RepPair(t)-1, RepPair(t)+1, ground);
+    if (sz <= 0)
+      return sz;
+    return sz+2;
+} else {
+  int sz = 0;
+  Functor fun = FunctorOfTerm(t);
+      
+  if (IsExtensionFunctor(fun))
+    return 1+ SizeOfExtension(t);
+  
+   sz = sz_ground_complex_term(RepAppl(t),
+			    RepAppl(t)+
+			    ArityOfFunctor(fun),
+			    ground);
+   if (sz <= 0)
+     return sz;
+   return 1+ArityOfFunctor(fun)+sz;
+  }  
+}
+
 static Int var_in_complex_term(register CELL *pt0,
 			       register CELL *pt0_end,
 			       Term v)
@@ -1346,22 +1421,23 @@ static Int TermHash(Term t1, Int depth_lim, Int k)
   Int i;
   if (IsVarTerm(t1)) {
     return(-1);
-  } else if (IsAtomTerm(t1)) {
-    register char *s = AtomName(AtomOfTerm(t1));
-    for (i=0; s[i]; i++)
+  } else if (IsAtomOrIntTerm(t1)) {
+    if (IsAtomTerm(t1)) {
+      register char *s = AtomName(AtomOfTerm(t1));
+      for (i=0; s[i]; i++)
 	HASHADD(s[i]);
-    return(k);
+      return k;
+    } else {
+      HASHADD(IntOfTerm(t1));
+      return k;
+    }
   } else if (IsPairTerm(t1)) {
     HASHADD('.');
     depth_lim--;
     if (depth_lim == 0) return(TRUE);
     k = TermHash(HeadOfTerm(t1),depth_lim,k);
-    if (k < 0) return(-1);
-    k = TermHash(TailOfTerm(t1),depth_lim,k);
-    return(k);
-  } else if (IsIntTerm(t1)) {
-    HASHADD(IntOfTerm(t1));
-    return(k);
+    if (k < 0) return k;
+    return TermHash(TailOfTerm(t1),depth_lim,k);
   } else {
     Functor f = FunctorOfTerm(t1);
 
@@ -1385,17 +1461,18 @@ static Int TermHash(Term t1, Int depth_lim, Int k)
 	return(-1);
       }
     } else {
-      int ar = ArityOfFunctor(FunctorOfTerm(t1));
-      int res = TRUE;
-      register char *s = AtomName(NameOfFunctor(f));
+      int ar;
+      char *s;
 
-      depth_lim--;
-      if (depth_lim == 0) return(TRUE);
+      s = AtomName(NameOfFunctor(f));
       for (i=0; s[i]; i++) 
 	HASHADD(s[i]);
-      for (i=1; i<=ar && res; i++) {
+      depth_lim--;
+      if (depth_lim == 0) return k;
+      ar = ArityOfFunctor(f);
+      for (i=1; i<=ar; i++) {
 	k = TermHash(ArgOfTerm(i,t1),depth_lim,k);
-	if (k == -1 ) return(-1);
+	if (k < 0) return k;
       }
       return(k);
     }
@@ -1681,7 +1758,7 @@ p_variant(void) /* variant terms t1 and t2	 */
   }
  error:
   if (out == -1) {
-    if (!Yap_gc(2, ENV, P)) {
+    if (!Yap_gcl((ASP-H)*sizeof(CELL), 2, ENV, P)) {
       Yap_Error(OUT_OF_STACK_ERROR, TermNil, "in variant");
       return FALSE;
     }
@@ -1970,10 +2047,31 @@ camacho_dum(void)
 
 #endif /* DEBUG */
 
+static Int
+p_unifiable(void)
+{
+  tr_fr_ptr trp;
+  Term tf = TermNil;
+  if (!Yap_unify(ARG1,ARG2)) {
+    return FALSE;
+  }
+  trp = TR;
+  while (trp != B->cp_tr) {
+    Term t[2];
+    --trp;
+    t[0] = TrailTerm(trp);
+    t[1] = *(CELL *)t[0];
+    tf = MkPairTerm(Yap_MkApplTerm(FunctorEq,2,t),tf);
+    RESET_VARIABLE(t[0]);
+  }
+  return Yap_unify(ARG3, tf);
+}
+
 void Yap_InitUtilCPreds(void)
 {
   Term cm = CurrentModule;
   Yap_InitCPred("copy_term", 2, p_copy_term, 0);
+  Yap_InitCPred("duplicate_term", 2, p_duplicate_term, 0);
   Yap_InitCPred("copy_term_nat", 2, p_copy_term_no_delays, 0);
   Yap_InitCPred("ground", 1, p_ground, SafePredFlag);
   Yap_InitCPred("$variables_in_term", 3, p_variables_in_term, HiddenPredFlag);
@@ -1985,6 +2083,7 @@ void Yap_InitUtilCPreds(void)
   Yap_InitCPred("term_hash", 4, GvNTermHash, SafePredFlag);
   Yap_InitCPred("variant", 2, p_variant, 0);
   Yap_InitCPred("subsumes", 2, p_subsumes, SafePredFlag);
+  Yap_InitCPred("protected_unifiable", 3, p_unifiable, 0);
   CurrentModule = cm;
 #ifdef DEBUG
   Yap_InitCPred("$force_trail_expansion", 1, p_force_trail_expansion, SafePredFlag|HiddenPredFlag);
