@@ -10,7 +10,7 @@
 * File:		Heap.h         						 *
 * mods:									 *
 * comments:	Heap Init Structure					 *
-* version:      $Id: Heap.h,v 1.94 2006/03/30 01:11:10 vsc Exp $	 *
+* version:      $Id: Heap.h,v 1.132 2008/05/28 17:18:35 vsc Exp $	 *
 *************************************************************************/
 
 /* information that can be stored in Code Space */
@@ -19,15 +19,41 @@
 #ifndef HEAP_H
 #define HEAP_H 1
 
+#if defined(YAPOR) || defined(THREADS)
+#define WL wl[worker_id]
+#else
+#define WL wl
+#endif
+
+#if defined(THREADS)
+#define RINFO rinfo[worker_id]
+#else
+#define RINFO rinfo
+#endif
+
 #ifndef INT_KEYS_DEFAULT_SIZE
 #define INT_KEYS_DEFAULT_SIZE 256
 #endif
 
+#if USE_DL_MALLOC
+
+#define MAX_DLMALLOC_HOLES 32
+
+typedef struct memory_hole {
+  ADDR start;
+  ADDR end;
+} memory_hole_type;
+#endif
+
 
 #define GC_MAVARS_HASH_SIZE 512
 
 typedef struct gc_ma_hash_entry_struct {
   UInt timestmp;
+#ifdef TABLING
+  tr_fr_ptr loc;
+  struct gc_ma_hash_entry_struct *more;
+#endif
   CELL* addr;
   struct gc_ma_hash_entry_struct *next;
 } gc_ma_hash_entry;
@@ -56,20 +82,23 @@ typedef struct scratch_block_struct {
 typedef struct restore_info {
   Int cl_diff,
     g_diff,
+    g_diff0,
     h_diff,
     l_diff,
     tr_diff,
     x_diff,
     delay_diff;
   CELL    *old_ASP, *old_LCL0;
+  CELL    *g_split;
   tr_fr_ptr old_TR;  
   CELL    *old_GlobalBase, *old_H, *old_H0;
+  CELL    *old_DelayTop, *current_DelayTop;
   ADDR     old_TrailBase, old_TrailTop;
   ADDR     old_HeapBase, old_HeapTop;
 } restoreinfo;
 
-#if THREADS
-extern struct restore_info rinfo[MAX_WORKERS];
+#if defined(THREADS)
+extern struct restore_info rinfo[MAX_THREADS];
 #else
 extern struct restore_info rinfo;
 #endif
@@ -85,6 +114,10 @@ typedef struct worker_local_struct {
 #ifdef USE_GMP
   mpz_t  big_tmp;
 #endif
+  union CONSULT_OBJ *consultsp;
+  union CONSULT_OBJ *consultbase;
+  union CONSULT_OBJ *consultlow;
+  UInt   consultcapacity;
   UInt   active_signals;
   UInt   i_pred_arity;
   yamop *prof_end;
@@ -103,14 +136,13 @@ typedef struct worker_local_struct {
   unsigned int      gc_calls;	/* number of times GC has been called */ 
   Int      tot_gc_time; /* total time spent in GC */
   YAP_ULONG_LONG      tot_gc_recovered; /* number of heap objects in all garbage collections */
+  Int      last_gc_time, last_ss_time; /* number of heap objects in all garbage collections */
 /* in a single gc */
 #if defined(YAPOR) || defined(THREADS)
   /* otherwise, use global variables for speed */
   unsigned long int   tot_marked, tot_oldies;	/* number of heap objects marked */
-#if DEBUG
 #ifdef COROUTINING
   unsigned long int   tot_smarked;
-#endif
 #endif
   struct choicept *wl_current_B;
 #if defined(TABLING) || defined(SBA)
@@ -131,6 +163,7 @@ typedef struct worker_local_struct {
   int disc_trail_entries;
   gc_ma_hash_entry Gc_ma_hash_table[GC_MAVARS_HASH_SIZE];
   gc_ma_hash_entry *Gc_ma_h_top;
+  gc_ma_hash_entry *Gc_ma_h_list;
   UInt Gc_timestamp;    /* an unsigned int */
   ADDR  DB_vec, DB_vec0;
   struct RB_red_blk_node *DB_root, *DB_nil;
@@ -138,14 +171,21 @@ typedef struct worker_local_struct {
   jmp_buf  gc_restore; /* where to jump if garbage collection crashes */
   struct array_entry *dynamic_arrays;
   struct static_array_entry *static_arrays;
+  struct global_entry *global_variables;
+  int allow_restart;
+  Term global_arena;
+  UInt global_arena_overflows; 
+  Term global_delay_arena;
   yamop trust_lu_code[3];
 } worker_local;
 
 #ifdef THREADS
 typedef struct thandle {
   int in_use;
+  int zombie;
   UInt ssize;
   UInt tsize;
+  void *stack_address;
   Term tdetach;
   Term  cmod;
   struct DB_TERM *tgoal;
@@ -156,6 +196,11 @@ typedef struct thandle {
   struct pred_entry *local_preds;
   pthread_t handle;
   int ref_count;
+#ifdef LOW_LEVEL_TRACER
+  long long int thread_inst_count;
+  int been_here1;
+  int been_here2;
+#endif
   pthread_mutex_t tlock;
 #if HAVE_GETRUSAGE
   struct timeval *start_of_timesp;
@@ -168,8 +213,15 @@ typedef int   (*Agc_hook)(Atom);
 
 typedef struct various_codes {
   special_functors funcs;
+  UInt hole_size;
   struct malloc_state *av_;
-  ADDR hole_start, hole_end;
+
+  UInt clause_space, index_space_Tree, index_space_EXT, index_space_SW;
+  UInt lu_clause_space, lu_index_space_Tree, lu_index_space_CP, lu_index_space_EXT, lu_index_space_SW;
+#if USE_DL_MALLOC
+  struct memory_hole memory_holes[MAX_DLMALLOC_HOLES];
+  UInt nof_memory_holes;
+#endif
   Int heap_used;
   Int heap_max;
   ADDR heap_top;
@@ -178,10 +230,13 @@ typedef struct various_codes {
 #if defined(YAPOR) || defined(THREADS)
   lockvar  bgl;		 /* protect long critical regions   */
   lockvar  free_blocks_lock;     /* protect the list of free blocks */
-  worker_local wl[MAX_WORKERS];
+  worker_local wl[MAX_AGENTS];
 #else
   worker_local wl;
 #endif
+#ifdef BEAM
+  yamop beam_retry_code;
+#endif /* BEAM */
 #ifdef YAPOR
   int seq_def;
   yamop getwork_code;
@@ -196,8 +251,10 @@ typedef struct various_codes {
 #endif /* TABLING */
   OPCODE expand_op_code;
   yamop *expand_clauses_first, *expand_clauses_last;
+  UInt expand_clauses;
 #if defined(YAPOR) || defined(THREADS)
   lockvar expand_clauses_list_lock;
+  lockvar op_list_lock;
 #endif
   yamop comma_code[5];
   yamop failcode[1];
@@ -232,18 +289,14 @@ typedef struct various_codes {
     struct pred_entry *p;
     struct pred_entry *p0;
   } env_for_yes_code; /* sla */
-  yamop yescode[1];
-  yamop nocode[1];
-  yamop rtrycode[1];
+  struct yami yescode;
+  struct yami nocode;
+  struct yami rtrycode;
   struct {
     OPREG arity;
     struct yami *clause;
     Functor func;
   } clausecode[1];
-  union CONSULT_OBJ *consultsp;
-  union CONSULT_OBJ *consultbase;
-  union CONSULT_OBJ *consultlow;
-  UInt   consultcapacity;
 #if HAVE_LIBREADLINE
   char *readline_buf, *readline_pos;
 #endif
@@ -257,6 +310,7 @@ typedef struct various_codes {
   struct pred_entry  *creep_code;
   struct pred_entry  *undef_code;
   struct pred_entry  *spy_code;
+  UInt new_cps, live_cps, dirty_cps, freed_cps;
   int   system_profiling;
   int   system_call_counting;
   int   system_pred_goal_expansion_all;
@@ -272,6 +326,7 @@ typedef struct various_codes {
   char prompt[MAX_PROMPT];
   OPCODE undef_op; 
   OPCODE index_op; 
+  OPCODE lockpred_op; 
   OPCODE fail_op; 
   yamop *retry_recorded_k_code,
     *retry_c_recordedp_code;
@@ -287,7 +342,7 @@ typedef struct various_codes {
   Int   yap_flags_field[NUMBER_OF_YAP_FLAGS];
   char *char_conversion_table;
   char *char_conversion_table2;
-#if THREADS
+#if defined(THREADS) || defined(YAPOR)
   unsigned int n_of_threads;      /* number of threads and processes in system */
   unsigned int n_of_threads_created;      /* number of threads created since start */
   UInt  threads_total_time;      /* total run time for dead threads */
@@ -298,18 +353,20 @@ typedef struct various_codes {
   lockvar  dead_static_clauses_lock;        /* protect DeadStaticClauses */
   lockvar  dead_mega_clauses_lock;        /* protect DeadMegaClauses */
   lockvar  dead_static_indices_lock;        /* protect DeadStaticIndices */
+  lockvar  dbterms_list_lock;        /* protect DBTermList */
   int      heap_top_owner;
 #ifdef LOW_LEVEL_TRACER
   lockvar  low_level_trace_lock;
 #endif
 #endif
   unsigned int size_of_overflow;
-  Term  module_name[MaxModules];
-  struct pred_entry *module_pred[MaxModules];
-  SMALLUNSGN   no_of_modules;
+  struct mod_entry *current_modules;
+  struct operator_entry *op_list;
+  struct hold_entry *global_hold_entry;
   struct static_clause *dead_static_clauses;
   struct static_mega_clause *dead_mega_clauses;
   struct static_index *dead_static_indices;
+  struct dbterm_list *dbterms_list;
   Atom
     atom_abol,
     atom_alarm,
@@ -404,6 +461,7 @@ typedef struct various_codes {
     functor_cut_by,
     functor_eq,
     functor_execute_in_mod,
+    functor_execute2_in_mod,
     functor_execute_within,
     functor_g_atom,
     functor_g_atomic,
@@ -422,6 +480,7 @@ typedef struct various_codes {
 #ifdef MULTI_ASSIGNMENT_VARIABLES
     functor_mutable,
 #endif
+    functor_nb_queue,
     functor_not,
     functor_or,
     functor_portray,
@@ -447,13 +506,20 @@ typedef struct various_codes {
     idb_module,
     attributes_module,
     charsio_module,
-    terms_module;
+    terms_module,
+    system_module,
+    readutil_module,
+    hacks_module,
+    arg_module,
+    globals_module,
+    swi_module;
   void *last_wtime;
   struct pred_entry *pred_goal_expansion;
   struct pred_entry *pred_meta_call;
   struct pred_entry *pred_dollar_catch;
   struct pred_entry *pred_recorded_with_key;
   struct pred_entry *pred_log_upd_clause;
+  struct pred_entry *pred_log_upd_clause_erase;
   struct pred_entry *pred_log_upd_clause0;
   struct pred_entry *pred_static_clause;
   struct pred_entry *pred_throw;
@@ -464,9 +530,7 @@ typedef struct various_codes {
   struct logic_upd_index *db_erased_ilist;
   UInt expand_clauses_sz;
   struct stream_desc *yap_streams;
-#ifdef DEBUG
   int    debugger_output_msg;
-#endif
   UInt n_of_file_aliases;
   UInt sz_of_file_aliases;
   struct AliasDescS * file_aliases;
@@ -478,11 +542,18 @@ typedef struct various_codes {
 #endif /* LOW_PROF */
   struct reduction_counters call_counters;
   char *yap_lib_dir;
+  UInt      agc_last_call; /* amount of space recovered in all garbage collections */
+  UInt      agc_threshold; /* amount of space recovered in all garbage collections */
   Agc_hook  agc_hook;
   void *foreign_code_loaded;
   ADDR  foreign_code_base;
   ADDR  foreign_code_top;
   ADDR  foreign_code_max;
+  struct pred_entry **pred_hash;
+#if defined(YAPOR) || defined(THREADS)
+  rwlock_t pred_hash_rw_lock;
+#endif
+  UInt  preds_in_hash_table, pred_hash_table_size;
   int parser_error_style;
   int   compiler_profiling;
   int   compiler_call_counting;
@@ -499,10 +570,13 @@ typedef struct various_codes {
 #endif /* YAPOR || TABLING */
 #ifdef THREADS
   lockvar  thread_handles_lock;        /* protect ThreadManipulation */
-  struct thandle thread_handle[MAX_WORKERS];
+  struct thandle thread_handle[MAX_THREADS];
 #endif
   UInt n_of_atoms;
   UInt atom_hash_table_size;
+  UInt wide_atom_hash_table_size;
+  UInt n_of_wide_atoms;
+  AtomHashEntry *wide_hash_chain;
   AtomHashEntry *hash_chain;
 } all_heap_codes;
 
@@ -512,10 +586,28 @@ struct various_codes *Yap_heap_regs;
 #define Yap_heap_regs  ((all_heap_codes *)HEAP_INIT_BASE)
 #endif
 
+#define  Yap_HoleSize            Yap_heap_regs->hole_size
 #define  Yap_av                  Yap_heap_regs->av_
-#define  Yap_hole_start          Yap_heap_regs->hole_start
-#define  Yap_hole_end            Yap_heap_regs->hole_end
+#define  Yap_ClauseSpace         Yap_heap_regs->clause_space
+#define  Yap_IndexSpace_Tree     Yap_heap_regs->index_space_Tree
+#define  Yap_IndexSpace_EXT      Yap_heap_regs->index_space_EXT
+#define  Yap_IndexSpace_SW       Yap_heap_regs->index_space_SW
+#define  Yap_LUClauseSpace       Yap_heap_regs->lu_clause_space
+#define  Yap_LUIndexSpace_Tree   Yap_heap_regs->lu_index_space_Tree
+#define  Yap_LUIndexSpace_CP     Yap_heap_regs->lu_index_space_CP
+#define  Yap_LUIndexSpace_EXT    Yap_heap_regs->lu_index_space_EXT
+#define  Yap_LUIndexSpace_SW     Yap_heap_regs->lu_index_space_SW
+#define  Yap_MemoryHoles         Yap_heap_regs->memory_holes
+#define  Yap_NOfMemoryHoles      Yap_heap_regs->nof_memory_holes
+#if USE_DL_MALLOC || (USE_SYSTEM_MALLOC && HAVE_MALLINFO)
+#define  HeapUsed                Yap_givemallinfo()
+#ifdef YAPOR
+#define  HeapUsedLock            Yap_heap_regs->heap_used_lock
+#endif
+#else
 #define  HeapUsed                Yap_heap_regs->heap_used
+#define  HeapUsedLock            Yap_heap_regs->heap_used_lock
+#endif
 #define  HeapMax                 Yap_heap_regs->heap_max
 #define  HeapTop                 Yap_heap_regs->heap_top
 #define  HeapLim                 Yap_heap_regs->heap_lim
@@ -525,6 +617,9 @@ struct various_codes *Yap_heap_regs;
 #define  GETWORK_SEQ             (&(Yap_heap_regs->getwork_seq_code))
 #define  GETWORK_FIRST_TIME      (&(Yap_heap_regs->getwork_first_time_code))
 #endif /* YAPOR */
+#ifdef BEAM
+#define  BEAM_RETRY_CODE         ((yamop *)&(Yap_heap_regs->beam_retry_code)) 
+#endif /* BEAM */
 #ifdef TABLING
 #define  LOAD_ANSWER              ((yamop *)&(Yap_heap_regs->table_load_answer_code))
 #define  TRY_ANSWER               ((yamop *)&(Yap_heap_regs->table_try_answer_code))
@@ -535,12 +630,14 @@ struct various_codes *Yap_heap_regs;
 #define  ExpandClausesFirst       Yap_heap_regs->expand_clauses_first
 #define  ExpandClausesLast        Yap_heap_regs->expand_clauses_last
 #define  ExpandClausesListLock    Yap_heap_regs->expand_clauses_list_lock
+#define  Yap_ExpandClauses        Yap_heap_regs->expand_clauses
+#define  OpListLock               Yap_heap_regs->op_list_lock
 #define  COMMA_CODE               Yap_heap_regs->comma_code
 #define  FAILCODE                 Yap_heap_regs->failcode
 #define  TRUSTFAILCODE            Yap_heap_regs->trustfailcode
-#define  YESCODE                  Yap_heap_regs->yescode
-#define  NOCODE                   Yap_heap_regs->nocode
-#define  RTRYCODE                 Yap_heap_regs->rtrycode
+#define  YESCODE                  (&Yap_heap_regs->yescode)
+#define  NOCODE                   (&Yap_heap_regs->nocode)
+#define  RTRYCODE                 (&Yap_heap_regs->rtrycode)
 #define  DUMMYCODE                Yap_heap_regs->dummycode
 #define  CLAUSECODE               Yap_heap_regs->clausecode
 #define  INVISIBLECHAIN           Yap_heap_regs->invisiblechain
@@ -565,6 +662,7 @@ struct various_codes *Yap_heap_regs;
 #define  yap_flags                Yap_heap_regs->yap_flags_field
 #define  UNDEF_OPCODE             Yap_heap_regs->undef_op
 #define  INDEX_OPCODE             Yap_heap_regs->index_op
+#define  LOCKPRED_OPCODE          Yap_heap_regs->lockpred_op
 #define  FAIL_OPCODE              Yap_heap_regs->fail_op
 #ifdef THREADS
 #define  ThreadHandlesLock	  Yap_heap_regs->thread_handles_lock
@@ -573,6 +671,9 @@ struct various_codes *Yap_heap_regs;
 #define  NOfAtoms                 Yap_heap_regs->n_of_atoms
 #define  AtomHashTableSize        Yap_heap_regs->atom_hash_table_size
 #define  HashChain                Yap_heap_regs->hash_chain
+#define  NOfWideAtoms             Yap_heap_regs->n_of_wide_atoms
+#define  WideAtomHashTableSize    Yap_heap_regs->wide_atom_hash_table_size
+#define  WideHashChain            Yap_heap_regs->wide_hash_chain
 #define  INT_KEYS_SIZE            Yap_heap_regs->int_keys_size
 #define  INT_KEYS_TIMESTAMP       Yap_heap_regs->int_keys_timestamp
 #define  INT_KEYS                 Yap_heap_regs->IntKeys
@@ -583,9 +684,8 @@ struct various_codes *Yap_heap_regs;
 #define  INT_BB_KEYS              Yap_heap_regs->IntBBKeys
 #define  CharConversionTable      Yap_heap_regs->char_conversion_table
 #define  CharConversionTable2     Yap_heap_regs->char_conversion_table2
-#define  ModuleName               Yap_heap_regs->module_name
-#define  ModulePred               Yap_heap_regs->module_pred
-#define  NoOfModules              Yap_heap_regs->no_of_modules
+#define  CurrentModules		  Yap_heap_regs->current_modules
+#define  OpList                   Yap_heap_regs->op_list
 #define  AtomAbol                 Yap_heap_regs->atom_abol
 #define  AtomAlarm                Yap_heap_regs->atom_alarm
 #define  AtomAppend               Yap_heap_regs->atom_append
@@ -674,6 +774,7 @@ struct various_codes *Yap_heap_regs;
 #define  FunctorCutBy             Yap_heap_regs->functor_cut_by
 #define  FunctorEq                Yap_heap_regs->functor_eq
 #define  FunctorExecuteInMod      Yap_heap_regs->functor_execute_in_mod
+#define  FunctorExecute2InMod     Yap_heap_regs->functor_execute2_in_mod
 #define  FunctorExecuteWithin     Yap_heap_regs->functor_execute_within
 #define  FunctorGAtom             Yap_heap_regs->functor_g_atom
 #define  FunctorGAtomic           Yap_heap_regs->functor_g_atomic
@@ -692,6 +793,7 @@ struct various_codes *Yap_heap_regs;
 #ifdef MULTI_ASSIGNMENT_VARIABLES
 #define  FunctorMutable           Yap_heap_regs->functor_mutable
 #endif
+#define  FunctorNBQueue           Yap_heap_regs->functor_nb_queue
 #define  FunctorNot               Yap_heap_regs->functor_not
 #define  FunctorOr                Yap_heap_regs->functor_or
 #define  FunctorPortray           Yap_heap_regs->functor_portray
@@ -717,11 +819,18 @@ struct various_codes *Yap_heap_regs;
 #define  ATTRIBUTES_MODULE        Yap_heap_regs->attributes_module
 #define  CHARSIO_MODULE           Yap_heap_regs->charsio_module
 #define  TERMS_MODULE             Yap_heap_regs->terms_module
+#define  SYSTEM_MODULE            Yap_heap_regs->system_module
+#define  READUTIL_MODULE          Yap_heap_regs->readutil_module
+#define  HACKS_MODULE             Yap_heap_regs->hacks_module
+#define  GLOBALS_MODULE           Yap_heap_regs->globals_module
+#define  ARG_MODULE	          Yap_heap_regs->arg_module
+#define  SWI_MODULE               Yap_heap_regs->swi_module
 #define  PredGoalExpansion        Yap_heap_regs->pred_goal_expansion
 #define  PredMetaCall             Yap_heap_regs->pred_meta_call
 #define  PredDollarCatch          Yap_heap_regs->pred_dollar_catch
 #define  PredRecordedWithKey      Yap_heap_regs->pred_recorded_with_key
 #define  PredLogUpdClause         Yap_heap_regs->pred_log_upd_clause
+#define  PredLogUpdClauseErase    Yap_heap_regs->pred_log_upd_clause_erase
 #define  PredLogUpdClause0        Yap_heap_regs->pred_log_upd_clause0
 #define  PredStaticClause         Yap_heap_regs->pred_static_clause
 #define  PredThrow                Yap_heap_regs->pred_throw
@@ -750,54 +859,16 @@ struct various_codes *Yap_heap_regs;
 #define  PredEntriesCounterOn     Yap_heap_regs->call_counters.reductions_retries_on
 #define  RetriesCounterOn         Yap_heap_regs->call_counters.retries_on
 #define  Yap_LibDir               Yap_heap_regs->yap_lib_dir
+#define  AGcLastCall              Yap_heap_regs->agc_last_call
+#define  AGcThreshold             Yap_heap_regs->agc_threshold
 #define  AGCHook                  Yap_heap_regs->agc_hook
 #define  ParserErrorStyle         Yap_heap_regs->parser_error_style
 #ifdef  COROUTINING
 #define  WakeUpCode               Yap_heap_regs->wake_up_code
 #endif
 #if defined(YAPOR) || defined(THREADS)
-/* The old stack pointers */
-#define  OldASP                   rinfo[worker_id].old_ASP
-#define  OldLCL0                  rinfo[worker_id].old_LCL0
-#define  OldTR                    rinfo[worker_id].old_TR
-#define  OldGlobalBase            rinfo[worker_id].old_GlobalBase
-#define  OldH                     rinfo[worker_id].old_H
-#define  OldH0                    rinfo[worker_id].old_H0
-#define  OldTrailBase             rinfo[worker_id].old_TrailBase
-#define  OldTrailTop              rinfo[worker_id].old_TrailTop
-#define  OldHeapBase              rinfo[worker_id].old_HeapBase
-#define  OldHeapTop               rinfo[worker_id].old_HeapTop
-#define  ClDiff                   rinfo[worker_id].cl_diff
-#define  GDiff                    rinfo[worker_id].g_diff
-#define  HDiff                    rinfo[worker_id].h_diff
-#define  LDiff                    rinfo[worker_id].l_diff
-#define  TrDiff                   rinfo[worker_id].tr_diff
-#define  XDiff                    rinfo[worker_id].x_diff
-#define  DelayDiff                rinfo[worker_id].delay_diff
-#define  FormatInfo               Yap_heap_regs->wl[worker_id].f_info
-#define  ScannerStack             Yap_heap_regs->wl[worker_id].scanner_stack
-#define  ScannerExtraBlocks       Yap_heap_regs->wl[worker_id].scanner_extra_blocks
 #define  SignalLock               Yap_heap_regs->wl[worker_id].signal_lock
 #define  WPP                      Yap_heap_regs->wl[worker_id].wpp
-#define  UncaughtThrow            Yap_heap_regs->wl[worker_id].uncaught_throw
-#define  DoingUndefp              Yap_heap_regs->wl[worker_id].doing_undefp
-#define  Yap_BigTmp               Yap_heap_regs->wl[worker_id].big_tmp
-#define  ActiveSignals            Yap_heap_regs->wl[worker_id].active_signals
-#define  IPredArity               Yap_heap_regs->wl[worker_id].i_pred_arity
-#define  ProfEnd                  Yap_heap_regs->wl[worker_id].prof_end
-#define  StartLine                Yap_heap_regs->wl[worker_id].start_line
-#define  ScratchPad               Yap_heap_regs->wl[worker_id].scratchpad
-#ifdef  COROUTINING
-#define  WokenGoals               Yap_heap_regs->wl[worker_id].woken_goals
-#define  AttsMutableList          Yap_heap_regs->wl[worker_id].atts_mutable_list
-#endif
-/* support for generations with backtracking */
-#define  GcCalls                  Yap_heap_regs->wl[worker_id].gc_calls
-#define  GcGeneration             Yap_heap_regs->wl[worker_id].gc_generation
-#define  GcPhase                  Yap_heap_regs->wl[worker_id].gc_phase
-#define  GcCurrentPhase           Yap_heap_regs->wl[worker_id].gc_current_phase
-#define  TotGcTime                Yap_heap_regs->wl[worker_id].tot_gc_time
-#define  TotGcRecovered           Yap_heap_regs->wl[worker_id].tot_gc_recovered
 #define  total_marked             Yap_heap_regs->wl[worker_id].tot_marked
 #define  total_oldies             Yap_heap_regs->wl[worker_id].tot_oldies
 #if DEBUG
@@ -818,6 +889,7 @@ struct various_codes *Yap_heap_regs;
 #endif /* GC_NO_TAGS */
 #define  gc_ma_hash_table         Yap_heap_regs->wl[worker_id].Gc_ma_hash_table
 #define  gc_ma_h_top              Yap_heap_regs->wl[worker_id].Gc_ma_h_top
+#define  gc_ma_h_list             Yap_heap_regs->wl[worker_id].Gc_ma_h_list
 #define  gc_timestamp             Yap_heap_regs->wl[worker_id].Gc_timestamp
 #define  cont_top0                Yap_heap_regs->wl[worker_id].conttop0
 #define  db_vec                   Yap_heap_regs->wl[worker_id].DB_vec
@@ -825,54 +897,68 @@ struct various_codes *Yap_heap_regs;
 #define  db_root                  Yap_heap_regs->wl[worker_id].DB_root
 #define  db_nil                   Yap_heap_regs->wl[worker_id].DB_nil
 #define  cont_top                 Yap_heap_regs->wl[worker_id].conttop
-#define  Yap_gc_restore           Yap_heap_regs->wl[worker_id].gc_restore
-#define  TrustLUCode              Yap_heap_regs->wl[worker_id].trust_lu_code
-#define  DynamicArrays            Yap_heap_regs->wl[worker_id].dynamic_arrays
-#define  StaticArrays             Yap_heap_regs->wl[worker_id].static_arrays
-#else
-#define  OldASP                   rinfo.old_ASP
-#define  OldLCL0                  rinfo.old_LCL0
-#define  OldTR                    rinfo.old_TR
-#define  OldGlobalBase            rinfo.old_GlobalBase
-#define  OldH                     rinfo.old_H
-#define  OldH0                    rinfo.old_H0
-#define  OldTrailBase             rinfo.old_TrailBase
-#define  OldTrailTop              rinfo.old_TrailTop
-#define  OldHeapBase              rinfo.old_HeapBase
-#define  OldHeapTop               rinfo.old_HeapTop
-#define  ClDiff                   rinfo.cl_diff
-#define  GDiff                    rinfo.g_diff
-#define  HDiff                    rinfo.h_diff
-#define  LDiff                    rinfo.l_diff
-#define  TrDiff                   rinfo.tr_diff
-#define  XDiff                    rinfo.x_diff
-#define  DelayDiff                rinfo.delay_diff
-#define  FormatInfo               Yap_heap_regs->wl.f_info
-#define  ScannerStack             Yap_heap_regs->wl.scanner_stack
-#define  ScannerExtraBlocks       Yap_heap_regs->wl.scanner_extra_blocks
-#define  Yap_BigTmp               Yap_heap_regs->wl.big_tmp
-#define  ActiveSignals            Yap_heap_regs->wl.active_signals
-#define  IPredArity               Yap_heap_regs->wl.i_pred_arity
-#define  ProfEnd                  Yap_heap_regs->wl.prof_end
-#define  UncaughtThrow            Yap_heap_regs->wl.uncaught_throw
-#define  DoingUndefp              Yap_heap_regs->wl.doing_undefp
-#define  StartLine                Yap_heap_regs->wl.start_line
-#define  ScratchPad               Yap_heap_regs->wl.scratchpad
+#endif
+#define  OldASP                   RINFO.old_ASP
+#define  OldLCL0                  RINFO.old_LCL0
+#define  OldTR                    RINFO.old_TR
+#define  OldGlobalBase            RINFO.old_GlobalBase
+#define  OldH                     RINFO.old_H
+#define  OldH0                    RINFO.old_H0
+#define  OldTrailBase             RINFO.old_TrailBase
+#define  OldTrailTop              RINFO.old_TrailTop
+#define  OldHeapBase              RINFO.old_HeapBase
+#define  OldHeapTop               RINFO.old_HeapTop
+#define  OldDelayTop              RINFO.old_DelayTop
+#define  CurrentDelayTop          RINFO.current_DelayTop
+#define  ClDiff                   RINFO.cl_diff
+#define  GDiff                    RINFO.g_diff
+#define  GDiff0                   RINFO.g_diff0
+#define  GSplit			  RINFO.g_split
+#define  HDiff                    RINFO.h_diff
+#define  LDiff                    RINFO.l_diff
+#define  TrDiff                   RINFO.tr_diff
+#define  XDiff                    RINFO.x_diff
+#define  DelayDiff                RINFO.delay_diff
+/* current consult stack */
+#define  ConsultSp                Yap_heap_regs->WL.consultsp
+/* top of consult stack  */
+#define  ConsultBase              Yap_heap_regs->WL.consultbase
+/* low-water mark for consult  */
+#define  ConsultLow               Yap_heap_regs->WL.consultlow
+/* current maximum number of cells in consult stack */
+#define  ConsultCapacity          Yap_heap_regs->WL.consultcapacity
+#define  FormatInfo               Yap_heap_regs->WL.f_info
+#define  ScannerStack             Yap_heap_regs->WL.scanner_stack
+#define  ScannerExtraBlocks       Yap_heap_regs->WL.scanner_extra_blocks
+#define  Yap_BigTmp               Yap_heap_regs->WL.big_tmp
+#define  ActiveSignals            Yap_heap_regs->WL.active_signals
+#define  IPredArity               Yap_heap_regs->WL.i_pred_arity
+#define  ProfEnd                  Yap_heap_regs->WL.prof_end
+#define  UncaughtThrow            Yap_heap_regs->WL.uncaught_throw
+#define  DoingUndefp              Yap_heap_regs->WL.doing_undefp
+#define  StartLine                Yap_heap_regs->WL.start_line
+#define  ScratchPad               Yap_heap_regs->WL.scratchpad
 #ifdef  COROUTINING
-#define  WokenGoals               Yap_heap_regs->wl.woken_goals
-#define  AttsMutableList          Yap_heap_regs->wl.atts_mutable_list
-#endif
-#define  GcGeneration             Yap_heap_regs->wl.gc_generation
-#define  GcPhase                  Yap_heap_regs->wl.gc_phase
-#define  GcCurrentPhase           Yap_heap_regs->wl.gc_current_phase
-#define  GcCalls                  Yap_heap_regs->wl.gc_calls
-#define  TotGcTime                Yap_heap_regs->wl.tot_gc_time
-#define  TotGcRecovered           Yap_heap_regs->wl.tot_gc_recovered
-#define  Yap_gc_restore           Yap_heap_regs->wl.gc_restore
-#define  TrustLUCode              Yap_heap_regs->wl.trust_lu_code
-#define  DynamicArrays            Yap_heap_regs->wl.dynamic_arrays
-#define  StaticArrays             Yap_heap_regs->wl.static_arrays
+#define  WokenGoals               Yap_heap_regs->WL.woken_goals
+#define  AttsMutableList          Yap_heap_regs->WL.atts_mutable_list
 #endif
+#define  GcGeneration             Yap_heap_regs->WL.gc_generation
+#define  GcPhase                  Yap_heap_regs->WL.gc_phase
+#define  GcCurrentPhase           Yap_heap_regs->WL.gc_current_phase
+#define  GcCalls                  Yap_heap_regs->WL.gc_calls
+#define  TotGcTime                Yap_heap_regs->WL.tot_gc_time
+#define  TotGcRecovered           Yap_heap_regs->WL.tot_gc_recovered
+#define  LastGcTime               Yap_heap_regs->WL.last_gc_time
+#define  LastSSTime               Yap_heap_regs->WL.last_ss_time
+#define  Yap_gc_restore           Yap_heap_regs->WL.gc_restore
+#define  TrustLUCode              Yap_heap_regs->WL.trust_lu_code
+#define  DynamicArrays            Yap_heap_regs->WL.dynamic_arrays
+#define  StaticArrays             Yap_heap_regs->WL.static_arrays
+#define  GlobalVariables          Yap_heap_regs->WL.global_variables
+#define  GlobalArena              Yap_heap_regs->WL.global_arena
+#define  GlobalArenaOverflows     Yap_heap_regs->WL.global_arena_overflows
+#define  Yap_AllowRestart         Yap_heap_regs->WL.allow_restart
+#define  GlobalDelayArena         Yap_heap_regs->WL.global_delay_arena
 #define  profiling                Yap_heap_regs->compiler_profiling
 #define  call_counting            Yap_heap_regs->compiler_call_counting
 #define  compile_arrays           Yap_heap_regs->compiler_compile_arrays
@@ -883,9 +969,17 @@ struct various_codes *Yap_heap_regs;
 #define  ForeignCodeTop           Yap_heap_regs->foreign_code_top;
 #define  ForeignCodeMax           Yap_heap_regs->foreign_code_max;
 #define  ForeignCodeLoaded        Yap_heap_regs->foreign_code_loaded
+#define  PredHash                 Yap_heap_regs->pred_hash
+#define  PredHashRWLock           Yap_heap_regs->pred_hash_rw_lock
+#define  PredsInHashTable         Yap_heap_regs->preds_in_hash_table
+#define  PredHashTableSize        Yap_heap_regs->pred_hash_table_size
+#define  PredHashInitialSize      1039L
+#define  PredHashIncrement        7919L
 #define  ParserErrorStyle         Yap_heap_regs->parser_error_style
+#define  GlobalHoldEntry          Yap_heap_regs->global_hold_entry
 #define  DeadStaticClauses        Yap_heap_regs->dead_static_clauses
 #define  DeadMegaClauses          Yap_heap_regs->dead_mega_clauses
+#define  DBTermsList              Yap_heap_regs->dbterms_list
 #define  DeadStaticIndices        Yap_heap_regs->dead_static_indices
 #define  SizeOfOverflow           Yap_heap_regs->size_of_overflow
 #define  LastWtimePtr             Yap_heap_regs->last_wtime
@@ -898,14 +992,21 @@ struct various_codes *Yap_heap_regs;
 #define  NOfThreads               Yap_heap_regs->n_of_threads
 #define  NOfThreadsCreated        Yap_heap_regs->n_of_threads_created
 #define  ThreadsTotalTime         Yap_heap_regs->threads_total_time
-#define  HeapUsedLock             Yap_heap_regs->heap_used_lock
 #define  DeadStaticClausesLock    Yap_heap_regs->dead_static_clauses_lock
 #define  DeadMegaClausesLock      Yap_heap_regs->dead_mega_clauses_lock
+#define  DBTermsListLock          Yap_heap_regs->dbterms_list_lock
 #define  DeadStaticIndicesLock    Yap_heap_regs->dead_static_indices_lock
+#define  ModulesLock		  Yap_heap_regs->modules_lock
 #endif
 #define  CreepCode                Yap_heap_regs->creep_code
 #define  UndefCode                Yap_heap_regs->undef_code
 #define  SpyCode                  Yap_heap_regs->spy_code
+#ifdef DEBUG
+#define  Yap_NewCps		  Yap_heap_regs->new_cps
+#define  Yap_LiveCps		  Yap_heap_regs->live_cps
+#define  Yap_DirtyCps		  Yap_heap_regs->dirty_cps
+#define  Yap_FreedCps		  Yap_heap_regs->freed_cps
+#endif
 #if defined(YAPOR) || defined(TABLING)
 #define  GLOBAL		          Yap_heap_regs->global
 #define  REMOTE                   Yap_heap_regs->remote
@@ -923,22 +1024,18 @@ struct various_codes *Yap_heap_regs;
 /* initially allow for files with up to 1024 predicates. This number
    is extended whenever needed */
 #define  InitialConsultCapacity    1024
-/* current consult stack */
-#define  ConsultSp                (Yap_heap_regs->consultsp                   )
-/* top of consult stack  */
-#define  ConsultBase              (Yap_heap_regs->consultbase                 )
-/* low-water mark for consult  */
-#define  ConsultLow               (Yap_heap_regs->consultlow	          )
-/* current maximum number of cells in consult stack */
-#define  ConsultCapacity          (Yap_heap_regs->consultcapacity             )
 #if HAVE_LIBREADLINE
 #define  ReadlineBuf              Yap_heap_regs->readline_buf
 #define  ReadlinePos              Yap_heap_regs->readline_pos
 #endif
 
+
+#if (defined(USE_SYSTEM_MALLOC) && HAVE_MALLINFO)||USE_DL_MALLOC
+UInt STD_PROTO(Yap_givemallinfo, (void));
+#endif
+
 ADDR    STD_PROTO(Yap_ExpandPreAllocCodeSpace, (UInt, void *));
 #define Yap_ReleasePreAllocCodeSpace(x)
-#if USE_SYSTEM_MALLOC||USE_DL_MALLOC
 ADDR    STD_PROTO(Yap_InitPreAllocCodeSpace, (void));
 EXTERN inline ADDR
 Yap_PreAllocCodeSpace(void)
@@ -947,15 +1044,8 @@ Yap_PreAllocCodeSpace(void)
   if (ptr) return ptr;
   return Yap_InitPreAllocCodeSpace();
 }
-#else
-EXTERN inline ADDR
-Yap_PreAllocCodeSpace(void)
-{
-  return Addr(HeapTop) + sizeof(CELL);
-}
-#endif
 
-#if THREADS
+#ifdef THREADS
 Prop STD_PROTO(Yap_NewThreadPred, (PredEntry *));
 Prop STD_PROTO(Yap_NewPredPropByFunctor, (Functor, Term));
 
