@@ -10,7 +10,7 @@
 * File:		Yap.h.m4						 *
 * mods:									 *
 * comments:	main header file for YAP				 *
-* version:      $Id: Yap.h,v 1.13 2006/03/09 15:52:04 tiagosoares Exp $	 *
+* version:      $Id: Yap.h,v 1.37 2008/05/23 10:02:13 vsc Exp $	 *
 *************************************************************************/
 
 #include "config.h"
@@ -53,7 +53,7 @@
 #undef TRAILING_REQUIRES_BRANCH
 #endif /* YAPOR || TABLING */
 
-#if ANALYST
+#ifdef ANALYST
 #ifdef USE_THREADED_CODE
 #undef USE_THREADED_CODE
 #endif
@@ -65,7 +65,7 @@
 #endif
 #endif
 
-#if SUPPORT_THREADS || SUPPORT_CONDOR
+#if defined(SUPPORT_THREADS) || defined(SUPPORT_CONDOR)
 #define USE_SYSTEM_MALLOC 1
 #endif
 
@@ -82,7 +82,7 @@
 #undef  USE_THREADED_CODE
 #endif
 #define inline __inline
-#define YAP_VERSION "Yap-5.1.1"
+#define YAP_VERSION "Yap-5.1.3"
 #define BIN_DIR "c:\\Yap\\bin"
 #define LIB_DIR "c:\\Yap\\lib\\Yap"
 #define SHARE_DIR "c:\\Yap\\share\\Yap"
@@ -141,12 +141,6 @@
 #include <pthread.h>
 #endif
 
-#if defined(YAPOR) || defined(THREADS)
-#define MAX_WORKERS (sizeof(unsigned long) * 8)
-#else
-#define MAX_WORKERS 1
-#endif /* YAPOR */
-
 #if SIZEOF_INT_P==4
 
 #if SIZEOF_INT==4
@@ -171,7 +165,7 @@
 
 #elif SIZEOF_INT_P==8
 
-#   if SIZEOF_INT==8
+#if SIZEOF_INT==8
 /*   */ typedef int Int;
 /*   */ typedef unsigned int UInt;
 
@@ -193,7 +187,7 @@
 
 #   elif SIZEOF_INT==4
 /*   */ typedef int Short;
-/*   */ typedef short int UShort;
+/*   */ typedef unsigned int UShort;
 
 #   else
 #	error Yap requires integer types half the size of a pointer
@@ -225,7 +219,7 @@ typedef unsigned long int YAP_ULONG_LONG;
 #define LOW_PROF 1
 #endif
 
-#if DEBUG
+#ifdef DEBUG
 extern char Yap_Option[20];
 #endif
 
@@ -236,14 +230,17 @@ extern char Yap_Option[20];
 #define MMAP_ADDR 0x42000000
 #endif
 
-#if !IN_SECOND_QUADRANT
+#if !defined(IN_SECOND_QUADRANT)
 #if __linux__ || __FreeBSD__ || __NetBSD__ || mips || __APPLE__
 #if defined(YAPOR) && defined(__alpha)
 
 #define MMAP_ADDR 0x40000000
-#elif mips
+#elif defined(mips)
 #define MMAP_ADDR 0x02000000
-#elif __APPLE__
+#elif defined(__APPLE__) && __LP64__
+// this address is high enough that it is likely not to confuse Apple's malloc debugger; lowest possible is 0x100200000
+#define MMAP_ADDR 0x200000000
+#elif defined(__APPLE__) && !__LP64__
 #define MMAP_ADDR 0x20000000
 #else
 #define MMAP_ADDR 0x08800000
@@ -253,11 +250,11 @@ extern char Yap_Option[20];
 #endif
 #endif /* !IN_SECOND_QUADRANT */
 
-#ifdef THREADS
+#ifdef USE_SYSTEM_MALLOC
 #define HEAP_INIT_BASE  0L
 #define AtomBase        NULL
 #else
-#if defined(MMAP_ADDR) && (USE_MMAP || USE_SHMAT) && !__simplescalar__
+#if defined(MMAP_ADDR) && (defined(USE_MMAP) || USE_SHMAT) && !defined(__simplescalar__)
 #define HEAP_INIT_BASE  (MMAP_ADDR)
 #define AtomBase        ((char *)MMAP_ADDR)
 #else
@@ -329,7 +326,7 @@ typedef CELL Term;
 typedef pthread_mutex_t lockvar;
 typedef pthread_rwlock_t rwlock_t;
 #include <pthread_locks.h>
-#elif defined(i386)
+#elif defined(i386) || defined(__x86_64__)
 typedef volatile int lockvar;
 #include <x86_locks.h>
 #elif defined(sparc) || defined(__sparc)
@@ -391,6 +388,12 @@ typedef pthread_rwlock_t rwlock_t;
 #ifdef __alpha
 #include <alpha_locks_funcs.h>
 #endif
+#ifdef YAPOR
+#define MAX_AGENTS MAX_WORKERS
+#endif
+#ifdef THREADS
+#define MAX_AGENTS MAX_THREADS
+#endif
 #endif
 
 /************ variables	concerned with Error Handling *************/
@@ -445,6 +448,7 @@ typedef enum
   EVALUATION_ERROR_UNDERFLOW,
   EVALUATION_ERROR_ZERO_DIVISOR,
   EXISTENCE_ERROR_ARRAY,
+  EXISTENCE_ERROR_KEY,
   EXISTENCE_ERROR_SOURCE_SINK,
   EXISTENCE_ERROR_STREAM,
   INSTANTIATION_ERROR,
@@ -474,6 +478,8 @@ typedef enum
   REPRESENTATION_ERROR_CHARACTER,
   REPRESENTATION_ERROR_CHARACTER_CODE,
   REPRESENTATION_ERROR_MAX_ARITY,
+  RESOURCE_ERROR_MAX_THREADS,
+  RESOURCE_ERROR_MEMORY,
   RETRY_COUNTER_UNDERFLOW,
   SYNTAX_ERROR,
   SYSTEM_ERROR,
@@ -513,7 +519,6 @@ typedef enum
   YAP_TO_CHARS_FLAG = 7,
   LANGUAGE_MODE_FLAG = 8,
   STRICT_ISO_FLAG = 9,
-  SPY_CREEP_FLAG = 10,
   SOURCE_MODE_FLAG = 11,
   CHARACTER_ESCAPE_FLAG = 12,
   WRITE_QUOTED_STRING_FLAG = 13,
@@ -521,8 +526,13 @@ typedef enum
   HALT_AFTER_CONSULT_FLAG = 15,
   FAST_BOOT_FLAG = 16,
   STACK_DUMP_ON_ERROR_FLAG = 17,
-  INDEXING_MODE_FLAG = 18,
-  TABLING_MODE_FLAG = 19
+  GENERATE_DEBUG_INFO_FLAG = 18,
+  INDEXING_MODE_FLAG = 19,
+  TABLING_MODE_FLAG = 20,
+  VARS_CAN_HAVE_QUOTE_FLAG = 21,
+  QUIET_MODE_FLAG = 22,
+  /* let this be the last one */
+  LAST_FLAG = 23
 } yap_flags;
 
 #define STRING_AS_CHARS		0
@@ -562,10 +572,12 @@ typedef enum
   YAP_BREAK_SIGNAL = 0x2000,	/* received break signal */
   YAP_STACK_DUMP_SIGNAL = 0x4000,	/* received stack dump signal */
   YAP_STATISTICS_SIGNAL = 0x8000,	/* received statistics */
-  YAP_DELAY_CREEP_SIGNAL = 0x10000	/* received a creep but should not do it */
+  YAP_DELAY_CREEP_SIGNAL = 0x10000,	/* received a creep but should not do it */
+  YAP_AGC_SIGNAL = 0x20000,	/* call atom garbage collector asap */
+  YAP_PIPE_SIGNAL = 0x40000	/* call atom garbage collector asap */
 } yap_signals;
 
-#define NUMBER_OF_YAP_FLAGS  TABLING_MODE_FLAG + 1
+#define NUMBER_OF_YAP_FLAGS  LAST_FLAG
 
 /************************  prototypes **********************************/
 
@@ -638,7 +650,7 @@ typedef enum
 	if you place things in the lower addresses (power to the libc people).
 */
 
-#if (defined(_AIX) || defined(_WIN32) || defined(__APPLE__) || defined(sparc) || defined(__sparc) || defined(mips) || defined(__FreeBSD__) || defined(_POWER) || defined(__POWERPC__) || defined(__linux__) || defined(IN_SECOND_QUADRANT) || defined(__CYGWIN__))
+#if (defined(_AIX) || (defined(__APPLE__) && !defined(__LP64__)) || defined(_WIN32) || defined(sparc) || defined(__sparc) || defined(mips) || defined(__FreeBSD__) || defined(_POWER) || defined(__POWERPC__) || defined(__linux__) || defined(IN_SECOND_QUADRANT) || defined(__CYGWIN__))
 #define USE_LOW32_TAGS 1
 #endif
 
@@ -726,7 +738,7 @@ typedef struct thread_globs
 
 } tglobs;
 
-extern struct thread_globs Yap_thread_gl[MAX_WORKERS];
+extern struct thread_globs Yap_thread_gl[MAX_THREADS];
 
 
 #define    Yap_LocalBase  Yap_thread_gl[worker_id].local_base
@@ -772,7 +784,7 @@ VarOfTerm (Term t)
 }
 
 
-#if SBA
+#ifdef SBA
 
 inline EXTERN Term MkVarTerm (void);
 
@@ -836,7 +848,7 @@ FunctorOfTerm (Term t)
 }
 
 
-#if USE_SYSTEM_MALLOC || USE_DL_MALLOC
+#if (defined(USE_SYSTEM_MALLOC) || USE_DL_MALLOC) && !defined(__LP64__)
 #if USE_LOW32_TAGS
 
 inline EXTERN Term MkAtomTerm (Atom);
@@ -844,7 +856,7 @@ inline EXTERN Term MkAtomTerm (Atom);
 inline EXTERN Term
 MkAtomTerm (Atom a)
 {
-  return (Term) ((AtomTag | (CELL) (a)));
+  return (Term) (AtomTag | (CELL) (a));
 }
 
 
@@ -865,7 +877,7 @@ inline EXTERN Term MkAtomTerm (Atom);
 inline EXTERN Term
 MkAtomTerm (Atom a)
 {
-  return (Term) (TAGGEDA (AtomTag, (CELL) (a)));
+  return (Term) (TAGGEDA ((CELL)AtomTag, (CELL) (a)));
 }
 
 
@@ -1030,7 +1042,7 @@ IntegerOfTerm (Term t)
 
 /*************** unification routines ***********************************/
 
-#if SBA
+#ifdef SBA
 #include "sbaamiops.h"
 #else
 #include "amiops.h"
@@ -1102,6 +1114,7 @@ TailOfTermCell (Term t)
 
 /*************** variables concerned with atoms	table *******************/
 #define	MaxHash	    1001
+#define	MaxWideHash (MaxHash/10+1)
 
 #define FAIL_RESTORE  0
 #define DO_EVERYTHING 1
@@ -1151,10 +1164,11 @@ typedef enum
   CCallMode = 0x1000,		/* In c Call */
   UnifyMode = 0x2000,		/* In Unify Code */
   UserCCallMode = 0x4000,	/* In User C-call Code */
-  MallocMode = 0x8000	/* Doing malloc, realloc, free */
+  MallocMode = 0x8000,		/* Doing malloc, realloc, free */
+  SystemMode = 0x10000,		/* in system mode */
 } prolog_exec_mode;
 
-extern prolog_exec_mode Yap_PrologMode;
+extern Int Yap_PrologMode;
 extern int Yap_CritLocks;
 
 /************** Access to yap initial arguments ***************************/
@@ -1168,9 +1182,9 @@ extern int Yap_PrologShouldHandleInterrupts;
 
 /******************* number of modules ****************************/
 
-#define MaxModules	256
+#define DefaultMaxModules	256
 
-#if YAPOR
+#ifdef YAPOR
 #define YAPEnterCriticalSection()                                        \
 	{                                                                \
           if (worker_id != GLOBAL_LOCKS_who_locked_heap) {               \
@@ -1197,10 +1211,10 @@ extern int Yap_PrologShouldHandleInterrupts;
             UNLOCK(GLOBAL_LOCKS_heap_access);                            \
           }                                                              \
         }
-#elif THREADS
+#elif defined(THREADS)
 #define YAPEnterCriticalSection()                                        \
 	{                                                                \
-          LOCK(BGL);                                              \
+          /* LOCK(BGL); */                                              \
           Yap_PrologMode |= CritMode;                                   \
         }
 #define YAPLeaveCriticalSection()                                        \
@@ -1214,7 +1228,7 @@ extern int Yap_PrologShouldHandleInterrupts;
 	    Yap_PrologMode &= ~AbortMode;                             \
 	    Yap_Error(PURE_ABORT, 0, "");                             \
           }                                                            \
-          UNLOCK(BGL);                                           \
+          /* UNLOCK(BGL); */                                           \
         }
 #else
 #define YAPEnterCriticalSection()                                        \
@@ -1255,7 +1269,7 @@ typedef struct TIMED_MAVAR
 
 /********* while debugging you may need some info ***********************/
 
-#if EMACS
+#ifdef EMACS
 extern char emacs_tmp[], emacs_tmp2[];
 #endif
 
@@ -1265,7 +1279,7 @@ extern char emacs_tmp[], emacs_tmp2[];
 #include "opt.macros.h"
 #endif /* YAPOR || TABLING */
 
-#if SBA
+#ifdef SBA
 #include "sbaunify.h"
 #endif
 
