@@ -11,8 +11,30 @@
 * File:		YapOpcodes.h						 *
 * comments:	Central Table with all YAP opcodes                       *
 *									 *
-* Last rev:     $Date: 2006/03/24 16:34:21 $							 *
+* Last rev:     $Date: 2008/03/25 22:03:13 $							 *
 * $Log: YapOpcodes.h,v $
+* Revision 1.44  2008/03/25 22:03:13  vsc
+* fix some icc warnings
+*
+* Revision 1.43  2007/11/26 23:43:09  vsc
+* fixes to support threads and assert correctly, even if inefficiently.
+*
+* Revision 1.42  2007/11/07 09:25:27  vsc
+* speedup meta-calls
+*
+* Revision 1.41  2007/11/06 17:02:12  vsc
+* compile ground terms away.
+*
+* Revision 1.40  2006/10/10 14:08:17  vsc
+* small fixes on threaded implementation.
+*
+* Revision 1.39  2006/09/20 20:03:51  vsc
+* improve indexing on floats
+* fix sending large lists to DB
+*
+* Revision 1.38  2006/04/27 14:13:24  rslopes
+* *** empty log message ***
+*
 * Revision 1.37  2006/03/24 16:34:21  rslopes
 * New update to BEAM engine.
 * BEAM now uses YAP Indexing (JITI)
@@ -172,9 +194,10 @@
     OPCODE(get_4atoms              ,cccc),	  /* peephole */
     OPCODE(get_5atoms              ,ccccc),	  /* peephole */
     OPCODE(get_6atoms              ,cccccc),	  /* peephole */
-    OPCODE(get_float               ,xc),		
-    OPCODE(get_longint             ,xc),		
+    OPCODE(get_float               ,xd),		
+    OPCODE(get_longint             ,xi),		
     OPCODE(get_bigint              ,xc),		
+    OPCODE(get_dbterm              ,xc),		
     OPCODE(get_list                ,x),		
     OPCODE(get_struct              ,xf),		
     OPCODE(unify_x_var             ,ox),	
@@ -182,9 +205,10 @@
     OPCODE(unify_x_val             ,ox),	
     OPCODE(unify_y_val             ,oy),	
     OPCODE(unify_atom              ,oc),		
-    OPCODE(unify_float             ,oc),		
+    OPCODE(unify_float             ,od),		
     OPCODE(unify_longint           ,oc),		
     OPCODE(unify_bigint            ,oc),		
+    OPCODE(unify_dbterm            ,oc),		
     OPCODE(unify_list              ,o),		
     OPCODE(unify_struct            ,of),	
     OPCODE(put_x_var               ,xx),		
@@ -194,6 +218,8 @@
     OPCODE(put_unsafe              ,yx),		
     OPCODE(put_xx_val              ,xxxx), /* peephole */
     OPCODE(put_atom                ,xc),		
+    OPCODE(put_float               ,xd),		
+    OPCODE(put_longint             ,xi),		
     OPCODE(put_list                ,x),		
     OPCODE(put_struct              ,xf),		
     OPCODE(write_x_var             ,x),	
@@ -201,6 +227,8 @@
     OPCODE(write_x_val             ,x),	
     OPCODE(write_y_val             ,y),	
     OPCODE(write_atom              ,c),		
+    OPCODE(write_float             ,d),		
+    OPCODE(write_longint           ,i),		
     OPCODE(write_list              ,e),		
     OPCODE(write_struct            ,f),	
     OPCODE(pop                     ,e),		
@@ -216,6 +244,7 @@
     OPCODE(or_last                 ,p),
 #endif /* YAPOR */
 #ifdef BEAM 
+    OPCODE(retry_eam               ,e),		
     OPCODE(run_eam                 ,os),  
 #endif
     OPCODE(call_cpred              ,sla),		
@@ -237,6 +266,14 @@
     OPCODE(retry4                  ,l),		
     OPCODE(trust                   ,ld),		
     OPCODE(try_in                  ,l),
+    OPCODE(enter_lu_pred           ,Ill),
+    OPCODE(try_logical             ,lld),		
+    OPCODE(retry_logical           ,lld),		
+    OPCODE(trust_logical           ,lld),	     
+    OPCODE(count_retry_logical     ,lld),		
+    OPCODE(count_trust_logical     ,lld),	     
+    OPCODE(profiled_retry_logical  ,lld),		
+    OPCODE(profiled_trust_logical  ,lld),	     
     OPCODE(jump_if_var             ,l),	
     OPCODE(jump_if_nonvar          ,xll),	
     OPCODE(switch_on_cons          ,sssl),	
@@ -254,6 +291,7 @@
     OPCODE(index_blob              ,e),	
     OPCODE(trust_fail              ,e),		
     OPCODE(index_pred              ,e),
+    OPCODE(lock_pred               ,e),
     OPCODE(expand_index            ,e),		
     OPCODE(expand_clauses          ,sp),
     OPCODE(save_b_x                ,x),		
@@ -295,9 +333,10 @@
     OPCODE(unify_l_x_val           ,ox),	
     OPCODE(unify_l_y_val           ,oy),	
     OPCODE(unify_l_atom            ,oc),	
-    OPCODE(unify_l_float           ,oc),	
-    OPCODE(unify_l_longint         ,oc),	
+    OPCODE(unify_l_float           ,od),	
+    OPCODE(unify_l_longint         ,oi),	
     OPCODE(unify_l_bigint          ,oc),	
+    OPCODE(unify_l_dbterm          ,oc),	
     OPCODE(unify_l_void            ,o),	
     OPCODE(unify_l_n_voids         ,os),	
     OPCODE(unify_l_x_loc           ,ox),	
@@ -308,6 +347,8 @@
     OPCODE(unify_x_val_write       ,ox),	
     OPCODE(unify_y_val_write       ,oy),	
     OPCODE(unify_atom_write        ,oc),	
+    OPCODE(unify_float_write       ,od),	
+    OPCODE(unify_longint_write       ,oi),	
     OPCODE(unify_n_atoms_write     ,osc),
     OPCODE(unify_list_write        ,o),	
     OPCODE(unify_x_var2_write      ,oxx),	
@@ -321,6 +362,8 @@
     OPCODE(unify_l_x_val_write     ,ox),
     OPCODE(unify_l_y_val_write     ,oy),
     OPCODE(unify_l_atom_write      ,oc),	
+    OPCODE(unify_l_float_write     ,od),	
+    OPCODE(unify_l_longint_write   ,oi),	
     OPCODE(unify_l_void_write      ,o),	
     OPCODE(unify_l_n_voids_write   ,os),
     OPCODE(unify_l_x_loc_write     ,ox),	
@@ -346,13 +389,10 @@
     OPCODE(count_retry_and_mark    ,ld),
     OPCODE(lock_lu	           ,p),
     OPCODE(unlock_lu	           ,e),
-    OPCODE(enter_lu_pred           ,Ill),
-    OPCODE(stale_lu_index          ,Ill),
-    OPCODE(trust_logical_pred      ,l),
     OPCODE(alloc_for_logical_pred  ,EC),
     OPCODE(unify_idb_term          ,e),
     OPCODE(copy_idb_term           ,e),
-#if THREADS
+#if defined(THREADS)
     OPCODE(thread_local            ,e),
 #endif
 #ifdef SFUNC
@@ -454,6 +494,7 @@
     OPCODE(p_func2f_yx             ,yxx),
     OPCODE(p_func2f_yy             ,yyx),
     OPCODE(p_execute               ,sla),
+    OPCODE(p_execute2              ,sla),
     OPCODE(p_execute_tail          ,e)
 
 
