@@ -17,7 +17,7 @@
 
 /* This code can only be defined *after* including Regs.h!!! */
 
-#if USE_OFFSETS
+#ifdef USE_OFFSETS
 
 inline EXTERN Atom AbsAtom (AtomEntry * p);
 
@@ -188,19 +188,171 @@ IsFunctorProperty (int flags)
 	bb 00	functor entry 
 	ff df	sparse functor
 	ff ex	arithmetic property
+	ff f6   hold
 	ff f7   array
+	ff f8   wide atom
 	ff fa   module property
 	ff fb   blackboard property
 	ff fc	value property
+	ff fd	global property
 	ff ff	op property
 */
 
-/*	Module property 						*/
+
+/*	Global Variable property						*/
+typedef struct global_entry
+{
+  Prop NextOfPE;		/* used to chain properties             */
+  PropFlags KindOfPE;		/* kind of property                     */
+#if defined(YAPOR) || defined(THREADS)
+  rwlock_t GRWLock;		/* a simple lock to protect this entry */
+#if THREADS
+  unsigned int owner_id;        /* owner thread */
+#endif
+#endif
+  struct AtomEntryStruct *AtomOfGE; /* parent atom for deletion */
+  struct global_entry *NextGE; /* linked list of global entries */
+  Term  global;	/* index in module table                */
+} GlobalEntry;
+
+
+#if USE_OFFSETS_IN_PROPS
+
+inline EXTERN GlobalEntry *RepGlobalProp (Prop p);
+
+inline EXTERN GlobalEntry *
+RepGlobalProp (Prop p)
+{
+  return (GlobalEntry *) (AtomBase + Unsigned (p));
+}
+
+
+
+inline EXTERN Prop AbsGlobalProp (GlobalEntry * p);
+
+inline EXTERN Prop
+AbsGlobalProp (GlobalEntry * p)
+{
+  return (Prop) (Addr (p) - AtomBase);
+}
+
+
+#else
+
+inline EXTERN GlobalEntry *RepGlobalProp (Prop p);
+
+inline EXTERN GlobalEntry *
+RepGlobalProp (Prop p)
+{
+  return (GlobalEntry *) (p);
+}
+
+inline EXTERN Prop AbsGlobalProp (GlobalEntry * p);
+
+inline EXTERN Prop
+AbsGlobalProp (GlobalEntry * p)
+{
+  return (Prop) (p);
+}
+
+
+#endif
+
+#define GlobalProperty	((PropFlags)0xfffd)
+
+inline EXTERN PropFlags IsGlobalProperty (int);
+
+inline EXTERN PropFlags
+IsGlobalProperty (int flags)
+{
+  return (PropFlags) ((flags == GlobalProperty));
+}
+
+
+/*	Wide Atom property 						*/
 typedef struct
 {
   Prop NextOfPE;		/* used to chain properties             */
   PropFlags KindOfPE;		/* kind of property                     */
-  SMALLUNSGN IndexOfMod;	/* index in module table                */
+  UInt  SizeOfAtom;	        /* index in module table                */
+} WideAtomEntry;
+
+#if USE_OFFSETS_IN_PROPS
+
+inline EXTERN WideAtomEntry *RepWideAtomProp (Prop p);
+
+inline EXTERN WideAtomEntry *
+RepWideAtomProp (Prop p)
+{
+  return (WideAtomEntry *) (AtomBase + Unsigned (p));
+}
+
+
+
+inline EXTERN Prop AbsWideAtomProp (WideAtomEntry * p);
+
+inline EXTERN Prop
+AbsWideAtomProp (WideAtomEntry * p)
+{
+  return (Prop) (Addr (p) - AtomBase);
+}
+
+
+#else
+
+inline EXTERN WideAtomEntry *RepWideAtomProp (Prop p);
+
+inline EXTERN WideAtomEntry *
+RepWideAtomProp (Prop p)
+{
+  return (WideAtomEntry *) (p);
+}
+
+
+
+inline EXTERN Prop AbsWideAtomProp (WideAtomEntry * p);
+
+inline EXTERN Prop
+AbsWideAtomProp (WideAtomEntry * p)
+{
+  return (Prop) (p);
+}
+
+
+#endif
+
+#define WideAtomProperty	((PropFlags)0xfff8)
+
+
+inline EXTERN PropFlags IsWideAtomProperty (int);
+
+inline EXTERN PropFlags
+IsWideAtomProperty (int flags)
+{
+  return (PropFlags) ((flags == WideAtomProperty));
+}
+
+inline EXTERN int IsWideAtom (Atom);
+
+inline EXTERN int
+IsWideAtom (Atom at)
+{
+  return RepAtom(at)->PropsOfAE &&
+    IsWideAtomProperty(RepWideAtomProp(RepAtom(at)->PropsOfAE)->KindOfPE);
+}
+
+
+/*	Module property 						*/
+typedef struct mod_entry
+{
+  Prop NextOfPE;		 /* used to chain properties            */
+  PropFlags KindOfPE;		 /* kind of property                    */
+  struct pred_entry *PredForME;  /* index in module table               */
+  Atom   AtomOfME;		 /* module's name	                */
+#if defined(YAPOR) || defined(THREADS)
+  rwlock_t ModRWLock;		/* a read-write lock to protect the entry */
+#endif
+  struct  mod_entry *NextME;   /* next module                         */
 } ModEntry;
 
 #if USE_OFFSETS_IN_PROPS
@@ -261,14 +413,16 @@ IsModProperty (int flags)
 
 
 /*	    operator property entry structure				*/
-typedef struct
+typedef struct operator_entry
 {
   Prop NextOfPE;		/* used to chain properties     */
   PropFlags KindOfPE;		/* kind of property             */
 #if defined(YAPOR) || defined(THREADS)
   rwlock_t OpRWLock;		/* a read-write lock to protect the entry */
 #endif
+  Atom OpName;			/* atom name		        */
   Term OpModule;		/* module of predicate          */
+  struct operator_entry  *OpNext; /* next in list of operators  */
   BITS16 Prefix, Infix, Posfix;	/* precedences                  */
 } OpEntry;
 #if USE_OFFSETS_IN_PROPS
@@ -514,6 +668,8 @@ typedef enum
   TabledPredFlag = 0x00000040L,	/* is tabled */
   SequentialPredFlag = 0x00000020L,	/* may not create parallel choice points! */
   ProfiledPredFlag = 0x00000010L,	/* pred is being profiled   */
+  MyddasPredFlag = 0x00000008L,	/* Myddas Imported pred  */
+  ModuleTransparentPredFlag = 0x00000004L,	/* ModuleTransparent pred  */
 } pred_flag;
 
 /* profile data */
@@ -528,6 +684,15 @@ typedef struct
 #endif
 } profile_data;
 
+typedef enum {
+  LUCALL_EXEC,
+  LUCALL_ASSERT,
+  LUCALL_RETRACT
+} timestamp_type;
+
+#define TIMESTAMP_EOT   ((UInt)(~0L))
+#define TIMESTAMP_RESET (TIMESTAMP_EOT-1024)
+
 typedef struct pred_entry
 {
   Prop NextOfPE;		/* used to chain properties             */
@@ -557,7 +722,6 @@ typedef struct pred_entry
     struct mfile *file_srcs;	/* for multifile predicates */
   } src;
 #if defined(YAPOR) || defined(THREADS)
-  rwlock_t PRWLock;		/* a simple lock to protect this entry */
   lockvar PELock;		/* a simple lock to protect expansion */
 #endif
 #ifdef TABLING
@@ -567,6 +731,8 @@ typedef struct pred_entry
   struct Predicates *beamTable;
 #endif
   Term ModuleOfPred;		/* module for this definition           */
+  UInt TimeStampOfPred;
+  timestamp_type LastCallOfPred; 
   /* This must be at an odd number of cells, otherwise it
      will not be aligned on RISC machines */
   profile_data StatisticsForPred;	/* enable profiling for predicate  */
@@ -633,20 +799,23 @@ IsPredProperty (int flags)
 /* There are several flags for code and data base entries */
 typedef enum
 {
-  MegaMask = 0x200000,		/* informs this is a mega clause */
-  FactMask = 0x100000,		/* informs this is a fact */
-  SwitchRootMask = 0x80000,	/* informs this is the root for the index tree */
-  SwitchTableMask = 0x40000,	/* informs this is a switch table */
-  HasBlobsMask = 0x20000,	/* informs this has blobs which may be in use */
-  ProfFoundMask = 0x10000,	/* informs this clause is being counted by profiler */
-  DynamicMask = 0x8000,		/* informs this is a dynamic predicate */
-  InUseMask = 0x4000,		/* informs this block is being used */
-  ErasedMask = 0x2000,		/* informs this block has been erased */
-  IndexMask = 0x1000,		/* informs this is indexing code */
-  DBClMask = 0x0800,		/* informs this is a data base structure */
-  LogUpdRuleMask = 0x0400,	/* informs the code is for a log upd rule with env */
-  LogUpdMask = 0x0200,		/* informs this is a logic update index. */
-  StaticMask = 0x0100		/* dealing with static predicates */
+  FuncSwitchMask = 0x800000,	/* is a switch of functors */
+  HasDBTMask = 0x400000,	/* includes a pointer to a DBTerm */
+  MegaMask = 0x200000,		/* mega clause */
+  FactMask = 0x100000,		/* a fact */
+  SwitchRootMask = 0x80000,	/* root for the index tree */
+  SwitchTableMask = 0x40000,	/* switch table */
+  HasBlobsMask = 0x20000,	/* blobs which may be in use */
+  ProfFoundMask = 0x10000,	/* clause is being counted by profiler */
+  DynamicMask = 0x8000,		/* dynamic predicate */
+  InUseMask = 0x4000,		/* this block is being used */
+  ErasedMask = 0x2000,		/* this block has been erased */
+  IndexMask = 0x1000,		/* indexing code */
+  DBClMask = 0x0800,		/* data base structure */
+  LogUpdRuleMask = 0x0400,	/* code is for a log upd rule with env */
+  LogUpdMask = 0x0200,		/* logic update index. */
+  StaticMask = 0x0100,		/* static predicates */
+  DirtyMask = 0x0080		/* LUIndices  */
 /* other flags belong to DB */
 } dbentry_flags;
 
@@ -655,7 +824,10 @@ typedef enum
 typedef struct DB_TERM
 {
 #ifdef COROUTINING
-  CELL attachments;		/* attached terms */
+  union {
+    CELL attachments;		/* attached terms */
+    struct DB_TERM *NextDBT;
+  } ag;
 #endif
   struct DB_STRUCT **DBRefs;	/* pointer to other references     */
   CELL NOfCells;		/* Size of Term                         */
@@ -663,6 +835,18 @@ typedef struct DB_TERM
   Term Contents[MIN_ARRAY];	/* stored term                      */
 } DBTerm;
 
+inline EXTERN DBTerm *TermToDBTerm(Term);
+
+inline EXTERN DBTerm *TermToDBTerm(Term X)
+{
+  if (IsPairTerm(X)) {
+    return(DBTerm *)((char *)RepPair(X) - (CELL) &(((DBTerm *) NULL)->Contents));
+  } else {
+    return(DBTerm *)((char *)RepAppl(X) - (CELL) &(((DBTerm *) NULL)->Contents));
+  }
+}
+
+
 /* The ordering of the first 3 fields should be compatible with lu_clauses */
 typedef struct DB_STRUCT
 {
@@ -929,6 +1113,69 @@ IsBBProperty (int flags)
 }
 
 
+/*		hold property entry structure				*/
+typedef struct hold_entry
+{
+  Prop NextOfPE;		/* used to chain properties             */
+  PropFlags KindOfPE;		/* kind of property                     */
+} HoldEntry;
+
+#if USE_OFFSETS_IN_PROPS
+
+inline EXTERN HoldEntry *RepHoldProp (Prop p);
+
+inline EXTERN HoldEntry *
+RepHoldProp (Prop p)
+{
+  return (HoldEntry *) (AtomBase + Unsigned (p));
+}
+
+
+
+inline EXTERN Prop AbsHoldProp (HoldEntry * p);
+
+inline EXTERN Prop
+AbsHoldProp (HoldEntry * p)
+{
+  return (Prop) (Addr (p) - AtomBase);
+}
+
+
+#else
+
+inline EXTERN HoldEntry *RepHoldProp (Prop p);
+
+inline EXTERN HoldEntry *
+RepHoldProp (Prop p)
+{
+  return (HoldEntry *) (p);
+}
+
+
+
+inline EXTERN Prop AbsHoldProp (HoldEntry * p);
+
+inline EXTERN Prop
+AbsHoldProp (HoldEntry * p)
+{
+  return (Prop) (p);
+}
+
+
+#endif
+#define	HoldProperty  0xfff6
+
+/* only unary and binary expressions are acceptable */
+
+inline EXTERN PropFlags IsHoldProperty (int);
+
+inline EXTERN PropFlags
+IsHoldProperty (int flags)
+{
+  return (PropFlags) ((flags == HoldProperty));
+}
+
+
 
 
 /*		array property entry structure				*/
@@ -1109,11 +1356,12 @@ IsArrayProperty (int flags)
 
 /* cdmgr.c */
 int STD_PROTO (Yap_RemoveIndexation, (PredEntry *));
+void STD_PROTO (Yap_UpdateTimestamps, (PredEntry *));
 
 /* dbase.c */
 void STD_PROTO (Yap_ErDBE, (DBRef));
 DBTerm *STD_PROTO (Yap_StoreTermInDB, (Term, int));
-DBTerm *STD_PROTO (Yap_StoreTermInDBPlusExtraSpace, (Term, UInt));
+DBTerm *STD_PROTO (Yap_StoreTermInDBPlusExtraSpace, (Term, UInt, UInt *));
 Term STD_PROTO (Yap_FetchTermFromDB, (DBTerm *));
 void STD_PROTO (Yap_ReleaseTermFromDB, (DBTerm *));
 
@@ -1124,38 +1372,87 @@ Atom STD_PROTO (Yap_GetOp, (OpEntry *, int *, int));
 Prop STD_PROTO (Yap_GetAProp, (Atom, PropFlags));
 Prop STD_PROTO (Yap_GetAPropHavingLock, (AtomEntry *, PropFlags));
 
-#if THREADS
+#ifdef THREADS
 EXTERN inline PredEntry *STD_PROTO (Yap_GetThreadPred, (PredEntry *));
 #endif
 
+typedef enum
+{
+  PROLOG_MODULE = 0,
+  USER_MODULE = 1,
+  IDB_MODULE = 2,
+  ATTRIBUTES_MODULE = 3,
+  CHARSIO_MODULE = 4,
+  TERMS_MODULE = 5
+} default_modules;
+
+#include "Heap.h"
+
+EXTERN inline UInt STD_PROTO(PRED_HASH, (FunctorEntry *, Term, UInt));
+
+EXTERN inline UInt
+PRED_HASH(FunctorEntry *fe, Term cur_mod, UInt size)
+{
+  return (((CELL)fe+cur_mod)>>2) % size;
+}
+
+EXTERN inline Prop STD_PROTO(GetPredPropByFuncHavingLock, (FunctorEntry *, Term));
+
 EXTERN inline Prop
-PredPropByFunc (Functor f, Term cur_mod)
+GetPredPropByFuncHavingLock (FunctorEntry *fe, Term cur_mod)
+{
+  PredEntry *p;
+
+  if (!(p = RepPredProp(fe->PropsOfFE))) {
+    return NIL;
+  }
+  if ((p->ModuleOfPred == cur_mod || !(p->ModuleOfPred))) {
+#ifdef THREADS
+    /* Thread Local Predicates */
+    if (p->PredFlags & ThreadLocalPredFlag) {
+      return AbsPredProp (Yap_GetThreadPred (p));
+    }
+#endif
+    return AbsPredProp(p);
+  }
+  if (p->NextOfPE) {
+    UInt hash = PRED_HASH(fe,cur_mod,PredHashTableSize);
+    READ_LOCK(PredHashRWLock);
+    p = PredHash[hash];
+    
+    while (p) {
+      if (p->FunctorOfPred == fe &&
+	  p->ModuleOfPred == cur_mod)
+	{
+#ifdef THREADS
+	  /* Thread Local Predicates */
+	  if (p->PredFlags & ThreadLocalPredFlag) {
+	    READ_UNLOCK(PredHashRWLock);
+	    return AbsPredProp (Yap_GetThreadPred (p));
+	  }
+#endif
+	  READ_UNLOCK(PredHashRWLock);
+	  return AbsPredProp(p);
+	}
+      p = RepPredProp(p->NextOfPE);
+    }
+    READ_UNLOCK(PredHashRWLock);
+  }
+  return NIL;
+}
+
+EXTERN inline Prop
+PredPropByFunc (Functor fe, Term cur_mod)
 /* get predicate entry for ap/arity; create it if neccessary.              */
 {
   Prop p0;
-  FunctorEntry *fe = (FunctorEntry *) f;
 
   WRITE_LOCK (fe->FRWLock);
-  p0 = fe->PropsOfFE;
-  while (p0)
-    {
-      PredEntry *p = RepPredProp (p0);
-      if (			/* p->KindOfPE != 0 || only props */
-	   (p->ModuleOfPred == cur_mod || !(p->ModuleOfPred)))
-	{
-#if THREADS
-	  /* Thread Local Predicates */
-	  if (p->PredFlags & ThreadLocalPredFlag)
-	    {
-	      WRITE_UNLOCK (fe->FRWLock);
-	      return AbsPredProp (Yap_GetThreadPred (p));
-	    }
-#endif
-	  WRITE_UNLOCK (fe->FRWLock);
-	  return (p0);
-	}
-      p0 = p->NextOfPE;
-    }
+  p0 = GetPredPropByFuncHavingLock(fe, cur_mod);
+  if (p0) {
+    WRITE_UNLOCK (fe->FRWLock);
+    return p0;
+  }
   return Yap_NewPredPropByFunctor (fe, cur_mod);
 }
 
@@ -1174,7 +1471,7 @@ PredPropByAtom (Atom at, Term cur_mod)
       if (pe->KindOfPE == PEProp &&
 	  (pe->ModuleOfPred == cur_mod || !pe->ModuleOfPred))
 	{
-#if THREADS
+#ifdef THREADS
 	  /* Thread Local Predicates */
 	  if (pe->PredFlags & ThreadLocalPredFlag)
 	    {
@@ -1190,12 +1487,3 @@ PredPropByAtom (Atom at, Term cur_mod)
   return Yap_NewPredPropByAtom (ae, cur_mod);
 }
 
-typedef enum
-{
-  PROLOG_MODULE = 0,
-  USER_MODULE = 1,
-  IDB_MODULE = 2,
-  ATTRIBUTES_MODULE = 3,
-  CHARSIO_MODULE = 4,
-  TERMS_MODULE = 5
-} default_modules;
