@@ -12,8 +12,46 @@
 * File:		rclause.h						 *
 * comments:	walk through a clause					 *
 *									 *
-* Last rev:     $Date: 2005/12/17 03:25:39 $,$Author: vsc $						 *
+* Last rev:     $Date: 2008/05/12 14:04:23 $,$Author: vsc $						 *
 * $Log: rclause.h,v $
+* Revision 1.26  2008/05/12 14:04:23  vsc
+* updates to restore
+*
+* Revision 1.25  2008/04/01 08:42:46  vsc
+* fix restore and small VISTA thingies
+*
+* Revision 1.24  2008/03/25 22:03:14  vsc
+* fix some icc warnings
+*
+* Revision 1.23  2008/01/28 18:12:36  vsc
+* fix small bug in restore opcode
+*
+* Revision 1.22  2008/01/23 17:57:55  vsc
+* valgrind it!
+* enable atom garbage collection.
+*
+* Revision 1.21  2007/11/26 23:43:09  vsc
+* fixes to support threads and assert correctly, even if inefficiently.
+*
+* Revision 1.20  2007/11/07 09:25:27  vsc
+* speedup meta-calls
+*
+* Revision 1.19  2007/11/06 17:02:12  vsc
+* compile ground terms away.
+*
+* Revision 1.18  2006/11/27 17:42:03  vsc
+* support for UNICODE, and other bug fixes.
+*
+* Revision 1.17  2006/10/10 14:08:17  vsc
+* small fixes on threaded implementation.
+*
+* Revision 1.16  2006/09/20 20:03:51  vsc
+* improve indexing on floats
+* fix sending large lists to DB
+*
+* Revision 1.15  2006/04/27 14:13:24  rslopes
+* *** empty log message ***
+*
 * Revision 1.14  2005/12/17 03:25:39  vsc
 * major changes to support online event-based profiling
 * improve error discovery and restart on scanner.
@@ -161,8 +199,22 @@ restore_opcodes(yamop *pc)
       pc->u.ld.d = PtoOpAdjust(pc->u.ld.d);
       pc = NEXTOP(pc,ld);
       break;
+    case _try_logical:
+    case _retry_logical:
+    case _count_retry_logical:
+    case _profiled_retry_logical:
+      pc->u.lld.n = PtoOpAdjust(pc->u.lld.n);
+      pc->u.lld.d = PtoLUClauseAdjust(pc->u.lld.d);
+      pc = pc->u.lld.n;
+      break;
+    case _trust_logical:
+    case _count_trust_logical:
+    case _profiled_trust_logical:
+      pc->u.lld.n = PtoOpAdjust(pc->u.lld.n);
+      pc->u.lld.d = PtoLUClauseAdjust(pc->u.lld.d);
+      pc->u.lld.t.block = PtoLUIndexAdjust(pc->u.lld.t.block);
+      return;
     case _enter_lu_pred:
-    case _stale_lu_index:
       pc->u.Ill.I = (LogUpdIndex *)PtoOpAdjust((yamop *)(pc->u.Ill.I));
       pc->u.Ill.l1 = PtoOpAdjust(pc->u.Ill.l1);
       pc->u.Ill.l2 = PtoOpAdjust(pc->u.Ill.l2);
@@ -187,7 +239,6 @@ restore_opcodes(yamop *pc)
       pc->u.pp.p0 = PtoPredAdjust(pc->u.pp.p0);
       pc = NEXTOP(pc,pp);
       break;
-    case _trust_logical_pred:
     case _jump:
     case _move_back:
     case _skip:
@@ -214,6 +265,9 @@ restore_opcodes(yamop *pc)
       /* instructions type EC */
     case _alloc_for_logical_pred:
       pc->u.EC.ClBase = (struct logic_upd_clause *)PtoOpAdjust((yamop *)pc->u.EC.ClBase);
+#if defined(THREADS) || defined(YAPOR)
+      pc->u.EC.p = PtoPredAdjust(pc->u.EC.p);
+#endif
       pc = NEXTOP(pc,EC);
       break;
       /* instructions type e */
@@ -233,7 +287,11 @@ restore_opcodes(yamop *pc)
     case _write_l_list:
     case _pop:
     case _index_pred:
-#if THREADS
+    case _lock_pred:
+#ifdef BEAM
+    case _retry_eam:
+#endif
+#ifdef THREADS
     case _thread_local:
 #endif
     case _expand_index:
@@ -321,7 +379,7 @@ restore_opcodes(yamop *pc)
       break;
     case _expand_clauses:
       Yap_Error(SYSTEM_ERROR, TermNil,
-	    "Invalid Opcode expand_clauses at %p", pc);
+		"Invalid Opcode expand_clauses at %p", pc);
       break;
       /* instructions type y */
     case _save_b_y:
@@ -351,9 +409,12 @@ restore_opcodes(yamop *pc)
       /* instructions type sla */      
     case _p_execute_tail:
     case _p_execute:
-      pc->u.sla.sla_u.p = PtoPredAdjust(pc->u.sla.sla_u.p);
+    case _p_execute2:
       if (pc->u.sla.sla_u.mod != 0) {
-	pc->u.sla.sla_u.mod = AtomTermAdjust(pc->u.sla.sla_u.mod);
+	if (IsAtomTerm(pc->u.sla.sla_u.mod))
+	  pc->u.sla.sla_u.mod = AtomTermAdjust(pc->u.sla.sla_u.mod);
+	else
+	  pc->u.sla.sla_u.p = PtoPredAdjust(pc->u.sla.sla_u.p);	  
       }
       pc->u.sla.p0 = PtoPredAdjust(pc->u.sla.p0);
       if (pc->u.sla.bmap != NULL) {
@@ -422,11 +483,20 @@ restore_opcodes(yamop *pc)
       pc->u.yx.y = YAdjust(pc->u.yx.y);
       pc = NEXTOP(pc,yx);
       break;
+      /* instructions type xd */
+    case _get_float:
+    case _put_float:
+      pc->u.xd.x = XAdjust(pc->u.xd.x);
+      pc = NEXTOP(pc,xd);
+      break;
+      /* instructions type xd */
+    case _get_longint:
+    case _put_longint:
+      pc->u.xi.x = XAdjust(pc->u.xi.x);
+      pc = NEXTOP(pc,xi);
+      break;
       /* instructions type xc */
     case _get_atom:
-    case _put_atom:
-    case _get_float:
-    case _get_longint:
     case _get_bigint:
       pc->u.xc.x = XAdjust(pc->u.xc.x);
       {
@@ -438,7 +508,24 @@ restore_opcodes(yamop *pc)
       }
       pc = NEXTOP(pc,xc);
       break;
-      /* instructions type cc */
+    case _put_atom:
+      pc->u.xc.x = XAdjust(pc->u.xc.x);
+      {
+	Term t = pc->u.xc.c;
+	if (IsAtomTerm(t))
+	  pc->u.xc.c = AtomTermAdjust(t);
+	else if (IsApplTerm(t))
+	  pc->u.xc.c = BlobTermAdjust(t);
+	else if (IsPairTerm(t))
+	  pc->u.xc.c = CodeComposedTermAdjust(t);
+      }
+      pc = NEXTOP(pc,xc);
+      break;
+    case _get_dbterm:
+      pc->u.xc.x = XAdjust(pc->u.xc.x);
+      pc->u.xc.c = CodeComposedTermAdjust(pc->u.xc.c);
+      pc = NEXTOP(pc,xc);
+      break;
     case _get_2atoms:
       {
 	Term t = pc->u.cc.c1;
@@ -685,15 +772,33 @@ restore_opcodes(yamop *pc)
       pc->u.os.opcw = Yap_opcode(Yap_op_from_opcode(pc->u.os.opcw));
       pc = NEXTOP(pc,os);
       break;
+      /* instructions type od */
+    case _unify_float:
+    case _unify_l_float:
+    case _unify_float_write:
+    case _unify_l_float_write:
+      pc->u.oc.opcw = Yap_opcode(Yap_op_from_opcode(pc->u.oc.opcw));
+      pc = NEXTOP(pc,od);
+      break;
+      /* instructions type oi */
+    case _unify_longint:
+    case _unify_l_longint:
+    case _unify_longint_write:
+    case _unify_l_longint_write:
+      pc->u.oc.opcw = Yap_opcode(Yap_op_from_opcode(pc->u.oc.opcw));
+      pc = NEXTOP(pc,oi);
+      break;
       /* instructions type oc */
+    case _unify_dbterm:
+    case _unify_l_dbterm:
+      pc->u.oc.opcw = Yap_opcode(Yap_op_from_opcode(pc->u.oc.opcw));
+      pc->u.oc.c = CodeComposedTermAdjust(pc->u.oc.c);
+      pc = NEXTOP(pc,oc);
+      break;
     case _unify_atom_write:
     case _unify_atom:
     case _unify_l_atom_write:
     case _unify_l_atom:
-    case _unify_float:
-    case _unify_l_float:
-    case _unify_longint:
-    case _unify_l_longint:
     case _unify_bigint:
     case _unify_l_bigint:
       pc->u.oc.opcw = Yap_opcode(Yap_op_from_opcode(pc->u.oc.opcw));
@@ -737,6 +842,14 @@ restore_opcodes(yamop *pc)
 #endif /* TABLING */
       pc = NEXTOP(pc,s);
       break;
+      /* instructions type d */
+    case _write_float:
+      pc = NEXTOP(pc,d);
+      break;
+      /* instructions type i */
+    case _write_longint:
+      pc = NEXTOP(pc,i);
+      break;
       /* instructions type c */
    case _write_atom:
       {
@@ -745,6 +858,8 @@ restore_opcodes(yamop *pc)
 	    pc->u.c.c = AtomTermAdjust(t);
 	else if (IsApplTerm(t))
 	  pc->u.c.c = BlobTermAdjust(t);
+	else if (IsPairTerm(t))
+	  pc->u.c.c = CodeComposedTermAdjust(t);
       }
       pc = NEXTOP(pc,c);
       break;
@@ -849,7 +964,7 @@ restore_opcodes(yamop *pc)
 	for (j = 0; j < i; j++) {
 	  Functor oldfunc = (Functor)(oldcode[0]);
 	  CODEADDR oldjmp = (CODEADDR)(oldcode[1]);
-	  if (oldfunc != NULL) {
+	  if (oldfunc) {
 	    oldcode[0] = (CELL)FuncAdjust(oldfunc);
 	  }
 	  oldcode[1] = (CELL)CodeAddrAdjust(oldjmp);
@@ -864,12 +979,12 @@ restore_opcodes(yamop *pc)
       {
 	int             i, j;
 	CELL            *oldcode;
-#if !USE_OFFSETS
+#if !defined(USE_OFFSETS)
 	CELL            *startcode;
 #endif
 
 	i = pc->u.sssl.s;
-#if !USE_OFFSETS
+#if !defined(USE_OFFSETS)
 	startcode =
 #endif
 	  oldcode = (CELL *)(pc->u.sssl.l = PtoOpAdjust(pc->u.sssl.l));
