@@ -11,8 +11,119 @@
 * File:		rheap.h							 *
 * comments:	walk through heap code					 *
 *									 *
-* Last rev:     $Date: 2006/03/22 20:07:28 $,$Author: vsc $						 *
+* Last rev:     $Date: 2008/05/12 22:31:37 $,$Author: vsc $						 *
 * $Log: rheap.h,v $
+* Revision 1.98  2008/05/12 22:31:37  vsc
+* fix previous fixes
+*
+* Revision 1.97  2008/05/12 14:04:23  vsc
+* updates to restore
+*
+* Revision 1.96  2008/04/11 16:58:17  ricroc
+* yapor: seq_def initialization
+*
+* Revision 1.95  2008/04/06 12:06:48  vsc
+* more small fixes
+*
+* Revision 1.94  2008/04/06 11:53:02  vsc
+*  fix some restore bugs
+*
+* Revision 1.93  2008/04/04 09:10:02  vsc
+* restore was restoring twice
+*
+* Revision 1.92  2008/04/03 11:34:47  vsc
+* fix restorebb in cases entry key is not an atom (obs from Nicos
+* Angelopoulos)
+*
+* Revision 1.91  2008/04/01 15:31:43  vsc
+* more saved state fixes
+*
+* Revision 1.90  2008/04/01 14:09:43  vsc
+* improve restore
+*
+* Revision 1.89  2008/04/01 09:41:05  vsc
+* more fixes to restore
+*
+* Revision 1.88  2008/04/01 08:42:46  vsc
+* fix restore and small VISTA thingies
+*
+* Revision 1.87  2008/03/25 22:03:14  vsc
+* fix some icc warnings
+*
+* Revision 1.86  2008/03/25 16:45:53  vsc
+* make or-parallelism compile again
+*
+* Revision 1.85  2008/02/12 17:03:52  vsc
+* SWI-portability changes
+*
+* Revision 1.84  2008/02/07 21:39:51  vsc
+* fix case where predicate is for an integer (DBEntry).
+*
+* Revision 1.83  2008/01/23 17:57:55  vsc
+* valgrind it!
+* enable atom garbage collection.
+*
+* Revision 1.82  2007/12/05 12:17:23  vsc
+* improve JT
+* fix graph compatibility with SICStus
+* re-export declaration.
+*
+* Revision 1.81  2007/11/26 23:43:09  vsc
+* fixes to support threads and assert correctly, even if inefficiently.
+*
+* Revision 1.80  2007/11/07 09:35:53  vsc
+* small fix
+*
+* Revision 1.79  2007/11/07 09:25:27  vsc
+* speedup meta-calls
+*
+* Revision 1.78  2007/11/06 17:02:12  vsc
+* compile ground terms away.
+*
+* Revision 1.77  2007/10/10 09:44:24  vsc
+* some more fixes to make YAP swi compatible
+* fix absolute_file_name (again)
+* fix setarg
+*
+* Revision 1.76  2007/09/28 23:18:17  vsc
+* handle learning from interpretations.
+*
+* Revision 1.75  2007/04/10 22:13:21  vsc
+* fix max modules limitation
+*
+* Revision 1.74  2007/03/22 11:12:21  vsc
+* make sure that YAP_Restart does not restart a failed goal.
+*
+* Revision 1.73  2007/02/18 00:26:36  vsc
+* fix atom garbage collector (although it is still off by default)
+* make valgrind feel better
+*
+* Revision 1.72  2007/01/08 08:27:19  vsc
+* fix restore (Trevor)
+* make indexing a bit faster on IDB
+*
+* Revision 1.71  2006/11/27 17:42:03  vsc
+* support for UNICODE, and other bug fixes.
+*
+* Revision 1.70  2006/08/25 19:50:35  vsc
+* global data structures
+*
+* Revision 1.69  2006/08/22 16:12:46  vsc
+* global variables
+*
+* Revision 1.68  2006/08/02 18:18:30  vsc
+* preliminary support for readutil library (SWI compatible).
+*
+* Revision 1.67  2006/05/17 18:38:11  vsc
+* make system library use true file name
+*
+* Revision 1.66  2006/04/28 15:48:33  vsc
+* do locking on streams
+*
+* Revision 1.65  2006/04/28 13:23:23  vsc
+* fix number of overflow bugs affecting threaded version
+* make current_op faster.
+*
 * Revision 1.64  2006/03/22 20:07:28  vsc
 * take better care of zombies
 *
@@ -149,23 +260,35 @@ do_clean_susp_clauses(yamop *ipc) {
 
 #include "rclause.h"
 
-/* Restoring the heap */
-
 /* adjusts terms stored in the data base, when they have no variables */
 static Term 
 AdjustDBTerm(Term trm, Term *p_base)
 {
+  if (IsVarTerm(trm))
+    return CodeVarAdjust(trm);
   if (IsAtomTerm(trm))
     return AtomTermAdjust(trm);
   if (IsPairTerm(trm)) {
     Term           *p;
+    Term out;
 
     p = PtoHeapCellAdjust(RepPair(trm));
+    out = AbsPair(p);
+  loop:
     if (p >= p_base) {
       p[0] = AdjustDBTerm(p[0], p);
-      p[1] = AdjustDBTerm(p[1], p);
+      if (IsPairTerm(p[1])) {
+	/* avoid term recursion with very deep lists */
+	Term *newp = PtoHeapCellAdjust(RepPair(p[1]));
+	p[1] = AbsPair(newp);
+	p_base = p;
+	p = newp;
+	goto loop;
+      } else {
+	p[1] = AdjustDBTerm(p[1], p);
+      }
     }
-    return AbsPair(p);
+    return out;
   }
   if (IsApplTerm(trm)) {
     Term           *p;
@@ -193,12 +316,17 @@ AdjustDBTerm(Term trm, Term *p_base)
 }
 
 static void
-RestoreDBTerm(DBTerm *dbr)
+RestoreDBTerm(DBTerm *dbr, int attachments)
 {
+  if (attachments) {
 #ifdef COROUTINING
-  if (dbr->attachments)
-    dbr->attachments = AdjustDBTerm(dbr->attachments, dbr->Contents);
+    if (dbr->ag.attachments)
+      dbr->ag.attachments = AdjustDBTerm(dbr->ag.attachments, dbr->Contents);
 #endif
+  } else {
+    if (dbr->ag.NextDBT)
+      dbr->ag.NextDBT = DBTermAdjust(dbr->ag.NextDBT);
+  }    
   if (dbr->DBRefs !=  NULL) {
     DBRef          *cp;
     DBRef            tm;
@@ -211,6 +339,8 @@ RestoreDBTerm(DBTerm *dbr)
   dbr->Entry = AdjustDBTerm(dbr->Entry, dbr->Contents);
 }
 
+/* Restoring the heap */
+
 /* Restores a prolog clause, in its compiled form */
 static void 
 RestoreStaticClause(StaticClause *cl)
@@ -274,7 +404,7 @@ RestoreLUClause(LogUpdClause *cl, PredEntry *pp)
   }
   if (cl->ClSource) {
     cl->ClSource = DBTermAdjust(cl->ClSource);
-    RestoreDBTerm(cl->ClSource);
+    RestoreDBTerm(cl->ClSource, TRUE);
   }
   if (cl->ClPrev) {
     cl->ClPrev = PtoLUCAdjust(cl->ClPrev);
@@ -286,21 +416,43 @@ RestoreLUClause(LogUpdClause *cl, PredEntry *pp)
   restore_opcodes(cl->ClCode);
 }
 
+static void
+RestoreDBTermEntry(struct dbterm_list *dbl) {
+  DBTerm *dbt;
+
+  if (dbl->dbterms)
+    dbt = dbl->dbterms = DBTermAdjust(dbl->dbterms);
+  else
+    return;
+  dbl->clause_code = PtoOpAdjust(dbl->clause_code);
+  if (dbl->next_dbl)
+    dbl->next_dbl = PtoDBTLAdjust(dbl->next_dbl);
+  dbl->p = PredEntryAdjust(dbl->p);
+  while (dbt) {
+    RestoreDBTerm(dbt, FALSE);
+    dbt = dbt->ag.NextDBT;
+  }
+}
+
 static void 
-CleanLUIndex(LogUpdIndex *idx)
+CleanLUIndex(LogUpdIndex *idx, int recurse)
 {
-  idx->ClRefCount = 0;
   INIT_LOCK(idx->ClLock);
   idx->ClPred = PtoPredAdjust(idx->ClPred);
   if (idx->ParentIndex)
     idx->ParentIndex = LUIndexAdjust(idx->ParentIndex);
+  if (idx->PrevSiblingIndex) {
+    idx->PrevSiblingIndex = LUIndexAdjust(idx->PrevSiblingIndex);
+  }
   if (idx->SiblingIndex) {
     idx->SiblingIndex = LUIndexAdjust(idx->SiblingIndex);
-    CleanLUIndex(idx->SiblingIndex);
+    if (recurse)
+      CleanLUIndex(idx->SiblingIndex, TRUE);
   }
   if (idx->ChildIndex) {
     idx->ChildIndex = LUIndexAdjust(idx->ChildIndex);
-    CleanLUIndex(idx->ChildIndex);
+    if (recurse)
+      CleanLUIndex(idx->ChildIndex, TRUE);
   }
   if (!(idx->ClFlags & SwitchTableMask)) {
     restore_opcodes(idx->ClCode);
@@ -308,16 +460,18 @@ CleanLUIndex(LogUpdIndex *idx)
 }
 
 static void 
-CleanSIndex(StaticIndex *idx)
+CleanSIndex(StaticIndex *idx, int recurse)
 {
   idx->ClPred = PtoPredAdjust(idx->ClPred);
   if (idx->SiblingIndex) {
     idx->SiblingIndex = SIndexAdjust(idx->SiblingIndex);
-    CleanSIndex(idx->SiblingIndex);
+    if (recurse)
+      CleanSIndex(idx->SiblingIndex, TRUE);
   }
   if (idx->ChildIndex) {
     idx->ChildIndex = SIndexAdjust(idx->ChildIndex);
-    CleanSIndex(idx->ChildIndex);
+    if (recurse)
+      CleanSIndex(idx->ChildIndex, TRUE);
   }
   if (!(idx->ClFlags & SwitchTableMask)) {
     restore_opcodes(idx->ClCode);
@@ -329,11 +483,9 @@ static void
 restore_codes(void)
 {
   Yap_heap_regs->heap_top = AddrAdjust(OldHeapTop);
-  if (Yap_heap_regs->heap_lim) {
-    Yap_heap_regs->heap_lim = AddrAdjust(Yap_heap_regs->heap_lim);
-  }
 #ifdef YAPOR
-  Yap_heap_regs->seq_def = TRUE;
+  /* ricroc: if in use overrides startup file settings
+  Yap_heap_regs->seq_def = TRUE; */
   Yap_heap_regs->getwork_code.opc = Yap_opcode(_getwork);
   INIT_YAMOP_LTT(&(Yap_heap_regs->getwork_code), 0);
   Yap_heap_regs->getwork_seq_code.opc = Yap_opcode(_getwork_seq);
@@ -376,16 +528,17 @@ restore_codes(void)
   Yap_heap_regs->trustfailcode->opc = Yap_opcode(_trust_fail);
 
   Yap_heap_regs->env_for_yes_code.op = Yap_opcode(_call);
-  Yap_heap_regs->yescode->opc = Yap_opcode(_Ystop);
+  Yap_heap_regs->yescode.opc = Yap_opcode(_Ystop);
   Yap_heap_regs->undef_op = Yap_opcode(_undef_p);
   Yap_heap_regs->index_op = Yap_opcode(_index_pred);
+  Yap_heap_regs->lockpred_op = Yap_opcode(_lock_pred);
   Yap_heap_regs->fail_op = Yap_opcode(_op_fail);
-  Yap_heap_regs->nocode->opc = Yap_opcode(_Nstop);
+  Yap_heap_regs->nocode.opc = Yap_opcode(_Nstop);
+  Yap_heap_regs->rtrycode.opc = Yap_opcode(_retry_and_mark);
 #ifdef YAPOR
   INIT_YAMOP_LTT(&(Yap_heap_regs->nocode), 1);
   INIT_YAMOP_LTT(&(Yap_heap_regs->rtrycode), 1);
 #endif /* YAPOR */
-  ((yamop *)(&Yap_heap_regs->rtrycode))->opc = Yap_opcode(_retry_and_mark);
   if (((yamop *)(&Yap_heap_regs->rtrycode))->u.ld.d != NIL)
     ((yamop *)(&Yap_heap_regs->rtrycode))->u.ld.d =
       PtoOpAdjust(((yamop *)(&Yap_heap_regs->rtrycode))->u.ld.d);
@@ -404,24 +557,26 @@ restore_codes(void)
 	(Functor)AtomAdjust((Atom)(Yap_heap_regs->clausecode->func));
     }
   }
+#if !defined(THREADS) && !defined(YAPOR)
   /* restore consult stack. It consists of heap pointers, so it
      is easy to fix.
   */
-  Yap_heap_regs->consultlow  = 
-    ConsultObjAdjust(Yap_heap_regs->consultlow);
-  Yap_heap_regs->consultbase =
-    ConsultObjAdjust(Yap_heap_regs->consultbase);
-  Yap_heap_regs->consultsp   =
-    ConsultObjAdjust(Yap_heap_regs->consultsp);
+  Yap_heap_regs->wl.consultlow  = 
+    ConsultObjAdjust(Yap_heap_regs->wl.consultlow);
+  Yap_heap_regs->wl.consultbase =
+    ConsultObjAdjust(Yap_heap_regs->wl.consultbase);
+  Yap_heap_regs->wl.consultsp   =
+    ConsultObjAdjust(Yap_heap_regs->wl.consultsp);
   {
     /* we assume all pointers have the same size */
-    register consult_obj *pt = Yap_heap_regs->consultsp;
+    register consult_obj *pt = Yap_heap_regs->wl.consultsp;
     while (pt <
-	   Yap_heap_regs->consultlow+Yap_heap_regs->consultcapacity) {
+	   Yap_heap_regs->wl.consultlow+Yap_heap_regs->wl.consultcapacity) {
       pt->p = PropAdjust(pt->p);
       pt ++;
     }
   }
+#endif
 #if USE_THREADED_CODE
   Yap_heap_regs->op_rtable = (opentry *)
     CodeAddrAdjust((CODEADDR)(Yap_heap_regs->op_rtable));
@@ -430,6 +585,10 @@ restore_codes(void)
     Yap_heap_regs->atprompt =
       AtomAdjust(Yap_heap_regs->atprompt);
   }
+  if (Yap_heap_regs->current_modules) {
+    Yap_heap_regs->current_modules = (struct mod_entry *)
+      AddrAdjust((ADDR)Yap_heap_regs->current_modules);
+  }
   if (Yap_heap_regs->char_conversion_table) {
     Yap_heap_regs->char_conversion_table = (char *)
       AddrAdjust((ADDR)Yap_heap_regs->char_conversion_table);
@@ -438,6 +597,10 @@ restore_codes(void)
     Yap_heap_regs->char_conversion_table2 = (char *)
       AddrAdjust((ADDR)Yap_heap_regs->char_conversion_table2);
   }
+  if (Yap_heap_regs->op_list) {
+    Yap_heap_regs->op_list = (struct operator_entry *)
+      AddrAdjust((ADDR)Yap_heap_regs->op_list);
+  }
   if (Yap_heap_regs->dead_static_clauses) {
     StaticClause *sc = PtoStCAdjust(Yap_heap_regs->dead_static_clauses);
     Yap_heap_regs->dead_static_clauses = sc;
@@ -454,11 +617,19 @@ restore_codes(void)
       mc = mc->ClNext;
     }
   }
+  if (Yap_heap_regs->dbterms_list) {
+    struct dbterm_list *dbl = PtoDBTLAdjust(Yap_heap_regs->dbterms_list);
+    Yap_heap_regs->dbterms_list = dbl;
+    while (dbl) {
+      RestoreDBTermEntry(dbl);
+      dbl = dbl->next_dbl;
+    }
+  }
   if (Yap_heap_regs->dead_static_indices) {
     StaticIndex *si = (StaticIndex *)AddrAdjust((ADDR)(Yap_heap_regs->dead_static_indices));
     Yap_heap_regs->dead_static_indices = si;
     while (si) {
-      CleanSIndex(si);
+      CleanSIndex(si, FALSE);
       si = si->SiblingIndex;
     }
   }
@@ -473,7 +644,7 @@ restore_codes(void)
       for (i = 0; i < Yap_heap_regs->int_keys_size; i++) {
 	if (Yap_heap_regs->IntKeys[i] != NIL) {
 	  Prop p0 = Yap_heap_regs->IntKeys[i] = PropAdjust(Yap_heap_regs->IntKeys[i]);
-	  RestoreEntries(RepProp(p0));
+	  RestoreEntries(RepProp(p0), TRUE);
 	}
       }
     }
@@ -505,20 +676,25 @@ restore_codes(void)
       for (i = 0; i < Yap_heap_regs->int_bb_keys_size; i++) {
 	if (Yap_heap_regs->IntBBKeys[i] != NIL) {
 	  Prop p0 = Yap_heap_regs->IntBBKeys[i] = PropAdjust(Yap_heap_regs->IntBBKeys[i]);
-	  RestoreEntries(RepProp(p0));
+	  RestoreEntries(RepProp(p0), TRUE);
 	}
       }
     }
   }
-  {
-    /* adjust atoms in atom table */
-    unsigned int i = 0;
-
-    for (i = 0; i < Yap_heap_regs->no_of_modules; i++) {
-      Yap_heap_regs->module_name[i] = AtomTermAdjust(Yap_heap_regs->module_name[i]);
-      if (Yap_heap_regs->module_pred[i]) {
-	Yap_heap_regs->module_pred[i] = PtoPredAdjust(Yap_heap_regs->module_pred[i]);
-      }
+  if (Yap_heap_regs->db_erased_list) {
+    LogUpdClause *lcl = Yap_heap_regs->db_erased_list = 
+      PtoLUCAdjust(Yap_heap_regs->db_erased_list);
+    while (lcl) {
+      RestoreLUClause(lcl, FALSE);
+      lcl = lcl->ClNext;
+    }
+  }
+  if (Yap_heap_regs->db_erased_ilist) {
+    LogUpdIndex *icl = Yap_heap_regs->db_erased_ilist = 
+      LUIndexAdjust(Yap_heap_regs->db_erased_ilist);
+    while (icl) {
+      CleanLUIndex(icl, FALSE);
+      icl = icl->SiblingIndex;
     }
   }
   Yap_heap_regs->atom_abol = AtomAdjust(Yap_heap_regs->atom_abol);
@@ -609,6 +785,7 @@ restore_codes(void)
   Yap_heap_regs->functor_csult = FuncAdjust(Yap_heap_regs->functor_csult);
   Yap_heap_regs->functor_eq = FuncAdjust(Yap_heap_regs->functor_eq);
   Yap_heap_regs->functor_execute_in_mod = FuncAdjust(Yap_heap_regs->functor_execute_in_mod);
+  Yap_heap_regs->functor_execute2_in_mod = FuncAdjust(Yap_heap_regs->functor_execute2_in_mod);
   Yap_heap_regs->functor_execute_within = FuncAdjust(Yap_heap_regs->functor_execute_within);
   Yap_heap_regs->functor_g_atom = FuncAdjust(Yap_heap_regs->functor_g_atom);
   Yap_heap_regs->functor_g_atomic = FuncAdjust(Yap_heap_regs->functor_g_atomic);
@@ -627,6 +804,7 @@ restore_codes(void)
 #ifdef MULTI_ASSIGNMENT_VARIABLES
   Yap_heap_regs->functor_mutable = FuncAdjust(Yap_heap_regs->functor_mutable);
 #endif
+  Yap_heap_regs->functor_nb_queue = FuncAdjust(Yap_heap_regs->functor_nb_queue);
   Yap_heap_regs->functor_not = FuncAdjust(Yap_heap_regs->functor_not);
   Yap_heap_regs->functor_or = FuncAdjust(Yap_heap_regs->functor_or);
   Yap_heap_regs->functor_portray = FuncAdjust(Yap_heap_regs->functor_portray);
@@ -651,51 +829,66 @@ restore_codes(void)
   Yap_heap_regs->attributes_module = AtomTermAdjust(Yap_heap_regs->attributes_module);
   Yap_heap_regs->charsio_module = AtomTermAdjust(Yap_heap_regs->charsio_module);
   Yap_heap_regs->terms_module = AtomTermAdjust(Yap_heap_regs->terms_module);
-  if (Yap_heap_regs->file_aliases != NULL) {
+  Yap_heap_regs->system_module = AtomTermAdjust(Yap_heap_regs->system_module);
+  Yap_heap_regs->readutil_module = AtomTermAdjust(Yap_heap_regs->readutil_module);
+  Yap_heap_regs->hacks_module = AtomTermAdjust(Yap_heap_regs->hacks_module);
+  Yap_heap_regs->globals_module = AtomTermAdjust(Yap_heap_regs->globals_module);
+  Yap_heap_regs->arg_module = AtomTermAdjust(Yap_heap_regs->arg_module);
+  Yap_heap_regs->swi_module = AtomTermAdjust(Yap_heap_regs->swi_module);
+  Yap_heap_regs->global_hold_entry = HoldEntryAdjust(Yap_heap_regs->global_hold_entry);
+  if (Yap_heap_regs->yap_streams != NULL) {
+    int sno;
+
     Yap_heap_regs->yap_streams =
       (struct stream_desc *)AddrAdjust((ADDR)Yap_heap_regs->yap_streams);
+    for (sno = 0; sno < MaxStreams; ++sno) {
+      if (Stream[sno].status & Free_Stream_f)
+	continue;
+      if (Stream[sno].status & (Socket_Stream_f|Pipe_Stream_f|InMemory_Stream_f)) 
+	continue;
+      Stream[sno].u.file.user_name = AtomTermAdjust(Stream[sno].u.file.user_name);
+      Stream[sno].u.file.name = AtomAdjust(Stream[sno].u.file.name);
+    }    
   }
   if (Yap_heap_regs->file_aliases != NULL) {
+    int i;
+
     Yap_heap_regs->file_aliases =
       (struct AliasDescS *)AddrAdjust((ADDR)Yap_heap_regs->file_aliases);
+    for (i = 0; i < NOfFileAliases; i++)
+      FileAliases[i].name = AtomAdjust(FileAliases[i].name);
+  }
+  if (Yap_heap_regs->yap_lib_dir) {
+    Yap_heap_regs->yap_lib_dir =
+      (char *)AddrAdjust((ADDR)Yap_heap_regs->yap_lib_dir);
   }
-  Yap_heap_regs->yap_lib_dir =
-    (char *)AddrAdjust((ADDR)Yap_heap_regs->yap_lib_dir);
   Yap_heap_regs->pred_goal_expansion =
-    (PredEntry *)AddrAdjust((ADDR)Yap_heap_regs->pred_goal_expansion);
+    PredEntryAdjust(Yap_heap_regs->pred_goal_expansion);
   Yap_heap_regs->pred_meta_call =
-    (PredEntry *)AddrAdjust((ADDR)Yap_heap_regs->pred_meta_call);
+    PredEntryAdjust(Yap_heap_regs->pred_meta_call);
   Yap_heap_regs->pred_dollar_catch =
-    (PredEntry *)AddrAdjust((ADDR)Yap_heap_regs->pred_dollar_catch);
+    PredEntryAdjust(Yap_heap_regs->pred_dollar_catch);
   Yap_heap_regs->pred_recorded_with_key =
-    (PredEntry *)AddrAdjust((ADDR)Yap_heap_regs->pred_recorded_with_key);
+    PredEntryAdjust(Yap_heap_regs->pred_recorded_with_key);
   Yap_heap_regs->pred_log_upd_clause =
-    (PredEntry *)AddrAdjust((ADDR)Yap_heap_regs->pred_log_upd_clause);
+    PredEntryAdjust(Yap_heap_regs->pred_log_upd_clause);
+  Yap_heap_regs->pred_log_upd_clause_erase =
+    PredEntryAdjust(Yap_heap_regs->pred_log_upd_clause_erase);
   Yap_heap_regs->pred_log_upd_clause0 =
-    (PredEntry *)AddrAdjust((ADDR)Yap_heap_regs->pred_log_upd_clause0);
+    PredEntryAdjust(Yap_heap_regs->pred_log_upd_clause0);
   Yap_heap_regs->pred_static_clause =
-    (PredEntry *)AddrAdjust((ADDR)Yap_heap_regs->pred_static_clause);
+    PredEntryAdjust(Yap_heap_regs->pred_static_clause);
   Yap_heap_regs->pred_throw =
-    (PredEntry *)AddrAdjust((ADDR)Yap_heap_regs->pred_throw);
+    PredEntryAdjust(Yap_heap_regs->pred_throw);
   Yap_heap_regs->pred_handle_throw =
-    (PredEntry *)AddrAdjust((ADDR)Yap_heap_regs->pred_handle_throw);
-#if DEBUG
-  if (Yap_heap_regs->db_erased_list) {
-    Yap_heap_regs->db_erased_list = 
-      PtoLUCAdjust(Yap_heap_regs->db_erased_list);
-  }
-  if (Yap_heap_regs->db_erased_ilist) {
-    Yap_heap_regs->db_erased_ilist = 
-      LUIndexAdjust(Yap_heap_regs->db_erased_ilist);
-  }
-#endif
+    PredEntryAdjust(Yap_heap_regs->pred_handle_throw);
   if (Yap_heap_regs->undef_code != NULL)
     Yap_heap_regs->undef_code = (PredEntry *)PtoHeapCellAdjust((CELL *)(Yap_heap_regs->undef_code));
   if (Yap_heap_regs->creep_code != NULL)
     Yap_heap_regs->creep_code = (PredEntry *)PtoHeapCellAdjust((CELL *)(Yap_heap_regs->creep_code));
   if (Yap_heap_regs->spy_code != NULL)
     Yap_heap_regs->spy_code = (PredEntry *)PtoHeapCellAdjust((CELL *)(Yap_heap_regs->spy_code));
-#if !defined(THREADS)
+#if !defined(THREADS) && !defined(YAPOR)
   if (Yap_heap_regs->wl.scratchpad.ptr) {
     Yap_heap_regs->wl.scratchpad.ptr =
       (char *)AddrAdjust((ADDR)Yap_heap_regs->wl.scratchpad.ptr);
@@ -709,7 +902,7 @@ restore_codes(void)
 #ifdef COROUTINING
   if (Yap_heap_regs->wake_up_code != NULL)
     Yap_heap_regs->wake_up_code = (PredEntry *)PtoHeapCellAdjust((CELL *)(Yap_heap_regs->wake_up_code));
-#if !defined(THREADS)
+#if !defined(THREADS) && !defined(YAPOR)
   Yap_heap_regs->wl.atts_mutable_list =
     AbsAppl(PtoGloAdjust(RepAppl(Yap_heap_regs->wl.atts_mutable_list)));
   if (Yap_heap_regs->wl.dynamic_arrays) {
@@ -720,6 +913,24 @@ restore_codes(void)
     Yap_heap_regs->wl.static_arrays =
       PtoArraySAdjust(Yap_heap_regs->wl.static_arrays);
   }
+  if (Yap_heap_regs->wl.global_variables) {
+    Yap_heap_regs->wl.global_variables =
+      PtoGlobalEAdjust(Yap_heap_regs->wl.global_variables);
+  }
+  if (Yap_heap_regs->wl.global_arena) {
+    if (IsAtomTerm(Yap_heap_regs->wl.global_arena)) {
+      Yap_heap_regs->wl.global_arena =
+	AtomTermAdjust(Yap_heap_regs->wl.global_arena);
+    } else {
+      Yap_heap_regs->wl.global_arena =
+	AbsAppl(PtoGloAdjust(RepAppl(Yap_heap_regs->wl.global_arena)));
+    }
+  }
+  if (Yap_heap_regs->wl.global_delay_arena) {
+    Yap_heap_regs->wl.global_delay_arena =
+      GlobalAdjust(Yap_heap_regs->wl.global_delay_arena);
+  }
+  Yap_heap_regs->wl.allow_restart = FALSE;
 #endif
 #endif
   if (Yap_heap_regs->last_wtime != NULL)
@@ -734,6 +945,8 @@ restore_codes(void)
     PtoPredAdjust(Yap_heap_regs->logdb_erased_marker->ClPred);
   Yap_heap_regs->hash_chain = 
     (AtomHashEntry *)PtoHeapCellAdjust((CELL *)(Yap_heap_regs->hash_chain));
+  Yap_heap_regs->wide_hash_chain = 
+    (AtomHashEntry *)PtoHeapCellAdjust((CELL *)(Yap_heap_regs->wide_hash_chain));
 }
 
 
@@ -753,7 +966,7 @@ RestoreDBEntry(DBRef dbr)
   else
     fprintf(stderr, " a var\n");
 #endif
-  RestoreDBTerm(&(dbr->DBT));
+  RestoreDBTerm(&(dbr->DBT), TRUE);
   if (dbr->Parent) {
     dbr->Parent = (DBProp)AddrAdjust((ADDR)(dbr->Parent));
   }
@@ -841,15 +1054,26 @@ CleanClauses(yamop *First, yamop *Last, PredEntry *pp)
 
 /* Restores a DB structure, as it was saved in the heap */
 static void 
-RestoreBB(BlackBoardEntry *pp)
+RestoreBB(BlackBoardEntry *pp, int int_key)
 {
   Term t = pp->Element;
   if (t) {
-    if (!IsVarTerm(t) && !IsAtomicTerm(t)) {
-      RestoreLUClause((LogUpdClause *)DBRefOfTerm(t),NULL);
+    if (!IsVarTerm(t)) {
+      if (IsAtomicTerm(t)) {
+	if (IsAtomTerm(t)) {
+	  pp->Element = AtomTermAdjust(t);
+	}
+      } else {
+	RestoreLUClause((LogUpdClause *)DBRefOfTerm(t),NULL);
+      }
     }
   }
-  pp->KeyOfBB = AtomAdjust(pp->KeyOfBB);
+  if (!int_key) {
+    pp->KeyOfBB = AtomAdjust(pp->KeyOfBB);
+  }
+  if (pp->ModuleOfBB) {
+    pp->ModuleOfBB = AtomTermAdjust(pp->ModuleOfBB);
+  }
 }
 
 static void
@@ -878,8 +1102,6 @@ restore_static_array(StaticArrayEntry *ae)
 	    *base++ = (AtomEntry *)LocalAddrAdjust((ADDR)reg);
 	  } else if (IsOldGlobal((CELL)reg)) {
 	    *base++ = (AtomEntry *)GlobalAddrAdjust((ADDR)reg);
-	  } else if (IsOldDelay((CELL)reg)) {
-	    *base++ = (AtomEntry *)DelayAddrAdjust((ADDR)reg);
 	  } else if (IsOldTrail((CELL)reg)) {
 	    *base++ = (AtomEntry *)TrailAddrAdjust((ADDR)reg);
 	  } else {
@@ -962,7 +1184,7 @@ restore_static_array(StaticArrayEntry *ae)
 	  } else {
 	    DBTerm *db = (DBTerm *)RepAppl(reg);
 	    db = DBTermAdjust(db);
-	    RestoreDBTerm(db);
+	    RestoreDBTerm(db, TRUE);
 	    base->tstore = AbsAppl((CELL *)db);
 	  }
 	}
@@ -981,7 +1203,7 @@ restore_static_array(StaticArrayEntry *ae)
 	    base++;
 	  } else {
 	    *base++ = reg = DBTermAdjust(reg);
-	    RestoreDBTerm(reg);
+	    RestoreDBTerm(reg, TRUE);
 	  }
 	}
       }
@@ -1006,7 +1228,9 @@ CleanCode(PredEntry *pp)
   }
   if (pp->ArityOfPE) {
     if (pp->ModuleOfPred == IDB_MODULE) {
-      if (pp->PredFlags & AtomDBPredFlag) {
+      if (pp->PredFlags & NumberDBPredFlag) {
+	/* it's an integer, do nothing */
+      } else if (pp->PredFlags & AtomDBPredFlag) {
 	pp->FunctorOfPred = (Functor)AtomAdjust((Atom)(pp->FunctorOfPred));
       } else {
 	pp->FunctorOfPred = FuncAdjust(pp->FunctorOfPred);
@@ -1017,14 +1241,19 @@ CleanCode(PredEntry *pp)
   } else {
     pp->FunctorOfPred = (Functor)AtomAdjust((Atom)(pp->FunctorOfPred));
   }
-  if (pp->ModuleOfPred != IDB_MODULE) {
-    if (pp->src.OwnerFile && pp->ModuleOfPred != IDB_MODULE)
-      pp->src.OwnerFile = AtomAdjust(pp->src.OwnerFile);
-  }
   if (!(pp->PredFlags & NumberDBPredFlag)) {
-    pp->src.OwnerFile = AtomAdjust(pp->src.OwnerFile);
+    if (pp->PredFlags & MultiFileFlag) {
+      if (pp->src.file_srcs)
+	pp->src.file_srcs = MFileAdjust(pp->src.file_srcs);
+    } else {
+      if (pp->src.OwnerFile)
+	pp->src.OwnerFile = AtomAdjust(pp->src.OwnerFile);
+    }
   }
   pp->OpcodeOfPred = Yap_opcode(Yap_op_from_opcode(pp->OpcodeOfPred));
+  if (pp->NextPredOfModule) {
+    pp->NextPredOfModule = PtoPredAdjust(pp->NextPredOfModule);
+  }
   if (pp->PredFlags & (AsmPredFlag|CPredFlag)) {
     /* assembly */
     if (pp->CodeOfPred) {
@@ -1041,8 +1270,6 @@ CleanCode(PredEntry *pp)
     pp->CodeOfPred =PtoOpAdjust(pp->CodeOfPred);
     pp->cs.p_code.TrueCodeOfPred = PtoOpAdjust(pp->cs.p_code.TrueCodeOfPred);
     pp->cs.p_code.ExpandCode = Yap_opcode(_expand_index);
-    if (pp->NextPredOfModule)
-      pp->NextPredOfModule = PtoPredAdjust(pp->NextPredOfModule);
     flag = pp->PredFlags;
     FirstC = pp->cs.p_code.FirstClause;
     LastC = pp->cs.p_code.LastClause;
@@ -1059,9 +1286,9 @@ CleanCode(PredEntry *pp)
       fprintf(stderr, "Correcting indexed code\n");
 #endif
       if (flag & LogUpdatePredFlag) {
-	CleanLUIndex(ClauseCodeToLogUpdIndex(pp->cs.p_code.TrueCodeOfPred));
+	CleanLUIndex(ClauseCodeToLogUpdIndex(pp->cs.p_code.TrueCodeOfPred), TRUE);
       } else {
-	CleanSIndex(ClauseCodeToStaticIndex(pp->cs.p_code.TrueCodeOfPred));
+	CleanSIndex(ClauseCodeToStaticIndex(pp->cs.p_code.TrueCodeOfPred), TRUE);
       } 
     } else if (flag & DynamicPredFlag) {
 #ifdef	DEBUG_RESTORE2
@@ -1078,7 +1305,7 @@ CleanCode(PredEntry *pp)
  * if we find code or data bases 
  */
 static void 
-RestoreEntries(PropEntry *pp)
+RestoreEntries(PropEntry *pp, int int_key)
 {
   while (!EndOfPAEntr(pp)) {
     switch(pp->KindOfPE) {
@@ -1092,9 +1319,10 @@ RestoreEntries(PropEntry *pp)
 	  AtomAdjust(fe->NameOfFE);
 	p0 = fe->PropsOfFE =
 	  PropAdjust(fe->PropsOfFE);
-	while (!EndOfPAEntr(p0)) {
+	if (!EndOfPAEntr(p0)) {
+	  /* at most one property */
 	  CleanCode(RepPredProp(p0));
-	  p0 = RepPredProp(p0)->NextOfPE =
+	  RepPredProp(p0)->NextOfPE =
 	    PropAdjust(RepPredProp(p0)->NextOfPE);
 	}
       }
@@ -1137,8 +1365,6 @@ RestoreEntries(PropEntry *pp)
 	      ae->ValueOfVE = AbsAppl(PtoHeapCellAdjust(ptr));
 	    } else if (IsOldLocalInTRPtr(ptr)) {
 	      ae->ValueOfVE = AbsAppl(PtoLocAdjust(ptr));
-	    } else if (IsOldDelayPtr(ptr)) {
-	      ae->ValueOfVE = AbsAppl(PtoDelayAdjust(ptr));
 	    } else if (IsOldTrailPtr(ptr)) {
 	      ae->ValueOfVE = AbsAppl(CellPtoTRAdjust(ptr));
 	    }
@@ -1163,8 +1389,7 @@ RestoreEntries(PropEntry *pp)
 	DBEntry *de = (DBEntry *) pp;
 	de->NextOfPE =
 	  PropAdjust(de->NextOfPE);
-	if (HDiff)
-	  RestoreDB(de);
+	RestoreDB(de);
       }
       break;
     case BBProperty:
@@ -1172,8 +1397,41 @@ RestoreEntries(PropEntry *pp)
 	BlackBoardEntry *bb = (BlackBoardEntry *) pp;
 	bb->NextOfPE =
 	  PropAdjust(bb->NextOfPE);
-	if (HDiff)
-	  RestoreBB(bb);
+	RestoreBB(bb, int_key);
+      }
+      break;
+    case GlobalProperty:
+      {
+	GlobalEntry *gb = (GlobalEntry *) pp;
+	Term gbt = gb->global;
+
+	gb->NextOfPE =
+	  PropAdjust(gb->NextOfPE);
+	gb->AtomOfGE =
+	  AtomEntryAdjust(gb->AtomOfGE);
+	if (gb->NextGE) {
+	  gb->NextGE =
+	    GlobalEntryAdjust(gb->NextGE);
+	}
+	if (IsVarTerm(gbt)) {
+	  CELL *gbp = VarOfTerm(gbt);
+	  if (IsOldGlobalPtr(gbp))
+	    gbp = PtoGloAdjust(gbp);
+	  else
+	    gbp = CellPtoHeapAdjust(gbp);
+	  gb->global = (CELL)gbp;
+	} else if (IsPairTerm(gbt)) {
+	  gb->global = AbsPair(PtoGloAdjust(RepPair(gbt)));
+	} else if (IsApplTerm(gbt)) {
+	  CELL *gbp = RepAppl(gbt);
+	  if (IsOldGlobalPtr(gbp))
+	    gbp = PtoGloAdjust(gbp);
+	  else
+	    gbp = CellPtoHeapAdjust(gbp);
+	  gb->global = AbsAppl(gbp);
+	} else if (IsAtomTerm(gbt)) {
+	  gb->global = AtomTermAdjust(gbt);
+	} /* numbers need no adjusting */
       }
       break;
     case OpProperty:
@@ -1183,8 +1441,25 @@ RestoreEntries(PropEntry *pp)
 	 opp->OpModule = AtomTermAdjust(opp->OpModule);
 	}
       }
-    case ExpProperty:
     case ModProperty:
+      {
+	ModEntry *me = (ModEntry *)pp;
+	if (me->NextOfPE) {
+	  me->NextOfPE =
+	    PropAdjust(me->NextOfPE);
+	}
+	if (me->PredForME) {
+	  me->PredForME =
+	  PtoPredAdjust(me->PredForME);
+	}
+	me->AtomOfME =
+	  AtomAdjust(me->AtomOfME);
+	if (me->NextME)
+	  me->NextME = (struct mod_entry *)
+	    AddrAdjust((ADDR)me->NextME);
+      }
+      break;      
+    case ExpProperty:
       pp->NextOfPE =
 	PropAdjust(pp->NextOfPE);
       break;
@@ -1198,3 +1473,21 @@ RestoreEntries(PropEntry *pp)
   }
 }
 
+static void
+RestoreAtom(AtomEntry *at)
+{
+  AtomEntry *nat;
+
+#ifdef DEBUG_RESTORE2			/* useful during debug */
+  if (IsWideAtom(AbsAtom(at)))
+    fprintf(errout, "Restoring %S\n", at->WStrOfAE);
+  else
+    fprintf(errout, "Restoring %s\n", at->StrOfAE);
+#endif
+  at->PropsOfAE = PropAdjust(at->PropsOfAE);
+  RestoreEntries(RepProp(at->PropsOfAE), FALSE);
+  /* cannot use AtomAdjust without breaking agc */
+  nat = RepAtom(at->NextOfAE);
+  if (nat)
+    at->NextOfAE = AbsAtom(AtomEntryAdjust(nat));
+}
