@@ -15,7 +15,7 @@
 *									 *
 *************************************************************************/
 
-#if defined MYDDAS_MYSQL && defined CUT_C
+#if defined MYDDAS_MYSQL
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -24,11 +24,12 @@
 #include "Yap.h"
 #include "Yatom.h"
 #include "cut_c.h"
-#include "myddas_util.h"
+#include "myddas.h"
 #ifdef MYDDAS_STATS
 #include "myddas_structs.h"
 #include "myddas_statistics.h"
 #endif
+#include "myddas_wkb2prolog.h"
 
 #define IS_SQL_INT(FIELD) FIELD == FIELD_TYPE_INT24    || \
 	                  FIELD == FIELD_TYPE_LONG     || \
@@ -40,6 +41,7 @@
 	                    FIELD == FIELD_TYPE_DOUBLE     || \
 	                    FIELD == FIELD_TYPE_FLOAT 
 
+#define IS_SQL_GEOMETRY(FIELD) FIELD == FIELD_TYPE_GEOMETRY
 
 static Int null_id = 0;
 
@@ -52,7 +54,6 @@ STATIC_PROTO(Int c_db_my_table_write,(void));
 STATIC_PROTO(Int c_db_my_row,(void));
 STATIC_PROTO(Int c_db_my_row_cut,(void));
 STATIC_PROTO(Int c_db_my_get_fields_properties,(void));
-STATIC_PROTO(Int c_db_my_number_of_fields_in_query,(void));
 STATIC_PROTO(Int c_db_my_get_next_result_set,(void));
 STATIC_PROTO(Int c_db_my_get_database,(void));
 STATIC_PROTO(Int c_db_my_change_database,(void));
@@ -65,14 +66,11 @@ void Yap_InitMYDDAS_MySQLPreds(void)
   /* db_number_of_fields: Relation x Connection x NumberOfFields */
   Yap_InitCPred("c_db_my_number_of_fields",3, c_db_my_number_of_fields, 0);  
   
-  /* db_number_of_fields_in_query: SQLQuery x Connection x NumberOfFields */
-  Yap_InitCPred("c_db_my_number_of_fields_in_query",3, c_db_my_number_of_fields_in_query, 0);  
-  
   /* db_get_attributes_types: Relation x TypesList */
   Yap_InitCPred("c_db_my_get_attributes_types", 3, c_db_my_get_attributes_types,  0);  
   
   /* db_query: SQLQuery x ResultSet x Connection */
-  Yap_InitCPred("c_db_my_query", 4, c_db_my_query, 0);  
+  Yap_InitCPred("c_db_my_query", 5, c_db_my_query, 0);  
   
   /* db_disconnect: Connection */
   Yap_InitCPred("c_db_my_disconnect", 1,c_db_my_disconnect, 0);
@@ -133,12 +131,16 @@ c_db_my_connect(void) {
 
   conn = mysql_init(NULL);
   if (conn == NULL) {
-    printf("erro no init\n");
+#ifdef DEBUG
+    printf("ERROR: ** c_db_my_connect ** error on init\n");
+#endif
     return FALSE;
   }
   
   if (mysql_real_connect(conn, host, user, passwd, database, port, socket, CLIENT_MULTI_STATEMENTS) == NULL) {
-    printf("erro no connect\n");
+#ifdef DEBUG
+    printf("ERROR: ** c_db_my_connect ** error on connect\n");
+#endif
     return FALSE;
   }
 
@@ -147,10 +149,12 @@ c_db_my_connect(void) {
   else
     {
       /* Criar um novo no na lista de ligacoes*/
-      //new = add_connection(&TOP,conn,NULL);
       new = myddas_util_add_connection(conn,NULL);
+      
       if (new == NULL){
-	printf("Erro ao alocar memoria para lista\n");
+#ifdef DEBUG
+	printf("ERROR: ** c_db_my_connect ** Error allocating memory\n");
+#endif
 	return FALSE;
       }
       return TRUE;
@@ -164,6 +168,7 @@ c_db_my_query(void) {
   Term arg_result_set = Deref(ARG2);
   Term arg_conn = Deref(ARG3);
   Term arg_mode = Deref(ARG4);
+  Term arg_arity = Deref(ARG5);
   
   char *sql = AtomName(AtomOfTerm(arg_sql_query));
   char *mode = AtomName(AtomOfTerm(arg_mode));
@@ -171,27 +176,30 @@ c_db_my_query(void) {
   
   MYSQL_RES *res_set;
   
-  Int length=strlen(sql);
+  MyddasInt length=strlen(sql);
 
 #ifdef MYDDAS_STATS 
   MYDDAS_UTIL_CONNECTION node = myddas_util_search_connection(conn);
-
+  MyddasULInt count = 0;
+  
   /* Count the number of querys made to the server */
-  UInt number_querys;
+  MyddasULInt number_querys;
   MYDDAS_STATS_CON_GET_NUMBER_QUERIES_MADE(node,number_querys);
   MYDDAS_STATS_CON_SET_NUMBER_QUERIES_MADE(node,++number_querys);
+  MYDDAS_STATS_CON_GET_NUMBER_QUERIES_MADE_COUNT(node,count);
+  MYDDAS_STATS_CON_SET_NUMBER_QUERIES_MADE_COUNT(node,++count);
 
-  /* Measure time spent by the MySQL Server 
-     processing the SQL Query */ 
+  /* Measure time spent by the MySQL Server
+     processing the SQL Query */
   MYDDAS_STATS_TIME start,end,total_time,diff;
   start = myddas_stats_walltime();
 #endif 
   
-  /* executar a query SQL */
+  /* Send query to server and process it */
   if (mysql_real_query(conn, sql, length) != 0)
     {
 #ifdef DEBUG
-      printf("Erro na query! %s\n",sql);
+      printf("ERROR: **c_db_my_query** Error on query! %s\n",sql);
 #endif 
       return FALSE;
     }
@@ -205,15 +213,20 @@ c_db_my_query(void) {
   myddas_stats_subtract_time(diff,end,start);
   diff = myddas_stats_time_copy_to_final(diff);
   
-  free(end);
-  free(start);
+  MYDDAS_FREE(end,struct myddas_stats_time_struct);
+  MYDDAS_FREE(start,struct myddas_stats_time_struct);
   
   MYDDAS_STATS_CON_GET_TOTAL_TIME_DBSERVER(node,total_time);
-  
-  /* Automacally updates the MYDDAS_STRUCTURE */
+    /* Automacally updates the MYDDAS_STRUCTURE */
   myddas_stats_add_time(total_time,diff,total_time);
+  MYDDAS_STATS_CON_GET_TOTAL_TIME_DBSERVER_COUNT(node,count);
+  MYDDAS_STATS_CON_SET_TOTAL_TIME_DBSERVER_COUNT(node,++count);
   
-  myddas_stats_move_time(diff,node->lastTimeofDBServer);
+  MYDDAS_STATS_TIME time = NULL;
+  MYDDAS_STATS_CON_GET_LAST_TIME_DBSERVER(node,time);
+  myddas_stats_move_time(diff,time);
+  MYDDAS_STATS_CON_GET_LAST_TIME_DBSERVER_COUNT(node,count);
+  MYDDAS_STATS_CON_SET_LAST_TIME_DBSERVER_COUNT(node,++count);
 #endif 
   
   /* guardar os tuplos do lado do cliente */
@@ -238,16 +251,21 @@ c_db_my_query(void) {
     myddas_stats_subtract_time(diff,end,start);
     diff = myddas_stats_time_copy_to_final(diff);
     
-    free(end);
-    free(start);
+    MYDDAS_FREE(end,struct myddas_stats_time_struct);
+    MYDDAS_FREE(start,struct myddas_stats_time_struct);
     
     MYDDAS_STATS_CON_GET_TOTAL_TIME_TRANSFERING(node,total_time);
-    
     /* Automacally updates the MYDDAS_STRUCTURE */
     myddas_stats_add_time(total_time,diff,total_time);
+    MYDDAS_STATS_CON_GET_TOTAL_TIME_TRANSFERING_COUNT(node,count);
+    MYDDAS_STATS_CON_SET_TOTAL_TIME_TRANSFERING_COUNT(node,++count);
         
-    myddas_stats_move_time(diff,node->lastFromDBServer);
-    
+    time = NULL;
+    MYDDAS_STATS_CON_GET_LAST_TIME_TRANSFERING(node,time);
+    MYDDAS_STATS_CON_GET_LAST_TIME_TRANSFERING_COUNT(node,count);
+    MYDDAS_STATS_CON_SET_LAST_TIME_TRANSFERING_COUNT(node,++count);
+    myddas_stats_move_time(diff,time);
+   
     /* Measure the number of Rows returned from the server */
     if (res_set != NULL)
       {
@@ -255,17 +273,19 @@ c_db_my_query(void) {
 	   returns a NULL pointer*/
 	
 	/* This is only works if we use mysql_store_result */
-	UInt numberRows = mysql_num_rows(res_set);
-	UInt rows;
+	MyddasUInt numberRows = mysql_num_rows(res_set);
+	MyddasUInt rows;
 	
 	MYDDAS_STATS_CON_GET_TOTAL_ROWS(node,rows);
 	numberRows = numberRows + rows;
 	MYDDAS_STATS_CON_SET_TOTAL_ROWS(node,numberRows);
-      
+	MYDDAS_STATS_CON_GET_TOTAL_ROWS_COUNT(node,count);
+	MYDDAS_STATS_CON_SET_TOTAL_ROWS_COUNT(node,++count);
+
 	/* Calculate the ammount of data sent by the server */
-	UInt total,number_fields = mysql_num_fields(res_set);
+	MyddasUInt total,number_fields = mysql_num_fields(res_set);
 	MYSQL_ROW row;
-	UInt i;
+	MyddasULInt i;
 	total=0;
 	while ((row = mysql_fetch_row(res_set)) != NULL){
 	  mysql_field_seek(res_set,0);
@@ -276,11 +296,15 @@ c_db_my_query(void) {
 	  }
 	}
 	MYDDAS_STATS_CON_SET_LAST_BYTES_TRANSFERING_FROM_DBSERVER(node,total);
-	UInt bytes;
+	MYDDAS_STATS_CON_GET_LAST_BYTES_TRANSFERING_FROM_DBSERVER_COUNT(node,count);
+	MYDDAS_STATS_CON_SET_LAST_BYTES_TRANSFERING_FROM_DBSERVER_COUNT(node,++count);
 	
+	MyddasUInt bytes = 0;
 	MYDDAS_STATS_CON_GET_TOTAL_BYTES_TRANSFERING_FROM_DBSERVER(node,bytes);
 	total = total + bytes;
 	MYDDAS_STATS_CON_SET_TOTAL_BYTES_TRANSFERING_FROM_DBSERVER(node,total);
+	MYDDAS_STATS_CON_GET_TOTAL_BYTES_TRANSFERING_FROM_DBSERVER_COUNT(node,count);
+	MYDDAS_STATS_CON_SET_TOTAL_BYTES_TRANSFERING_FROM_DBSERVER_COUNT(node,++count);
 	mysql_data_seek(res_set,0);
       }
 #endif 
@@ -291,10 +315,16 @@ c_db_my_query(void) {
       //INSERT statements don't return any res_set
       if (mysql_field_count(conn) == 0)
 	return TRUE;
-      printf("Query vazia!\n");
+#ifdef DEBUG
+      printf("Empty Query!\n");
+#endif
       return FALSE;
     }
   
+  if (!Yap_unify(arg_arity, MkIntegerTerm(mysql_num_fields(res_set)))){
+    return FALSE;
+  }
+  
   if (!Yap_unify(arg_result_set, MkIntegerTerm((Int) res_set)))
     {
       mysql_free_result(res_set);
@@ -325,7 +355,7 @@ c_db_my_number_of_fields(void) {
   if (mysql_query(conn, sql) != 0)
     {
 #ifdef DEBUG
-      printf("Erro na query! %s\n",sql);
+      printf("ERROR: **c_db_my_number_of_fields** Error on the query! %s\n",sql);
 #endif
       return FALSE;
     }
@@ -334,7 +364,7 @@ c_db_my_number_of_fields(void) {
   if ((res_set = mysql_store_result(conn)) == NULL)
     {
 #ifdef DEBUG
-      printf("Erro na query! %s\n",sql);
+      printf("ERROR: **c_db_my_number_of_fields** Error storing the query! %s\n",sql);
 #endif
       
       return FALSE;
@@ -366,8 +396,10 @@ c_db_my_get_attributes_types(void) {
 
   sprintf(sql,"DESCRIBE `%s`",relation);
 
+  Int length = strlen(sql);
+
   /* executar a query SQL */
-  if (mysql_query(conn, sql) != 0)
+  if (mysql_real_query(conn, sql, length) != 0)
     {
 #ifdef DEBUG
       printf("Erro na query! %s\n",sql);
@@ -377,7 +409,9 @@ c_db_my_get_attributes_types(void) {
   /* guardar os tuplos do lado do cliente */
   if ((res_set = mysql_store_result(conn)) == NULL)
   {
+#ifdef DEBUG
       printf("Query vazia!\n");
+#endif
       return FALSE;
   }
 
@@ -451,9 +485,10 @@ c_db_my_row_cut(void) {
 static Int
 c_db_my_row(void) {
 #ifdef MYDDAS_STATS
-  /* Measure time used by the 
-     c_db_my_row function */
+/*   Measure time used by the  */
+/*      c_db_my_row function */
   MYDDAS_STATS_TIME start,end,total_time,diff;
+  MyddasULInt count = 0;
   start = myddas_stats_walltime();
 #endif 
   Term arg_result_set = Deref(ARG1);
@@ -470,7 +505,7 @@ c_db_my_row(void) {
   Int i, arity;
   
   arity = IntegerOfTerm(arg_arity);
-
+  
   while(TRUE)
     {
       if ((row = mysql_fetch_row(res_set)) != NULL)
@@ -504,6 +539,11 @@ c_db_my_row(void) {
 		      if (!Yap_unify(head, MkFloatTerm(atof(row[i]))))
 			continue;
 		    }
+		  else if (IS_SQL_GEOMETRY(field->type))
+		    {
+		      if (!Yap_unify(head, wkb2prolog(row[i])))
+			continue;
+		    }
 		  else 
 		    {
 		      if (!Yap_unify(head, MkAtomTerm(Yap_LookupAtom(row[i]))))
@@ -518,20 +558,21 @@ c_db_my_row(void) {
 	  myddas_stats_subtract_time(diff,end,start);
 	  diff = myddas_stats_time_copy_to_final(diff);
 	  
-	  free(end);
-	  free(start);
+	  MYDDAS_FREE(end,struct myddas_stats_time_struct);
+	  MYDDAS_FREE(start,struct myddas_stats_time_struct);
 	  
 	  MYDDAS_STATS_GET_DB_ROW_FUNCTION(total_time);
 	  myddas_stats_add_time(total_time,diff,total_time);
-	  
-	  free(diff);
+	  MYDDAS_STATS_GET_DB_ROW_FUNCTION_COUNT(count);
+	  MYDDAS_STATS_SET_DB_ROW_FUNCTION_COUNT(++count);
+	  	  
+	  MYDDAS_FREE(diff,struct myddas_stats_time_struct);
 #endif /* MYDDAS_STATS */
 	  return TRUE;
 	}
       else 
 	{
 	  mysql_free_result(res_set);
-	  cut_fail();
 #ifdef MYDDAS_STATS
 	  end = myddas_stats_walltime();
 	  
@@ -539,56 +580,21 @@ c_db_my_row(void) {
 	  myddas_stats_subtract_time(diff,end,start);
 	  diff = myddas_stats_time_copy_to_final(diff);
 	  
-	  free(end);
-	  free(start);
+	  MYDDAS_FREE(end,struct myddas_stats_time_struct);
+	  MYDDAS_FREE(start,struct myddas_stats_time_struct);
 	  
 	  MYDDAS_STATS_GET_DB_ROW_FUNCTION(total_time);
 	  myddas_stats_add_time(total_time,diff,total_time);
+	  MYDDAS_STATS_GET_DB_ROW_FUNCTION_COUNT(count);
+	  MYDDAS_STATS_SET_DB_ROW_FUNCTION_COUNT(++count);
 	  
-	  free(diff);
+	  MYDDAS_FREE(diff,struct myddas_stats_time_struct);
 #endif /* MYDDAS_STATS */
-	  return FALSE;
+	  cut_fail();  /* This macro already does a return FALSE */
 	}
-    }
-}
-
-
-/* Mudar esta funcao de forma a nao fazer a consulta, pois 
- no predicate db_sql_selet vai fazer duas vezes a mesma consutla*/ 
-static Int 
-c_db_my_number_of_fields_in_query(void) {
-  Term arg_query = Deref(ARG1);
-  Term arg_conn = Deref(ARG2);
-  Term arg_fields = Deref(ARG3);
-
-  char *query = AtomName(AtomOfTerm(arg_query));
-  MYSQL *conn = (MYSQL *) (IntegerOfTerm(arg_conn));
-
-  MYSQL_RES *res_set;
-
-  /* executar a query SQL */
-  if (mysql_query(conn, query) != 0)
-    {
-#ifdef DEBUG
-      printf("Erro na query! %s\n",query);
-#endif
-      return FALSE;
+      
     }
   
-  /* guardar os tuplos do lado do cliente */
-  if ((res_set = mysql_store_result(conn)) == NULL)
-    {
-#ifdef DEBUG
-      printf("Erro na query! %s\n",query);
-#endif
-      return FALSE;
-    }
-  
-  if (!Yap_unify(arg_fields, MkIntegerTerm(mysql_num_fields(res_set)))){
-    return FALSE;
-  }
-  mysql_free_result(res_set);
-  return TRUE;  
 }
 
 static Int 
@@ -611,8 +617,10 @@ c_db_my_get_fields_properties(void) {
      query*/
   sprintf (sql,"SELECT * FROM `%s` LIMIT 0",relacao);
 
+  Int length=strlen(sql);
+
   /* executar a query SQL */
-  if (mysql_query(conn, sql) != 0)
+  if (mysql_real_query(conn, sql, length) != 0)
     {
 #ifdef DEBUG
       printf("Erro na query! %s\n",sql);
@@ -714,4 +722,4 @@ c_db_my_change_database(void) {
   return TRUE;
 }
 
-#endif /*MYDDAS_MYSQL && CUT_C*/
+#endif /* MYDDAS_MYSQL */
