@@ -22,7 +22,7 @@
 #include <string.h>
 #include "Yap.h"
 #include "Yatom.h"
-#include "myddas_util.h"
+#include "myddas.h"
 #include "cut_c.h"
 #include <sql.h>
 #include <sqlucode.h>
@@ -302,8 +302,8 @@ c_db_odbc_query(void) {
   Term arg_bind_list = Deref(ARG4);
   Term arg_conn = Deref(ARG5);
 
-  char *sql = AtomName(AtomOfTerm(arg_sql_query));
-
+  SQLCHAR *sql = AtomName(AtomOfTerm(arg_sql_query));
+  
   
   SQLHDBC  hdbc =(SQLHDBC) (IntegerOfTerm(arg_conn));
   SQLHSTMT hstmt;
