@@ -5,7 +5,7 @@
                                                                
   Copyright:   R. Rocha and NCC - University of Porto, Portugal
   File:        opt.preds.c
-  version:     $Id: opt.preds.c,v 1.25 2005/08/10 21:36:34 ricroc Exp $   
+  version:     $Id: opt.preds.c,v 1.29 2008/04/11 16:26:19 ricroc Exp $   
                                                                      
 **********************************************************************/
 
@@ -48,28 +48,29 @@ static qg_ans_fr_ptr actual_answer;
 
 #ifdef YAPOR
 static realtime current_time(void);
-static int p_yapor_on(void);
-static int p_start_yapor(void);
-static int p_sequential(void);
-static int p_default_sequential(void);
-static int p_execution_mode(void);
-static int p_performance(void);
-static int p_parallel_new_answer(void);
-static int p_parallel_yes_answer(void);
+static Int p_yapor_on(void);
+static Int p_start_yapor(void);
+static Int p_sequential(void);
+static Int p_default_sequential(void);
+static Int p_execution_mode(void);
+static Int p_performance(void);
+static Int p_parallel_new_answer(void);
+static Int p_parallel_yes_answer(void);
 static int parallel_new_answer_putchar(int sno, int ch);
 static void show_answers(void);
 static void answer_to_stdout(char *answer);
-static int p_or_statistics(void);
+static Int p_or_statistics(void);
 #endif /* YAPOR */
 
 #ifdef TABLING
-static int p_table(void);
-static int p_tabling_mode(void);
-static int p_abolish_table(void);
-static int p_abolish_all_tables(void);
-static int p_show_table(void);
-static int p_table_statistics(void);
-static int p_tabling_statistics(void);
+static Int p_table(void);
+static Int p_tabling_mode(void);
+static Int p_abolish_table(void);
+static Int p_abolish_all_tables(void);
+static Int p_show_tabled_predicates(void);
+static Int p_show_table(void);
+static Int p_table_statistics(void);
+static Int p_tabling_statistics(void);
 #endif /* TABLING */
 
 #if defined(YAPOR) && defined(TABLING)
@@ -128,6 +129,7 @@ void Yap_init_optyap_preds(void) {
   Yap_InitCPred("$c_tabling_mode", 3, p_tabling_mode, SafePredFlag|SyncPredFlag|HiddenPredFlag);
   Yap_InitCPred("$c_abolish_table", 2, p_abolish_table, SafePredFlag|SyncPredFlag|HiddenPredFlag);
   Yap_InitCPred("abolish_all_tables", 0, p_abolish_all_tables, SafePredFlag|SyncPredFlag);
+  Yap_InitCPred("show_tabled_predicates", 0, p_show_tabled_predicates, SafePredFlag|SyncPredFlag);
   Yap_InitCPred("$c_show_table", 2, p_show_table, SafePredFlag|SyncPredFlag|HiddenPredFlag);
   Yap_InitCPred("$c_table_statistics", 2, p_table_statistics, SafePredFlag|SyncPredFlag|HiddenPredFlag);
   Yap_InitCPred("tabling_statistics", 0, p_tabling_statistics, SafePredFlag|SyncPredFlag);
@@ -171,13 +173,13 @@ realtime current_time(void) {
 
 
 static
-int p_yapor_on(void) {
+Int p_yapor_on(void) {
   return (PARALLEL_EXECUTION_MODE);
 }
 
 
 static
-int p_start_yapor(void) {
+Int p_start_yapor(void) {
 #ifdef TIMESTAMP_CHECK
   GLOBAL_timestamp = 0;
 #endif /* TIMESTAMP_CHECK */
@@ -189,6 +191,7 @@ int p_start_yapor(void) {
   BITMAP_clear(GLOBAL_bm_pruning_workers);
 #endif /* TABLING_INNER_CUTS */
   make_root_choice_point();
+  GLOBAL_performance_mode &= ~PERFORMANCE_IN_EXECUTION;
   GLOBAL_execution_time = current_time();
   BITMAP_clear(GLOBAL_bm_finished_workers);
   PUT_IN_EXECUTING(worker_id);
@@ -197,7 +200,7 @@ int p_start_yapor(void) {
 
 
 static
-int p_sequential(void) {
+Int p_sequential(void) {
   Term t, mod;
   PredEntry *pe;
 
@@ -221,7 +224,7 @@ int p_sequential(void) {
 
 
 static
-int p_default_sequential(void) {
+Int p_default_sequential(void) {
   Term t;
   t = Deref(ARG1);
   if (IsVarTerm(t)) {
@@ -250,7 +253,7 @@ int p_default_sequential(void) {
 
 
 static
-int p_execution_mode(void) {
+Int p_execution_mode(void) {
   Term t;
   t = Deref(ARG1);
   if (IsVarTerm(t)) {
@@ -279,7 +282,7 @@ int p_execution_mode(void) {
 
 
 static
-int p_performance(void) {
+Int p_performance(void) {
   Term t;
   realtime one_worker_execution_time = 0;
   int i;
@@ -354,7 +357,7 @@ int p_performance(void) {
 
 
 static
-int p_parallel_new_answer(void) {
+Int p_parallel_new_answer(void) {
   or_fr_ptr leftmost_or_fr;
 
   length_answer = 0;
@@ -376,7 +379,7 @@ int p_parallel_new_answer(void) {
 
 
 static
-int p_parallel_yes_answer(void) {
+Int p_parallel_yes_answer(void) {
   GLOBAL_answers = YES_ANSWER;
   return (TRUE);
 }
@@ -421,9 +424,7 @@ void show_answers(void) {
   }
   fprintf(Yap_stderr, " (in %f seconds) ]\n\n", GLOBAL_execution_time);
 
-  if (GLOBAL_performance_mode & PERFORMANCE_IN_EXECUTION) {
-    GLOBAL_performance_mode &= ~PERFORMANCE_IN_EXECUTION;
-  } else if (GLOBAL_performance_mode == PERFORMANCE_ON) {
+  if (GLOBAL_performance_mode == PERFORMANCE_ON) {
     for (i = GLOBAL_number_goals; i > 0; i--) {
       if (GLOBAL_best_times(i) > GLOBAL_execution_time) {
         if (i + 1 < MAX_BEST_TIMES)
@@ -488,7 +489,7 @@ void answer_to_stdout(char *answer) {
 
 
 static
-int p_or_statistics(void) {
+Int p_or_statistics(void) {
 #ifdef SHM_MEMORY_ALLOC_SCHEME
   long pages_in_use = 0, bytes_in_use = 0;
 
@@ -521,7 +522,7 @@ int p_or_statistics(void) {
 
 #ifdef TABLING
 static
-int p_table(void) {
+Int p_table(void) {
   Term mod, t;
   PredEntry *pe;
   tab_ent_ptr tab_ent;
@@ -557,7 +558,7 @@ int p_table(void) {
 
 
 static
-int p_tabling_mode(void) {
+Int p_tabling_mode(void) {
   Term mod, t, val;
   tab_ent_ptr tab_ent;
 
@@ -633,7 +634,7 @@ int p_tabling_mode(void) {
 
 
 static
-int p_abolish_table(void) {
+Int p_abolish_table(void) {
   Term mod, t;
   tab_ent_ptr tab_ent;
   sg_hash_ptr hash;
@@ -660,7 +661,7 @@ int p_abolish_table(void) {
 
 
 static
-int p_abolish_all_tables(void) {
+Int p_abolish_all_tables(void) {
   tab_ent_ptr tab_ent;
   sg_hash_ptr hash;
   sg_node_ptr sg_node;
@@ -682,7 +683,37 @@ int p_abolish_all_tables(void) {
 
 
 static
-int p_show_table(void) {
+Int p_show_tabled_predicates(void) {
+  tab_ent_ptr tab_ent;
+  PredEntry *pred;
+  char *name;
+  Int arity;
+
+  fprintf(Yap_stderr, "Tabled predicates\n");
+  tab_ent = GLOBAL_root_tab_ent;
+  if (tab_ent == NULL)
+    fprintf(Yap_stderr, "  none\n");
+  else 
+    while(tab_ent) {
+      pred = TabEnt_pe(tab_ent);
+      arity = pred->ArityOfPE;
+      if (arity == 0)
+	name = RepAtom((Atom)pred->FunctorOfPred)->StrOfAE;
+      else
+	name = RepAtom(NameOfFunctor((pred->FunctorOfPred)))->StrOfAE;
+#if SHORT_INTS
+      fprintf(Yap_stderr, "  %s/%ld\n", name, arity);
+#else
+      fprintf(Yap_stderr, "  %s/%d\n", name, arity);
+#endif /* SHORT_INTS */
+      tab_ent = TabEnt_next(tab_ent);
+    }
+  return (TRUE);
+}
+
+
+static
+Int p_show_table(void) {
   Term mod, t;
   tab_ent_ptr tab_ent;
   Atom at;
@@ -704,7 +735,7 @@ int p_show_table(void) {
 
 
 static
-int p_table_statistics(void) {
+Int p_table_statistics(void) {
   Term mod, t;
   tab_ent_ptr tab_ent;
   Atom at;
@@ -727,7 +758,7 @@ int p_table_statistics(void) {
 
 
 static
-int p_tabling_statistics(void) {
+Int p_tabling_statistics(void) {
 #ifdef SHM_MEMORY_ALLOC_SCHEME
   long pages_in_use = 0, bytes_in_use = 0;
 
