@@ -5,7 +5,7 @@
                                                                
   Copyright:   R. Rocha and NCC - University of Porto, Portugal
   File:        tab.macros.h
-  version:     $Id: tab.macros.h,v 1.19 2005/08/05 14:55:03 vsc Exp $   
+  version:     $Id: tab.macros.h,v 1.22 2008/05/23 18:28:58 ricroc Exp $   
                                                                      
 **********************************************************************/
 
@@ -65,11 +65,9 @@ STD_PROTO(static inline tg_sol_fr_ptr CUT_prune_tg_solution_frames, (tg_sol_fr_p
 #define LOAD_CP(CP)            ((struct loader_choicept *)(CP))
 #define IS_BATCHED_GEN_CP(CP)  (GEN_CP(CP)->cp_dep_fr == NULL)
 
-
-#define TAG_AS_ANSWER_LEAF_NODE(NODE)  TrNode_parent(NODE) = (ans_node_ptr)((unsigned int)TrNode_parent(NODE) | 0x1)
-#define UNTAG_ANSWER_LEAF_NODE(NODE)   ((ans_node_ptr)((unsigned int)NODE & 0xfffffffe))
-#define IS_ANSWER_LEAF_NODE(NODE)      ((unsigned int)TrNode_parent(NODE) & 0x1)
-
+#define TAG_AS_ANSWER_LEAF_NODE(NODE)  TrNode_parent(NODE) = (ans_node_ptr)((unsigned long int)TrNode_parent(NODE) | 0x1)
+#define UNTAG_ANSWER_LEAF_NODE(NODE)   ((ans_node_ptr)((unsigned long int)NODE & ~(0x1)))
+#define IS_ANSWER_LEAF_NODE(NODE)      ((unsigned long int)TrNode_parent(NODE) & 0x1)
 
 #define STACK_NOT_EMPTY(STACK, STACK_BASE)  STACK != STACK_BASE
 #define STACK_PUSH_UP(ITEM, STACK)          *--STACK = (CELL)(ITEM)
@@ -106,8 +104,7 @@ STD_PROTO(static inline tg_sol_fr_ptr CUT_prune_tg_solution_frames, (tg_sol_fr_p
         ((CELL) TERM) >= GLOBAL_table_var_enumerator(0) &&                 \
         ((CELL) TERM) <= GLOBAL_table_var_enumerator(MAX_TABLE_VARS - 1)
 
-
-#define HASH_TABLE_LOCK(NODE)  ((((unsigned int)NODE) >> 5) & (TABLE_LOCK_BUCKETS - 1))
+#define HASH_TABLE_LOCK(NODE)  ((((unsigned long int)NODE) >> 5) & (TABLE_LOCK_BUCKETS - 1))
 #define LOCK_TABLE(NODE)         LOCK(GLOBAL_table_lock(HASH_TABLE_LOCK(NODE)))
 #define UNLOCK_TABLE(NODE)     UNLOCK(GLOBAL_table_lock(HASH_TABLE_LOCK(NODE)))
 
@@ -245,7 +242,7 @@ STD_PROTO(static inline tg_sol_fr_ptr CUT_prune_tg_solution_frames, (tg_sol_fr_p
         { SgFr_init_yapor_fields(SG_FR);                           \
           SgFr_state(SG_FR) = evaluating;                          \
           SgFr_next(SG_FR) = LOCAL_top_sg_fr;                      \
-          LOCAL_top_sg_fr = sg_fr;                                 \
+          LOCAL_top_sg_fr = SG_FR;                                 \
 	}
 
 
@@ -257,8 +254,9 @@ STD_PROTO(static inline tg_sol_fr_ptr CUT_prune_tg_solution_frames, (tg_sol_fr_p
         DepFr_leader_cp(DEP_FR) = NORM_CP(LEADER_CP);                                                  \
         DepFr_cons_cp(DEP_FR) = NORM_CP(CONS_CP);                                                      \
         /* start with TrNode_child(DepFr_last_answer(DEP_FR)) pointing to SgFr_first_answer(SG_FR) */  \
-        DepFr_last_answer(DEP_FR) = (ans_node_ptr)((int)(SG_FR) +                                      \
-          (int)(&SgFr_first_answer((sg_fr_ptr)DEP_FR)) - (int)(&TrNode_child((ans_node_ptr)DEP_FR)));  \
+        DepFr_last_answer(DEP_FR) = (ans_node_ptr)((unsigned long int)(SG_FR) +                        \
+                                    (unsigned long int)(&SgFr_first_answer((sg_fr_ptr)DEP_FR)) -       \
+                                    (unsigned long int)(&TrNode_child((ans_node_ptr)DEP_FR)));         \
         DepFr_next(DEP_FR) = NEXT
 
 
@@ -549,29 +547,28 @@ void abolish_incomplete_subgoals(choiceptr prune_cp) {
       /* no answers --> ready */
       SgFr_state(sg_fr) = ready;
       UNLOCK(SgFr_lock(sg_fr));
+    } else if (SgFr_first_answer(sg_fr) == SgFr_answer_trie(sg_fr)) {
+      /* yes answer --> complete */
+      /* at this point the subgoal should be already completed (early completion)  */ 
+      /* SgFr_state(sg_fr) = complete; */
+      UNLOCK(SgFr_lock(sg_fr));
     } else {
-      if (SgFr_first_answer(sg_fr) == SgFr_answer_trie(sg_fr)) {
-	/* yes answer --> complete */
-	SgFr_state(sg_fr) = complete;
-	UNLOCK(SgFr_lock(sg_fr));
-      } else {
-	/* answers --> incomplete/ready */
+      /* answers --> incomplete/ready */
 #ifdef INCOMPLETE_TABLING
-	SgFr_state(sg_fr) = incomplete;
-	UNLOCK(SgFr_lock(sg_fr));
+      SgFr_state(sg_fr) = incomplete;
+      UNLOCK(SgFr_lock(sg_fr));
 #else
-	ans_node_ptr node;
-	SgFr_state(sg_fr) = ready;
-	free_answer_hash_chain(SgFr_hash_chain(sg_fr));
-	SgFr_hash_chain(sg_fr) = NULL;
-	SgFr_first_answer(sg_fr) = NULL;
-	SgFr_last_answer(sg_fr) = NULL;
-	node = TrNode_child(SgFr_answer_trie(sg_fr));
-	TrNode_child(SgFr_answer_trie(sg_fr)) = NULL;
-	UNLOCK(SgFr_lock(sg_fr));
-	free_answer_trie_branch(node);
+      ans_node_ptr node;
+      SgFr_state(sg_fr) = ready;
+      free_answer_hash_chain(SgFr_hash_chain(sg_fr));
+      SgFr_hash_chain(sg_fr) = NULL;
+      SgFr_first_answer(sg_fr) = NULL;
+      SgFr_last_answer(sg_fr) = NULL;
+      node = TrNode_child(SgFr_answer_trie(sg_fr));
+      TrNode_child(SgFr_answer_trie(sg_fr)) = NULL;
+      UNLOCK(SgFr_lock(sg_fr));
+      free_answer_trie_branch(node);
 #endif /* INCOMPLETE_TABLING */
-      }
     }
 #ifdef LIMIT_TABLING
     insert_into_global_sg_fr_list(sg_fr);
