@@ -5,7 +5,7 @@
                                                                
   Copyright:   R. Rocha and NCC - University of Porto, Portugal
   File:        tab.tries.C
-  version:     $Id: tab.tries.c,v 1.19 2006/04/12 14:57:59 ricroc Exp $   
+  version:     $Id: tab.tries.c,v 1.24 2008/05/20 18:25:37 ricroc Exp $   
                                                                      
 **********************************************************************/
 
@@ -725,9 +725,9 @@ sg_fr_ptr subgoal_search(yamop *preg, CELL **Yaddr) {
 	if (f == FunctorDouble) {
 	  volatile Float dbl = FloatOfTerm(t);
 	  volatile Term *t_dbl = (Term *)((void *) &dbl);
-#if SIZEOF_DOUBLE == 2 * SIZEOF_LONG_INT
+#if SIZEOF_DOUBLE == 2 * SIZEOF_INT_P
 	  current_sg_node = subgoal_trie_node_check_insert(tab_ent, current_sg_node, *(t_dbl + 1));
-#endif /* SIZEOF_DOUBLE x SIZEOF_LONG_INT */
+#endif /* SIZEOF_DOUBLE x SIZEOF_INT_P */
 	  current_sg_node = subgoal_trie_node_check_insert(tab_ent, current_sg_node, *t_dbl);
 	} else if (f == FunctorLongInt) {
 	  Int li = LongIntOfTerm(t);
@@ -755,7 +755,7 @@ sg_fr_ptr subgoal_search(yamop *preg, CELL **Yaddr) {
   } else {
     sg_fr = (sg_fr_ptr) TrNode_sg_fr(current_sg_node);
 #ifdef LIMIT_TABLING
-    if (SgFr_state(sg_fr) == ready) {
+    if (SgFr_state(sg_fr) <= ready) {  /* incomplete or ready */
       remove_from_global_sg_fr_list(sg_fr);
     }
 #endif /* LIMIT_TABLING */
@@ -828,9 +828,9 @@ ans_node_ptr answer_search(sg_fr_ptr sg_fr, CELL *subs_ptr) {
 	  volatile Float dbl = FloatOfTerm(t);
 	  volatile Term *t_dbl = (Term *)((void *) &dbl);
 	  current_ans_node = answer_trie_node_check_insert(sg_fr, current_ans_node, AbsAppl((Term *)f), _trie_retry_null);
-#if SIZEOF_DOUBLE == 2 * SIZEOF_LONG_INT
+#if SIZEOF_DOUBLE == 2 * SIZEOF_INT_P
 	  current_ans_node = answer_trie_node_check_insert(sg_fr, current_ans_node, *(t_dbl + 1), _trie_retry_extension);
-#endif /* SIZEOF_DOUBLE x SIZEOF_LONG_INT */
+#endif /* SIZEOF_DOUBLE x SIZEOF_INT_P */
 	  current_ans_node = answer_trie_node_check_insert(sg_fr, current_ans_node, *t_dbl, _trie_retry_extension);
 	  current_ans_node = answer_trie_node_check_insert(sg_fr, current_ans_node, AbsAppl((Term *)f), _trie_retry_float);
 	} else if (f == FunctorLongInt) {
@@ -863,7 +863,7 @@ ans_node_ptr answer_search(sg_fr_ptr sg_fr, CELL *subs_ptr) {
 
 void load_answer_trie(ans_node_ptr ans_node, CELL *subs_ptr) {
   CELL *stack_vars_base, *stack_vars, *stack_terms_base, *stack_terms;
-  int subs_arity, i, n_vars = MAX_TABLE_VARS;
+  int subs_arity, i, n_vars = -1;
   Term t;
 
   if ((subs_arity = *subs_ptr) == 0)
@@ -881,14 +881,15 @@ void load_answer_trie(ans_node_ptr ans_node, CELL *subs_ptr) {
   do {
     if (IsVarTerm(t)) {
       int var_index = VarIndexOfTableTerm(t);
-      if (n_vars == MAX_TABLE_VARS) {
-	stack_vars += var_index;
-	STACK_CHECK_EXPAND(stack_terms, stack_vars, stack_terms_base);
-      }
-      if (var_index < n_vars) {
+      if(var_index > n_vars) {
+	for (i = var_index; i > n_vars; i--)
+	  stack_vars_base[i] = 0; 
 	n_vars = var_index;
-	stack_vars_base[var_index] = MkVarTerm();
-      }
+	stack_vars = stack_terms_base + n_vars;
+	STACK_CHECK_EXPAND(stack_terms, stack_vars, stack_terms_base);
+      }  
+      if (stack_vars_base[var_index] == 0)
+	stack_vars_base[var_index] = MkVarTerm(); 
       STACK_PUSH_UP(stack_vars_base[var_index], stack_terms);
       STACK_CHECK_EXPAND(stack_terms, stack_vars, stack_terms_base);
     } else if (IsAtomOrIntTerm(t)) {
@@ -907,11 +908,11 @@ void load_answer_trie(ans_node_ptr ans_node, CELL *subs_ptr) {
 	t = TrNode_entry(ans_node);
 	ans_node = TrNode_parent(ans_node);
 	*t_dbl = t;
-#if SIZEOF_DOUBLE == 2 * SIZEOF_LONG_INT
+#if SIZEOF_DOUBLE == 2 * SIZEOF_INT_P
 	t = TrNode_entry(ans_node);
 	ans_node = TrNode_parent(ans_node);
 	*(t_dbl + 1) = t;
-#endif /* SIZEOF_DOUBLE x SIZEOF_LONG_INT */
+#endif /* SIZEOF_DOUBLE x SIZEOF_INT_P */
 	ans_node = TrNode_parent(ans_node);
 	t = MkFloatTerm(dbl);
 	STACK_PUSH_UP(t, stack_terms);
@@ -1114,12 +1115,35 @@ int traverse_table(tab_ent_ptr tab_ent, Atom pred_atom, int show_table) {
   TrStat_ans_max_depth = -1;
   TrStat_ans_min_depth = -1;
   if (sg_node) {
-    char str[STR_ARRAY_SIZE];
-    int str_index = sprintf(str, "  ?- %s(", AtomName(pred_atom));
-    int arity[ARITY_ARRAY_SIZE];
-    arity[0] = 1;
-    arity[1] = TabEnt_arity(tab_ent);
-    return traverse_subgoal_trie(sg_node, str, str_index, arity, 1, TRAVERSE_NORMAL);
+    if (TabEnt_arity(tab_ent)) {
+      char str[STR_ARRAY_SIZE];
+      int str_index = sprintf(str, "  ?- %s(", AtomName(pred_atom));
+      int arity[ARITY_ARRAY_SIZE];
+      arity[0] = 1;
+      arity[1] = TabEnt_arity(tab_ent);
+      return traverse_subgoal_trie(sg_node, str, str_index, arity, 1, TRAVERSE_NORMAL);
+    } else {
+      sg_fr_ptr sg_fr = (sg_fr_ptr) sg_node;
+      TrStat_subgoals++;
+      TrStat_sg_linear_nodes = TrStat_sg_min_depth = TrStat_sg_max_depth = 0;
+      SHOW_TABLE("  ?- %s.\n", AtomName(pred_atom));
+      TrStat_ans_nodes++;
+      TrStat_ans_max_depth = TrStat_ans_min_depth = 0;
+      if (SgFr_first_answer(sg_fr) == NULL) {
+	if (SgFr_state(sg_fr) < complete) {
+	  TrStat_sg_incomplete++;
+	  SHOW_TABLE("    ---> INCOMPLETE\n");
+	} else {
+	  TrStat_answers_no++;
+	  SHOW_TABLE("    NO\n");
+	}
+      } else if (SgFr_first_answer(sg_fr) == SgFr_answer_trie(sg_fr)) {
+	TrStat_answers_yes++;
+	TrStat_answers++;
+	SHOW_TABLE("    TRUE\n");
+      }
+    }
+    return TRUE;
   }
   SHOW_TABLE("  empty\n");
   return TRUE;
@@ -1276,7 +1300,7 @@ int traverse_subgoal_trie(sg_node_ptr sg_node, char *str, int str_index, int *ar
   }
 
   /* test the node type */
-#if SIZEOF_DOUBLE == 2 * SIZEOF_LONG_INT
+#if SIZEOF_DOUBLE == 2 * SIZEOF_INT_P
   if (mode == TRAVERSE_FLOAT) {
     arity[0]++;
     arity[arity[0]] = (int) t;
@@ -1287,10 +1311,12 @@ int traverse_subgoal_trie(sg_node_ptr sg_node, char *str, int str_index, int *ar
     *t_dbl = t;
     *(t_dbl + 1) = (Term) arity[arity[0]];
     arity[0]--;
-#else /* SIZEOF_DOUBLE == SIZEOF_LONG_INT */
+#else /* SIZEOF_DOUBLE == SIZEOF_INT_P */
   if (mode == TRAVERSE_FLOAT) {
-    Float dbl = (Float) t;
-#endif /* SIZEOF_DOUBLE x SIZEOF_LONG_INT */
+    volatile Float dbl;
+    volatile Term *t_dbl = (Term *)((void *) &dbl);
+    *t_dbl = t;
+#endif /* SIZEOF_DOUBLE x SIZEOF_INT_P */
     str_index += sprintf(& str[str_index], "%.15g", dbl);
     while (arity[0]) {
       if (arity[arity[0]] > 0) {
@@ -1316,7 +1342,11 @@ int traverse_subgoal_trie(sg_node_ptr sg_node, char *str, int str_index, int *ar
     mode = TRAVERSE_NORMAL;
   } else if (mode == TRAVERSE_LONG) {
     Int li = (Int) t;
+#if SHORT_INTS
+    str_index += sprintf(& str[str_index], "%ld", li);
+#else
     str_index += sprintf(& str[str_index], "%d", li);
+#endif /* SHORT_INTS */
     while (arity[0]) {
       if (arity[arity[0]] > 0) {
 	arity[arity[0]]--;
@@ -1340,8 +1370,12 @@ int traverse_subgoal_trie(sg_node_ptr sg_node, char *str, int str_index, int *ar
     }
     mode = TRAVERSE_NORMAL;
   } else if (IsVarTerm(t)) {
+#if SHORT_INTS
+    str_index += sprintf(& str[str_index], "VAR%ld", VarIndexOfTableTerm(t));
+#else
     str_index += sprintf(& str[str_index], "VAR%d", VarIndexOfTableTerm(t));
-    while (arity[0]) {
+#endif /* SHORT_INTS */
+   while (arity[0]) {
       if (arity[arity[0]] > 0) {
 	arity[arity[0]]--;
 	if (arity[arity[0]] == 0) {
@@ -1363,7 +1397,11 @@ int traverse_subgoal_trie(sg_node_ptr sg_node, char *str, int str_index, int *ar
       }
     }
   } else if (IsIntTerm(t)) {
+#if SHORT_INTS
+    str_index += sprintf(& str[str_index], "%ld", IntOfTerm(t));
+#else
     str_index += sprintf(& str[str_index], "%d", IntOfTerm(t));
+#endif /* SHORT_INTS */
     while (arity[0]) {
       if (arity[arity[0]] > 0) {
 	arity[arity[0]]--;
@@ -1456,19 +1494,19 @@ int traverse_subgoal_trie(sg_node_ptr sg_node, char *str, int str_index, int *ar
     } else if (depth > TrStat_sg_max_depth) {
       TrStat_sg_max_depth = depth;
     }
-    if (SgFr_state(sg_fr) < complete) {
-      TrStat_sg_incomplete++;
-      SHOW_TABLE("%s. ---> INCOMPLETE\n", str);
-    } else {
-      SHOW_TABLE("%s.\n", str);
-    }
+    SHOW_TABLE("%s.\n", str);
     TrStat_ans_nodes++;
     if (SgFr_first_answer(sg_fr) == NULL) {
       if (TrStat_ans_max_depth < 0)
 	TrStat_ans_max_depth = 0;
       TrStat_ans_min_depth = 0;
-      TrStat_answers_no++;
-      SHOW_TABLE("    NO ANSWERS\n");
+      if (SgFr_state(sg_fr) < complete) {
+	TrStat_sg_incomplete++;
+	SHOW_TABLE("    ---> INCOMPLETE\n");
+      } else {
+	TrStat_answers_no++;
+	SHOW_TABLE("    NO\n");
+      }
     } else if (SgFr_first_answer(sg_fr) == SgFr_answer_trie(sg_fr)) {
       if (TrStat_ans_max_depth < 0)
 	TrStat_ans_max_depth = 0;
@@ -1482,6 +1520,10 @@ int traverse_subgoal_trie(sg_node_ptr sg_node, char *str, int str_index, int *ar
       answer_arity[0] = 0;
       if (! traverse_answer_trie(TrNode_child(SgFr_answer_trie(sg_fr)), answer_str, 0, answer_arity, 0, 1, TRAVERSE_NORMAL))
 	return FALSE;
+      if (SgFr_state(sg_fr) < complete) {
+	TrStat_sg_incomplete++;
+	SHOW_TABLE("    ---> INCOMPLETE\n");
+      }
     }
   }
 
@@ -1536,7 +1578,7 @@ int traverse_answer_trie(ans_node_ptr ans_node, char *str, int str_index, int *a
 
   /* test the node type */
   if (mode == TRAVERSE_FLOAT) {
-#if SIZEOF_DOUBLE == 2 * SIZEOF_LONG_INT
+#if SIZEOF_DOUBLE == 2 * SIZEOF_INT_P
     arity[0]++;
     arity[arity[0]] = (int) t;
     mode = TRAVERSE_FLOAT2;
@@ -1546,9 +1588,11 @@ int traverse_answer_trie(ans_node_ptr ans_node, char *str, int str_index, int *a
     *t_dbl = t;
     *(t_dbl + 1) = (Term) arity[arity[0]];
     arity[0]--;
-#else /* SIZEOF_DOUBLE == SIZEOF_LONG_INT */
-    Float dbl = (Float) t;
-#endif /* SIZEOF_DOUBLE x SIZEOF_LONG_INT */
+#else /* SIZEOF_DOUBLE == SIZEOF_INT_P */
+    volatile Float dbl;
+    volatile Term *t_dbl = (Term *)((void *) &dbl);
+    *t_dbl = t;
+#endif /* SIZEOF_DOUBLE x SIZEOF_INT_P */
     str_index += sprintf(& str[str_index], "%.15g", dbl);
     while (arity[0]) {
       if (arity[arity[0]] > 0) {
@@ -1576,7 +1620,11 @@ int traverse_answer_trie(ans_node_ptr ans_node, char *str, int str_index, int *a
     mode = TRAVERSE_NORMAL;
   } else if (mode == TRAVERSE_LONG) {
     Int li = (Int) t;
+#if SHORT_INTS
+    str_index += sprintf(& str[str_index], "%ld", li);
+#else
     str_index += sprintf(& str[str_index], "%d", li);
+#endif /* SHORT_INTS */
     while (arity[0]) {
       if (arity[arity[0]] > 0) {
 	arity[arity[0]]--;
@@ -1602,7 +1650,11 @@ int traverse_answer_trie(ans_node_ptr ans_node, char *str, int str_index, int *a
   } else if (mode == TRAVERSE_LONG_END) {
     mode = TRAVERSE_NORMAL;
   } else if (IsVarTerm(t)) {
+#if SHORT_INTS
+    str_index += sprintf(& str[str_index], "ANSVAR%ld", VarIndexOfTableTerm(t));
+#else
     str_index += sprintf(& str[str_index], "ANSVAR%d", VarIndexOfTableTerm(t));
+#endif /* SHORT_INTS */
     while (arity[0]) {
       if (arity[arity[0]] > 0) {
 	arity[arity[0]]--;
@@ -1625,7 +1677,11 @@ int traverse_answer_trie(ans_node_ptr ans_node, char *str, int str_index, int *a
       }
     }
   } else if (IsIntTerm(t)) {
+#if SHORT_INTS
+    str_index += sprintf(& str[str_index], "%ld", IntOfTerm(t));
+#else
     str_index += sprintf(& str[str_index], "%d", IntOfTerm(t));
+#endif /* SHORT_INTS */
     while (arity[0]) {
       if (arity[arity[0]] > 0) {
 	arity[arity[0]]--;
