@@ -78,18 +78,6 @@ static void PROTO(do_bootfile, (char *));
 static void PROTO(do_top_goal,(YAP_Term));
 static void PROTO(exec_top_level,(int, YAP_init_args *));
 
-#ifndef LIGHT
-void PROTO (exit, (int));
-#endif
-
-#ifdef LIGHT
-#include <unix.h>
-int
-  _main (int, char **);
-#else
-int PROTO (main, (int, char **));
-#endif
-
 #ifdef DEBUG
 static int output_msg;
 #endif
@@ -117,8 +105,7 @@ static char *def_value[MAXDEFS];
 static int  def_c=0;
 /* End preprocessor code */
 
-
-#if USE_MYPUTC
+#ifdef USE_MYPUTC
 static void
 myputc (int ch)
 {
@@ -131,7 +118,7 @@ mygetc (void)
 {
   int ch;
   if (eof_found)
-    return (EOF);
+    return EOF;
   ch = getc (bootfile);
   if (ch == EOF)
     eof_found = TRUE;
@@ -141,7 +128,7 @@ mygetc (void)
 #endif
     yap_lineno++;
   }
-  return (ch);
+  return ch;
 }
 
 static void
@@ -152,7 +139,7 @@ do_top_goal (YAP_Term Goal)
     fprintf(stderr,"Entering absmi\n");
 #endif
   /* PlPutc(0,'a'); PlPutc(0,'\n'); */
-  YAP_RunGoal(Goal);
+  YAP_RunGoalOnce(Goal);
 }
 
 /* do initial boot by consulting the file boot.yap */
@@ -234,8 +221,10 @@ print_usage(void)
   fprintf(stderr,"\n[ Valid switches for command line arguments: ]\n");
   fprintf(stderr,"  -?   Shows this screen\n");
   fprintf(stderr,"  -b   Boot file \n");
+  fprintf(stderr,"  -f   Do not consult startup prolog files\n");
   fprintf(stderr,"  -g   Run Goal Before Top-Level \n");
   fprintf(stderr,"  -z   Run Goal Before Top-Level \n");
+  fprintf(stderr,"  -q   start with informational messages off\n");
   fprintf(stderr,"  -l   load Prolog file\n");
   fprintf(stderr,"  -L   run Prolog file and exit\n");
   fprintf(stderr,"  -p   extra path for file-search-path\n");
@@ -248,16 +237,16 @@ print_usage(void)
 #ifdef TABLING
   fprintf(stderr,"  -ts  Maximum table space area in Mbytes (default: unlimited)\n");
 #endif /* TABLING */
-#ifdef YAPOR
+#if defined(ENV_COPY) || defined(ACOW) || defined(SBA)
   fprintf(stderr,"  -w   Number of workers (default: %d)\n",
 	  DEFAULT_NUMBERWORKERS);
   fprintf(stderr,"  -sl  Loop scheduler executions before look for hiden shared work (default: %d)\n", 
           DEFAULT_SCHEDULERLOOP);
   fprintf(stderr,"  -d   Value of delayed release of load (default: %d)\n",
 	  DEFAULT_DELAYEDRELEASELOAD);
+#endif /* ENV_COPY || ACOW || SBA */
   /* nf: Preprocessor */		
-  fprintf(stderr,"  -DVar=Name : persistent definition\n");
-#endif /* YAPOR */
+  /* fprintf(stderr,"  -DVar=Name   Persistent definition\n"); */
   fprintf(stderr,"\n");
 }
 
@@ -273,7 +262,7 @@ static int
 parse_yap_arguments(int argc, char *argv[], YAP_init_args *iap)
 {
   char *p;
-#if USE_SYSTEM_MALLOC
+#ifdef USE_SYSTEM_MALLOC
   int BootMode = YAP_FULL_BOOT_FROM_PROLOG;
 #else
   int BootMode = YAP_BOOT_FROM_SAVED_CODE;
@@ -296,12 +285,17 @@ parse_yap_arguments(int argc, char *argv[], YAP_init_args *iap)
           case '?':
 	    print_usage();
             exit(EXIT_SUCCESS);
+          case 'q':
+	    iap->QuietMode = TRUE;
+	    break;
+#if defined(ENV_COPY) || defined(ACOW) || defined(SBA)
 	  case 'w':
 	    ssize = &(iap->NumberWorkers);
 	    goto GetSize;
           case 'd':
             ssize = &(iap->DelayedReleaseLoad);
 	    goto GetSize;
+#endif /* ENV_COPY || ACOW || SBA */
 #ifdef USE_SOCKET
           case 'c':          /* running as client */
 	    {
@@ -386,10 +380,12 @@ parse_yap_arguments(int argc, char *argv[], YAP_init_args *iap)
 	  case 's':
 	  case 'S':
 	    ssize = &(iap->StackSize);
+#if defined(ENV_COPY) || defined(ACOW) || defined(SBA)
 	    if (p[1] == 'l') {
 	      p++;
 	      ssize = &(iap->SchedulerLoop);
 	    }
+#endif /* ENV_COPY || ACOW || SBA */
 	    goto GetSize;
 	  case 'h':
 	  case 'H':
@@ -398,10 +394,12 @@ parse_yap_arguments(int argc, char *argv[], YAP_init_args *iap)
 	  case 't':
 	  case 'T':
 	    ssize = &(iap->TrailSize);
+#ifdef TABLING
 	    if (p[1] == 's') {
 	      p++;
 	      ssize = &(iap->MaxTableSpaceSize);
 	    }
+#endif /* TABLING */
 	  GetSize:
 	    if (*++p == '\0')
 	      {
@@ -433,47 +431,34 @@ parse_yap_arguments(int argc, char *argv[], YAP_init_args *iap)
 	    break;
 #endif
 	  case 'L':
-           p++;
-           iap->HaltAfterConsult = TRUE;
-           if (*p != '\0') {
-             iap->YapPrologRCFile = p;
-             break;
-           } else {
-             if (--argc == 0) {
-               fprintf(stderr,
-                       " [ YAP unrecoverable error: missing file name with option 'L' ]\n");
-               exit(1);
-             }
-             p = *++argv;
-             if (p[0] == '-' && p[1] == '-' && p[2] == '\0') {
-               if (--argc == 0) {
-                 fprintf(stderr,
-                         " [ YAP unrecoverable error: missing filename with option 'L' ]\n");
-                 exit(1);
-               }
-               p = *++argv;
-               iap->YapPrologRCFile = p;
-               argc = 1;
-
-             } else {
-               iap->YapPrologRCFile = p;
-             }
-           }
-           break;
+	    iap->QuietMode = TRUE;
+	    iap->HaltAfterConsult = TRUE;
 	  case 'l':
-	    if ((*argv)[0] == '\0') 
-	      iap->YapPrologRCFile = *argv;
-	    else {
-	      argc--;
-	      if (argc == 0) {
-		fprintf(stderr," [ YAP unrecoverable error: missing file name with option 'l' ]\n");
-		exit(EXIT_FAILURE);
-	      }
-	      argv++;
-	      iap->YapPrologRCFile = *argv;
-	    }
-	    break;
-	    /* run goal before top-level */
+           p++;
+	   if (!*++argv) {
+	     fprintf(stderr,
+		     "%% YAP unrecoverable error: missing boot file name\n");
+	     exit(1);
+	   } else {
+	     iap->YapPrologRCFile = *argv;
+	     argc--;
+	   }
+           if (*p) {
+	     /* we have something, usually, of the form:
+		-L --
+		FileName
+		ExtraArgs
+	     */
+	     /* being called from a script */
+	     while (*p && (*p == ' ' || *p == '\t'))
+	       p++;
+	     if (p[0] == '-' && p[1] == '-') {
+	       /* ignore what is next */
+	       argc = 1;
+	     }
+	   }
+	   break;
+	   /* run goal before top-level */
 	  case 'g':
 	    if ((*argv)[0] == '\0') 
 	      iap->YapPrologGoal = *argv;
@@ -566,7 +551,7 @@ parse_yap_arguments(int argc, char *argv[], YAP_init_args *iap)
       goto myddas_error;
     }
 #endif
-  return(BootMode);
+  return BootMode;
 }
 
 static char boot_file[256];
@@ -605,11 +590,11 @@ init_standard_system(int argc, char *argv[], YAP_init_args *iap)
 #endif  
   iap->ErrorNo = 0;
   iap->ErrorCause = NULL;
+  iap->QuietMode = FALSE;
 
   BootMode = parse_yap_arguments(argc,argv,iap);
 
   if (BootMode == YAP_FULL_BOOT_FROM_PROLOG) {
-
 #if HAVE_STRNCAT
     strncpy(boot_file, PL_SRC_DIR, 256);
 #else
@@ -680,17 +665,17 @@ exec_top_level(int BootMode, YAP_init_args *iap)
 	fgoal = YAP_MkFunctor(YAP_FullLookupAtom("$silent_bootstrap"), 1);
 	goal = YAP_MkApplTerm(fgoal, 1, as);
 	/* launch consult */
-	YAP_RunGoal(goal);
+	YAP_RunGoalOnce(goal);
 	/* set default module to user */
 	as[0] = YAP_MkAtomTerm(YAP_LookupAtom("user"));
 	fgoal = YAP_MkFunctor(YAP_FullLookupAtom("module"), 1);
 	goal = YAP_MkApplTerm(fgoal, 1, as);
-	YAP_RunGoal(goal);
+	YAP_RunGoalOnce(goal);
       }
       YAP_PutValue(livegoal, YAP_MkAtomTerm (YAP_FullLookupAtom("$true")));
       
     }
-      /* the top-level is now ready */
+  /* the top-level is now ready */
 
   /* read it before case someone, that is, Ashwin, hides
      the atom false away ;-).
@@ -727,14 +712,14 @@ main (int argc, char **argv)
     // load the module
     YAP_Term mod_arg[1];
     mod_arg[0] = YAP_MkAtomTerm(YAP_LookupAtom("ypp"));
-    YAP_RunGoal(YAP_MkApplTerm(YAP_MkFunctor(YAP_LookupAtom("use_module"),1), 1, mod_arg)); 
+    YAP_RunGoalOnce(YAP_MkApplTerm(YAP_MkFunctor(YAP_LookupAtom("use_module"),1), 1, mod_arg)); 
     // process the definitions
     for(i=0;i<def_c;++i) {
       YAP_Term t_args[2],t_goal;
       t_args[0] = YAP_MkAtomTerm(YAP_LookupAtom(def_var[i]));
       t_args[1] = YAP_MkAtomTerm(YAP_LookupAtom(def_value[i])); 
       t_goal  = YAP_MkApplTerm(YAP_MkFunctor(YAP_LookupAtom("ypp_define"),2), 2, t_args); 
-      YAP_RunGoal(t_goal);
+      YAP_RunGoalOnce(t_goal);
     }
   }
   YAP_ClearExceptions();
