@@ -8,8 +8,40 @@
 *									 *
 **************************************************************************
 *									 *
-* $Id: sys.c,v 1.25 2006/01/17 14:10:42 vsc Exp $									 *
+* $Id: sys.c,v 1.35 2008/05/23 13:16:13 vsc Exp $									 *
 * mods:		$Log: sys.c,v $
+* mods:		Revision 1.35  2008/05/23 13:16:13  vsc
+* mods:		fix sys.c for win32
+* mods:		
+* mods:		Revision 1.34  2008/05/22 23:25:21  vsc
+* mods:		add tmp_file/2
+* mods:		
+* mods:		Revision 1.33  2007/10/05 18:24:30  vsc
+* mods:		fix garbage collector and fix LeaveGoal
+* mods:		
+* mods:		Revision 1.32  2007/05/07 12:11:39  vsc
+* mods:		fix mktime fix
+* mods:		
+* mods:		Revision 1.31  2007/05/07 11:21:29  vsc
+* mods:		mktime needs to know if daylight time savings are on
+* mods:		(obs from Bernd Gutmann).
+* mods:		
+* mods:		Revision 1.30  2007/05/02 11:16:43  vsc
+* mods:		small fixes to sys.c
+* mods:		
+* mods:		Revision 1.29  2006/10/10 14:08:17  vsc
+* mods:		small fixes on threaded implementation.
+* mods:		
+* mods:		Revision 1.28  2006/05/25 16:28:28  vsc
+* mods:		include thread_sleep functionality.
+* mods:		
+* mods:		Revision 1.27  2006/05/17 18:38:11  vsc
+* mods:		make system library use true file name
+* mods:		
+* mods:		Revision 1.26  2006/04/25 03:23:40  vsc
+* mods:		fix ! in debugger (execute_clause)
+* mods:		improve system/1 and execute/1
+* mods:		
 * mods:		Revision 1.25  2006/01/17 14:10:42  vsc
 * mods:		YENV may be an HW register (breaks some tabling code)
 * mods:		All YAAM instructions are now brackedted, so Op introduced an { and EndOp introduces an }. This is because Ricardo assumes that.
@@ -57,9 +89,7 @@
 
 #include "config.h"
 #include "YapInterface.h"
-#if STDC_HEADERS
 #include <stdlib.h>
-#endif
 #if HAVE_UNISTD_H
 #include <unistd.h>
 #endif
@@ -76,6 +106,9 @@
 #if HAVE_FCNTL_H
 #include <fcntl.h>
 #endif
+#if HAVE_MATH_H
+#include <math.h>
+#endif
 #if HAVE_UNISTD_H
 #include <unistd.h>
 #endif
@@ -172,6 +205,7 @@ sysmktime(void)
   loc.tm_hour = YAP_IntOfTerm(YAP_ARG4);
   loc.tm_min = YAP_IntOfTerm(YAP_ARG5);
   loc.tm_sec = YAP_IntOfTerm(YAP_ARG6);
+  loc.tm_isdst = daylight;
 
   if ((tim = mktime(&loc)) < 0) {
     return YAP_Unify(YAP_ARG8, YAP_MkIntTerm(errno));
@@ -250,7 +284,7 @@ list_directory(void)
 #if HAVE_STRNCAT
   strncat(bs, "/*", BUF_SIZE);
 #else
-  strncat(bs, "/*");
+  strcat(bs, "/*");
 #endif
   if ((hFile = _findfirst(bs, &c_file)) == -1L) {
     return(YAP_Unify(YAP_ARG2,tf));
@@ -349,7 +383,7 @@ dir_separator(void)
 static int
 file_property(void)
 {
-  char *fd;
+  const char *fd;
 #if HAVE_LSTAT 
   struct stat buf;
 
@@ -458,15 +492,49 @@ p_mktemp(void)
 }
 
 static int
-p_tpmnam(void)
+p_tmpnam(void)
 {
 #if HAVE_TMPNAM
-  return(YAP_Unify(YAP_ARG1,YAP_MkAtomTerm(YAP_LookupAtom(tmpnam(NULL)))));
+  char buf[L_tmpnam], *s;
+  if (!(s = tmpnam(buf)))
+    return FALSE;
+  return YAP_Unify(YAP_ARG1,YAP_MkAtomTerm(YAP_LookupAtom(s)));
 #else
   return FALSE;
 #endif
 }
 
+static int
+p_tmpdir(void)
+{
+#if defined(__MINGW32__) || _MSC_VER
+  char buf[512];
+  DWORD out = GetTempPath(512, buf);
+  if (!out) {
+    return(YAP_Unify(YAP_ARG2, WinError()));
+  }
+  if (out > 511) {
+    char *nbuf = malloc(out+1);
+    if (!nbuf)
+      return YAP_Unify(YAP_ARG2, YAP_MkAtomTerm(YAP_LookupAtom("no malloc memory")));
+    out = GetTempPath(512, nbuf);
+    if (!out) {
+      return YAP_Unify(YAP_ARG2, WinError());
+    }
+    return  YAP_Unify(YAP_ARG1,YAP_MkAtomTerm(YAP_LookupAtom(nbuf)));
+  }
+  return  YAP_Unify(YAP_ARG1,YAP_MkAtomTerm(YAP_LookupAtom(buf)));
+#else
+  char *s;
+  if ((s = getenv("TMPDIR")))
+    return  YAP_Unify(YAP_ARG1,YAP_MkAtomTerm(YAP_LookupAtom(s)));
+#ifdef P_tmpdir
+  return  YAP_Unify(YAP_ARG1,YAP_MkAtomTerm(YAP_LookupAtom(P_tmpdir)));
+#endif
+  return  YAP_Unify(YAP_ARG1,YAP_MkAtomTerm(YAP_LookupAtom("/tmp")));
+#endif
+}
+
 /* return YAP's environment */
 static int
 p_environ(void)
@@ -576,7 +644,7 @@ execute_command(void)
   StartupInfo.hStdError = errf;
   /* got stdin, stdout and error as I like it */
   if (CreateProcess(NULL,
-		    YAP_AtomName(YAP_AtomOfTerm(YAP_ARG1)),
+		    (char *)YAP_AtomName(YAP_AtomOfTerm(YAP_ARG1)),
 		    NULL,
 		    NULL,
 		    TRUE,
@@ -692,12 +760,12 @@ do_system(void)
 #if HAVE_SYSTEM
   int sys = system(command);
   if (sys < 0) {
-    return(YAP_Unify(YAP_ARG3,YAP_MkIntTerm(errno)));
+    return YAP_Unify(YAP_ARG3,YAP_MkIntTerm(errno));
   }
-  return(YAP_Unify(YAP_ARG2, YAP_MkIntTerm(sys)));
+  return YAP_Unify(YAP_ARG2, YAP_MkIntTerm(sys));
 #else
-  YAP_Error(0,0L,"system not available in this configuration");
-  return(FALSE);
+  YAP_Error(0,0L,"system not available in this configuration, trying %s", command);
+  return FALSE;
 #endif
 }
 
@@ -708,6 +776,9 @@ static int
 do_shell(void)
 {
 #if defined(__MINGW32__) || _MSC_VER
+  YAP_Error(0,0L,"system not available in this configuration");
+  return(FALSE);
+#elif HAVE_SYSTEM
   char *buf = YAP_AllocSpaceFromYap(BUF_SIZE);
   int sys;
 
@@ -716,29 +787,23 @@ do_shell(void)
     return(FALSE);
   }
 #if HAVE_STRNCPY
-  strncpy(YAP_AtomName(YAP_AtomOfTerm(YAP_ARG1)), buf, BUF_SIZE);
-  strncpy(" ", buf, BUF_SIZE);
-  strncpy(YAP_AtomName(YAP_AtomOfTerm(YAP_ARG2)), buf, BUF_SIZE);
-  strncpy(" ", buf, BUF_SIZE);
-  strncpy(YAP_AtomName(YAP_AtomOfTerm(YAP_ARG3)), buf, BUF_SIZE);
+  strncpy(buf, YAP_AtomName(YAP_AtomOfTerm(YAP_ARG1)), BUF_SIZE);
+  strncpy(buf, " ", BUF_SIZE);
+  strncpy(buf, YAP_AtomName(YAP_AtomOfTerm(YAP_ARG2)), BUF_SIZE);
+  strncpy(buf, " ", BUF_SIZE);
+  strncpy(buf, YAP_AtomName(YAP_AtomOfTerm(YAP_ARG3)), BUF_SIZE);
 #else
-  strcpy(YAP_AtomName(YAP_AtomOfTerm(YAP_ARG1)), buf);
-  strcpy(" ", buf);
-  strcpy(YAP_AtomName(YAP_AtomOfTerm(YAP_ARG2)), buf);
-  strcpy(" ", buf);
-  strcpy(YAP_AtomName(YAP_AtomOfTerm(YAP_ARG3)), buf);
+  strcpy(buf, YAP_AtomName(YAP_AtomOfTerm(YAP_ARG1)));
+  strcpy(buf, YAP_AtomName(YAP_AtomOfTerm(YAP_ARG2)));
+  strcpy(buf, " ");
+  strcpy(buf, YAP_AtomName(YAP_AtomOfTerm(YAP_ARG3)));
 #endif
-#if HAVE_SYSTEM
   sys = system(buf);
   YAP_FreeSpaceFromYap(buf);
   if (sys < 0) {
-    return(YAP_Unify(YAP_ARG5,YAP_MkIntTerm(errno)));
+    return YAP_Unify(YAP_ARG5,YAP_MkIntTerm(errno));
   }
-  return(YAP_Unify(YAP_ARG4, YAP_MkIntTerm(sys)));
-#else
-  YAP_Error(0,0L,"system not available in this configuration");
-  return(FALSE);
-#endif
+  return YAP_Unify(YAP_ARG4, YAP_MkIntTerm(sys));
 #else
   char *cptr[4];
   int t;
@@ -750,17 +815,19 @@ do_shell(void)
   cptr[3]= NULL;
   t = fork();
   if (t < 0) {
-    return(YAP_Unify(YAP_ARG5,YAP_MkIntTerm(errno)));
+    return YAP_Unify(YAP_ARG5,YAP_MkIntTerm(errno));
   } else if (t == 0) {
-    t = execvp(YAP_AtomName(YAP_AtomOfTerm(YAP_ARG1)),cptr);
-    return(t);
+    t = execvp(cptr[0],cptr);
+    return t;
   } else {
     t = wait(&sys);
+    fprintf(stderr,"after wait %x:%x\n",t,sys);
     if (t < 0) {
-      return(YAP_Unify(YAP_ARG5,YAP_MkIntTerm(errno)));
+      return YAP_Unify(YAP_ARG5,YAP_MkIntTerm(errno));
     }
   }
-  return(YAP_Unify(YAP_ARG4, YAP_MkIntTerm(sys)));
+  fprintf(stderr,"after wait %x\n", sys);
+  return YAP_Unify(YAP_ARG4, YAP_MkIntTerm(sys));
 #endif
 }
 
@@ -841,7 +908,7 @@ static int
 p_sleep(void)
 {
   YAP_Term ts = YAP_ARG1;
-  long int secs = 0, usecs = 0, out;
+  unsigned long int secs = 0, usecs = 0, out;
   if (YAP_IsIntTerm(ts)) {
     secs = YAP_IntOfTerm(ts);
   }  else if (YAP_IsFloatTerm(ts)) {
@@ -849,25 +916,39 @@ p_sleep(void)
     if (tfl > 1.0)
       secs = tfl;
     else
-      usecs = tfl*1000;
+      usecs = tfl*1000000;
   }
 #if defined(__MINGW32__) || _MSC_VER
-  if (secs) usecs = secs*1000;
+  if (secs) usecs = secs*1000 + usecs/1000;
   Sleep(usecs);
+  /* no errors possible */
   out = 0;
-#else
-#if HAVE_USLEEP
+#elif HAVE_NANOSLEEP
+  {
+    struct timespec req;
+    if (YAP_IsFloatTerm(ts)) {
+      double tfl = YAP_FloatOfTerm(ts);
+     
+      req.tv_nsec = (tfl-floor(tfl))*1000000000;
+      req.tv_sec = rint(tfl);
+    } else {
+      req.tv_nsec = 0;
+      req.tv_sec = secs;
+    }
+    out = nanosleep(&req, NULL);
+  }
+#elif HAVE_USLEEP
   if (usecs > 0) {
-    usleep(usecs);
-    out = 0;
+    out = usleep(usecs);
   } else
-#endif
-#if HAVE_SLEEP
+#elif HAVE_SLEEP
     {
       out = sleep(secs);
     }
+#else
+  YAP_Error(0,0L,"sleep not available in this configuration");
+  return FALSE:
 #endif
-#endif /* defined(__MINGW32__) || _MSC_VER */
   return(YAP_Unify(YAP_ARG2, YAP_MkIntTerm(out)));
 }
 
@@ -952,17 +1033,18 @@ static int
 error_message(void)
 {
 #if HAVE_STRERROR
-  return(YAP_Unify(YAP_ARG2,YAP_MkAtomTerm(YAP_LookupAtom(strerror(YAP_IntOfTerm(YAP_ARG1))))));
+  return YAP_Unify(YAP_ARG2,YAP_MkAtomTerm(YAP_LookupAtom(strerror(YAP_IntOfTerm(YAP_ARG1)))));
 #else
-#if HAVE_STRERROR
-  return(YAP_Unify(YAP_ARG2,YAP_ARG1));
-#endif
+  return YAP_Unify(YAP_ARG2,YAP_ARG1);
 #endif
 }
 
 void
 init_sys(void)
 {
+#if HAVE_MKTIME
+  tzset();
+#endif
   YAP_UserCPredicate("datime", datime, 2);
   YAP_UserCPredicate("mktime", sysmktime, 8);
   YAP_UserCPredicate("list_directory", list_directory, 3);
@@ -982,7 +1064,8 @@ init_sys(void)
   YAP_UserCPredicate("pid", pid, 2);
   YAP_UserCPredicate("kill", p_kill, 3);
   YAP_UserCPredicate("mktemp", p_mktemp, 3);
-  YAP_UserCPredicate("tmpnam", p_tpmnam, 2);
+  YAP_UserCPredicate("tmpnam", p_tmpnam, 2);
+  YAP_UserCPredicate("tmpdir", p_tmpdir, 2);
   YAP_UserCPredicate("rename_file", rename_file, 3);
   YAP_UserCPredicate("sleep", p_sleep, 2);
   YAP_UserCPredicate("error_message", error_message, 2);
