@@ -14,12 +14,17 @@
 #include	<string.h>
 #include	<stdio.h>
 
-#include	<yap2swi.h>
+#include	<SWI-Prolog.h>
+
+#ifdef USE_GMP
+#include <gmp.h>
+#endif
 
 #define BUF_SIZE 256
 #define TMP_BUF_SIZE 2*BUF_SIZE
+#define BUF_RINGS 16
 
-char buffers[TMP_BUF_SIZE+BUF_SIZE*4];
+char buffers[TMP_BUF_SIZE+BUF_SIZE*BUF_RINGS];
 static int buf_index = 0;
 
 static char *
@@ -27,19 +32,17 @@ alloc_ring_buf(void)
 {
   int ret = buf_index;
   buf_index++;
-  if (buf_index == 4)
+  if (buf_index == BUF_RINGS)
     buf_index = 0;
   return buffers+(TMP_BUF_SIZE+ret*BUF_SIZE);
 }
 
-/* SWI: void PL_agc_hook(void)
-   YAP: NO EQUIVALENT */
+/* SWI: void PL_agc_hook(void) */
 
-/* dummy function for now (until Vitor comes through!)*/
 X_API PL_agc_hook_t
 PL_agc_hook(PL_agc_hook_t entry)
 {
-  return entry;
+  return (PL_agc_hook_t)YAP_AGCRegisterHook((YAP_agc_hook)entry);
 }
 
 /* SWI: char* PL_atom_chars(atom_t atom)
@@ -130,7 +133,7 @@ X_API int PL_get_atom_chars(term_t ts, char **a)  /* SAM check type */
   term-types should converted and the second how the argument is
   stored. Below is a specification of these constants. BUF_RING
   implies, if the data is not static (as from an atom), the data is
-  copied to the next buffer from a ring of four (4) buffers. This is a
+  copied to the next buffer from a ring of sixteen (16) buffers. This is a
   convenient way of converting multiple arguments passed to a foreign
   predicate to C-strings. If BUF_MALLOC is used, the data must be
   freed using free() when not needed any longer.
@@ -204,9 +207,14 @@ X_API int PL_get_chars(term_t l, char **sp, unsigned flags)
   }
   *sp = tmp;
   if (YAP_IsAtomTerm(t)) {
+    YAP_Atom at = YAP_AtomOfTerm(t);
     if (!(flags & (CVT_ATOM|CVT_ATOMIC|CVT_ALL)))
       return 0;
-    *sp = (char *)YAP_AtomName(YAP_AtomOfTerm(t));
+    if (YAP_IsWideAtom(at))
+      /* will this always work? */
+      snprintf(*sp,BUF_SIZE,"%ls",YAP_WideAtomName(at));
+    else
+      *sp = (char *)YAP_AtomName(YAP_AtomOfTerm(t));
     return 1;
   } else if (YAP_IsIntTerm(t)) {
     if (!(flags & (CVT_INTEGER|CVT_NUMBER|CVT_ATOMIC|CVT_ALL)))
@@ -228,7 +236,7 @@ X_API int PL_get_chars(term_t l, char **sp, unsigned flags)
     *bf = '\0';
   }
   if (flags & BUF_MALLOC) {
-    char *nbf = malloc(strlen(tmp));
+    char *nbf = YAP_AllocSpaceFromYap(strlen(tmp)+1);
     if (nbf == NULL)
       return 0;
     strncpy(nbf,tmp,BUF_SIZE);
@@ -237,6 +245,57 @@ X_API int PL_get_chars(term_t l, char **sp, unsigned flags)
   return 1;
 }
 
+X_API int PL_get_nchars(term_t l, size_t *len, char **sp, unsigned flags)
+{
+  int out = PL_get_chars(l, sp, flags);
+  if (!out) return out;
+  *len = strlen(*sp);
+  return out;
+}
+
+
+/* same as get_chars, but works on buffers of wide chars */
+X_API int PL_get_wchars(term_t l, size_t *len, wchar_t **wsp, unsigned flags)
+{
+  if (YAP_IsAtomTerm(l)) {
+    YAP_Atom at = YAP_AtomOfTerm(l);
+
+    if (!(flags & (CVT_ATOM|CVT_ATOMIC|CVT_ALL)))
+      return 0;
+    if (YAP_IsWideAtom(at))
+      /* will this always work? */
+      *wsp = (wchar_t *)YAP_WideAtomName(at);
+  } else {
+    char *sp;
+    int res = PL_get_chars(l, &sp, ((flags & ~(BUF_MALLOC|BUF_DISCARDABLE))|BUF_RING));
+    size_t sz;
+
+    if (!res)
+      return FALSE;
+    sz = wcstombs(sp,NULL,BUF_SIZE);
+    if (flags & BUF_MALLOC) {
+      wchar_t *nbf = (wchar_t *)YAP_AllocSpaceFromYap(sz+1);
+      if (nbf == NULL)
+	return 0;
+      *wsp = nbf;
+    } else if (flags & BUF_DISCARDABLE) {
+      wchar_t *buf = (wchar_t *)buffers;
+
+      if (wcstombs(sp,buf,BUF_SIZE) == -1)
+	return 0;
+      *wsp = buf;
+    } else {
+      wchar_t *tmp = (wchar_t *)alloc_ring_buf();
+      if (wcstombs(sp, tmp, BUF_SIZE) == -1)
+	return 0;
+      *wsp = tmp;
+    }
+    return res;
+  }
+  return 0;
+}
+
+
 /* SWI: int PL_get_functor(term_t t, functor_t *f)
    YAP: YAP_Functor YAP_FunctorOfTerm(Term) */
 X_API int PL_get_functor(term_t ts, functor_t *f)
@@ -282,6 +341,27 @@ X_API int PL_get_integer(term_t ts, int *i)
   return 1;
 }
 
+/* SWI: int PL_get_bool(term_t t, int *i)
+   YAP: long int  YAP_AtomOfTerm(Term) */
+X_API int PL_get_bool(term_t ts, int *i)
+{
+  YAP_Term t = YAP_GetFromSlot(ts);
+  char *sp;
+
+  if (!YAP_IsAtomTerm(t) )
+    return 0;
+  sp = (char *)YAP_AtomName(YAP_AtomOfTerm(t));
+  if (!strcmp(sp,"true")) {
+    *sp = TRUE;
+    return 1;    
+  }
+  if (!strcmp(sp,"false")) {
+    *sp = FALSE;
+    return 1;    
+  }
+  return 0;
+}
+
 X_API int PL_get_long(term_t ts, long *i)
 {
   YAP_Term t = YAP_GetFromSlot(ts);
@@ -299,6 +379,37 @@ X_API int PL_get_long(term_t ts, long *i)
   return 1;
 }
 
+
+X_API int PL_get_int64(term_t ts, int64_t *i)
+{
+  YAP_Term t = YAP_GetFromSlot(ts);
+  if (!YAP_IsIntTerm(t) ) {
+    if (YAP_IsFloatTerm(t)) {
+      double dbl = YAP_FloatOfTerm(t);
+      if (dbl - (int64_t)dbl == 0.0) {
+	*i = (int64_t)dbl;
+	return 1;
+      }
+#if USE_GMP
+    } else if (YAP_IsBigNumTerm(t)) {
+      MP_INT g;
+      char s[64];
+      YAP_BigNumOfTerm(t, (void *)&g);
+      if (mpz_sizeinbase(&g,2) > 64) {
+	return 0;
+      }
+      mpz_get_str (s, 10, &g);
+      sscanf(s, "%lld", (long long int *)i);
+      return 1;
+#endif
+    }
+    return 0;
+  }
+  *i = YAP_IntOfTerm(t);
+  return 1;
+}
+
+
 X_API int PL_get_list(term_t ts, term_t h, term_t tl)
 {
   YAP_Term t = YAP_GetFromSlot(ts);
@@ -419,6 +530,46 @@ X_API atom_t PL_new_atom(const char *c)
   return (atom_t)YAP_LookupAtom((char *)c);
 }
 
+X_API atom_t PL_new_atom_wchars(int len, const wchar_t *c)
+{
+  atom_t at;
+  int i;
+
+  for (i=0;i<len;i++) {
+    if (c[i] > 255) break;
+  }
+  if (i!=len) {
+    wchar_t *nbf = (wchar_t *)YAP_AllocSpaceFromYap((len+1)*sizeof(wchar_t));
+    for (i=0;i<len;i++)
+      nbf[i] = c[i];
+    nbf[len]='\0';
+    at = (atom_t)YAP_LookupWideAtom(nbf);
+    YAP_FreeSpaceFromYap(nbf);
+  } else {
+    char *nbf = (char *)YAP_AllocSpaceFromYap((len+1)*sizeof(char));
+    for (i=0;i<len;i++)
+      nbf[i] = c[i];
+    nbf[len]='\0';
+    at = (atom_t)YAP_LookupAtom(nbf);
+    YAP_FreeSpaceFromYap(nbf);
+  }
+  return at;
+}
+
+X_API char *PL_atom_nchars(atom_t name, size_t *sp)
+{
+  *sp = YAP_AtomNameLength((YAP_Atom)name);
+  return (char *)YAP_AtomName((YAP_Atom)name);
+}
+
+X_API wchar_t *PL_atom_wchars(atom_t name, size_t *sp)
+{
+  if (!YAP_IsWideAtom((YAP_Atom)name))
+    return NULL;
+  *sp = YAP_AtomNameLength((YAP_Atom)name);
+  return (wchar_t *)YAP_WideAtomName((YAP_Atom)name);
+}
+
 X_API functor_t PL_new_functor(atom_t name, int arity)
 {
   functor_t f;
@@ -532,6 +683,18 @@ X_API void PL_put_integer(term_t t, long n)
   YAP_PutInSlot(t,YAP_MkIntTerm(n));
 }
 
+X_API void PL_put_int64(term_t t, int64_t n)
+{
+#if USE_GMP
+  char s[64];
+  MP_INT rop;
+
+  sprintf(s, "%lld", (long long int)n);
+  mpz_init_set_str (&rop, s, 10);
+  YAP_PutInSlot(t,YAP_MkBigNumTerm((void *)&rop));
+#endif
+}
+
 X_API void PL_put_list(term_t t)
 {
   YAP_PutInSlot(t,YAP_MkNewPairTerm());
@@ -623,6 +786,24 @@ X_API int PL_unify_integer(term_t t, long n)
   return YAP_Unify(YAP_GetFromSlot(t),iterm);
 }
 
+/* SWI: int PL_unify_integer(term_t ?t, long n)
+   YAP long int  unify(YAP_Term* a, Term* b) */
+X_API int PL_unify_int64(term_t t, int64_t n)
+{	
+#if USE_GMP
+  YAP_Term iterm;
+  char s[64];
+  MP_INT rop;
+
+  sprintf(s, "%lld", (long long int)n);
+  mpz_init_set_str (&rop, s, 10);
+  iterm = YAP_MkBigNumTerm((void *)&rop);
+  return YAP_Unify(YAP_GetFromSlot(t),iterm);
+#else
+  return FALSE;
+#endif
+}
+
 /* SWI: int PL_unify_list(term_t ?t, term_t +h, term_t -t)
    YAP long int  unify(YAP_Term* a, Term* b) */
 X_API int PL_unify_list(term_t t, term_t h, term_t tail)
@@ -814,23 +995,25 @@ X_API int PL_unify_term(term_t l,...)
   }
   va_end (ap);
   ptr = (arg_types *)buffers;
-  return YAP_Unify(YAP_GetFromSlot(l),get_term(&ptr));
+    return YAP_Unify(YAP_GetFromSlot(l),get_term(&ptr));
 }
 
 /* end PL_unify_* functions =============================*/
 
-/* SWI: void PL_register_atom(atom_t atom)
-   YAP: NO EQUIVALENT */
+/* SWI: void PL_register_atom(atom_t atom) */
 /* SAM TO DO */
 X_API void PL_register_atom(atom_t atom)
 {
+  extern int Yap_AtomGetHold(atom_t atom);
+  Yap_AtomGetHold(atom);
 }
 
-/* SWI: void PL_unregister_atom(atom_t atom)
-   YAP: NO EQUIVALENT */
+/* SWI: void PL_unregister_atom(atom_t atom) */
 /* SAM TO DO */
 X_API void PL_unregister_atom(atom_t atom)
 {
+  extern int Yap_AtomReleaseHold(atom_t atom);
+  Yap_AtomReleaseHold(atom);
 }
 
 X_API int PL_get_string_chars(term_t t, char **s, int *len)
@@ -1055,6 +1238,13 @@ PL_is_initialised(int *argc, char ***argv)
   return TRUE;
 }
 
+X_API atom_t PL_module_name(module_t m)
+{
+  YAP_Atom at = YAP_AtomOfTerm((YAP_Term)m);
+  YAP_CreateModule(at);
+  return (atom_t)at;
+}
+
 X_API predicate_t PL_pred(functor_t f, module_t m)
 {
   if (YAP_IsAtomTerm(f)) {
@@ -1092,6 +1282,7 @@ open_query execution;
 
 X_API qid_t PL_open_query(module_t ctx, int flags, predicate_t p, term_t t0)
 {
+  YAP_Atom yname;
   atom_t name;
   unsigned long int  arity;
   YAP_Module  m;
@@ -1103,7 +1294,8 @@ X_API qid_t PL_open_query(module_t ctx, int flags, predicate_t p, term_t t0)
   }
   execution.open=1;
   execution.state=0;
-  YAP_PredicateInfo(p, (YAP_Atom *)&name, &arity, &m);
+  YAP_PredicateInfo(p, &yname, &arity, &m);
+  name = (atom_t)yname;
   t[0] = YAP_ModuleName(m);
   if (arity == 0) {
     t[1] = YAP_MkAtomTerm((YAP_Atom)name);
@@ -1167,13 +1359,33 @@ X_API int PL_call(term_t tp, module_t m)
 
 X_API void PL_register_extensions(PL_extension *ptr)
 {
-  /* ignore flags for now */
   while(ptr->predicate_name != NULL) {
-    YAP_UserCPredicateWithArgs(ptr->predicate_name,(YAP_Bool (*)(void))ptr->function,ptr->arity,YAP_CurrentModule());
+    if (ptr->flags & (PL_FA_NOTRACE|PL_FA_NONDETERMINISTIC|PL_FA_VARARGS|PL_FA_CREF)) {
+      YAP_Error(0,YAP_MkIntTerm(ptr->flags),"non-implemented flag %x when creating predicates", ptr->flags);
+      return;      
+    }      
+    if (ptr->flags & PL_FA_TRANSPARENT)
+      YAP_UserCPredicateWithArgs(ptr->predicate_name,(YAP_Bool (*)(void))ptr->function,ptr->arity,YAP_MkAtomTerm(YAP_LookupAtom("prolog")));
+    else
+      YAP_UserCPredicateWithArgs(ptr->predicate_name,(YAP_Bool (*)(void))ptr->function,ptr->arity,YAP_CurrentModule());
     ptr++;
   }
 }
 
+X_API void PL_register_foreign_in_module(const char *module, const char *name, int arity, foreign_t (*function)(void), int flags)
+{
+    if (flags & (PL_FA_NOTRACE|PL_FA_NONDETERMINISTIC|PL_FA_VARARGS|PL_FA_CREF)) {
+      YAP_Error(0,YAP_MkIntTerm(flags),"non-implemented flag %x when creating predicates", flags);
+      return;      
+    }      
+  if (flags & PL_FA_TRANSPARENT)
+    YAP_UserCPredicateWithArgs(name,(YAP_Bool (*)(void))function,arity,YAP_MkAtomTerm(YAP_LookupAtom("prolog")));
+  else if (module == NULL)
+    YAP_UserCPredicateWithArgs(name,(YAP_Bool (*)(void))function,arity,YAP_CurrentModule());
+  else
+    YAP_UserCPredicateWithArgs(name,(YAP_Bool (*)(void))function,arity,YAP_MkAtomTerm(YAP_LookupAtom(module)));
+}
+
 X_API void PL_load_extensions(PL_extension *ptr)
 {
   /* ignore flags for now */
@@ -1266,19 +1478,32 @@ PL_destroy_engine(PL_engine_t e)
 X_API int
 PL_set_engine(PL_engine_t engine, PL_engine_t *old)
 {
-  int cwid = YAP_ThreadSelf();
+  long int cwid = YAP_ThreadSelf();
   if (*old) *old = (PL_engine_t)cwid;
   if (engine == PL_ENGINE_CURRENT)
     return PL_ENGINE_SET;
   if (engine < 0) /* should really check if engine does not exist */
     return PL_ENGINE_INVAL;
-  if (!(YAP_ThreadAttachEngine((int)engine))) {
+  if (!(YAP_ThreadAttachEngine((long int)engine))) {
     return PL_ENGINE_INUSE;
   }
   return PL_ENGINE_SET;
 }
 
 
+X_API void *
+PL_malloc(int sz)
+{
+  return YAP_AllocSpaceFromYap(sz);
+}
+
+X_API void
+PL_free(void *obj)
+{
+  return YAP_FreeSpaceFromYap(obj);
+}
+
+
 /* note: fprintf may be called from anywhere, so please don't try
    to be smart and allocate stack from somewhere else */
 X_API int Sprintf(char *format,...)
@@ -1341,8 +1566,10 @@ SWI_ctime(void)
 #endif
 }
 
+void Yap_swi_install(void);
+
 void
-swi_install(void)
+Yap_swi_install(void)
 {
   YAP_UserCPredicate("ctime", SWI_ctime, 2);
 }
